structure HOL : sig
  type 'a equal
  val equal : 'a equal -> 'a -> 'a -> bool
  val eq : 'a equal -> 'a -> 'a -> bool
  datatype 'a itself = Type
end = struct

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

fun eq A_ a b = equal A_ a b;

datatype 'a itself = Type;

end; (*struct HOL*)

structure Set : sig
  val insert : 'a HOL.equal -> 'a -> ('a -> bool) -> 'a -> bool
  val member : 'a -> ('a -> bool) -> bool
end = struct

fun insert A_ y a x = HOL.eq A_ y x orelse a x;

fun member x a = a x;

end; (*struct Set*)

structure Orderings : sig
  type 'a ord
  val less_eq : 'a ord -> 'a -> 'a -> bool
  val less : 'a ord -> 'a -> 'a -> bool
  type 'a preorder
  val ord_preorder : 'a preorder -> 'a ord
  type 'a order
  val preorder_order : 'a order -> 'a preorder
  type 'a bot
  val order_bot : 'a bot -> 'a order
  val bot : 'a bot -> 'a
  val less_eq_bool : bool -> bool -> bool
  val less_bool : bool -> bool -> bool
  val ord_bool : bool ord
  val preorder_bool : bool preorder
  val order_bool : bool order
  val bot_boola : bool
  val bot_bool : bool bot
  val bot_fun : 'b bot -> 'a -> 'b
  type 'a linorder
  val order_linorder : 'a linorder -> 'a order
end = struct

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
val less_eq = #less_eq : 'a ord -> 'a -> 'a -> bool;
val less = #less : 'a ord -> 'a -> 'a -> bool;

type 'a preorder = {ord_preorder : 'a ord};
val ord_preorder = #ord_preorder : 'a preorder -> 'a ord;

type 'a order = {preorder_order : 'a preorder};
val preorder_order = #preorder_order : 'a order -> 'a preorder;

type 'a bot = {order_bot : 'a order, bot : 'a};
val order_bot = #order_bot : 'a bot -> 'a order;
val bot = #bot : 'a bot -> 'a;

fun less_eq_bool true b = b
  | less_eq_bool false b = true;

fun less_bool true b = false
  | less_bool false b = b;

val ord_bool = {less_eq = less_eq_bool, less = less_bool} : bool ord;

val preorder_bool = {ord_preorder = ord_bool} : bool preorder;

val order_bool = {preorder_order = preorder_bool} : bool order;

val bot_boola : bool = false;

val bot_bool = {order_bot = order_bool, bot = bot_boola} : bool bot;

fun bot_fun B_ = (fn _ => bot B_);

type 'a linorder = {order_linorder : 'a order};
val order_linorder = #order_linorder : 'a linorder -> 'a order;

end; (*struct Orderings*)

structure Fun : sig
  val id : 'a -> 'a
  val fun_upd : 'a HOL.equal -> ('a -> 'b) -> 'a -> 'b -> 'a -> 'b
end = struct

fun id x = (fn xa => xa) x;

fun fun_upd A_ f a b = (fn x => (if HOL.eq A_ x a then b else f x));

end; (*struct Fun*)

structure More_List : sig
  val fold : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
end = struct

fun fold f [] = Fun.id
  | fold f (x :: xs) = fold f xs o f x;

end; (*struct More_List*)

structure Product_Type : sig
  val fst : 'a * 'b -> 'a
  val snd : 'a * 'b -> 'b
  val apsnd : ('a -> 'b) -> 'c * 'a -> 'c * 'b
  val scomp : ('a -> 'b * 'c) -> ('b -> 'c -> 'd) -> 'a -> 'd
  val map_pair : ('a -> 'b) -> ('c -> 'd) -> 'a * 'c -> 'b * 'd
  val equal_proda : 'a HOL.equal -> 'b HOL.equal -> 'a * 'b -> 'a * 'b -> bool
  val equal_prod : 'a HOL.equal -> 'b HOL.equal -> ('a * 'b) HOL.equal
  val equal_unita : unit -> unit -> bool
  val equal_unit : unit HOL.equal
  val equal_bool : bool -> bool -> bool
end = struct

fun fst (a, b) = a;

fun snd (a, b) = b;

fun apsnd f (x, y) = (x, f y);

fun scomp f g = (fn x => let
                           val (a, b) = f x;
                         in
                           g a b
                         end);

fun map_pair f g (a, b) = (f a, g b);

fun equal_proda A_ B_ (aa, ba) (a, b) = HOL.eq A_ aa a andalso HOL.eq B_ ba b;

fun equal_prod A_ B_ = {equal = equal_proda A_ B_} : ('a * 'b) HOL.equal;

fun equal_unita u v = true;

val equal_unit = {equal = equal_unita} : unit HOL.equal;

fun equal_bool p true = p
  | equal_bool p false = not p
  | equal_bool true p = p
  | equal_bool false p = not p;

end; (*struct Product_Type*)

structure Arith : sig
  val suc : IntInf.int -> IntInf.int
  type 'a one
  val one : 'a one -> 'a
  val one_inta : IntInf.int
  val one_int : IntInf.int one
  val minus_nat : IntInf.int -> IntInf.int -> IntInf.int
  val funpow : IntInf.int -> ('a -> 'a) -> 'a -> 'a
  val ord_nat : IntInf.int Orderings.ord
  type 'a zero
  val zero : 'a zero -> 'a
  val abs_int : IntInf.int -> IntInf.int
  val sgn_int : IntInf.int -> IntInf.int
  val zero_inta : IntInf.int
  val zero_int : IntInf.int zero
  type 'a times
  val times : 'a times -> 'a -> 'a -> 'a
  type 'a power
  val one_power : 'a power -> 'a one
  val times_power : 'a power -> 'a times
  val powera : 'a -> ('a -> 'a -> 'a) -> 'a -> IntInf.int -> 'a
  val power : 'a power -> 'a -> IntInf.int -> 'a
  type 'a minus
  val minus : 'a minus -> 'a -> 'a -> 'a
  val times_int : IntInf.int times
  val power_int : IntInf.int power
  val equal_nat : IntInf.int HOL.equal
  val preorder_nat : IntInf.int Orderings.preorder
  val order_nat : IntInf.int Orderings.order
  val divmod_int : IntInf.int -> IntInf.int -> IntInf.int * IntInf.int
  val div_int : IntInf.int -> IntInf.int -> IntInf.int
  val mod_int : IntInf.int -> IntInf.int -> IntInf.int
  val linorder_nat : IntInf.int Orderings.linorder
  type 'a zero_neq_one
  val one_zero_neq_one : 'a zero_neq_one -> 'a one
  val zero_zero_neq_one : 'a zero_neq_one -> 'a zero
  val zero_neq_one_int : IntInf.int zero_neq_one
end = struct

fun suc n = IntInf.+ (n, (1 : IntInf.int));

type 'a one = {one : 'a};
val one = #one : 'a one -> 'a;

val one_inta : IntInf.int = (1 : IntInf.int);

val one_int = {one = one_inta} : IntInf.int one;

fun minus_nat n m = IntInf.max (0, (IntInf.- (n, m)));

fun funpow n f =
  (if ((n : IntInf.int) = (0 : IntInf.int)) then Fun.id
    else f o funpow (minus_nat n (1 : IntInf.int)) f);

val ord_nat =
  {less_eq = (fn a => fn b => IntInf.<= (a, b)),
    less = (fn a => fn b => IntInf.< (a, b))}
  : IntInf.int Orderings.ord;

type 'a zero = {zero : 'a};
val zero = #zero : 'a zero -> 'a;

fun abs_int i = (if IntInf.< (i, (0 : IntInf.int)) then IntInf.~ i else i);

fun sgn_int i =
  (if ((i : IntInf.int) = (0 : IntInf.int)) then (0 : IntInf.int)
    else (if IntInf.< ((0 : IntInf.int), i) then (1 : IntInf.int)
           else IntInf.~ (1 : IntInf.int)));

val zero_inta : IntInf.int = (0 : IntInf.int);

val zero_int = {zero = zero_inta} : IntInf.int zero;

type 'a times = {times : 'a -> 'a -> 'a};
val times = #times : 'a times -> 'a -> 'a -> 'a;

type 'a power = {one_power : 'a one, times_power : 'a times};
val one_power = #one_power : 'a power -> 'a one;
val times_power = #times_power : 'a power -> 'a times;

fun powera one times a n =
  (if ((n : IntInf.int) = (0 : IntInf.int)) then one
    else times a (powera one times a (minus_nat n (1 : IntInf.int))));

fun power A_ = powera (one (one_power A_)) (times (times_power A_));

type 'a minus = {minus : 'a -> 'a -> 'a};
val minus = #minus : 'a minus -> 'a -> 'a -> 'a;

val times_int = {times = (fn a => fn b => IntInf.* (a, b))} : IntInf.int times;

val power_int = {one_power = one_int, times_power = times_int} :
  IntInf.int power;

val equal_nat = {equal = (fn a => fn b => ((a : IntInf.int) = b))} :
  IntInf.int HOL.equal;

val preorder_nat = {ord_preorder = ord_nat} : IntInf.int Orderings.preorder;

val order_nat = {preorder_order = preorder_nat} : IntInf.int Orderings.order;

fun divmod_int k l =
  (if ((k : IntInf.int) = (0 : IntInf.int))
    then ((0 : IntInf.int), (0 : IntInf.int))
    else (if ((l : IntInf.int) = (0 : IntInf.int)) then ((0 : IntInf.int), k)
           else Product_Type.apsnd (fn a => IntInf.* (sgn_int l, a))
                  (if (((sgn_int k) : IntInf.int) = (sgn_int l))
                    then IntInf.divMod (IntInf.abs k, IntInf.abs l)
                    else let
                           val (r, s) =
                             IntInf.divMod (IntInf.abs k, IntInf.abs l);
                         in
                           (if ((s : IntInf.int) = (0 : IntInf.int))
                             then (IntInf.~ r, (0 : IntInf.int))
                             else (IntInf.- (IntInf.~ r, (1 : IntInf.int)),
                                    IntInf.- (abs_int l, s)))
                         end)));

fun div_int a b = Product_Type.fst (divmod_int a b);

fun mod_int a b = Product_Type.snd (divmod_int a b);

val linorder_nat = {order_linorder = order_nat} : IntInf.int Orderings.linorder;

type 'a zero_neq_one = {one_zero_neq_one : 'a one, zero_zero_neq_one : 'a zero};
val one_zero_neq_one = #one_zero_neq_one : 'a zero_neq_one -> 'a one;
val zero_zero_neq_one = #zero_zero_neq_one : 'a zero_neq_one -> 'a zero;

val zero_neq_one_int =
  {one_zero_neq_one = one_int, zero_zero_neq_one = zero_int} :
  IntInf.int zero_neq_one;

end; (*struct Arith*)

structure List : sig
  val hd : 'a list -> 'a
  val tl : 'a list -> 'a list
  val map : ('a -> 'b) -> 'a list -> 'b list
  val nth : 'a list -> IntInf.int -> 'a
  val rev : 'a list -> 'a list
  val set : 'a HOL.equal -> 'a list -> 'a -> bool
  val upt : IntInf.int -> IntInf.int -> IntInf.int list
  val zip : 'a list -> 'b list -> ('a * 'b) list
  val maps : ('a -> 'b list) -> 'a list -> 'b list
  val null : 'a list -> bool
  val foldl : ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a
  val foldr : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
  val member : 'a HOL.equal -> 'a list -> 'a -> bool
  val remdups : 'a HOL.equal -> 'a list -> 'a list
  val remove1 : 'a HOL.equal -> 'a -> 'a list -> 'a list
  val distinct : 'a HOL.equal -> 'a list -> bool
  val list_all : ('a -> bool) -> 'a list -> bool
  val dropWhile : ('a -> bool) -> 'a list -> 'a list
  val equal_lista : 'a HOL.equal -> 'a list -> 'a list -> bool
  val equal_list : 'a HOL.equal -> ('a list) HOL.equal
  val replicate : IntInf.int -> 'a -> 'a list
  val size_list : 'a list -> IntInf.int
  val insort_key :
    'b Orderings.linorder -> ('a -> 'b) -> 'a -> 'a list -> 'a list
  val insort_insert_key :
    'b HOL.equal * 'b Orderings.linorder ->
      ('a -> 'b) -> 'a -> 'a list -> 'a list
end = struct

fun hd (x :: xs) = x;

fun tl [] = []
  | tl (x :: xs) = xs;

fun map f [] = []
  | map f (x :: xs) = f x :: map f xs;

fun nth (x :: xs) n =
  (if ((n : IntInf.int) = (0 : IntInf.int)) then x
    else nth xs (Arith.minus_nat n (1 : IntInf.int)));

fun rev xs = More_List.fold (fn a => fn b => a :: b) xs [];

fun set A_ (x :: xs) = Set.insert A_ x (set A_ xs)
  | set A_ [] = Orderings.bot_fun Orderings.bot_bool;

fun upt i j = (if IntInf.< (i, j) then i :: upt (Arith.suc i) j else []);

fun zip (x :: xs) (y :: ys) = (x, y) :: zip xs ys
  | zip xs [] = []
  | zip [] ys = [];

fun maps f [] = []
  | maps f (x :: xs) = f x @ maps f xs;

fun null [] = true
  | null (x :: xs) = false;

fun foldl f a [] = a
  | foldl f a (x :: xs) = foldl f (f a x) xs;

fun foldr f [] a = a
  | foldr f (x :: xs) a = f x (foldr f xs a);

fun member A_ [] y = false
  | member A_ (x :: xs) y = HOL.eq A_ x y orelse member A_ xs y;

fun remdups A_ [] = []
  | remdups A_ (x :: xs) =
    (if member A_ xs x then remdups A_ xs else x :: remdups A_ xs);

fun remove1 A_ x [] = []
  | remove1 A_ x (y :: xs) =
    (if HOL.eq A_ x y then xs else y :: remove1 A_ x xs);

fun distinct A_ [] = true
  | distinct A_ (x :: xs) = not (member A_ xs x) andalso distinct A_ xs;

fun list_all p [] = true
  | list_all p (x :: xs) = p x andalso list_all p xs;

fun dropWhile p [] = []
  | dropWhile p (x :: xs) = (if p x then dropWhile p xs else x :: xs);

fun equal_lista A_ (a :: lista) [] = false
  | equal_lista A_ [] (a :: lista) = false
  | equal_lista A_ (aa :: listaa) (a :: lista) =
    HOL.eq A_ aa a andalso equal_lista A_ listaa lista
  | equal_lista A_ [] [] = true;

fun equal_list A_ = {equal = equal_lista A_} : ('a list) HOL.equal;

fun replicate n x =
  (if ((n : IntInf.int) = (0 : IntInf.int)) then []
    else x :: replicate (Arith.minus_nat n (1 : IntInf.int)) x);

fun size_list [] = (0 : IntInf.int)
  | size_list (a :: lista) =
    IntInf.+ (size_list lista, Arith.suc (0 : IntInf.int));

fun insort_key B_ f x [] = [x]
  | insort_key B_ f x (y :: ys) =
    (if Orderings.less_eq
          ((Orderings.ord_preorder o Orderings.preorder_order o
             Orderings.order_linorder)
            B_)
          (f x) (f y)
      then x :: y :: ys else y :: insort_key B_ f x ys);

fun insort_insert_key (B1_, B2_) f x xs =
  (if member B1_ (map f xs) (f x) then xs else insort_key B2_ f x xs);

end; (*struct List*)

structure Card_Univ : sig
  type 'a card_UNIV
  val card_UNIV : 'a card_UNIV -> 'a HOL.itself -> IntInf.int
  val card_UNIV_nata : IntInf.int HOL.itself -> IntInf.int
  val card_UNIV_nat : IntInf.int card_UNIV
  val is_list_UNIV : 'a card_UNIV * 'a HOL.equal -> 'a list -> bool
end = struct

type 'a card_UNIV = {card_UNIV : 'a HOL.itself -> IntInf.int};
val card_UNIV = #card_UNIV : 'a card_UNIV -> 'a HOL.itself -> IntInf.int;

fun card_UNIV_nata x = (fn _ => (0 : IntInf.int)) x;

val card_UNIV_nat = {card_UNIV = card_UNIV_nata} : IntInf.int card_UNIV;

fun is_list_UNIV (A1_, A2_) xs =
  let
    val c = card_UNIV A1_ HOL.Type;
  in
    (if ((c : IntInf.int) = (0 : IntInf.int)) then false
      else (((List.size_list (List.remdups A2_ xs)) : IntInf.int) = c))
  end;

end; (*struct Card_Univ*)

structure Predicate : sig
  datatype 'a seq = Empty | Insert of 'a * 'a pred | Join of 'a pred * 'a seq
  and 'a pred = Seq of (unit -> 'a seq)
  val is_empty : 'a pred -> bool
  val null : 'a seq -> bool
  val singleton : 'a HOL.equal -> (unit -> 'a) -> 'a pred -> 'a
  val the_only : 'a HOL.equal -> (unit -> 'a) -> 'a seq -> 'a
  val not_unique : 'a pred -> 'a
  val the : 'a HOL.equal -> 'a pred -> 'a
  val bind : 'a pred -> ('a -> 'b pred) -> 'b pred
  val apply : ('a -> 'b pred) -> 'a seq -> 'b seq
  val eval : 'a HOL.equal -> 'a pred -> 'a -> bool
  val member : 'a HOL.equal -> 'a seq -> 'a -> bool
  val holds : unit pred -> bool
  val bot_pred : 'a pred
  val single : 'a -> 'a pred
  val adjunct : 'a pred -> 'a seq -> 'a seq
  val sup_pred : 'a pred -> 'a pred -> 'a pred
  val if_pred : bool -> unit pred
  val not_pred : unit pred -> unit pred
end = struct

datatype 'a seq = Empty | Insert of 'a * 'a pred | Join of 'a pred * 'a seq
and 'a pred = Seq of (unit -> 'a seq);

fun is_empty (Seq f) = null (f ())
and null Empty = true
  | null (Insert (x, p)) = false
  | null (Join (p, xq)) = is_empty p andalso null xq;

fun singleton A_ dfault (Seq f) =
  (case f () of Empty => dfault ()
    | Insert (x, p) =>
      (if is_empty p then x
        else let
               val y = singleton A_ dfault p;
             in
               (if HOL.eq A_ x y then x else dfault ())
             end)
    | Join (p, xq) =>
      (if is_empty p then the_only A_ dfault xq
        else (if null xq then singleton A_ dfault p
               else let
                      val x = singleton A_ dfault p;
                      val y = the_only A_ dfault xq;
                    in
                      (if HOL.eq A_ x y then x else dfault ())
                    end)))
and the_only A_ dfault (Insert (x, p)) =
  (if is_empty p then x
    else let
           val y = singleton A_ dfault p;
         in
           (if HOL.eq A_ x y then x else dfault ())
         end)
  | the_only A_ dfault (Join (p, xq)) =
    (if is_empty p then the_only A_ dfault xq
      else (if null xq then singleton A_ dfault p
             else let
                    val x = singleton A_ dfault p;
                    val y = the_only A_ dfault xq;
                  in
                    (if HOL.eq A_ x y then x else dfault ())
                  end));

fun not_unique _ = raise Fail "not_unique";

fun the A_ a = singleton A_ (fn _ => not_unique a) a;

fun bind (Seq g) f = Seq (fn _ => apply f (g ()))
and apply f Empty = Empty
  | apply f (Insert (x, p)) = Join (f x, Join (bind p f, Empty))
  | apply f (Join (p, xq)) = Join (bind p f, apply f xq);

fun eval A_ (Seq f) = member A_ (f ())
and member A_ Empty x = false
  | member A_ (Insert (y, p)) x = HOL.eq A_ x y orelse eval A_ p x
  | member A_ (Join (p, xq)) x = eval A_ p x orelse member A_ xq x;

fun holds p = eval Product_Type.equal_unit p ();

val bot_pred : 'a pred = Seq (fn _ => Empty);

fun single x = Seq (fn _ => Insert (x, bot_pred));

fun adjunct p Empty = Join (p, Empty)
  | adjunct p (Insert (x, q)) = Insert (x, sup_pred q p)
  | adjunct p (Join (q, xq)) = Join (q, adjunct p xq)
and sup_pred (Seq f) (Seq g) =
  Seq (fn _ =>
        (case f () of Empty => g ()
          | Insert (x, p) => Insert (x, sup_pred p (Seq g))
          | Join (p, xq) => adjunct (Seq g) (Join (p, xq))));

fun if_pred b = (if b then single () else bot_pred);

fun not_pred p =
  (if eval Product_Type.equal_unit p () then bot_pred else single ());

end; (*struct Predicate*)

structure Stringa : sig
  datatype nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
    Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC |
    NibbleD | NibbleE | NibbleF
  val equal_char : char HOL.equal
  val equal_literal : string HOL.equal
end = struct

datatype nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
  Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC | NibbleD
  | NibbleE | NibbleF;

val equal_char = {equal = (fn a => fn b => ((a : char) = b))} : char HOL.equal;

val equal_literal = {equal = (fn a => fn b => ((a : string) = b))} :
  string HOL.equal;

end; (*struct Stringa*)

structure Type : sig
  datatype ty = Void | Boolean | Integer | Nt | Class of string | Array of ty
  val object : string
  val thread : string
  val equal_tya : ty -> ty -> bool
  val equal_ty : ty HOL.equal
  val is_refT_i : ty -> unit Predicate.pred
  val is_refT : ty -> bool
  val throwable : string
  val ground_type : ty -> ty
  val is_class_type_of_i_o : ty -> string Predicate.pred
end = struct

datatype ty = Void | Boolean | Integer | Nt | Class of string | Array of ty;

val object : string = "java/lang/Object";

val thread : string = "java/lang/Thread";

fun equal_tya (Array ty) (Class literal) = false
  | equal_tya (Class literal) (Array ty) = false
  | equal_tya (Array ty) Nt = false
  | equal_tya Nt (Array ty) = false
  | equal_tya (Class literal) Nt = false
  | equal_tya Nt (Class literal) = false
  | equal_tya (Array ty) Integer = false
  | equal_tya Integer (Array ty) = false
  | equal_tya (Class literal) Integer = false
  | equal_tya Integer (Class literal) = false
  | equal_tya Nt Integer = false
  | equal_tya Integer Nt = false
  | equal_tya (Array ty) Boolean = false
  | equal_tya Boolean (Array ty) = false
  | equal_tya (Class literal) Boolean = false
  | equal_tya Boolean (Class literal) = false
  | equal_tya Nt Boolean = false
  | equal_tya Boolean Nt = false
  | equal_tya Integer Boolean = false
  | equal_tya Boolean Integer = false
  | equal_tya (Array ty) Void = false
  | equal_tya Void (Array ty) = false
  | equal_tya (Class literal) Void = false
  | equal_tya Void (Class literal) = false
  | equal_tya Nt Void = false
  | equal_tya Void Nt = false
  | equal_tya Integer Void = false
  | equal_tya Void Integer = false
  | equal_tya Boolean Void = false
  | equal_tya Void Boolean = false
  | equal_tya (Array tya) (Array ty) = equal_tya tya ty
  | equal_tya (Class literala) (Class literal) = ((literala : string) = literal)
  | equal_tya Nt Nt = true
  | equal_tya Integer Integer = true
  | equal_tya Boolean Boolean = true
  | equal_tya Void Void = true;

val equal_ty = {equal = equal_tya} : ty HOL.equal;

fun is_refT_i x =
  Predicate.sup_pred
    (case x of Void => Predicate.bot_pred | Boolean => Predicate.bot_pred
      | Integer => Predicate.bot_pred | Nt => Predicate.single ()
      | Class _ => Predicate.bot_pred | Array _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case x of Void => Predicate.bot_pred | Boolean => Predicate.bot_pred
        | Integer => Predicate.bot_pred | Nt => Predicate.bot_pred
        | Class _ => Predicate.single () | Array _ => Predicate.bot_pred)
      (case x of Void => Predicate.bot_pred | Boolean => Predicate.bot_pred
        | Integer => Predicate.bot_pred | Nt => Predicate.bot_pred
        | Class _ => Predicate.bot_pred | Array _ => Predicate.single ()));

fun is_refT x1 = Predicate.holds (is_refT_i x1);

val throwable : string = "java/lang/Throwable";

fun ground_type (Array t) = ground_type t
  | ground_type Void = Void
  | ground_type Boolean = Boolean
  | ground_type Integer = Integer
  | ground_type Nt = Nt
  | ground_type (Class v) = Class v;

fun is_class_type_of_i_o x =
  (case x of Void => Predicate.bot_pred | Boolean => Predicate.bot_pred
    | Integer => Predicate.bot_pred | Nt => Predicate.bot_pred
    | Class a => Predicate.single a | Array _ => Predicate.single object);

end; (*struct Type*)

structure Numeral_Type : sig
  type 'a bit0
  type num1
end = struct

datatype 'a bit0 = EMPTY__

datatype num1 = EMPTY__

end; (*struct Numeral_Type*)

structure Type_Length : sig
  type 'a len0
  val len_of : 'a len0 -> 'a HOL.itself -> IntInf.int
  type 'a len
  val len0_len : 'a len -> 'a len0
  val len_of_bit0 : 'a len0 -> 'a Numeral_Type.bit0 HOL.itself -> IntInf.int
  val len0_bit0 : 'a len0 -> 'a Numeral_Type.bit0 len0
  val len_bit0 : 'a len -> 'a Numeral_Type.bit0 len
  val len_of_num1 : Numeral_Type.num1 HOL.itself -> IntInf.int
  val len0_num1 : Numeral_Type.num1 len0
  val len_num1 : Numeral_Type.num1 len
end = struct

type 'a len0 = {len_of : 'a HOL.itself -> IntInf.int};
val len_of = #len_of : 'a len0 -> 'a HOL.itself -> IntInf.int;

type 'a len = {len0_len : 'a len0};
val len0_len = #len0_len : 'a len -> 'a len0;

fun len_of_bit0 A_ x = IntInf.* ((2 : IntInf.int), len_of A_ HOL.Type);

fun len0_bit0 A_ = {len_of = len_of_bit0 A_} : 'a Numeral_Type.bit0 len0;

fun len_bit0 A_ = {len0_len = len0_bit0 (len0_len A_)} :
  'a Numeral_Type.bit0 len;

fun len_of_num1 x = (1 : IntInf.int);

val len0_num1 = {len_of = len_of_num1} : Numeral_Type.num1 len0;

val len_num1 = {len0_len = len0_num1} : Numeral_Type.num1 len;

end; (*struct Type_Length*)

structure FinFun : sig
  datatype ('a, 'b) finfun = Finfun_const of 'b |
    Finfun_update_code of ('a, 'b) finfun * 'a * 'b
  val code_abort : (unit -> 'a) -> 'a
  val finfun_All_except :
    'a Card_Univ.card_UNIV * 'a HOL.equal ->
      'a list -> ('a, bool) finfun -> bool
  val finfun_All :
    'a Card_Univ.card_UNIV * 'a HOL.equal -> ('a, bool) finfun -> bool
  val finfun_comp : ('a -> 'b) -> ('c, 'a) finfun -> ('c, 'b) finfun
  val finfun_update :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) finfun -> 'a -> 'b -> ('a, 'b) finfun
  val finfun_apply : 'a HOL.equal -> ('a, 'b) finfun -> 'a -> 'b
  val finfun_Diag :
    'a HOL.equal -> 'b HOL.equal -> 'c HOL.equal ->
      ('a, 'b) finfun -> ('a, 'c) finfun -> ('a, ('b * 'c)) finfun
  val equal_finfuna :
    'a Card_Univ.card_UNIV * 'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) finfun -> ('a, 'b) finfun -> bool
  val equal_finfun :
    'a Card_Univ.card_UNIV * 'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) finfun HOL.equal
  val finfun_default : 'a Card_Univ.card_UNIV -> ('a, 'b) finfun -> 'b
  val finfun_to_list :
    'a Card_Univ.card_UNIV * 'a HOL.equal * 'a Orderings.linorder ->
      'b HOL.equal -> ('a, 'b) finfun -> 'a list
end = struct

datatype ('a, 'b) finfun = Finfun_const of 'b |
  Finfun_update_code of ('a, 'b) finfun * 'a * 'b;

fun code_abort _ = raise Fail "code_abort";

fun finfun_All_except (A1_, A2_) aa (Finfun_update_code (f, a, b)) =
  (List.member A2_ aa a orelse b) andalso
    finfun_All_except (A1_, A2_) (a :: aa) f
  | finfun_All_except (A1_, A2_) a (Finfun_const b) =
    b orelse Card_Univ.is_list_UNIV (A1_, A2_) a;

fun finfun_All (A1_, A2_) = finfun_All_except (A1_, A2_) [];

fun finfun_comp g (Finfun_update_code (f, a, b)) =
  Finfun_update_code (finfun_comp g f, a, g b)
  | finfun_comp g (Finfun_const c) = Finfun_const (g c);

fun finfun_update A_ B_ (Finfun_update_code (f, aa, ba)) a b =
  (if HOL.eq A_ aa a then finfun_update A_ B_ f aa b
    else Finfun_update_code (finfun_update A_ B_ f a b, aa, ba))
  | finfun_update A_ B_ (Finfun_const ba) a b =
    (if HOL.eq B_ ba b then Finfun_const ba
      else Finfun_update_code (Finfun_const ba, a, b));

fun finfun_apply A_ (Finfun_update_code (f, aa, b)) a =
  (if HOL.eq A_ aa a then b else finfun_apply A_ f a)
  | finfun_apply A_ (Finfun_const b) a = b;

fun finfun_Diag A_ B_ C_ (Finfun_update_code (f, a, b)) g =
  finfun_update A_ (Product_Type.equal_prod B_ C_) (finfun_Diag A_ B_ C_ f g) a
    (b, finfun_apply A_ g a)
  | finfun_Diag A_ B_ C_ (Finfun_const b) (Finfun_update_code (g, a, c)) =
    Finfun_update_code (finfun_Diag A_ B_ C_ (Finfun_const b) g, a, (b, c))
  | finfun_Diag A_ B_ C_ (Finfun_const b) (Finfun_const c) =
    Finfun_const (b, c);

fun equal_finfuna (A1_, A2_) B_ f g =
  finfun_All (A1_, A2_)
    (finfun_comp (fn (a, b) => HOL.eq B_ a b) (finfun_Diag A2_ B_ B_ f g));

fun equal_finfun (A1_, A2_) B_ = {equal = equal_finfuna (A1_, A2_) B_} :
  ('a, 'b) finfun HOL.equal;

fun finfun_default A_ (Finfun_update_code (f, a, b)) = finfun_default A_ f
  | finfun_default A_ (Finfun_const c) =
    (if (((Card_Univ.card_UNIV A_ HOL.Type) : IntInf.int) = (0 : IntInf.int))
      then c else (raise Fail "undefined"));

fun finfun_to_list (A1_, A2_, A3_) B_ (Finfun_update_code (f, a, b)) =
  (if HOL.eq B_ b (finfun_default A1_ f)
    then List.remove1 A2_ a (finfun_to_list (A1_, A2_, A3_) B_ f)
    else List.insort_insert_key (A2_, A3_) (fn x => x) a
           (finfun_to_list (A1_, A2_, A3_) B_ f))
  | finfun_to_list (A1_, A2_, A3_) B_ (Finfun_const c) =
    (if (((Card_Univ.card_UNIV A1_ HOL.Type) : IntInf.int) = (0 : IntInf.int))
      then []
      else code_abort
             (fn _ => finfun_to_list (A1_, A2_, A3_) B_ (Finfun_const c)));

end; (*struct FinFun*)

structure Bit : sig
  datatype bit = Zero_bit | One_bit
end = struct

datatype bit = Zero_bit | One_bit;

end; (*struct Bit*)

structure Bit_Operations : sig
  val bitOR_bit : Bit.bit -> Bit.bit -> Bit.bit
  val bitAND_bit : Bit.bit -> Bit.bit -> Bit.bit
  val bitNOT_bit : Bit.bit -> Bit.bit
  val bitXOR_bit : Bit.bit -> Bit.bit -> Bit.bit
end = struct

fun bitOR_bit Bit.Zero_bit y = y
  | bitOR_bit Bit.One_bit y = Bit.One_bit;

fun bitAND_bit Bit.Zero_bit y = Bit.Zero_bit
  | bitAND_bit Bit.One_bit y = y;

fun bitNOT_bit Bit.Zero_bit = Bit.One_bit
  | bitNOT_bit Bit.One_bit = Bit.Zero_bit;

fun bitXOR_bit Bit.Zero_bit y = y
  | bitXOR_bit Bit.One_bit y = bitNOT_bit y;

end; (*struct Bit_Operations*)

structure Bit_Representation : sig
  val bitval : 'a Arith.zero_neq_one -> Bit.bit -> 'a
  val bit : IntInf.int -> Bit.bit -> IntInf.int
  val bin_last : IntInf.int -> Bit.bit
  val bin_rest : IntInf.int -> IntInf.int
  val sbintrunc : IntInf.int -> IntInf.int -> IntInf.int
end = struct

fun bitval A_ =
  (fn a =>
    (case a of Bit.Zero_bit => Arith.zero (Arith.zero_zero_neq_one A_)
      | Bit.One_bit => Arith.one (Arith.one_zero_neq_one A_)));

fun bit k b = IntInf.+ (IntInf.+ (bitval Arith.zero_neq_one_int b, k), k);

fun bin_last w =
  (if (((Arith.mod_int w (2 : IntInf.int)) : IntInf.int) = (0 : IntInf.int))
    then Bit.Zero_bit else Bit.One_bit);

fun bin_rest w = Arith.div_int w (2 : IntInf.int);

fun sbintrunc n bin =
  (if ((n : IntInf.int) = (0 : IntInf.int))
    then (case bin_last bin of Bit.Zero_bit => (0 : IntInf.int)
           | Bit.One_bit => IntInf.~ (1 : IntInf.int))
    else bit (sbintrunc (Arith.minus_nat n (1 : IntInf.int)) (bin_rest bin))
           (bin_last bin));

end; (*struct Bit_Representation*)

structure Bit_Int : sig
  val bin_rec :
    'a -> 'a -> (IntInf.int -> Bit.bit -> 'a -> 'a) -> IntInf.int -> 'a
  val bitOR_int : IntInf.int -> IntInf.int -> IntInf.int
  val bitAND_int : IntInf.int -> IntInf.int -> IntInf.int
  val bitNOT_int : IntInf.int -> IntInf.int
  val bitXOR_int : IntInf.int -> IntInf.int -> IntInf.int
end = struct

fun bin_rec f1 f2 f3 bin =
  (if ((bin : IntInf.int) = (0 : IntInf.int)) then f1
    else (if ((bin : IntInf.int) = (IntInf.~ (1 : IntInf.int))) then f2
           else f3 (Bit_Representation.bin_rest bin)
                  (Bit_Representation.bin_last bin)
                  (bin_rec f1 f2 f3 (Bit_Representation.bin_rest bin))));

fun bitOR_int x =
  bin_rec (fn xa => xa) (fn _ => IntInf.~ (1 : IntInf.int))
    (fn _ => fn b => fn s => fn y =>
      Bit_Representation.bit (s (Bit_Representation.bin_rest y))
        (Bit_Operations.bitOR_bit b (Bit_Representation.bin_last y)))
    x;

fun bitAND_int x =
  bin_rec (fn _ => (0 : IntInf.int)) (fn y => y)
    (fn _ => fn b => fn s => fn y =>
      Bit_Representation.bit (s (Bit_Representation.bin_rest y))
        (Bit_Operations.bitAND_bit b (Bit_Representation.bin_last y)))
    x;

fun bitNOT_int x =
  bin_rec (IntInf.~ (1 : IntInf.int)) (0 : IntInf.int)
    (fn _ => fn b => fn s =>
      Bit_Representation.bit s (Bit_Operations.bitNOT_bit b))
    x;

fun bitXOR_int x =
  bin_rec (fn xa => xa) bitNOT_int
    (fn _ => fn b => fn s => fn y =>
      Bit_Representation.bit (s (Bit_Representation.bin_rest y))
        (Bit_Operations.bitXOR_bit b (Bit_Representation.bin_last y)))
    x;

end; (*struct Bit_Int*)

structure Word : sig
  datatype 'a word = Word_of_int of IntInf.int
  val uint : 'a Type_Length.len0 -> 'a word -> IntInf.int
  val sint : 'a Type_Length.len -> 'a word -> IntInf.int
  val unat : 'a Type_Length.len0 -> 'a word -> IntInf.int
  val shiftl1 : 'a Type_Length.len0 -> 'a word -> 'a word
  val shiftr1 : 'a Type_Length.len0 -> 'a word -> 'a word
  val sshiftr1 : 'a Type_Length.len -> 'a word -> 'a word
  val sshiftr : 'a Type_Length.len -> 'a word -> IntInf.int -> 'a word
  val word_sle : 'a Type_Length.len -> 'a word -> 'a word -> bool
  val plus_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val zero_word : 'a Type_Length.len0 -> 'a word
  val bitOR_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val equal_word : 'a Type_Length.len0 -> 'a word -> 'a word -> bool
  val minus_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val times_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val word_sless : 'a Type_Length.len -> 'a word -> 'a word -> bool
  val bitAND_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val bitXOR_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val shiftl_word : 'a Type_Length.len0 -> 'a word -> IntInf.int -> 'a word
  val shiftr_word : 'a Type_Length.len0 -> 'a word -> IntInf.int -> 'a word
end = struct

datatype 'a word = Word_of_int of IntInf.int;

fun uint A_ (Word_of_int w) =
  Arith.mod_int w
    (Arith.power Arith.power_int (2 : IntInf.int)
      (Type_Length.len_of A_ HOL.Type));

fun sint A_ w =
  Bit_Representation.sbintrunc
    (Arith.minus_nat (Type_Length.len_of (Type_Length.len0_len A_) HOL.Type)
      (1 : IntInf.int))
    (uint (Type_Length.len0_len A_) w);

fun unat A_ w = IntInf.max (0, (uint A_ w));

fun shiftl1 A_ w =
  Word_of_int (Bit_Representation.bit (uint A_ w) Bit.Zero_bit);

fun shiftr1 A_ w = Word_of_int (Bit_Representation.bin_rest (uint A_ w));

fun sshiftr1 A_ w = Word_of_int (Bit_Representation.bin_rest (sint A_ w));

fun sshiftr A_ w n = Arith.funpow n (sshiftr1 A_) w;

fun word_sle A_ a b = IntInf.<= (sint A_ a, sint A_ b);

fun plus_word A_ a b = Word_of_int (IntInf.+ (uint A_ a, uint A_ b));

fun zero_word A_ = Word_of_int (0 : IntInf.int);

fun bitOR_word A_ a b = Word_of_int (Bit_Int.bitOR_int (uint A_ a) (uint A_ b));

fun equal_word A_ k l = (((uint A_ k) : IntInf.int) = (uint A_ l));

fun minus_word A_ a b = Word_of_int (IntInf.- (uint A_ a, uint A_ b));

fun times_word A_ a b = Word_of_int (IntInf.* (uint A_ a, uint A_ b));

fun word_sless A_ x y =
  word_sle A_ x y andalso not (equal_word (Type_Length.len0_len A_) x y);

fun bitAND_word A_ a b =
  Word_of_int (Bit_Int.bitAND_int (uint A_ a) (uint A_ b));

fun bitXOR_word A_ a b =
  Word_of_int (Bit_Int.bitXOR_int (uint A_ a) (uint A_ b));

fun shiftl_word A_ w n = Arith.funpow n (shiftl1 A_) w;

fun shiftr_word A_ w n = Arith.funpow n (shiftr1 A_) w;

end; (*struct Word*)

structure Value : sig
  datatype 'a vala = Unit | Null | Bool of bool |
    Intg of
      Numeral_Type.num1 Numeral_Type.bit0 Numeral_Type.bit0 Numeral_Type.bit0
        Numeral_Type.bit0
        Numeral_Type.bit0
        Word.word
    | Addr of 'a
  type 'a addr
  val hash_addr : 'a addr -> 'a -> IntInf.int
  val monitor_finfun_to_list :
    'a addr -> ('a, IntInf.int) FinFun.finfun -> 'a list
  val equal_vala : 'a HOL.equal -> 'a vala -> 'a vala -> bool
  val equal_val : 'a HOL.equal -> 'a vala HOL.equal
  val default_val : Type.ty -> 'a vala
end = struct

datatype 'a vala = Unit | Null | Bool of bool |
  Intg of
    Numeral_Type.num1 Numeral_Type.bit0 Numeral_Type.bit0 Numeral_Type.bit0
      Numeral_Type.bit0
      Numeral_Type.bit0
      Word.word
  | Addr of 'a;

type 'a addr =
  {hash_addr : 'a -> IntInf.int,
    monitor_finfun_to_list : ('a, IntInf.int) FinFun.finfun -> 'a list};
val hash_addr = #hash_addr : 'a addr -> 'a -> IntInf.int;
val monitor_finfun_to_list = #monitor_finfun_to_list :
  'a addr -> ('a, IntInf.int) FinFun.finfun -> 'a list;

fun equal_vala A_ (Addr addr) (Intg word) = false
  | equal_vala A_ (Intg word) (Addr addr) = false
  | equal_vala A_ (Addr addr) (Bool boola) = false
  | equal_vala A_ (Bool boola) (Addr addr) = false
  | equal_vala A_ (Intg word) (Bool boola) = false
  | equal_vala A_ (Bool boola) (Intg word) = false
  | equal_vala A_ (Addr addr) Null = false
  | equal_vala A_ Null (Addr addr) = false
  | equal_vala A_ (Intg word) Null = false
  | equal_vala A_ Null (Intg word) = false
  | equal_vala A_ (Bool boola) Null = false
  | equal_vala A_ Null (Bool boola) = false
  | equal_vala A_ (Addr addr) Unit = false
  | equal_vala A_ Unit (Addr addr) = false
  | equal_vala A_ (Intg word) Unit = false
  | equal_vala A_ Unit (Intg word) = false
  | equal_vala A_ (Bool boola) Unit = false
  | equal_vala A_ Unit (Bool boola) = false
  | equal_vala A_ Null Unit = false
  | equal_vala A_ Unit Null = false
  | equal_vala A_ (Addr addra) (Addr addr) = HOL.eq A_ addra addr
  | equal_vala A_ (Intg worda) (Intg word) =
    Word.equal_word
      (Type_Length.len0_bit0
        (Type_Length.len0_bit0
          (Type_Length.len0_bit0
            (Type_Length.len0_bit0
              (Type_Length.len0_bit0 Type_Length.len0_num1)))))
      worda word
  | equal_vala A_ (Bool boolaa) (Bool boola) =
    Product_Type.equal_bool boolaa boola
  | equal_vala A_ Null Null = true
  | equal_vala A_ Unit Unit = true;

fun equal_val A_ = {equal = equal_vala A_} : 'a vala HOL.equal;

fun default_val Type.Void = Unit
  | default_val Type.Boolean = Bool false
  | default_val Type.Integer =
    Intg (Word.zero_word
           (Type_Length.len0_bit0
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0 Type_Length.len0_num1))))))
  | default_val Type.Nt = Null
  | default_val (Type.Class c) = Null
  | default_val (Type.Array a) = Null;

end; (*struct Value*)

structure MM : sig
  val hash_addr_nat : IntInf.int -> IntInf.int
  val monitor_finfun_to_list_nat :
    (IntInf.int, IntInf.int) FinFun.finfun -> IntInf.int list
  val addr_nat : IntInf.int Value.addr
end = struct

fun hash_addr_nat x = x;

fun monitor_finfun_to_list_nat x =
  FinFun.finfun_to_list
    (Card_Univ.card_UNIV_nat, Arith.equal_nat, Arith.linorder_nat)
    Arith.equal_nat x;

val addr_nat =
  {hash_addr = hash_addr_nat,
    monitor_finfun_to_list = monitor_finfun_to_list_nat}
  : IntInf.int Value.addr;

end; (*struct MM*)

structure Aux : sig
  val distinct_fst : 'a HOL.equal -> ('a * 'b) list -> bool
end = struct

fun distinct_fst A_ = List.distinct A_ o List.map Product_Type.fst;

end; (*struct Aux*)

structure Err : sig
  datatype 'a err = Err | Ok of 'a
end = struct

datatype 'a err = Err | Ok of 'a;

end; (*struct Err*)

structure Option : sig
  val map : ('a -> 'b) -> 'a option -> 'b option
  val the : 'a option -> 'a
  val bind : 'a option -> ('a -> 'b option) -> 'b option
  val is_none : 'a option -> bool
  val equal_optiona : 'a HOL.equal -> 'a option -> 'a option -> bool
  val equal_option : 'a HOL.equal -> ('a option) HOL.equal
end = struct

fun map f (SOME x) = SOME (f x)
  | map f NONE = NONE;

fun the (SOME x) = x;

fun bind NONE f = NONE
  | bind (SOME x) f = f x;

fun is_none (SOME x) = false
  | is_none NONE = true;

fun equal_optiona A_ (SOME a) NONE = false
  | equal_optiona A_ NONE (SOME a) = false
  | equal_optiona A_ (SOME aa) (SOME a) = HOL.eq A_ aa a
  | equal_optiona A_ NONE NONE = true;

fun equal_option A_ = {equal = equal_optiona A_} : ('a option) HOL.equal;

end; (*struct Option*)

structure Map : sig
  val map_of : 'a HOL.equal -> ('a * 'b) list -> 'a -> 'b option
  val map_add : ('a -> 'b option) -> ('a -> 'b option) -> 'a -> 'b option
  val map_upds :
    'a HOL.equal -> ('a -> 'b option) -> 'a list -> 'b list -> 'a -> 'b option
end = struct

fun map_of A_ ((l, v) :: ps) k =
  (if HOL.eq A_ l k then SOME v else map_of A_ ps k)
  | map_of A_ [] k = NONE;

fun map_add m1 m2 = (fn x => (case m2 x of NONE => m1 x | SOME a => SOME a));

fun map_upds A_ m xs ys = map_add m (map_of A_ (List.rev (List.zip xs ys)));

end; (*struct Map*)

structure RBT_Impl : sig
  datatype color = R | B
  datatype ('a, 'b) rbt = Empty |
    Branch of color * ('a, 'b) rbt * 'a * 'b * ('a, 'b) rbt
  val balance : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val paint : color -> ('a, 'b) rbt -> ('a, 'b) rbt
  val balance_left : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val balance_right : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val combine : ('a, 'b) rbt -> ('a, 'b) rbt -> ('a, 'b) rbt
  val del : 'a Orderings.linorder -> 'a -> ('a, 'b) rbt -> ('a, 'b) rbt
  val del_from_left :
    'a Orderings.linorder ->
      'a -> ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val del_from_right :
    'a Orderings.linorder ->
      'a -> ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val ins :
    'a Orderings.linorder ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val delete : 'a Orderings.linorder -> 'a -> ('a, 'b) rbt -> ('a, 'b) rbt
  val insert_with_key :
    'a Orderings.linorder ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val insert : 'a Orderings.linorder -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val lookup : 'a Orderings.linorder -> ('a, 'b) rbt -> 'a -> 'b option
end = struct

datatype color = R | B;

datatype ('a, 'b) rbt = Empty |
  Branch of color * ('a, 'b) rbt * 'a * 'b * ('a, 'b) rbt;

fun balance (Branch (R, a, w, x, b)) s t (Branch (R, c, y, z, d)) =
  Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z Empty =
    Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, a, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z Empty =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance
    (Branch (R, Branch (B, va, vb, vc, vd), w, x, Branch (R, b, s, t, c))) y z
    Empty =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance
    (Branch (R, Branch (B, ve, vf, vg, vh), w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Branch (B, ve, vf, vg, vh), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance Empty w x (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, d))
  | balance Empty w x (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance Empty w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, va, vb, vc, vd))) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, ve, vf, vg, vh))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, ve, vf, vg, vh)))
  | balance Empty s t Empty = Branch (B, Empty, s, t, Empty)
  | balance Empty s t (Branch (B, va, vb, vc, vd)) =
    Branch (B, Empty, s, t, Branch (B, va, vb, vc, vd))
  | balance Empty s t (Branch (v, Empty, vb, vc, Empty)) =
    Branch (B, Empty, s, t, Branch (v, Empty, vb, vc, Empty))
  | balance Empty s t (Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty)) =
    Branch
      (B, Empty, s, t, Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty))
  | balance Empty s t (Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi))) =
    Branch
      (B, Empty, s, t, Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance Empty s t
    (Branch (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
    = Branch
        (B, Empty, s, t,
          Branch
            (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance (Branch (B, va, vb, vc, vd)) s t Empty =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Empty)
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (B, ve, vf, vg, vh)) =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Branch (B, ve, vf, vg, vh))
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (v, Empty, vf, vg, Empty))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t, Branch (v, Empty, vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty)) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm))) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t,
          Branch
            (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (v, Empty, vb, vc, Empty)) s t Empty =
    Branch (B, Branch (v, Empty, vb, vc, Empty), s, t, Empty)
  | balance (Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh))) s t Empty =
    Branch
      (B, Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh)), s, t, Empty)
  | balance (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty)) s t Empty =
    Branch
      (B, Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty), s, t, Empty)
  | balance
    (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)))
    s t Empty =
    Branch
      (B, Branch
            (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)),
        s, t, Empty)
  | balance (Branch (v, Empty, vf, vg, Empty)) s t (Branch (B, va, vb, vc, vd))
    = Branch
        (B, Branch (v, Empty, vf, vg, Empty), s, t, Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl))) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl)), s, t,
        Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty)) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty), s, t,
        Branch (B, va, vb, vc, vd))
  | balance
    (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)))
    s t (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch
            (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)),
        s, t, Branch (B, va, vb, vc, vd));

fun paint c Empty = Empty
  | paint c (Branch (uu, l, k, v, r)) = Branch (c, l, k, v, r);

fun balance_left (Branch (R, a, k, x, b)) s y c =
  Branch (R, Branch (B, a, k, x, b), s, y, c)
  | balance_left Empty k x (Branch (B, a, s, y, b)) =
    balance Empty k x (Branch (R, a, s, y, b))
  | balance_left (Branch (B, va, vb, vc, vd)) k x (Branch (B, a, s, y, b)) =
    balance (Branch (B, va, vb, vc, vd)) k x (Branch (R, a, s, y, b))
  | balance_left Empty k x (Branch (R, Branch (B, a, s, y, b), t, z, c)) =
    Branch (R, Branch (B, Empty, k, x, a), s, y, balance b t z (paint R c))
  | balance_left (Branch (B, va, vb, vc, vd)) k x
    (Branch (R, Branch (B, a, s, y, b), t, z, c)) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), k, x, a), s, y,
        balance b t z (paint R c))
  | balance_left Empty k x Empty = Empty
  | balance_left Empty k x (Branch (R, Empty, vb, vc, vd)) = Empty
  | balance_left Empty k x (Branch (R, Branch (R, ve, vf, vg, vh), vb, vc, vd))
    = Empty
  | balance_left (Branch (B, va, vb, vc, vd)) k x Empty = Empty
  | balance_left (Branch (B, va, vb, vc, vd)) k x
    (Branch (R, Empty, vf, vg, vh)) = Empty
  | balance_left (Branch (B, va, vb, vc, vd)) k x
    (Branch (R, Branch (R, vi, vj, vk, vl), vf, vg, vh)) = Empty;

fun balance_right a k x (Branch (R, b, s, y, c)) =
  Branch (R, a, k, x, Branch (B, b, s, y, c))
  | balance_right (Branch (B, a, k, x, b)) s y Empty =
    balance (Branch (R, a, k, x, b)) s y Empty
  | balance_right (Branch (B, a, k, x, b)) s y (Branch (B, va, vb, vc, vd)) =
    balance (Branch (R, a, k, x, b)) s y (Branch (B, va, vb, vc, vd))
  | balance_right (Branch (R, a, k, x, Branch (B, b, s, y, c))) t z Empty =
    Branch (R, balance (paint R a) k x b, s, y, Branch (B, c, t, z, Empty))
  | balance_right (Branch (R, a, k, x, Branch (B, b, s, y, c))) t z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, balance (paint R a) k x b, s, y,
        Branch (B, c, t, z, Branch (B, va, vb, vc, vd)))
  | balance_right Empty k x Empty = Empty
  | balance_right (Branch (R, va, vb, vc, Empty)) k x Empty = Empty
  | balance_right (Branch (R, va, vb, vc, Branch (R, ve, vf, vg, vh))) k x Empty
    = Empty
  | balance_right Empty k x (Branch (B, va, vb, vc, vd)) = Empty
  | balance_right (Branch (R, ve, vf, vg, Empty)) k x
    (Branch (B, va, vb, vc, vd)) = Empty
  | balance_right (Branch (R, ve, vf, vg, Branch (R, vi, vj, vk, vl))) k x
    (Branch (B, va, vb, vc, vd)) = Empty;

fun combine Empty x = x
  | combine (Branch (v, va, vb, vc, vd)) Empty = Branch (v, va, vb, vc, vd)
  | combine (Branch (R, a, k, x, b)) (Branch (R, c, s, y, d)) =
    (case combine b c
      of Empty => Branch (R, a, k, x, Branch (R, Empty, s, y, d))
      | Branch (R, b2, t, z, c2) =>
        Branch (R, Branch (R, a, k, x, b2), t, z, Branch (R, c2, s, y, d))
      | Branch (B, b2, t, z, c2) =>
        Branch (R, a, k, x, Branch (R, Branch (B, b2, t, z, c2), s, y, d)))
  | combine (Branch (B, a, k, x, b)) (Branch (B, c, s, y, d)) =
    (case combine b c
      of Empty => balance_left a k x (Branch (B, Empty, s, y, d))
      | Branch (R, b2, t, z, c2) =>
        Branch (R, Branch (B, a, k, x, b2), t, z, Branch (B, c2, s, y, d))
      | Branch (B, b2, t, z, c2) =>
        balance_left a k x (Branch (B, Branch (B, b2, t, z, c2), s, y, d)))
  | combine (Branch (B, va, vb, vc, vd)) (Branch (R, b, k, x, c)) =
    Branch (R, combine (Branch (B, va, vb, vc, vd)) b, k, x, c)
  | combine (Branch (R, a, k, x, b)) (Branch (B, va, vb, vc, vd)) =
    Branch (R, a, k, x, combine b (Branch (B, va, vb, vc, vd)));

fun del A_ x Empty = Empty
  | del A_ x (Branch (c, a, y, s, b)) =
    (if Orderings.less
          ((Orderings.ord_preorder o Orderings.preorder_order o
             Orderings.order_linorder)
            A_)
          x y
      then del_from_left A_ x a y s b
      else (if Orderings.less
                 ((Orderings.ord_preorder o Orderings.preorder_order o
                    Orderings.order_linorder)
                   A_)
                 y x
             then del_from_right A_ x a y s b else combine a b))
and del_from_left A_ x (Branch (B, lt, z, v, rt)) y s b =
  balance_left (del A_ x (Branch (B, lt, z, v, rt))) y s b
  | del_from_left A_ x Empty y s b = Branch (R, del A_ x Empty, y, s, b)
  | del_from_left A_ x (Branch (R, va, vb, vc, vd)) y s b =
    Branch (R, del A_ x (Branch (R, va, vb, vc, vd)), y, s, b)
and del_from_right A_ x a y s (Branch (B, lt, z, v, rt)) =
  balance_right a y s (del A_ x (Branch (B, lt, z, v, rt)))
  | del_from_right A_ x a y s Empty = Branch (R, a, y, s, del A_ x Empty)
  | del_from_right A_ x a y s (Branch (R, va, vb, vc, vd)) =
    Branch (R, a, y, s, del A_ x (Branch (R, va, vb, vc, vd)));

fun ins A_ f k v Empty = Branch (R, Empty, k, v, Empty)
  | ins A_ f k v (Branch (B, l, x, y, r)) =
    (if Orderings.less
          ((Orderings.ord_preorder o Orderings.preorder_order o
             Orderings.order_linorder)
            A_)
          k x
      then balance (ins A_ f k v l) x y r
      else (if Orderings.less
                 ((Orderings.ord_preorder o Orderings.preorder_order o
                    Orderings.order_linorder)
                   A_)
                 x k
             then balance l x y (ins A_ f k v r)
             else Branch (B, l, x, f k y v, r)))
  | ins A_ f k v (Branch (R, l, x, y, r)) =
    (if Orderings.less
          ((Orderings.ord_preorder o Orderings.preorder_order o
             Orderings.order_linorder)
            A_)
          k x
      then Branch (R, ins A_ f k v l, x, y, r)
      else (if Orderings.less
                 ((Orderings.ord_preorder o Orderings.preorder_order o
                    Orderings.order_linorder)
                   A_)
                 x k
             then Branch (R, l, x, y, ins A_ f k v r)
             else Branch (R, l, x, f k y v, r)));

fun delete A_ k t = paint B (del A_ k t);

fun insert_with_key A_ f k v t = paint B (ins A_ f k v t);

fun insert A_ = insert_with_key A_ (fn _ => fn _ => fn nv => nv);

fun lookup A_ Empty k = NONE
  | lookup A_ (Branch (uu, l, x, y, r)) k =
    (if Orderings.less
          ((Orderings.ord_preorder o Orderings.preorder_order o
             Orderings.order_linorder)
            A_)
          k x
      then lookup A_ l k
      else (if Orderings.less
                 ((Orderings.ord_preorder o Orderings.preorder_order o
                    Orderings.order_linorder)
                   A_)
                 x k
             then lookup A_ r k else SOME y));

end; (*struct RBT_Impl*)

structure RBT : sig
  datatype ('a, 'b) rbt = Rbt of ('a, 'b) RBT_Impl.rbt
  val empty : 'a Orderings.linorder -> ('a, 'b) rbt
  val impl_of : 'a Orderings.linorder -> ('a, 'b) rbt -> ('a, 'b) RBT_Impl.rbt
  val delete : 'a Orderings.linorder -> 'a -> ('a, 'b) rbt -> ('a, 'b) rbt
  val insert : 'a Orderings.linorder -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val lookup : 'a Orderings.linorder -> ('a, 'b) rbt -> 'a -> 'b option
end = struct

datatype ('a, 'b) rbt = Rbt of ('a, 'b) RBT_Impl.rbt;

fun empty A_ = Rbt RBT_Impl.Empty;

fun impl_of A_ (Rbt x) = x;

fun delete A_ k t = Rbt (RBT_Impl.delete A_ k (impl_of A_ t));

fun insert A_ k v t = Rbt (RBT_Impl.insert A_ k v (impl_of A_ t));

fun lookup A_ t = RBT_Impl.lookup A_ (impl_of A_ t);

end; (*struct RBT*)

structure Transitive_Closure_Table : sig
  val rtrancl_tab_FioB_i_i_i :
    'a HOL.equal ->
      ('a -> 'a Predicate.pred) -> 'a list -> 'a -> 'a -> unit Predicate.pred
  val rtranclp_FioB_i_i :
    'a HOL.equal -> ('a -> 'a Predicate.pred) -> 'a -> 'a -> unit Predicate.pred
end = struct

fun rtrancl_tab_FioB_i_i_i A_ r xa xb xc =
  Predicate.sup_pred
    (if HOL.eq A_ xb xc then Predicate.single () else Predicate.bot_pred)
    (Predicate.bind (Predicate.if_pred (not (List.member A_ xa xb)))
      (fn () =>
        Predicate.bind (r xb)
          (fn xaa =>
            Predicate.bind (rtrancl_tab_FioB_i_i_i A_ r (xb :: xa) xaa xc)
              (fn () => Predicate.single ()))));

fun rtranclp_FioB_i_i A_ r xa xb =
  Predicate.bind (rtrancl_tab_FioB_i_i_i A_ r [] xa xb)
    (fn () => Predicate.single ());

end; (*struct Transitive_Closure_Table*)

structure TypeRel : sig
  datatype 'a fmod_ext = Fmod_ext of bool * 'a
  datatype 'a prog =
    Program of
      (string *
        (string *
          ((string * (Type.ty * unit fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list))) list
  val class :
    'a prog ->
      string ->
        (string *
          ((string * (Type.ty * unit fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list)) option
  val eq_i_i : 'a HOL.equal -> 'a -> 'a -> unit Predicate.pred
  val classes :
    'a prog ->
      (string *
        (string *
          ((string * (Type.ty * unit fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list))) list
  val is_class : 'a prog -> string -> bool
  val is_type : 'a prog -> Type.ty -> bool
  val subcls1_i_i_o : 'a prog -> string -> string Predicate.pred
  val widen_i_i_ia : 'a prog -> Type.ty -> Type.ty -> unit Predicate.pred
  val widen : 'a prog -> Type.ty -> Type.ty -> bool
  val equal_fmod_exta : 'a HOL.equal -> 'a fmod_ext -> 'a fmod_ext -> bool
  val equal_fmod_ext : 'a HOL.equal -> 'a fmod_ext HOL.equal
  val fields_i_i_o :
    'a prog ->
      string ->
        (((string * string) * (Type.ty * unit fmod_ext)) list) Predicate.pred
  val fields :
    'a prog -> string -> ((string * string) * (Type.ty * unit fmod_ext)) list
  val methods_i_i_o :
    'a prog ->
      string ->
        (string -> ((Type.ty list * (Type.ty * 'a option)) * string) option)
          Predicate.pred
  val method_i_i_i_o_o_o_o :
    'a prog ->
      string ->
        string ->
          (Type.ty list * (Type.ty * ('a option * string))) Predicate.pred
  val method :
    'a HOL.equal ->
      'a prog ->
        string -> string -> string * (Type.ty list * (Type.ty * 'a option))
  val subcls_i_i_i : 'a prog -> string -> string -> unit Predicate.pred
  val subcls : 'a prog -> string -> string -> bool
  val is_class_i_i : 'a prog -> string -> unit Predicate.pred
  val widen_i_i_i : 'a prog -> Type.ty -> Type.ty -> unit Predicate.pred
  val mapP_FioB_i_o :
    ('a -> 'b Predicate.pred) -> 'a list -> ('b list) Predicate.pred
  val sees_field_hoaux_PPiiii_Poooo :
    (string * string) * (Type.ty * unit fmod_ext) ->
      (string * (string * (Type.ty * unit fmod_ext))) Predicate.pred
  val sees_field_i_i_i_o_o_i :
    'a prog ->
      string -> string -> string -> (Type.ty * unit fmod_ext) Predicate.pred
end = struct

datatype 'a fmod_ext = Fmod_ext of bool * 'a;

datatype 'a prog =
  Program of
    (string *
      (string *
        ((string * (Type.ty * unit fmod_ext)) list *
          (string * (Type.ty list * (Type.ty * 'a option))) list))) list;

fun class (Program p) = Map.map_of Stringa.equal_literal p;

fun eq_i_i A_ xa xb =
  (if HOL.eq A_ xa xb then Predicate.single () else Predicate.bot_pred);

fun classes (Program p) = p;

fun is_class p c = not (Option.is_none (class p c));

fun is_type p Type.Void = true
  | is_type p Type.Boolean = true
  | is_type p Type.Integer = true
  | is_type p Type.Nt = true
  | is_type p (Type.Class c) = is_class p c
  | is_type p (Type.Array a) =
    (case Type.ground_type a of Type.Void => true | Type.Boolean => true
      | Type.Integer => true | Type.Nt => false | Type.Class aa => is_class p aa
      | Type.Array _ => true);

fun subcls1_i_i_o x xa =
  Predicate.bind (Predicate.if_pred (not ((xa : string) = Type.object)))
    (fn () =>
      (case class x xa of NONE => Predicate.bot_pred
        | SOME (d, _) => Predicate.single d));

fun widen_i_i_ia x xa xb =
  Predicate.sup_pred
    (if Type.equal_tya xa xb then Predicate.single () else Predicate.bot_pred)
    (Predicate.sup_pred
      (case xa of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class c =>
          (case xb of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class d =>
              Predicate.bind
                (Transitive_Closure_Table.rtranclp_FioB_i_i
                  Stringa.equal_literal (subcls1_i_i_o x) c d)
                (fn () => Predicate.single ())
            | Type.Array _ => Predicate.bot_pred)
        | Type.Array _ => Predicate.bot_pred)
      (Predicate.sup_pred
        (case xa of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred
          | Type.Nt =>
            (case xb of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.single ()
              | Type.Array _ => Predicate.bot_pred)
          | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case xa of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred
            | Type.Nt =>
              (case xb of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.single ())
            | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case xa of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ =>
                (case xb of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class c =>
                    Predicate.bind (eq_i_i Stringa.equal_literal c Type.object)
                      (fn () => Predicate.single ())
                  | Type.Array _ => Predicate.bot_pred))
            (case xa of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array a =>
                (case xb of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array b =>
                    Predicate.bind (widen_i_i_ia x a b)
                      (fn () => Predicate.single ())))))));

fun widen x1 x2 x3 = Predicate.holds (widen_i_i_ia x1 x2 x3);

fun equal_fmod_exta A_ (Fmod_ext (volatilea, morea)) (Fmod_ext (volatile, more))
  = Product_Type.equal_bool volatilea volatile andalso HOL.eq A_ morea more;

fun equal_fmod_ext A_ = {equal = equal_fmod_exta A_} : 'a fmod_ext HOL.equal;

fun fields_i_i_o x xa =
  Predicate.sup_pred
    (Predicate.bind (Predicate.if_pred (not ((xa : string) = Type.object)))
      (fn () =>
        (case class x xa of NONE => Predicate.bot_pred
          | SOME (d, (fs, _)) =>
            Predicate.bind (fields_i_i_o x d)
              (fn xb =>
                Predicate.single
                  (List.map (fn (f, a) => ((f, xa), a)) fs @ xb)))))
    (case class x Type.object of NONE => Predicate.bot_pred
      | SOME (_, (fs, _)) =>
        (if ((xa : string) = Type.object)
          then Predicate.single
                 (List.map (fn (f, a) => ((f, Type.object), a)) fs)
          else Predicate.bot_pred));

fun fields p c =
  Predicate.the
    (List.equal_list
      (Product_Type.equal_prod
        (Product_Type.equal_prod Stringa.equal_literal Stringa.equal_literal)
        (Product_Type.equal_prod Type.equal_ty
          (equal_fmod_ext Product_Type.equal_unit))))
    (fields_i_i_o p c);

fun methods_i_i_o x xa =
  Predicate.sup_pred
    (case class x Type.object of NONE => Predicate.bot_pred
      | SOME (_, (_, ms)) =>
        (if ((xa : string) = Type.object)
          then Predicate.single
                 (Option.map (fn m => (m, Type.object)) o
                   Map.map_of Stringa.equal_literal ms)
          else Predicate.bot_pred))
    (Predicate.bind (Predicate.if_pred (not ((xa : string) = Type.object)))
      (fn () =>
        (case class x xa of NONE => Predicate.bot_pred
          | SOME (d, (_, ms)) =>
            Predicate.bind (methods_i_i_o x d)
              (fn xb =>
                Predicate.single
                  (Map.map_add xb
                    (Option.map (fn m => (m, xa)) o
                      Map.map_of Stringa.equal_literal ms))))));

fun method_i_i_i_o_o_o_o x xa xb =
  Predicate.bind (methods_i_i_o x xa)
    (fn xc =>
      (case xc xb of NONE => Predicate.bot_pred
        | SOME ((tv, (tc, p)), dc) => Predicate.single (tv, (tc, (p, dc)))));

fun method A_ p c m =
  Predicate.the
    (Product_Type.equal_prod Stringa.equal_literal
      (Product_Type.equal_prod (List.equal_list Type.equal_ty)
        (Product_Type.equal_prod Type.equal_ty (Option.equal_option A_))))
    (Predicate.bind (method_i_i_i_o_o_o_o p c m)
      (fn (ts, (t, (ma, d))) => Predicate.single (d, (ts, (t, ma)))));

fun subcls_i_i_i x xa xb =
  Predicate.bind
    (Transitive_Closure_Table.rtranclp_FioB_i_i Stringa.equal_literal
      (subcls1_i_i_o x) xa xb)
    (fn () => Predicate.single ());

fun subcls x1 x2 x3 = Predicate.holds (subcls_i_i_i x1 x2 x3);

fun is_class_i_i x xa =
  Predicate.bind (Predicate.if_pred (not (Option.is_none (class x xa))))
    (fn () => Predicate.single ());

fun widen_i_i_i p ta t =
  (if widen p ta t then Predicate.single () else Predicate.bot_pred);

fun mapP_FioB_i_o h x =
  Predicate.sup_pred
    (case x of [] => Predicate.single [] | _ :: _ => Predicate.bot_pred)
    (case x of [] => Predicate.bot_pred
      | z :: xu =>
        Predicate.bind (mapP_FioB_i_o h xu)
          (fn xa =>
            Predicate.bind (h z) (fn xaa => Predicate.single (xaa :: xa))));

fun sees_field_hoaux_PPiiii_Poooo ((x, xf), (xg, xh)) =
  Predicate.single (x, (xf, (xg, xh)));

fun sees_field_i_i_i_o_o_i x xa xb xc =
  Predicate.bind (fields_i_i_o x xa)
    (fn xd =>
      Predicate.bind (mapP_FioB_i_o sees_field_hoaux_PPiiii_Poooo xd)
        (fn xe =>
          (case Map.map_of Stringa.equal_literal xe xb
            of NONE => Predicate.bot_pred
            | SOME (dca, (tc, fp)) =>
              (if ((xc : string) = dca) then Predicate.single (tc, fp)
                else Predicate.bot_pred))));

end; (*struct TypeRel*)

structure Exceptions : sig
  val sys_xcpts_list : string list
  val wf_syscls : 'a TypeRel.prog -> bool
end = struct

val sys_xcpts_list : string list =
  ["java/lang/NullPointerException", "java/lang/ClassCastException",
    "java/lang/OutOfMemoryError", "java/lang/ArrayIndexOutOfBoundsException",
    "java/lang/ArrayStoreException", "java/lang/NegativeArraySizeException",
    "java/lang/ArithmeticException", "java/lang/IllegalMonitorStateException",
    "java/lang/IllegalThreadStateException", "java/lang/InterruptedException"];

fun wf_syscls p =
  List.list_all (TypeRel.is_class p)
    [Type.object, Type.throwable, Type.thread] andalso
    List.list_all (fn c => TypeRel.subcls p c Type.throwable) sys_xcpts_list;

end; (*struct Exceptions*)

structure StartConfig : sig
  val initialization_list : string list
  val create_initial_object :
    'b Value.addr ->
      ('a -> string -> 'a * 'b option) ->
        'a * ('b list * bool) -> string -> 'a * ('b list * bool)
  val start_heap_data :
    'b Value.addr ->
      'a -> ('a -> string -> 'a * 'b option) -> 'a * ('b list * bool)
  val start_addrs :
    'b Value.addr -> 'a -> ('a -> string -> 'a * 'b option) -> 'b list
  val addr_of_sys_xcpt :
    'b Value.addr -> 'a -> ('a -> string -> 'a * 'b option) -> string -> 'b
  val start_tid :
    'a Value.addr -> ('a -> 'b) -> 'c -> ('c -> string -> 'c * 'a option) -> 'b
  val start_heap : 'b Value.addr -> 'a -> ('a -> string -> 'a * 'b option) -> 'a
end = struct

val initialization_list : string list =
  Type.thread :: Exceptions.sys_xcpts_list;

fun create_initial_object B_ new_obj =
  (fn (h, (ads, b)) => fn c =>
    (if b then (case new_obj h c of (ha, NONE) => (ha, (ads, false))
                 | (ha, SOME a) => (ha, (ads @ [a], true)))
      else (h, (ads, false))));

fun start_heap_data B_ empty_heap new_obj =
  List.foldl (create_initial_object B_ new_obj) (empty_heap, ([], true))
    initialization_list;

fun start_addrs B_ empty_heap new_obj =
  Product_Type.fst (Product_Type.snd (start_heap_data B_ empty_heap new_obj));

fun addr_of_sys_xcpt B_ empty_heap new_obj c =
  Option.the
    (Map.map_of Stringa.equal_literal
      (List.zip initialization_list (start_addrs B_ empty_heap new_obj)) c);

fun start_tid A_ addr2thread_id empty_heap new_obj =
  addr2thread_id (addr_of_sys_xcpt A_ empty_heap new_obj Type.thread);

fun start_heap B_ empty_heap new_obj =
  Product_Type.fst (start_heap_data B_ empty_heap new_obj);

end; (*struct StartConfig*)

structure FWState : sig
  val shr : 'a * (('b * 'c) * ('d * 'e)) -> 'c
  val thr : 'a * (('b * 'c) * ('d * 'e)) -> 'b
  datatype wake_up_status = WSNotified | WSWokenUp
  datatype 'a wait_set_status = InWS of 'a | PostWS of wake_up_status
  datatype 'a conditional_action = Join of 'a | Yield
  datatype lock_action = Lock | Unlock | UnlockFail | ReleaseAcquire
  val wset : 'a * (('b * 'c) * ('d * 'e)) -> 'd
  val locks : 'a * (('b * 'c) * ('d * 'e)) -> 'a
  datatype ('a, 'b, 'c) new_thread_action = NewThread of 'a * 'b * 'c |
    ThreadExists of 'a * bool
  datatype ('a, 'b) wait_set_action = Suspend of 'b | Notify of 'b |
    NotifyAll of 'b | WakeUp of 'a | Notified | WokenUp
  datatype 'a interrupt_action = IsInterrupted of 'a * bool | Interrupt of 'a |
    ClearInterrupt of 'a
  val obs_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'f list
  val thr_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('b, 'c, 'd) new_thread_action list
  val cond_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'b conditional_action list
  val wset_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('b, 'e) wait_set_action list
  val locks_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('a, (lock_action list)) FinFun.finfun
  datatype ('a, 'b, 'c, 'd, 'e, 'f) thread_action =
    LockAction of (lock_action * 'a) |
    NewThreadAction of ('b, 'c, 'd) new_thread_action |
    ConditionalAction of 'b conditional_action |
    WaitSetAction of ('b, 'e) wait_set_action |
    InterruptAction of 'b interrupt_action | ObsAction of 'f
  val interrupts : 'a * (('b * 'c) * ('d * 'e)) -> 'e
  type 'a obs_action
  val interrupt_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'b interrupt_action list
  val convert_new_thread_action :
    ('a -> 'b) ->
      ('c, 'a, 'd) new_thread_action -> ('c, 'b, 'd) new_thread_action
  val convert_extTA :
    ('a -> 'b) ->
      ('c, (lock_action list)) FinFun.finfun *
        (('d, 'a, 'e) new_thread_action list *
          ('d conditional_action list *
            (('d, 'f) wait_set_action list *
              ('d interrupt_action list * 'g list)))) ->
        ('c, (lock_action list)) FinFun.finfun *
          (('d, 'b, 'e) new_thread_action list *
            ('d conditional_action list *
              (('d, 'f) wait_set_action list *
                ('d interrupt_action list * 'g list))))
  val ta_update_obs :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'f -> ('a, (lock_action list)) FinFun.finfun *
              (('b, 'c, 'd) new_thread_action list *
                ('b conditional_action list *
                  (('b, 'e) wait_set_action list *
                    ('b interrupt_action list * 'f list))))
  val equal_lock_actiona : lock_action -> lock_action -> bool
  val equal_lock_action : lock_action HOL.equal
  val ta_update_locks :
    'a HOL.equal ->
      ('a, (lock_action list)) FinFun.finfun *
        (('b, 'c, 'd) new_thread_action list *
          ('b conditional_action list *
            (('b, 'e) wait_set_action list *
              ('b interrupt_action list * 'f list)))) ->
        lock_action ->
          'a -> ('a, (lock_action list)) FinFun.finfun *
                  (('b, 'c, 'd) new_thread_action list *
                    ('b conditional_action list *
                      (('b, 'e) wait_set_action list *
                        ('b interrupt_action list * 'f list))))
  val ta_update_wait_set :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('b, 'e) wait_set_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list))))
  val ta_update_NewThread :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('b, 'c, 'd) new_thread_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list))))
  val ta_update_interrupt :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'b interrupt_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list))))
  val equal_wait_set_actiona :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) wait_set_action -> ('a, 'b) wait_set_action -> bool
  val equal_wait_set_action :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) wait_set_action HOL.equal
  val equal_wake_up_status : wake_up_status -> wake_up_status -> bool
  val equal_wait_set_statusa :
    'a HOL.equal -> 'a wait_set_status -> 'a wait_set_status -> bool
  val equal_wait_set_status : 'a HOL.equal -> 'a wait_set_status HOL.equal
  val ta_update_Conditional :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'b conditional_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list))))
  val equal_conditional_actiona :
    'a HOL.equal -> 'a conditional_action -> 'a conditional_action -> bool
  val equal_conditional_action : 'a HOL.equal -> 'a conditional_action HOL.equal
  val thread_action_to_thread_action :
    'a HOL.equal -> 'f obs_action ->
      ('a, 'b, 'c, 'd, 'e, 'f) thread_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list)))) ->
          ('a, (lock_action list)) FinFun.finfun *
            (('b, 'c, 'd) new_thread_action list *
              ('b conditional_action list *
                (('b, 'e) wait_set_action list *
                  ('b interrupt_action list * 'f list))))
end = struct

fun shr lstsmws =
  Product_Type.snd (Product_Type.fst (Product_Type.snd lstsmws));

fun thr lstsmws =
  Product_Type.fst (Product_Type.fst (Product_Type.snd lstsmws));

datatype wake_up_status = WSNotified | WSWokenUp;

datatype 'a wait_set_status = InWS of 'a | PostWS of wake_up_status;

datatype 'a conditional_action = Join of 'a | Yield;

datatype lock_action = Lock | Unlock | UnlockFail | ReleaseAcquire;

fun wset lstsmws =
  Product_Type.fst (Product_Type.snd (Product_Type.snd lstsmws));

fun locks lstsmws = Product_Type.fst lstsmws;

datatype ('a, 'b, 'c) new_thread_action = NewThread of 'a * 'b * 'c |
  ThreadExists of 'a * bool;

datatype ('a, 'b) wait_set_action = Suspend of 'b | Notify of 'b |
  NotifyAll of 'b | WakeUp of 'a | Notified | WokenUp;

datatype 'a interrupt_action = IsInterrupted of 'a * bool | Interrupt of 'a |
  ClearInterrupt of 'a;

fun obs_a x =
  (Product_Type.snd o Product_Type.snd o Product_Type.snd o Product_Type.snd o
    Product_Type.snd)
    x;

fun thr_a x = (Product_Type.fst o Product_Type.snd) x;

fun cond_a x = (Product_Type.fst o Product_Type.snd o Product_Type.snd) x;

fun wset_a x =
  (Product_Type.fst o Product_Type.snd o Product_Type.snd o Product_Type.snd) x;

fun locks_a x = Product_Type.fst x;

datatype ('a, 'b, 'c, 'd, 'e, 'f) thread_action =
  LockAction of (lock_action * 'a) |
  NewThreadAction of ('b, 'c, 'd) new_thread_action |
  ConditionalAction of 'b conditional_action |
  WaitSetAction of ('b, 'e) wait_set_action |
  InterruptAction of 'b interrupt_action | ObsAction of 'f;

fun interrupts lstsmws =
  Product_Type.snd (Product_Type.snd (Product_Type.snd lstsmws));

type 'a obs_action = {};

fun interrupt_a x =
  (Product_Type.fst o Product_Type.snd o Product_Type.snd o Product_Type.snd o
    Product_Type.snd)
    x;

fun convert_new_thread_action f (NewThread (t, x, m)) = NewThread (t, f x, m)
  | convert_new_thread_action f (ThreadExists (t, b)) = ThreadExists (t, b);

fun convert_extTA f ta =
  (locks_a ta,
    (List.map (convert_new_thread_action f) (thr_a ta),
      Product_Type.snd (Product_Type.snd ta)));

fun ta_update_obs (ls, (nts, (js, (wss, (is, obs))))) ob =
  (ls, (nts, (js, (wss, (is, obs @ [ob])))));

fun equal_lock_actiona ReleaseAcquire UnlockFail = false
  | equal_lock_actiona UnlockFail ReleaseAcquire = false
  | equal_lock_actiona ReleaseAcquire Unlock = false
  | equal_lock_actiona Unlock ReleaseAcquire = false
  | equal_lock_actiona UnlockFail Unlock = false
  | equal_lock_actiona Unlock UnlockFail = false
  | equal_lock_actiona ReleaseAcquire Lock = false
  | equal_lock_actiona Lock ReleaseAcquire = false
  | equal_lock_actiona UnlockFail Lock = false
  | equal_lock_actiona Lock UnlockFail = false
  | equal_lock_actiona Unlock Lock = false
  | equal_lock_actiona Lock Unlock = false
  | equal_lock_actiona ReleaseAcquire ReleaseAcquire = true
  | equal_lock_actiona UnlockFail UnlockFail = true
  | equal_lock_actiona Unlock Unlock = true
  | equal_lock_actiona Lock Lock = true;

val equal_lock_action = {equal = equal_lock_actiona} : lock_action HOL.equal;

fun ta_update_locks A_ (ls, (nts, (js, (wss, obs)))) lta l =
  (FinFun.finfun_update A_ (List.equal_list equal_lock_action) ls l
     (FinFun.finfun_apply A_ ls l @ [lta]),
    (nts, (js, (wss, obs))));

fun ta_update_wait_set (ls, (nts, (js, (wss, (is, obs))))) ws =
  (ls, (nts, (js, (wss @ [ws], (is, obs)))));

fun ta_update_NewThread (ls, (nts, (js, (wss, (is, obs))))) nt =
  (ls, (nts @ [nt], (js, (wss, (is, obs)))));

fun ta_update_interrupt (ls, (nts, (js, (wss, (is, obs))))) i =
  (ls, (nts, (js, (wss, (is @ [i], obs)))));

fun equal_wait_set_actiona A_ B_ WokenUp Notified = false
  | equal_wait_set_actiona A_ B_ Notified WokenUp = false
  | equal_wait_set_actiona A_ B_ WokenUp (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) WokenUp = false
  | equal_wait_set_actiona A_ B_ Notified (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) Notified = false
  | equal_wait_set_actiona A_ B_ WokenUp (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll w) WokenUp = false
  | equal_wait_set_actiona A_ B_ Notified (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll w) Notified = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll w) (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ WokenUp (Notify w) = false
  | equal_wait_set_actiona A_ B_ (Notify w) WokenUp = false
  | equal_wait_set_actiona A_ B_ Notified (Notify w) = false
  | equal_wait_set_actiona A_ B_ (Notify w) Notified = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) (Notify w) = false
  | equal_wait_set_actiona A_ B_ (Notify w) (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll wa) (Notify w) = false
  | equal_wait_set_actiona A_ B_ (Notify wa) (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ WokenUp (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend w) WokenUp = false
  | equal_wait_set_actiona A_ B_ Notified (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend w) Notified = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend w) (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll wa) (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend wa) (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ (Notify wa) (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend wa) (Notify w) = false
  | equal_wait_set_actiona A_ B_ (WakeUp ta) (WakeUp t) = HOL.eq A_ ta t
  | equal_wait_set_actiona A_ B_ (NotifyAll wa) (NotifyAll w) = HOL.eq B_ wa w
  | equal_wait_set_actiona A_ B_ (Notify wa) (Notify w) = HOL.eq B_ wa w
  | equal_wait_set_actiona A_ B_ (Suspend wa) (Suspend w) = HOL.eq B_ wa w
  | equal_wait_set_actiona A_ B_ WokenUp WokenUp = true
  | equal_wait_set_actiona A_ B_ Notified Notified = true;

fun equal_wait_set_action A_ B_ = {equal = equal_wait_set_actiona A_ B_} :
  ('a, 'b) wait_set_action HOL.equal;

fun equal_wake_up_status WSWokenUp WSNotified = false
  | equal_wake_up_status WSNotified WSWokenUp = false
  | equal_wake_up_status WSWokenUp WSWokenUp = true
  | equal_wake_up_status WSNotified WSNotified = true;

fun equal_wait_set_statusa A_ (PostWS wake_up_status) (InWS w) = false
  | equal_wait_set_statusa A_ (InWS w) (PostWS wake_up_status) = false
  | equal_wait_set_statusa A_ (PostWS wake_up_statusa) (PostWS wake_up_status) =
    equal_wake_up_status wake_up_statusa wake_up_status
  | equal_wait_set_statusa A_ (InWS wa) (InWS w) = HOL.eq A_ wa w;

fun equal_wait_set_status A_ = {equal = equal_wait_set_statusa A_} :
  'a wait_set_status HOL.equal;

fun ta_update_Conditional (ls, (nts, (js, (wss, (is, obs))))) j =
  (ls, (nts, (js @ [j], (wss, (is, obs)))));

fun equal_conditional_actiona A_ Yield (Join t) = false
  | equal_conditional_actiona A_ (Join t) Yield = false
  | equal_conditional_actiona A_ (Join ta) (Join t) = HOL.eq A_ ta t
  | equal_conditional_actiona A_ Yield Yield = true;

fun equal_conditional_action A_ = {equal = equal_conditional_actiona A_} :
  'a conditional_action HOL.equal;

fun thread_action_to_thread_action A_ F_ (LockAction (la, l)) ta =
  ta_update_locks A_ ta la l
  | thread_action_to_thread_action A_ F_ (NewThreadAction nt) ta =
    ta_update_NewThread ta nt
  | thread_action_to_thread_action A_ F_ (ConditionalAction ca) ta =
    ta_update_Conditional ta ca
  | thread_action_to_thread_action A_ F_ (WaitSetAction wa) ta =
    ta_update_wait_set ta wa
  | thread_action_to_thread_action A_ F_ (InterruptAction ia) ta =
    ta_update_interrupt ta ia
  | thread_action_to_thread_action A_ F_ (ObsAction ob) ta =
    ta_update_obs ta ob;

end; (*struct FWState*)

structure Heap : sig
  datatype addr_loc = CField of string * string | ACell of IntInf.int
  val typeof : 'a Value.vala -> Type.ty option
  val typeof_h :
    'b Value.addr ->
      ('a -> 'b -> Type.ty option) -> 'a -> 'b Value.vala -> Type.ty option
end = struct

datatype addr_loc = CField of string * string | ACell of IntInf.int;

fun typeof Value.Unit = SOME Type.Void
  | typeof Value.Null = SOME Type.Nt
  | typeof (Value.Bool b) = SOME Type.Boolean
  | typeof (Value.Intg i) = SOME Type.Integer
  | typeof (Value.Addr a) = NONE;

fun typeof_h B_ typeof_addr h (Value.Intg va) = typeof (Value.Intg va)
  | typeof_h B_ typeof_addr h (Value.Bool va) = typeof (Value.Bool va)
  | typeof_h B_ typeof_addr h Value.Null = typeof Value.Null
  | typeof_h B_ typeof_addr h Value.Unit = typeof Value.Unit
  | typeof_h B_ typeof_addr h (Value.Addr a) = typeof_addr h a;

end; (*struct Heap*)

structure Observable_Events : sig
  datatype ('a, 'b) obs_event =
    ExternalCall of 'a * string * 'a Value.vala list * 'a Value.vala |
    ReadMem of 'a * Heap.addr_loc * 'a Value.vala |
    WriteMem of 'a * Heap.addr_loc * 'a Value.vala | NewObj of 'a * string |
    NewArr of 'a * Type.ty * IntInf.int | ThreadStart of 'b | ThreadJoin of 'b |
    SyncLock of 'a | SyncUnlock of 'a | ObsInterrupt of 'b |
    ObsInterrupted of 'b
  val convert_RA :
    'a HOL.equal * 'a Value.addr ->
      ('a, IntInf.int) FinFun.finfun -> ('a, 'b) obs_event list
  val obs_action_obs_event : ('a, 'b) obs_event FWState.obs_action
end = struct

datatype ('a, 'b) obs_event =
  ExternalCall of 'a * string * 'a Value.vala list * 'a Value.vala |
  ReadMem of 'a * Heap.addr_loc * 'a Value.vala |
  WriteMem of 'a * Heap.addr_loc * 'a Value.vala | NewObj of 'a * string |
  NewArr of 'a * Type.ty * IntInf.int | ThreadStart of 'b | ThreadJoin of 'b |
  SyncLock of 'a | SyncUnlock of 'a | ObsInterrupt of 'b | ObsInterrupted of 'b;

fun convert_RA (A1_, A2_) ln =
  List.maps
    (fn ad => List.replicate (FinFun.finfun_apply A1_ ln ad) (SyncLock ad))
    (Value.monitor_finfun_to_list A2_ ln);

val obs_action_obs_event = {} : ('a, 'b) obs_event FWState.obs_action;

end; (*struct Observable_Events*)

structure ExternalCall : sig
  datatype 'a extCallRet = RetVal of 'a Value.vala | RetExc of 'a | RetStaySame
  val external_WT_defs_i_i_i_i :
    string -> string -> Type.ty list -> Type.ty -> unit Predicate.pred
  val external_WT_defs : string -> string -> Type.ty list -> Type.ty -> bool
  val heap_copy_loc_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o :
    'b Value.addr ->
      ('a -> 'b -> Heap.addr_loc -> 'b Value.vala Predicate.pred) ->
        ('a -> 'b -> Heap.addr_loc -> 'b Value.vala -> 'a Predicate.pred) ->
          'b -> 'b -> Heap.addr_loc ->
                        'a -> (('b, 'c) Observable_Events.obs_event list * 'a)
                                Predicate.pred
  val heap_copies_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o :
    'b Value.addr ->
      ('a -> 'b -> Heap.addr_loc -> 'b Value.vala Predicate.pred) ->
        ('a -> 'b -> Heap.addr_loc -> 'b Value.vala -> 'a Predicate.pred) ->
          'b -> 'b -> Heap.addr_loc list ->
                        'a -> (('b, 'c) Observable_Events.obs_event list * 'a)
                                Predicate.pred
  val heap_clone_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_o_o :
    'b Value.addr ->
      ('a -> string -> 'a * 'b option) ->
        ('a -> Type.ty -> IntInf.int -> 'a * 'b option) ->
          ('a -> 'b -> Type.ty option) ->
            ('a -> 'b -> IntInf.int) ->
              ('a -> 'b -> Heap.addr_loc -> 'b Value.vala Predicate.pred) ->
                ('a ->
                  'b -> Heap.addr_loc -> 'b Value.vala -> 'a Predicate.pred) ->
                  'c TypeRel.prog ->
                    'a -> 'b -> ('a * (('b, 'd)
 Observable_Events.obs_event list *
'b) option)
                                  Predicate.pred
  val red_external_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_i_o_o_o :
    'a HOL.equal * 'a Value.addr ->
      ('a -> 'b) ->
        ('b -> 'a) ->
          'c -> ('c -> string -> 'c * 'a option) ->
                  ('c -> Type.ty -> IntInf.int -> 'c * 'a option) ->
                    ('c -> 'a -> Type.ty option) ->
                      ('c -> 'a -> IntInf.int) ->
                        ('c ->
                          'a -> Heap.addr_loc ->
                                  'a Value.vala Predicate.pred) ->
                          ('c ->
                            'a -> Heap.addr_loc ->
                                    'a Value.vala -> 'c Predicate.pred) ->
                            'd TypeRel.prog ->
                              'b -> 'c -> 'a ->
    string ->
      'a Value.vala list ->
        ((('a, (FWState.lock_action list)) FinFun.finfun *
           (('b, (string * (string * 'a)), 'c) FWState.new_thread_action list *
             ('b FWState.conditional_action list *
               (('b, 'a) FWState.wait_set_action list *
                 ('b FWState.interrupt_action list *
                   ('a, 'b) Observable_Events.obs_event list))))) *
          ('a extCallRet * 'c))
          Predicate.pred
end = struct

datatype 'a extCallRet = RetVal of 'a Value.vala | RetExc of 'a | RetStaySame;

fun external_WT_defs_i_i_i_i x xa xb xc =
  Predicate.sup_pred
    (case xb
      of [] =>
        (case xc
          of Type.Void =>
            (if ((xa : string) = "start()V") andalso
                  ((x : string) = Type.thread)
              then Predicate.single () else Predicate.bot_pred)
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
      | _ :: _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb
        of [] =>
          (case xc
            of Type.Void =>
              (if ((xa : string) = "join()V") andalso
                    ((x : string) = Type.thread)
                then Predicate.single () else Predicate.bot_pred)
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.bot_pred)
        | _ :: _ => Predicate.bot_pred)
      (Predicate.sup_pred
        (case xb
          of [] =>
            (case xc
              of Type.Void =>
                (if ((xa : string) = "interrupt()V") andalso
                      ((x : string) = Type.thread)
                  then Predicate.single () else Predicate.bot_pred)
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | _ :: _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case xb
            of [] =>
              (case xc of Type.Void => Predicate.bot_pred
                | Type.Boolean =>
                  (if ((xa : string) = "isInterrupted()Z") andalso
                        ((x : string) = Type.thread)
                    then Predicate.single () else Predicate.bot_pred)
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.bot_pred)
            | _ :: _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case xb
              of [] =>
                (case xc
                  of Type.Void =>
                    (if ((xa : string) = "wait()V") andalso
                          ((x : string) = Type.object)
                      then Predicate.single () else Predicate.bot_pred)
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | _ :: _ => Predicate.bot_pred)
            (Predicate.sup_pred
              (case xb
                of [] =>
                  (case xc
                    of Type.Void =>
                      (if ((xa : string) = "notify()V") andalso
                            ((x : string) = Type.object)
                        then Predicate.single () else Predicate.bot_pred)
                    | Type.Boolean => Predicate.bot_pred
                    | Type.Integer => Predicate.bot_pred
                    | Type.Nt => Predicate.bot_pred
                    | Type.Class _ => Predicate.bot_pred
                    | Type.Array _ => Predicate.bot_pred)
                | _ :: _ => Predicate.bot_pred)
              (Predicate.sup_pred
                (case xb
                  of [] =>
                    (case xc
                      of Type.Void =>
                        (if ((xa : string) = "notifyAll()V") andalso
                              ((x : string) = Type.object)
                          then Predicate.single () else Predicate.bot_pred)
                      | Type.Boolean => Predicate.bot_pred
                      | Type.Integer => Predicate.bot_pred
                      | Type.Nt => Predicate.bot_pred
                      | Type.Class _ => Predicate.bot_pred
                      | Type.Array _ => Predicate.bot_pred)
                  | _ :: _ => Predicate.bot_pred)
                (Predicate.sup_pred
                  (case xb
                    of [] =>
                      (case xc of Type.Void => Predicate.bot_pred
                        | Type.Boolean => Predicate.bot_pred
                        | Type.Integer => Predicate.bot_pred
                        | Type.Nt => Predicate.bot_pred
                        | Type.Class xba =>
                          (if ((xba : string) = Type.object) andalso
                                (((xa : string) = "clone()Ljava/lang/Object;") andalso
                                  ((x : string) = Type.object))
                            then Predicate.single () else Predicate.bot_pred)
                        | Type.Array _ => Predicate.bot_pred)
                    | _ :: _ => Predicate.bot_pred)
                  (Predicate.sup_pred
                    (case xb
                      of [] =>
                        (case xc of Type.Void => Predicate.bot_pred
                          | Type.Boolean => Predicate.bot_pred
                          | Type.Integer =>
                            (if ((xa : string) = "hashCode()I") andalso
                                  ((x : string) = Type.object)
                              then Predicate.single () else Predicate.bot_pred)
                          | Type.Nt => Predicate.bot_pred
                          | Type.Class _ => Predicate.bot_pred
                          | Type.Array _ => Predicate.bot_pred)
                      | _ :: _ => Predicate.bot_pred)
                    (Predicate.sup_pred
                      (case xb of [] => Predicate.bot_pred
                        | Type.Void :: lista => Predicate.bot_pred
                        | Type.Boolean :: lista => Predicate.bot_pred
                        | [Type.Integer] =>
                          (case xc
                            of Type.Void =>
                              (if ((xa : string) = "~print(I)V") andalso
                                    ((x : string) = Type.object)
                                then Predicate.single ()
                                else Predicate.bot_pred)
                            | Type.Boolean => Predicate.bot_pred
                            | Type.Integer => Predicate.bot_pred
                            | Type.Nt => Predicate.bot_pred
                            | Type.Class _ => Predicate.bot_pred
                            | Type.Array _ => Predicate.bot_pred)
                        | Type.Integer :: _ :: _ => Predicate.bot_pred
                        | Type.Nt :: lista => Predicate.bot_pred
                        | Type.Class _ :: lista => Predicate.bot_pred
                        | Type.Array _ :: lista => Predicate.bot_pred)
                      (Predicate.sup_pred
                        (case xb
                          of [] =>
                            (case xc of Type.Void => Predicate.bot_pred
                              | Type.Boolean => Predicate.bot_pred
                              | Type.Integer => Predicate.bot_pred
                              | Type.Nt => Predicate.bot_pred
                              | Type.Class xba =>
                                (if ((xba : string) = Type.thread) andalso
                                      (((xa : string) = "~Thread.currentThread()Ljava/lang/Thread;") andalso
((x : string) = Type.object))
                                  then Predicate.single ()
                                  else Predicate.bot_pred)
                              | Type.Array _ => Predicate.bot_pred)
                          | _ :: _ => Predicate.bot_pred)
                        (Predicate.sup_pred
                          (case xb
                            of [] =>
                              (case xc of Type.Void => Predicate.bot_pred
                                | Type.Boolean =>
                                  (if ((xa : string) = "~Thread.interrupted()Z") andalso
((x : string) = Type.object)
                                    then Predicate.single ()
                                    else Predicate.bot_pred)
                                | Type.Integer => Predicate.bot_pred
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                            | _ :: _ => Predicate.bot_pred)
                          (case xb
                            of [] =>
                              (case xc
                                of Type.Void =>
                                  (if ((xa : string) = "~Thread.yield()V") andalso
((x : string) = Type.object)
                                    then Predicate.single ()
                                    else Predicate.bot_pred)
                                | Type.Boolean => Predicate.bot_pred
                                | Type.Integer => Predicate.bot_pred
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                            | _ :: _ => Predicate.bot_pred))))))))))));

fun external_WT_defs x1 x2 x3 x4 =
  Predicate.holds (external_WT_defs_i_i_i_i x1 x2 x3 x4);

fun heap_copy_loc_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o B_ heap_read heap_write x xa
  xb xc =
  Predicate.bind (heap_read xc x xb)
    (fn xd =>
      Predicate.bind (heap_write xc xa xb xd)
        (fn xca =>
          Predicate.single
            ([Observable_Events.ReadMem (x, xb, xd),
               Observable_Events.WriteMem (xa, xb, xd)],
              xca)));

fun heap_copies_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o B_ heap_read heap_write x xa xb
  xc =
  Predicate.sup_pred
    (case xb of [] => Predicate.single ([], xc) | _ :: _ => Predicate.bot_pred)
    (case xb of [] => Predicate.bot_pred
      | al :: als =>
        Predicate.bind
          (heap_copy_loc_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o B_ heap_read heap_write
            x xa al xc)
          (fn (ob, h) =>
            Predicate.bind
              (heap_copies_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o B_ heap_read
                heap_write x xa als h)
              (fn (obs, ha) => Predicate.single (ob @ obs, ha))));

fun heap_clone_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_o_o B_ x xa xb xc heap_read
  heap_write xd xe xf =
  Predicate.sup_pred
    (case xb xe xf of NONE => Predicate.bot_pred
      | SOME Type.Void => Predicate.bot_pred
      | SOME Type.Boolean => Predicate.bot_pred
      | SOME Type.Integer => Predicate.bot_pred
      | SOME Type.Nt => Predicate.bot_pred
      | SOME (Type.Class c) =>
        (case x xe c of (h, NONE) => Predicate.single (h, NONE)
          | (h, SOME _) => Predicate.bot_pred)
      | SOME (Type.Array _) => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb xe xf of NONE => Predicate.bot_pred
        | SOME Type.Void => Predicate.bot_pred
        | SOME Type.Boolean => Predicate.bot_pred
        | SOME Type.Integer => Predicate.bot_pred
        | SOME Type.Nt => Predicate.bot_pred
        | SOME (Type.Class _) => Predicate.bot_pred
        | SOME (Type.Array t) =>
          (case xa xe t (xc xe xf) of (h, NONE) => Predicate.single (h, NONE)
            | (h, SOME _) => Predicate.bot_pred))
      (Predicate.sup_pred
        (case xb xe xf of NONE => Predicate.bot_pred
          | SOME Type.Void => Predicate.bot_pred
          | SOME Type.Boolean => Predicate.bot_pred
          | SOME Type.Integer => Predicate.bot_pred
          | SOME Type.Nt => Predicate.bot_pred
          | SOME (Type.Class c) =>
            (case x xe c of (h, NONE) => Predicate.bot_pred
              | (h, SOME a) =>
                Predicate.bind (TypeRel.fields_i_i_o xd c)
                  (fn xg =>
                    Predicate.bind
                      (heap_copies_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o B_ heap_read
                        heap_write xf a
                        (List.map
                          (fn (aa, b) =>
                            let
                              val (f, d) = aa;
                            in
                              (fn _ => Heap.CField (d, f))
                            end
                              b)
                          xg)
                        h)
                      (fn (obs, ha) =>
                        Predicate.single
                          (ha, SOME (Observable_Events.NewObj (a, c) :: obs,
                                      a)))))
          | SOME (Type.Array _) => Predicate.bot_pred)
        (case xb xe xf of NONE => Predicate.bot_pred
          | SOME Type.Void => Predicate.bot_pred
          | SOME Type.Boolean => Predicate.bot_pred
          | SOME Type.Integer => Predicate.bot_pred
          | SOME Type.Nt => Predicate.bot_pred
          | SOME (Type.Class _) => Predicate.bot_pred
          | SOME (Type.Array t) =>
            (case xa xe t (xc xe xf) of (h, NONE) => Predicate.bot_pred
              | (h, SOME a) =>
                Predicate.bind (TypeRel.fields_i_i_o xd Type.object)
                  (fn xg =>
                    Predicate.bind
                      (heap_copies_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o B_ heap_read
                        heap_write xf a
                        (List.map
                           (fn (aa, b) =>
                             let
                               val (f, d) = aa;
                             in
                               (fn _ => Heap.CField (d, f))
                             end
                               b)
                           xg @
                          List.map Heap.ACell
                            (List.upt (0 : IntInf.int) (xc xe xf)))
                        h)
                      (fn (obs, ha) =>
                        Predicate.single
                          (ha, SOME (Observable_Events.NewArr
                                       (a, t, xc xe xf) ::
                                       obs,
                                      a))))))));

fun red_external_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_i_o_o_o (A1_, A2_)
  x xa xb xc xd xe xf heap_read heap_write xg xh xi xj xk xl =
  Predicate.sup_pred
    (case xl
      of [] =>
        (case xe xi xj of NONE => Predicate.bot_pred
          | SOME Type.Void => Predicate.bot_pred
          | SOME Type.Boolean => Predicate.bot_pred
          | SOME Type.Integer => Predicate.bot_pred
          | SOME Type.Nt => Predicate.bot_pred
          | SOME (Type.Class c) =>
            Predicate.bind
              (Transitive_Closure_Table.rtranclp_FioB_i_i Stringa.equal_literal
                (TypeRel.subcls1_i_i_o xg) c Type.thread)
              (fn () =>
                (if ((xk : string) = "start()V")
                  then Predicate.single
                         (FWState.thread_action_to_thread_action A1_
                            Observable_Events.obs_action_obs_event
                            (FWState.ObsAction
                              (Observable_Events.ThreadStart (x xj)))
                            (FWState.thread_action_to_thread_action A1_
                              Observable_Events.obs_action_obs_event
                              (FWState.NewThreadAction
                                (FWState.NewThread
                                  (x xj, (c, ("run()V", xj)), xi)))
                              (FinFun.Finfun_const [],
                                ([], ([], ([], ([], [])))))),
                           (RetVal Value.Unit, xi))
                  else Predicate.bot_pred))
          | SOME (Type.Array _) => Predicate.bot_pred)
      | _ :: _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xl
        of [] =>
          (case xe xi xj of NONE => Predicate.bot_pred
            | SOME Type.Void => Predicate.bot_pred
            | SOME Type.Boolean => Predicate.bot_pred
            | SOME Type.Integer => Predicate.bot_pred
            | SOME Type.Nt => Predicate.bot_pred
            | SOME (Type.Class c) =>
              Predicate.bind
                (Transitive_Closure_Table.rtranclp_FioB_i_i
                  Stringa.equal_literal (TypeRel.subcls1_i_i_o xg) c
                  Type.thread)
                (fn () =>
                  (if ((xk : string) = "start()V")
                    then Predicate.single
                           (FWState.thread_action_to_thread_action A1_
                              Observable_Events.obs_action_obs_event
                              (FWState.NewThreadAction
                                (FWState.ThreadExists (x xj, true)))
                              (FinFun.Finfun_const [],
                                ([], ([], ([], ([], []))))),
                             (RetExc
                                (StartConfig.addr_of_sys_xcpt A2_ xb xc
                                  "java/lang/IllegalThreadStateException"),
                               xi))
                    else Predicate.bot_pred))
            | SOME (Type.Array _) => Predicate.bot_pred)
        | _ :: _ => Predicate.bot_pred)
      (Predicate.sup_pred
        (case xl
          of [] =>
            (case xe xi xj of NONE => Predicate.bot_pred
              | SOME Type.Void => Predicate.bot_pred
              | SOME Type.Boolean => Predicate.bot_pred
              | SOME Type.Integer => Predicate.bot_pred
              | SOME Type.Nt => Predicate.bot_pred
              | SOME (Type.Class c) =>
                Predicate.bind
                  (Transitive_Closure_Table.rtranclp_FioB_i_i
                    Stringa.equal_literal (TypeRel.subcls1_i_i_o xg) c
                    Type.thread)
                  (fn () =>
                    (if ((xk : string) = "join()V")
                      then Predicate.single
                             (FWState.thread_action_to_thread_action A1_
                                Observable_Events.obs_action_obs_event
                                (FWState.ObsAction
                                  (Observable_Events.ThreadJoin (x xj)))
                                (FWState.thread_action_to_thread_action A1_
                                  Observable_Events.obs_action_obs_event
                                  (FWState.InterruptAction
                                    (FWState.IsInterrupted (xh, false)))
                                  (FWState.thread_action_to_thread_action A1_
                                    Observable_Events.obs_action_obs_event
                                    (FWState.ConditionalAction
                                      (FWState.Join (x xj)))
                                    (FinFun.Finfun_const [],
                                      ([], ([], ([], ([], []))))))),
                               (RetVal Value.Unit, xi))
                      else Predicate.bot_pred))
              | SOME (Type.Array _) => Predicate.bot_pred)
          | _ :: _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case xl
            of [] =>
              (case xe xi xj of NONE => Predicate.bot_pred
                | SOME Type.Void => Predicate.bot_pred
                | SOME Type.Boolean => Predicate.bot_pred
                | SOME Type.Integer => Predicate.bot_pred
                | SOME Type.Nt => Predicate.bot_pred
                | SOME (Type.Class c) =>
                  Predicate.bind
                    (Transitive_Closure_Table.rtranclp_FioB_i_i
                      Stringa.equal_literal (TypeRel.subcls1_i_i_o xg) c
                      Type.thread)
                    (fn () =>
                      (if ((xk : string) = "join()V")
                        then Predicate.single
                               (FWState.thread_action_to_thread_action A1_
                                  Observable_Events.obs_action_obs_event
                                  (FWState.ObsAction
                                    (Observable_Events.ObsInterrupted xh))
                                  (FWState.thread_action_to_thread_action A1_
                                    Observable_Events.obs_action_obs_event
                                    (FWState.InterruptAction
                                      (FWState.ClearInterrupt xh))
                                    (FWState.thread_action_to_thread_action A1_
                                      Observable_Events.obs_action_obs_event
                                      (FWState.InterruptAction
(FWState.IsInterrupted (xh, true)))
                                      (FinFun.Finfun_const [],
([], ([], ([], ([], []))))))),
                                 (RetExc
                                    (StartConfig.addr_of_sys_xcpt A2_ xb xc
                                      "java/lang/InterruptedException"),
                                   xi))
                        else Predicate.bot_pred))
                | SOME (Type.Array _) => Predicate.bot_pred)
            | _ :: _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case xl
              of [] =>
                (case xe xi xj of NONE => Predicate.bot_pred
                  | SOME Type.Void => Predicate.bot_pred
                  | SOME Type.Boolean => Predicate.bot_pred
                  | SOME Type.Integer => Predicate.bot_pred
                  | SOME Type.Nt => Predicate.bot_pred
                  | SOME (Type.Class c) =>
                    Predicate.bind
                      (Transitive_Closure_Table.rtranclp_FioB_i_i
                        Stringa.equal_literal (TypeRel.subcls1_i_i_o xg) c
                        Type.thread)
                      (fn () =>
                        (if ((xk : string) = "interrupt()V")
                          then Predicate.single
                                 (FWState.thread_action_to_thread_action A1_
                                    Observable_Events.obs_action_obs_event
                                    (FWState.ObsAction
                                      (Observable_Events.ObsInterrupt (x xj)))
                                    (FWState.thread_action_to_thread_action A1_
                                      Observable_Events.obs_action_obs_event
                                      (FWState.InterruptAction
(FWState.Interrupt (x xj)))
                                      (FWState.thread_action_to_thread_action
A1_ Observable_Events.obs_action_obs_event
(FWState.WaitSetAction (FWState.WakeUp (x xj)))
(FWState.thread_action_to_thread_action A1_
  Observable_Events.obs_action_obs_event
  (FWState.NewThreadAction (FWState.ThreadExists (x xj, true)))
  (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))))),
                                   (RetVal Value.Unit, xi))
                          else Predicate.bot_pred))
                  | SOME (Type.Array _) => Predicate.bot_pred)
              | _ :: _ => Predicate.bot_pred)
            (Predicate.sup_pred
              (case xl
                of [] =>
                  (case xe xi xj of NONE => Predicate.bot_pred
                    | SOME Type.Void => Predicate.bot_pred
                    | SOME Type.Boolean => Predicate.bot_pred
                    | SOME Type.Integer => Predicate.bot_pred
                    | SOME Type.Nt => Predicate.bot_pred
                    | SOME (Type.Class c) =>
                      Predicate.bind
                        (Transitive_Closure_Table.rtranclp_FioB_i_i
                          Stringa.equal_literal (TypeRel.subcls1_i_i_o xg) c
                          Type.thread)
                        (fn () =>
                          (if ((xk : string) = "interrupt()V")
                            then Predicate.single
                                   (FWState.thread_action_to_thread_action A1_
                                      Observable_Events.obs_action_obs_event
                                      (FWState.NewThreadAction
(FWState.ThreadExists (x xj, false)))
                                      (FinFun.Finfun_const [],
([], ([], ([], ([], []))))),
                                     (RetVal Value.Unit, xi))
                            else Predicate.bot_pred))
                    | SOME (Type.Array _) => Predicate.bot_pred)
                | _ :: _ => Predicate.bot_pred)
              (Predicate.sup_pred
                (case xl
                  of [] =>
                    (case xe xi xj of NONE => Predicate.bot_pred
                      | SOME Type.Void => Predicate.bot_pred
                      | SOME Type.Boolean => Predicate.bot_pred
                      | SOME Type.Integer => Predicate.bot_pred
                      | SOME Type.Nt => Predicate.bot_pred
                      | SOME (Type.Class c) =>
                        Predicate.bind
                          (Transitive_Closure_Table.rtranclp_FioB_i_i
                            Stringa.equal_literal (TypeRel.subcls1_i_i_o xg) c
                            Type.thread)
                          (fn () =>
                            (if ((xk : string) = "isInterrupted()Z")
                              then Predicate.single
                                     (FWState.thread_action_to_thread_action A1_
Observable_Events.obs_action_obs_event
(FWState.ObsAction (Observable_Events.ObsInterrupted (x xj)))
(FWState.thread_action_to_thread_action A1_
  Observable_Events.obs_action_obs_event
  (FWState.InterruptAction (FWState.IsInterrupted (x xj, true)))
  (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))),
                                       (RetVal (Value.Bool true), xi))
                              else Predicate.bot_pred))
                      | SOME (Type.Array _) => Predicate.bot_pred)
                  | _ :: _ => Predicate.bot_pred)
                (Predicate.sup_pred
                  (case xl
                    of [] =>
                      (case xe xi xj of NONE => Predicate.bot_pred
                        | SOME Type.Void => Predicate.bot_pred
                        | SOME Type.Boolean => Predicate.bot_pred
                        | SOME Type.Integer => Predicate.bot_pred
                        | SOME Type.Nt => Predicate.bot_pred
                        | SOME (Type.Class c) =>
                          Predicate.bind
                            (Transitive_Closure_Table.rtranclp_FioB_i_i
                              Stringa.equal_literal (TypeRel.subcls1_i_i_o xg) c
                              Type.thread)
                            (fn () =>
                              (if ((xk : string) = "isInterrupted()Z")
                                then Predicate.single
                                       (FWState.thread_action_to_thread_action
  A1_ Observable_Events.obs_action_obs_event
  (FWState.InterruptAction (FWState.IsInterrupted (x xj, false)))
  (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
 (RetVal (Value.Bool false), xi))
                                else Predicate.bot_pred))
                        | SOME (Type.Array _) => Predicate.bot_pred)
                    | _ :: _ => Predicate.bot_pred)
                  (Predicate.sup_pred
                    (case xl
                      of [] =>
                        (if ((xk : string) = "wait()V")
                          then Predicate.single
                                 (FWState.thread_action_to_thread_action A1_
                                    Observable_Events.obs_action_obs_event
                                    (FWState.ObsAction
                                      (Observable_Events.ObsInterrupted xh))
                                    (FWState.thread_action_to_thread_action A1_
                                      Observable_Events.obs_action_obs_event
                                      (FWState.InterruptAction
(FWState.ClearInterrupt xh))
                                      (FWState.thread_action_to_thread_action
A1_ Observable_Events.obs_action_obs_event
(FWState.InterruptAction (FWState.IsInterrupted (xh, true)))
(FWState.thread_action_to_thread_action A1_
  Observable_Events.obs_action_obs_event (FWState.LockAction (FWState.Lock, xj))
  (FWState.thread_action_to_thread_action A1_
    Observable_Events.obs_action_obs_event
    (FWState.LockAction (FWState.Unlock, xj))
    (FinFun.Finfun_const [], ([], ([], ([], ([], []))))))))),
                                   (RetExc
                                      (StartConfig.addr_of_sys_xcpt A2_ xb xc
"java/lang/InterruptedException"),
                                     xi))
                          else Predicate.bot_pred)
                      | _ :: _ => Predicate.bot_pred)
                    (Predicate.sup_pred
                      (case xl
                        of [] =>
                          (if ((xk : string) = "wait()V")
                            then Predicate.single
                                   (FWState.thread_action_to_thread_action A1_
                                      Observable_Events.obs_action_obs_event
                                      (FWState.ObsAction
(Observable_Events.SyncUnlock xj))
                                      (FWState.thread_action_to_thread_action
A1_ Observable_Events.obs_action_obs_event
(FWState.InterruptAction (FWState.IsInterrupted (xh, false)))
(FWState.thread_action_to_thread_action A1_
  Observable_Events.obs_action_obs_event
  (FWState.LockAction (FWState.ReleaseAcquire, xj))
  (FWState.thread_action_to_thread_action A1_
    Observable_Events.obs_action_obs_event
    (FWState.LockAction (FWState.Lock, xj))
    (FWState.thread_action_to_thread_action A1_
      Observable_Events.obs_action_obs_event
      (FWState.LockAction (FWState.Unlock, xj))
      (FWState.thread_action_to_thread_action A1_
        Observable_Events.obs_action_obs_event
        (FWState.WaitSetAction (FWState.Suspend xj))
        (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))))))),
                                     (RetStaySame, xi))
                            else Predicate.bot_pred)
                        | _ :: _ => Predicate.bot_pred)
                      (Predicate.sup_pred
                        (case xl
                          of [] =>
                            (if ((xk : string) = "wait()V")
                              then Predicate.single
                                     (FWState.thread_action_to_thread_action A1_
Observable_Events.obs_action_obs_event
(FWState.LockAction (FWState.UnlockFail, xj))
(FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                                       (RetExc
  (StartConfig.addr_of_sys_xcpt A2_ xb xc
    "java/lang/IllegalMonitorStateException"),
 xi))
                              else Predicate.bot_pred)
                          | _ :: _ => Predicate.bot_pred)
                        (Predicate.sup_pred
                          (case xl
                            of [] =>
                              (if ((xk : string) = "wait()V")
                                then Predicate.single
                                       (FWState.thread_action_to_thread_action
  A1_ Observable_Events.obs_action_obs_event
  (FWState.WaitSetAction FWState.Notified)
  (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
 (RetVal Value.Unit, xi))
                                else Predicate.bot_pred)
                            | _ :: _ => Predicate.bot_pred)
                          (Predicate.sup_pred
                            (case xl
                              of [] =>
                                (if ((xk : string) = "wait()V")
                                  then Predicate.single
 (FWState.thread_action_to_thread_action A1_
    Observable_Events.obs_action_obs_event
    (FWState.ObsAction (Observable_Events.ObsInterrupted xh))
    (FWState.thread_action_to_thread_action A1_
      Observable_Events.obs_action_obs_event
      (FWState.InterruptAction (FWState.ClearInterrupt xh))
      (FWState.thread_action_to_thread_action A1_
        Observable_Events.obs_action_obs_event
        (FWState.WaitSetAction FWState.WokenUp)
        (FinFun.Finfun_const [], ([], ([], ([], ([], []))))))),
   (RetExc
      (StartConfig.addr_of_sys_xcpt A2_ xb xc "java/lang/InterruptedException"),
     xi))
                                  else Predicate.bot_pred)
                              | _ :: _ => Predicate.bot_pred)
                            (Predicate.sup_pred
                              (case xl
                                of [] =>
                                  (if ((xk : string) = "notify()V")
                                    then Predicate.single
   (FWState.thread_action_to_thread_action A1_
      Observable_Events.obs_action_obs_event
      (FWState.LockAction (FWState.Lock, xj))
      (FWState.thread_action_to_thread_action A1_
        Observable_Events.obs_action_obs_event
        (FWState.LockAction (FWState.Unlock, xj))
        (FWState.thread_action_to_thread_action A1_
          Observable_Events.obs_action_obs_event
          (FWState.WaitSetAction (FWState.Notify xj))
          (FinFun.Finfun_const [], ([], ([], ([], ([], []))))))),
     (RetVal Value.Unit, xi))
                                    else Predicate.bot_pred)
                                | _ :: _ => Predicate.bot_pred)
                              (Predicate.sup_pred
                                (case xl
                                  of [] =>
                                    (if ((xk : string) = "notify()V")
                                      then Predicate.single
     (FWState.thread_action_to_thread_action A1_
        Observable_Events.obs_action_obs_event
        (FWState.LockAction (FWState.UnlockFail, xj))
        (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
       (RetExc
          (StartConfig.addr_of_sys_xcpt A2_ xb xc
            "java/lang/IllegalMonitorStateException"),
         xi))
                                      else Predicate.bot_pred)
                                  | _ :: _ => Predicate.bot_pred)
                                (Predicate.sup_pred
                                  (case xl
                                    of [] =>
                                      (if ((xk : string) = "notifyAll()V")
then Predicate.single
       (FWState.thread_action_to_thread_action A1_
          Observable_Events.obs_action_obs_event
          (FWState.LockAction (FWState.Lock, xj))
          (FWState.thread_action_to_thread_action A1_
            Observable_Events.obs_action_obs_event
            (FWState.LockAction (FWState.Unlock, xj))
            (FWState.thread_action_to_thread_action A1_
              Observable_Events.obs_action_obs_event
              (FWState.WaitSetAction (FWState.NotifyAll xj))
              (FinFun.Finfun_const [], ([], ([], ([], ([], []))))))),
         (RetVal Value.Unit, xi))
else Predicate.bot_pred)
                                    | _ :: _ => Predicate.bot_pred)
                                  (Predicate.sup_pred
                                    (case xl
                                      of [] =>
(if ((xk : string) = "notifyAll()V")
  then Predicate.single
         (FWState.thread_action_to_thread_action A1_
            Observable_Events.obs_action_obs_event
            (FWState.LockAction (FWState.UnlockFail, xj))
            (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
           (RetExc
              (StartConfig.addr_of_sys_xcpt A2_ xb xc
                "java/lang/IllegalMonitorStateException"),
             xi))
  else Predicate.bot_pred)
                                      | _ :: _ => Predicate.bot_pred)
                                    (Predicate.sup_pred
                                      (case xl
of [] =>
  Predicate.bind
    (heap_clone_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_o_o A2_ xc xd xe xf heap_read
      heap_write xg xi xj)
    (fn a =>
      (case a of (h, NONE) => Predicate.bot_pred
        | (h, SOME (obs, aa)) =>
          (if ((xk : string) = "clone()Ljava/lang/Object;")
            then Predicate.single
                   ((FinFun.Finfun_const [], ([], ([], ([], ([], obs))))),
                     (RetVal (Value.Addr aa), h))
            else Predicate.bot_pred)))
| _ :: _ => Predicate.bot_pred)
                                      (Predicate.sup_pred
(case xl
  of [] =>
    Predicate.bind
      (heap_clone_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_o_o A2_ xc xd xe xf
        heap_read heap_write xg xi xj)
      (fn a =>
        (case a
          of (h, NONE) =>
            (if ((xk : string) = "clone()Ljava/lang/Object;")
              then Predicate.single
                     ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                       (RetExc
                          (StartConfig.addr_of_sys_xcpt A2_ xb xc
                            "java/lang/OutOfMemoryError"),
                         h))
              else Predicate.bot_pred)
          | (h, SOME _) => Predicate.bot_pred))
  | _ :: _ => Predicate.bot_pred)
(Predicate.sup_pred
  (case xl
    of [] =>
      (if ((xk : string) = "hashCode()I")
        then Predicate.single
               ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                 (RetVal
                    (Value.Intg (Word.Word_of_int (Value.hash_addr A2_ xj))),
                   xi))
        else Predicate.bot_pred)
    | _ :: _ => Predicate.bot_pred)
  (Predicate.sup_pred
    (if ((xk : string) = "~print(I)V")
      then Predicate.single
             (FWState.thread_action_to_thread_action A1_
                Observable_Events.obs_action_obs_event
                (FWState.ObsAction
                  (Observable_Events.ExternalCall
                    (xj, "~print(I)V", xl, Value.Unit)))
                (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
               (RetVal Value.Unit, xi))
      else Predicate.bot_pred)
    (Predicate.sup_pred
      (case xl
        of [] =>
          (if ((xk : string) = "~Thread.currentThread()Ljava/lang/Thread;")
            then Predicate.single
                   ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                     (RetVal (Value.Addr (xa xh)), xi))
            else Predicate.bot_pred)
        | _ :: _ => Predicate.bot_pred)
      (Predicate.sup_pred
        (case xl
          of [] =>
            (if ((xk : string) = "~Thread.interrupted()Z")
              then Predicate.single
                     (FWState.thread_action_to_thread_action A1_
                        Observable_Events.obs_action_obs_event
                        (FWState.ObsAction
                          (Observable_Events.ObsInterrupted xh))
                        (FWState.thread_action_to_thread_action A1_
                          Observable_Events.obs_action_obs_event
                          (FWState.InterruptAction (FWState.ClearInterrupt xh))
                          (FWState.thread_action_to_thread_action A1_
                            Observable_Events.obs_action_obs_event
                            (FWState.InterruptAction
                              (FWState.IsInterrupted (xh, true)))
                            (FinFun.Finfun_const [],
                              ([], ([], ([], ([], []))))))),
                       (RetVal (Value.Bool true), xi))
              else Predicate.bot_pred)
          | _ :: _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case xl
            of [] =>
              (if ((xk : string) = "~Thread.interrupted()Z")
                then Predicate.single
                       (FWState.thread_action_to_thread_action A1_
                          Observable_Events.obs_action_obs_event
                          (FWState.InterruptAction
                            (FWState.IsInterrupted (xh, false)))
                          (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                         (RetVal (Value.Bool false), xi))
                else Predicate.bot_pred)
            | _ :: _ => Predicate.bot_pred)
          (case xl
            of [] =>
              (if ((xk : string) = "~Thread.yield()V")
                then Predicate.single
                       (FWState.thread_action_to_thread_action A1_
                          Observable_Events.obs_action_obs_event
                          (FWState.ConditionalAction FWState.Yield)
                          (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                         (RetVal Value.Unit, xi))
                else Predicate.bot_pred)
            | _ :: _ => Predicate.bot_pred))))))))))))))))))))))));

end; (*struct ExternalCall*)

structure Sum_Type : sig
  datatype ('a, 'b) sum = Inl of 'a | Inr of 'b
end = struct

datatype ('a, 'b) sum = Inl of 'a | Inr of 'b;

end; (*struct Sum_Type*)

structure BinOp : sig
  datatype bop = Eq | NotEq | LessThan | LessOrEqual | GreaterThan |
    GreaterOrEqual | Add | Subtract | Mult | Div | Mod | BinAnd | BinOr | BinXor
    | ShiftLeft | ShiftRightZeros | ShiftRightSigned
  val binop_ShiftRightSigned :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_ShiftRightZeros :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_ShiftLeft :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_BinXor :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_BinOr :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_BinAnd :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val word_sdiv :
    'a Type_Length.len -> 'a Word.word -> 'a Word.word -> 'a Word.word
  val binop_Div :
    'b Value.addr ->
      'a -> ('a -> string -> 'a * 'b option) ->
              'b Value.vala ->
                'b Value.vala -> ('b Value.vala, 'b) Sum_Type.sum option
  val word_smod :
    'a Type_Length.len -> 'a Word.word -> 'a Word.word -> 'a Word.word
  val binop_Mod :
    'b Value.addr ->
      'a -> ('a -> string -> 'a * 'b option) ->
              'b Value.vala ->
                'b Value.vala -> ('b Value.vala, 'b) Sum_Type.sum option
  val binop_Mult :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_Subtract :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_Add :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_GreaterOrEqual :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_GreaterThan :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_LessOrEqual :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_LessThan :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop :
    'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> string -> 'a * 'b option) ->
              bop ->
                'b Value.vala ->
                  'b Value.vala -> ('b Value.vala, 'b) Sum_Type.sum option
  val equal_bop : bop -> bop -> bool
  val wT_binop_i_i_i_i_i :
    'a TypeRel.prog ->
      Type.ty -> bop -> Type.ty -> Type.ty -> unit Predicate.pred
  val wT_binop_i_i_i_i_o :
    'a TypeRel.prog -> Type.ty -> bop -> Type.ty -> Type.ty Predicate.pred
end = struct

datatype bop = Eq | NotEq | LessThan | LessOrEqual | GreaterThan |
  GreaterOrEqual | Add | Subtract | Mult | Div | Mod | BinAnd | BinOr | BinXor |
  ShiftLeft | ShiftRightZeros | ShiftRightSigned;

fun binop_ShiftRightSigned (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.sshiftr
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i1 (Word.unat
                  (Type_Length.len0_bit0
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                  (Word.bitAND_word
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                    i2 (Word.Word_of_int (31 : IntInf.int)))))))
  | binop_ShiftRightSigned Value.Unit v2 = NONE
  | binop_ShiftRightSigned Value.Null v2 = NONE
  | binop_ShiftRightSigned (Value.Bool v) v2 = NONE
  | binop_ShiftRightSigned (Value.Addr v) v2 = NONE
  | binop_ShiftRightSigned v1 Value.Unit = NONE
  | binop_ShiftRightSigned v1 Value.Null = NONE
  | binop_ShiftRightSigned v1 (Value.Bool v) = NONE
  | binop_ShiftRightSigned v1 (Value.Addr v) = NONE;

fun binop_ShiftRightZeros (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.shiftr_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 (Word.unat
                  (Type_Length.len0_bit0
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                  (Word.bitAND_word
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                    i2 (Word.Word_of_int (31 : IntInf.int)))))))
  | binop_ShiftRightZeros Value.Unit v2 = NONE
  | binop_ShiftRightZeros Value.Null v2 = NONE
  | binop_ShiftRightZeros (Value.Bool v) v2 = NONE
  | binop_ShiftRightZeros (Value.Addr v) v2 = NONE
  | binop_ShiftRightZeros v1 Value.Unit = NONE
  | binop_ShiftRightZeros v1 Value.Null = NONE
  | binop_ShiftRightZeros v1 (Value.Bool v) = NONE
  | binop_ShiftRightZeros v1 (Value.Addr v) = NONE;

fun binop_ShiftLeft (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.shiftl_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 (Word.unat
                  (Type_Length.len0_bit0
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                  (Word.bitAND_word
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                    i2 (Word.Word_of_int (31 : IntInf.int)))))))
  | binop_ShiftLeft Value.Unit v2 = NONE
  | binop_ShiftLeft Value.Null v2 = NONE
  | binop_ShiftLeft (Value.Bool v) v2 = NONE
  | binop_ShiftLeft (Value.Addr v) v2 = NONE
  | binop_ShiftLeft v1 Value.Unit = NONE
  | binop_ShiftLeft v1 Value.Null = NONE
  | binop_ShiftLeft v1 (Value.Bool v) = NONE
  | binop_ShiftLeft v1 (Value.Addr v) = NONE;

fun binop_BinXor (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.bitXOR_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_BinXor (Value.Bool b1) (Value.Bool b2) =
    SOME (Sum_Type.Inl (Value.Bool (not (Product_Type.equal_bool b1 b2))))
  | binop_BinXor Value.Unit v2 = NONE
  | binop_BinXor Value.Null v2 = NONE
  | binop_BinXor (Value.Bool v) Value.Unit = NONE
  | binop_BinXor (Value.Bool v) Value.Null = NONE
  | binop_BinXor (Value.Bool v) (Value.Intg va) = NONE
  | binop_BinXor (Value.Bool v) (Value.Addr va) = NONE
  | binop_BinXor (Value.Addr v) v2 = NONE
  | binop_BinXor v1 Value.Unit = NONE
  | binop_BinXor v1 Value.Null = NONE
  | binop_BinXor (Value.Intg va) (Value.Bool v) = NONE
  | binop_BinXor v1 (Value.Addr v) = NONE;

fun binop_BinOr (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.bitOR_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_BinOr (Value.Bool b1) (Value.Bool b2) =
    SOME (Sum_Type.Inl (Value.Bool (b1 orelse b2)))
  | binop_BinOr Value.Unit v2 = NONE
  | binop_BinOr Value.Null v2 = NONE
  | binop_BinOr (Value.Bool v) Value.Unit = NONE
  | binop_BinOr (Value.Bool v) Value.Null = NONE
  | binop_BinOr (Value.Bool v) (Value.Intg va) = NONE
  | binop_BinOr (Value.Bool v) (Value.Addr va) = NONE
  | binop_BinOr (Value.Addr v) v2 = NONE
  | binop_BinOr v1 Value.Unit = NONE
  | binop_BinOr v1 Value.Null = NONE
  | binop_BinOr (Value.Intg va) (Value.Bool v) = NONE
  | binop_BinOr v1 (Value.Addr v) = NONE;

fun binop_BinAnd (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.bitAND_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_BinAnd (Value.Bool b1) (Value.Bool b2) =
    SOME (Sum_Type.Inl (Value.Bool (b1 andalso b2)))
  | binop_BinAnd Value.Unit v2 = NONE
  | binop_BinAnd Value.Null v2 = NONE
  | binop_BinAnd (Value.Bool v) Value.Unit = NONE
  | binop_BinAnd (Value.Bool v) Value.Null = NONE
  | binop_BinAnd (Value.Bool v) (Value.Intg va) = NONE
  | binop_BinAnd (Value.Bool v) (Value.Addr va) = NONE
  | binop_BinAnd (Value.Addr v) v2 = NONE
  | binop_BinAnd v1 Value.Unit = NONE
  | binop_BinAnd v1 Value.Null = NONE
  | binop_BinAnd (Value.Intg va) (Value.Bool v) = NONE
  | binop_BinAnd v1 (Value.Addr v) = NONE;

fun word_sdiv A_ x y =
  let
    val xa = Word.sint A_ x;
    val ya = Word.sint A_ y;
    val negative =
      not (Product_Type.equal_bool (IntInf.< (xa, (0 : IntInf.int)))
            (IntInf.< (ya, (0 : IntInf.int))));
    val result = Arith.div_int (Arith.abs_int xa) (Arith.abs_int ya);
  in
    Word.Word_of_int (if negative then IntInf.~ result else result)
  end;

fun binop_Div B_ empty_heap new_obj v1 (Value.Addr v) = NONE
  | binop_Div B_ empty_heap new_obj v1 (Value.Bool v) = NONE
  | binop_Div B_ empty_heap new_obj v1 Value.Null = NONE
  | binop_Div B_ empty_heap new_obj v1 Value.Unit = NONE
  | binop_Div B_ empty_heap new_obj (Value.Addr v) v2 = NONE
  | binop_Div B_ empty_heap new_obj (Value.Bool v) v2 = NONE
  | binop_Div B_ empty_heap new_obj Value.Null v2 = NONE
  | binop_Div B_ empty_heap new_obj Value.Unit v2 = NONE
  | binop_Div B_ empty_heap new_obj (Value.Intg i1) (Value.Intg i2) =
    SOME (if Word.equal_word
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0
                       (Type_Length.len0_bit0 Type_Length.len0_num1)))))
               i2 (Word.zero_word
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0 Type_Length.len0_num1))))))
           then Sum_Type.Inr
                  (StartConfig.addr_of_sys_xcpt B_ empty_heap new_obj
                    "java/lang/ArithmeticException")
           else Sum_Type.Inl
                  (Value.Intg
                    (word_sdiv
                      (Type_Length.len_bit0
                        (Type_Length.len_bit0
                          (Type_Length.len_bit0
                            (Type_Length.len_bit0
                              (Type_Length.len_bit0 Type_Length.len_num1)))))
                      i1 i2)));

fun word_smod A_ x y =
  let
    val xa = Word.sint A_ x;
    val ya = Word.sint A_ y;
    val negative = IntInf.< (xa, (0 : IntInf.int));
    val result = Arith.mod_int (Arith.abs_int xa) (Arith.abs_int ya);
  in
    Word.Word_of_int (if negative then IntInf.~ result else result)
  end;

fun binop_Mod B_ empty_heap new_obj v1 (Value.Addr v) = NONE
  | binop_Mod B_ empty_heap new_obj v1 (Value.Bool v) = NONE
  | binop_Mod B_ empty_heap new_obj v1 Value.Null = NONE
  | binop_Mod B_ empty_heap new_obj v1 Value.Unit = NONE
  | binop_Mod B_ empty_heap new_obj (Value.Addr v) v2 = NONE
  | binop_Mod B_ empty_heap new_obj (Value.Bool v) v2 = NONE
  | binop_Mod B_ empty_heap new_obj Value.Null v2 = NONE
  | binop_Mod B_ empty_heap new_obj Value.Unit v2 = NONE
  | binop_Mod B_ empty_heap new_obj (Value.Intg i1) (Value.Intg i2) =
    SOME (if Word.equal_word
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0
                       (Type_Length.len0_bit0 Type_Length.len0_num1)))))
               i2 (Word.zero_word
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0 Type_Length.len0_num1))))))
           then Sum_Type.Inr
                  (StartConfig.addr_of_sys_xcpt B_ empty_heap new_obj
                    "java/lang/ArithmeticException")
           else Sum_Type.Inl
                  (Value.Intg
                    (word_smod
                      (Type_Length.len_bit0
                        (Type_Length.len_bit0
                          (Type_Length.len_bit0
                            (Type_Length.len_bit0
                              (Type_Length.len_bit0 Type_Length.len_num1)))))
                      i1 i2)));

fun binop_Mult (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.times_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_Mult Value.Unit v2 = NONE
  | binop_Mult Value.Null v2 = NONE
  | binop_Mult (Value.Bool v) v2 = NONE
  | binop_Mult (Value.Addr v) v2 = NONE
  | binop_Mult v1 Value.Unit = NONE
  | binop_Mult v1 Value.Null = NONE
  | binop_Mult v1 (Value.Bool v) = NONE
  | binop_Mult v1 (Value.Addr v) = NONE;

fun binop_Subtract (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.minus_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_Subtract Value.Unit v2 = NONE
  | binop_Subtract Value.Null v2 = NONE
  | binop_Subtract (Value.Bool v) v2 = NONE
  | binop_Subtract (Value.Addr v) v2 = NONE
  | binop_Subtract v1 Value.Unit = NONE
  | binop_Subtract v1 Value.Null = NONE
  | binop_Subtract v1 (Value.Bool v) = NONE
  | binop_Subtract v1 (Value.Addr v) = NONE;

fun binop_Add (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.plus_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_Add Value.Unit v2 = NONE
  | binop_Add Value.Null v2 = NONE
  | binop_Add (Value.Bool v) v2 = NONE
  | binop_Add (Value.Addr v) v2 = NONE
  | binop_Add v1 Value.Unit = NONE
  | binop_Add v1 Value.Null = NONE
  | binop_Add v1 (Value.Bool v) = NONE
  | binop_Add v1 (Value.Addr v) = NONE;

fun binop_GreaterOrEqual (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Bool
           (Word.word_sle
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i2 i1)))
  | binop_GreaterOrEqual Value.Unit v2 = NONE
  | binop_GreaterOrEqual Value.Null v2 = NONE
  | binop_GreaterOrEqual (Value.Bool v) v2 = NONE
  | binop_GreaterOrEqual (Value.Addr v) v2 = NONE
  | binop_GreaterOrEqual v1 Value.Unit = NONE
  | binop_GreaterOrEqual v1 Value.Null = NONE
  | binop_GreaterOrEqual v1 (Value.Bool v) = NONE
  | binop_GreaterOrEqual v1 (Value.Addr v) = NONE;

fun binop_GreaterThan (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Bool
           (Word.word_sless
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i2 i1)))
  | binop_GreaterThan Value.Unit v2 = NONE
  | binop_GreaterThan Value.Null v2 = NONE
  | binop_GreaterThan (Value.Bool v) v2 = NONE
  | binop_GreaterThan (Value.Addr v) v2 = NONE
  | binop_GreaterThan v1 Value.Unit = NONE
  | binop_GreaterThan v1 Value.Null = NONE
  | binop_GreaterThan v1 (Value.Bool v) = NONE
  | binop_GreaterThan v1 (Value.Addr v) = NONE;

fun binop_LessOrEqual (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Bool
           (Word.word_sle
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i1 i2)))
  | binop_LessOrEqual Value.Unit v2 = NONE
  | binop_LessOrEqual Value.Null v2 = NONE
  | binop_LessOrEqual (Value.Bool v) v2 = NONE
  | binop_LessOrEqual (Value.Addr v) v2 = NONE
  | binop_LessOrEqual v1 Value.Unit = NONE
  | binop_LessOrEqual v1 Value.Null = NONE
  | binop_LessOrEqual v1 (Value.Bool v) = NONE
  | binop_LessOrEqual v1 (Value.Addr v) = NONE;

fun binop_LessThan (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Bool
           (Word.word_sless
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i1 i2)))
  | binop_LessThan Value.Unit v2 = NONE
  | binop_LessThan Value.Null v2 = NONE
  | binop_LessThan (Value.Bool v) v2 = NONE
  | binop_LessThan (Value.Addr v) v2 = NONE
  | binop_LessThan v1 Value.Unit = NONE
  | binop_LessThan v1 Value.Null = NONE
  | binop_LessThan v1 (Value.Bool v) = NONE
  | binop_LessThan v1 (Value.Addr v) = NONE;

fun binop (B1_, B2_) empty_heap new_obj ShiftRightSigned a b =
  binop_ShiftRightSigned a b
  | binop (B1_, B2_) empty_heap new_obj ShiftRightZeros a b =
    binop_ShiftRightZeros a b
  | binop (B1_, B2_) empty_heap new_obj ShiftLeft a b = binop_ShiftLeft a b
  | binop (B1_, B2_) empty_heap new_obj BinXor a b = binop_BinXor a b
  | binop (B1_, B2_) empty_heap new_obj BinOr a b = binop_BinOr a b
  | binop (B1_, B2_) empty_heap new_obj BinAnd a b = binop_BinAnd a b
  | binop (B1_, B2_) empty_heap new_obj Div a b =
    binop_Div B2_ empty_heap new_obj a b
  | binop (B1_, B2_) empty_heap new_obj Mod a b =
    binop_Mod B2_ empty_heap new_obj a b
  | binop (B1_, B2_) empty_heap new_obj Mult a b = binop_Mult a b
  | binop (B1_, B2_) empty_heap new_obj Subtract a b = binop_Subtract a b
  | binop (B1_, B2_) empty_heap new_obj Add a b = binop_Add a b
  | binop (B1_, B2_) empty_heap new_obj GreaterOrEqual a b =
    binop_GreaterOrEqual a b
  | binop (B1_, B2_) empty_heap new_obj GreaterThan a b = binop_GreaterThan a b
  | binop (B1_, B2_) empty_heap new_obj LessOrEqual a b = binop_LessOrEqual a b
  | binop (B1_, B2_) empty_heap new_obj LessThan a b = binop_LessThan a b
  | binop (B1_, B2_) empty_heap new_obj NotEq v1 v2 =
    SOME (Sum_Type.Inl (Value.Bool (not (Value.equal_vala B1_ v1 v2))))
  | binop (B1_, B2_) empty_heap new_obj Eq v1 v2 =
    SOME (Sum_Type.Inl (Value.Bool (Value.equal_vala B1_ v1 v2)));

fun equal_bop ShiftRightSigned ShiftRightZeros = false
  | equal_bop ShiftRightZeros ShiftRightSigned = false
  | equal_bop ShiftRightSigned ShiftLeft = false
  | equal_bop ShiftLeft ShiftRightSigned = false
  | equal_bop ShiftRightZeros ShiftLeft = false
  | equal_bop ShiftLeft ShiftRightZeros = false
  | equal_bop ShiftRightSigned BinXor = false
  | equal_bop BinXor ShiftRightSigned = false
  | equal_bop ShiftRightZeros BinXor = false
  | equal_bop BinXor ShiftRightZeros = false
  | equal_bop ShiftLeft BinXor = false
  | equal_bop BinXor ShiftLeft = false
  | equal_bop ShiftRightSigned BinOr = false
  | equal_bop BinOr ShiftRightSigned = false
  | equal_bop ShiftRightZeros BinOr = false
  | equal_bop BinOr ShiftRightZeros = false
  | equal_bop ShiftLeft BinOr = false
  | equal_bop BinOr ShiftLeft = false
  | equal_bop BinXor BinOr = false
  | equal_bop BinOr BinXor = false
  | equal_bop ShiftRightSigned BinAnd = false
  | equal_bop BinAnd ShiftRightSigned = false
  | equal_bop ShiftRightZeros BinAnd = false
  | equal_bop BinAnd ShiftRightZeros = false
  | equal_bop ShiftLeft BinAnd = false
  | equal_bop BinAnd ShiftLeft = false
  | equal_bop BinXor BinAnd = false
  | equal_bop BinAnd BinXor = false
  | equal_bop BinOr BinAnd = false
  | equal_bop BinAnd BinOr = false
  | equal_bop ShiftRightSigned Mod = false
  | equal_bop Mod ShiftRightSigned = false
  | equal_bop ShiftRightZeros Mod = false
  | equal_bop Mod ShiftRightZeros = false
  | equal_bop ShiftLeft Mod = false
  | equal_bop Mod ShiftLeft = false
  | equal_bop BinXor Mod = false
  | equal_bop Mod BinXor = false
  | equal_bop BinOr Mod = false
  | equal_bop Mod BinOr = false
  | equal_bop BinAnd Mod = false
  | equal_bop Mod BinAnd = false
  | equal_bop ShiftRightSigned Div = false
  | equal_bop Div ShiftRightSigned = false
  | equal_bop ShiftRightZeros Div = false
  | equal_bop Div ShiftRightZeros = false
  | equal_bop ShiftLeft Div = false
  | equal_bop Div ShiftLeft = false
  | equal_bop BinXor Div = false
  | equal_bop Div BinXor = false
  | equal_bop BinOr Div = false
  | equal_bop Div BinOr = false
  | equal_bop BinAnd Div = false
  | equal_bop Div BinAnd = false
  | equal_bop Mod Div = false
  | equal_bop Div Mod = false
  | equal_bop ShiftRightSigned Mult = false
  | equal_bop Mult ShiftRightSigned = false
  | equal_bop ShiftRightZeros Mult = false
  | equal_bop Mult ShiftRightZeros = false
  | equal_bop ShiftLeft Mult = false
  | equal_bop Mult ShiftLeft = false
  | equal_bop BinXor Mult = false
  | equal_bop Mult BinXor = false
  | equal_bop BinOr Mult = false
  | equal_bop Mult BinOr = false
  | equal_bop BinAnd Mult = false
  | equal_bop Mult BinAnd = false
  | equal_bop Mod Mult = false
  | equal_bop Mult Mod = false
  | equal_bop Div Mult = false
  | equal_bop Mult Div = false
  | equal_bop ShiftRightSigned Subtract = false
  | equal_bop Subtract ShiftRightSigned = false
  | equal_bop ShiftRightZeros Subtract = false
  | equal_bop Subtract ShiftRightZeros = false
  | equal_bop ShiftLeft Subtract = false
  | equal_bop Subtract ShiftLeft = false
  | equal_bop BinXor Subtract = false
  | equal_bop Subtract BinXor = false
  | equal_bop BinOr Subtract = false
  | equal_bop Subtract BinOr = false
  | equal_bop BinAnd Subtract = false
  | equal_bop Subtract BinAnd = false
  | equal_bop Mod Subtract = false
  | equal_bop Subtract Mod = false
  | equal_bop Div Subtract = false
  | equal_bop Subtract Div = false
  | equal_bop Mult Subtract = false
  | equal_bop Subtract Mult = false
  | equal_bop ShiftRightSigned Add = false
  | equal_bop Add ShiftRightSigned = false
  | equal_bop ShiftRightZeros Add = false
  | equal_bop Add ShiftRightZeros = false
  | equal_bop ShiftLeft Add = false
  | equal_bop Add ShiftLeft = false
  | equal_bop BinXor Add = false
  | equal_bop Add BinXor = false
  | equal_bop BinOr Add = false
  | equal_bop Add BinOr = false
  | equal_bop BinAnd Add = false
  | equal_bop Add BinAnd = false
  | equal_bop Mod Add = false
  | equal_bop Add Mod = false
  | equal_bop Div Add = false
  | equal_bop Add Div = false
  | equal_bop Mult Add = false
  | equal_bop Add Mult = false
  | equal_bop Subtract Add = false
  | equal_bop Add Subtract = false
  | equal_bop ShiftRightSigned GreaterOrEqual = false
  | equal_bop GreaterOrEqual ShiftRightSigned = false
  | equal_bop ShiftRightZeros GreaterOrEqual = false
  | equal_bop GreaterOrEqual ShiftRightZeros = false
  | equal_bop ShiftLeft GreaterOrEqual = false
  | equal_bop GreaterOrEqual ShiftLeft = false
  | equal_bop BinXor GreaterOrEqual = false
  | equal_bop GreaterOrEqual BinXor = false
  | equal_bop BinOr GreaterOrEqual = false
  | equal_bop GreaterOrEqual BinOr = false
  | equal_bop BinAnd GreaterOrEqual = false
  | equal_bop GreaterOrEqual BinAnd = false
  | equal_bop Mod GreaterOrEqual = false
  | equal_bop GreaterOrEqual Mod = false
  | equal_bop Div GreaterOrEqual = false
  | equal_bop GreaterOrEqual Div = false
  | equal_bop Mult GreaterOrEqual = false
  | equal_bop GreaterOrEqual Mult = false
  | equal_bop Subtract GreaterOrEqual = false
  | equal_bop GreaterOrEqual Subtract = false
  | equal_bop Add GreaterOrEqual = false
  | equal_bop GreaterOrEqual Add = false
  | equal_bop ShiftRightSigned GreaterThan = false
  | equal_bop GreaterThan ShiftRightSigned = false
  | equal_bop ShiftRightZeros GreaterThan = false
  | equal_bop GreaterThan ShiftRightZeros = false
  | equal_bop ShiftLeft GreaterThan = false
  | equal_bop GreaterThan ShiftLeft = false
  | equal_bop BinXor GreaterThan = false
  | equal_bop GreaterThan BinXor = false
  | equal_bop BinOr GreaterThan = false
  | equal_bop GreaterThan BinOr = false
  | equal_bop BinAnd GreaterThan = false
  | equal_bop GreaterThan BinAnd = false
  | equal_bop Mod GreaterThan = false
  | equal_bop GreaterThan Mod = false
  | equal_bop Div GreaterThan = false
  | equal_bop GreaterThan Div = false
  | equal_bop Mult GreaterThan = false
  | equal_bop GreaterThan Mult = false
  | equal_bop Subtract GreaterThan = false
  | equal_bop GreaterThan Subtract = false
  | equal_bop Add GreaterThan = false
  | equal_bop GreaterThan Add = false
  | equal_bop GreaterOrEqual GreaterThan = false
  | equal_bop GreaterThan GreaterOrEqual = false
  | equal_bop ShiftRightSigned LessOrEqual = false
  | equal_bop LessOrEqual ShiftRightSigned = false
  | equal_bop ShiftRightZeros LessOrEqual = false
  | equal_bop LessOrEqual ShiftRightZeros = false
  | equal_bop ShiftLeft LessOrEqual = false
  | equal_bop LessOrEqual ShiftLeft = false
  | equal_bop BinXor LessOrEqual = false
  | equal_bop LessOrEqual BinXor = false
  | equal_bop BinOr LessOrEqual = false
  | equal_bop LessOrEqual BinOr = false
  | equal_bop BinAnd LessOrEqual = false
  | equal_bop LessOrEqual BinAnd = false
  | equal_bop Mod LessOrEqual = false
  | equal_bop LessOrEqual Mod = false
  | equal_bop Div LessOrEqual = false
  | equal_bop LessOrEqual Div = false
  | equal_bop Mult LessOrEqual = false
  | equal_bop LessOrEqual Mult = false
  | equal_bop Subtract LessOrEqual = false
  | equal_bop LessOrEqual Subtract = false
  | equal_bop Add LessOrEqual = false
  | equal_bop LessOrEqual Add = false
  | equal_bop GreaterOrEqual LessOrEqual = false
  | equal_bop LessOrEqual GreaterOrEqual = false
  | equal_bop GreaterThan LessOrEqual = false
  | equal_bop LessOrEqual GreaterThan = false
  | equal_bop ShiftRightSigned LessThan = false
  | equal_bop LessThan ShiftRightSigned = false
  | equal_bop ShiftRightZeros LessThan = false
  | equal_bop LessThan ShiftRightZeros = false
  | equal_bop ShiftLeft LessThan = false
  | equal_bop LessThan ShiftLeft = false
  | equal_bop BinXor LessThan = false
  | equal_bop LessThan BinXor = false
  | equal_bop BinOr LessThan = false
  | equal_bop LessThan BinOr = false
  | equal_bop BinAnd LessThan = false
  | equal_bop LessThan BinAnd = false
  | equal_bop Mod LessThan = false
  | equal_bop LessThan Mod = false
  | equal_bop Div LessThan = false
  | equal_bop LessThan Div = false
  | equal_bop Mult LessThan = false
  | equal_bop LessThan Mult = false
  | equal_bop Subtract LessThan = false
  | equal_bop LessThan Subtract = false
  | equal_bop Add LessThan = false
  | equal_bop LessThan Add = false
  | equal_bop GreaterOrEqual LessThan = false
  | equal_bop LessThan GreaterOrEqual = false
  | equal_bop GreaterThan LessThan = false
  | equal_bop LessThan GreaterThan = false
  | equal_bop LessOrEqual LessThan = false
  | equal_bop LessThan LessOrEqual = false
  | equal_bop ShiftRightSigned NotEq = false
  | equal_bop NotEq ShiftRightSigned = false
  | equal_bop ShiftRightZeros NotEq = false
  | equal_bop NotEq ShiftRightZeros = false
  | equal_bop ShiftLeft NotEq = false
  | equal_bop NotEq ShiftLeft = false
  | equal_bop BinXor NotEq = false
  | equal_bop NotEq BinXor = false
  | equal_bop BinOr NotEq = false
  | equal_bop NotEq BinOr = false
  | equal_bop BinAnd NotEq = false
  | equal_bop NotEq BinAnd = false
  | equal_bop Mod NotEq = false
  | equal_bop NotEq Mod = false
  | equal_bop Div NotEq = false
  | equal_bop NotEq Div = false
  | equal_bop Mult NotEq = false
  | equal_bop NotEq Mult = false
  | equal_bop Subtract NotEq = false
  | equal_bop NotEq Subtract = false
  | equal_bop Add NotEq = false
  | equal_bop NotEq Add = false
  | equal_bop GreaterOrEqual NotEq = false
  | equal_bop NotEq GreaterOrEqual = false
  | equal_bop GreaterThan NotEq = false
  | equal_bop NotEq GreaterThan = false
  | equal_bop LessOrEqual NotEq = false
  | equal_bop NotEq LessOrEqual = false
  | equal_bop LessThan NotEq = false
  | equal_bop NotEq LessThan = false
  | equal_bop ShiftRightSigned Eq = false
  | equal_bop Eq ShiftRightSigned = false
  | equal_bop ShiftRightZeros Eq = false
  | equal_bop Eq ShiftRightZeros = false
  | equal_bop ShiftLeft Eq = false
  | equal_bop Eq ShiftLeft = false
  | equal_bop BinXor Eq = false
  | equal_bop Eq BinXor = false
  | equal_bop BinOr Eq = false
  | equal_bop Eq BinOr = false
  | equal_bop BinAnd Eq = false
  | equal_bop Eq BinAnd = false
  | equal_bop Mod Eq = false
  | equal_bop Eq Mod = false
  | equal_bop Div Eq = false
  | equal_bop Eq Div = false
  | equal_bop Mult Eq = false
  | equal_bop Eq Mult = false
  | equal_bop Subtract Eq = false
  | equal_bop Eq Subtract = false
  | equal_bop Add Eq = false
  | equal_bop Eq Add = false
  | equal_bop GreaterOrEqual Eq = false
  | equal_bop Eq GreaterOrEqual = false
  | equal_bop GreaterThan Eq = false
  | equal_bop Eq GreaterThan = false
  | equal_bop LessOrEqual Eq = false
  | equal_bop Eq LessOrEqual = false
  | equal_bop LessThan Eq = false
  | equal_bop Eq LessThan = false
  | equal_bop NotEq Eq = false
  | equal_bop Eq NotEq = false
  | equal_bop ShiftRightSigned ShiftRightSigned = true
  | equal_bop ShiftRightZeros ShiftRightZeros = true
  | equal_bop ShiftLeft ShiftLeft = true
  | equal_bop BinXor BinXor = true
  | equal_bop BinOr BinOr = true
  | equal_bop BinAnd BinAnd = true
  | equal_bop Mod Mod = true
  | equal_bop Div Div = true
  | equal_bop Mult Mult = true
  | equal_bop Subtract Subtract = true
  | equal_bop Add Add = true
  | equal_bop GreaterOrEqual GreaterOrEqual = true
  | equal_bop GreaterThan GreaterThan = true
  | equal_bop LessOrEqual LessOrEqual = true
  | equal_bop LessThan LessThan = true
  | equal_bop NotEq NotEq = true
  | equal_bop Eq Eq = true;

fun wT_binop_i_i_i_i_i x xa xb xc xd =
  Predicate.sup_pred
    (case xb
      of Eq =>
        (case xd of Type.Void => Predicate.bot_pred
          | Type.Boolean =>
            Predicate.bind
              (Predicate.if_pred
                (TypeRel.widen x xa xc orelse TypeRel.widen x xc xa))
              (fn () => Predicate.single ())
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
      | NotEq => Predicate.bot_pred | LessThan => Predicate.bot_pred
      | LessOrEqual => Predicate.bot_pred | GreaterThan => Predicate.bot_pred
      | GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
      | Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
      | Div => Predicate.bot_pred | Mod => Predicate.bot_pred
      | BinAnd => Predicate.bot_pred | BinOr => Predicate.bot_pred
      | BinXor => Predicate.bot_pred | ShiftLeft => Predicate.bot_pred
      | ShiftRightZeros => Predicate.bot_pred
      | ShiftRightSigned => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb of Eq => Predicate.bot_pred
        | NotEq =>
          (case xd of Type.Void => Predicate.bot_pred
            | Type.Boolean =>
              Predicate.bind
                (Predicate.if_pred
                  (TypeRel.widen x xa xc orelse TypeRel.widen x xc xa))
                (fn () => Predicate.single ())
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.bot_pred)
        | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
        | GreaterThan => Predicate.bot_pred
        | GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
        | Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
        | Div => Predicate.bot_pred | Mod => Predicate.bot_pred
        | BinAnd => Predicate.bot_pred | BinOr => Predicate.bot_pred
        | BinXor => Predicate.bot_pred | ShiftLeft => Predicate.bot_pred
        | ShiftRightZeros => Predicate.bot_pred
        | ShiftRightSigned => Predicate.bot_pred)
      (Predicate.sup_pred
        (case xa of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer =>
            (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
              | LessThan =>
                (case xc of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer =>
                    (case xd of Type.Void => Predicate.bot_pred
                      | Type.Boolean => Predicate.single ()
                      | Type.Integer => Predicate.bot_pred
                      | Type.Nt => Predicate.bot_pred
                      | Type.Class _ => Predicate.bot_pred
                      | Type.Array _ => Predicate.bot_pred)
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | LessOrEqual => Predicate.bot_pred
              | GreaterThan => Predicate.bot_pred
              | GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
              | Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
              | Div => Predicate.bot_pred | Mod => Predicate.bot_pred
              | BinAnd => Predicate.bot_pred | BinOr => Predicate.bot_pred
              | BinXor => Predicate.bot_pred | ShiftLeft => Predicate.bot_pred
              | ShiftRightZeros => Predicate.bot_pred
              | ShiftRightSigned => Predicate.bot_pred)
          | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case xa of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer =>
              (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
                | LessThan => Predicate.bot_pred
                | LessOrEqual =>
                  (case xc of Type.Void => Predicate.bot_pred
                    | Type.Boolean => Predicate.bot_pred
                    | Type.Integer =>
                      (case xd of Type.Void => Predicate.bot_pred
                        | Type.Boolean => Predicate.single ()
                        | Type.Integer => Predicate.bot_pred
                        | Type.Nt => Predicate.bot_pred
                        | Type.Class _ => Predicate.bot_pred
                        | Type.Array _ => Predicate.bot_pred)
                    | Type.Nt => Predicate.bot_pred
                    | Type.Class _ => Predicate.bot_pred
                    | Type.Array _ => Predicate.bot_pred)
                | GreaterThan => Predicate.bot_pred
                | GreaterOrEqual => Predicate.bot_pred
                | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
                | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
                | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
                | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
                | ShiftLeft => Predicate.bot_pred
                | ShiftRightZeros => Predicate.bot_pred
                | ShiftRightSigned => Predicate.bot_pred)
            | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case xa of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xb of Eq => Predicate.bot_pred
                  | NotEq => Predicate.bot_pred | LessThan => Predicate.bot_pred
                  | LessOrEqual => Predicate.bot_pred
                  | GreaterThan =>
                    (case xc of Type.Void => Predicate.bot_pred
                      | Type.Boolean => Predicate.bot_pred
                      | Type.Integer =>
                        (case xd of Type.Void => Predicate.bot_pred
                          | Type.Boolean => Predicate.single ()
                          | Type.Integer => Predicate.bot_pred
                          | Type.Nt => Predicate.bot_pred
                          | Type.Class _ => Predicate.bot_pred
                          | Type.Array _ => Predicate.bot_pred)
                      | Type.Nt => Predicate.bot_pred
                      | Type.Class _ => Predicate.bot_pred
                      | Type.Array _ => Predicate.bot_pred)
                  | GreaterOrEqual => Predicate.bot_pred
                  | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
                  | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
                  | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
                  | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
                  | ShiftLeft => Predicate.bot_pred
                  | ShiftRightZeros => Predicate.bot_pred
                  | ShiftRightSigned => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
            (Predicate.sup_pred
              (case xa of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer =>
                  (case xb of Eq => Predicate.bot_pred
                    | NotEq => Predicate.bot_pred
                    | LessThan => Predicate.bot_pred
                    | LessOrEqual => Predicate.bot_pred
                    | GreaterThan => Predicate.bot_pred
                    | GreaterOrEqual =>
                      (case xc of Type.Void => Predicate.bot_pred
                        | Type.Boolean => Predicate.bot_pred
                        | Type.Integer =>
                          (case xd of Type.Void => Predicate.bot_pred
                            | Type.Boolean => Predicate.single ()
                            | Type.Integer => Predicate.bot_pred
                            | Type.Nt => Predicate.bot_pred
                            | Type.Class _ => Predicate.bot_pred
                            | Type.Array _ => Predicate.bot_pred)
                        | Type.Nt => Predicate.bot_pred
                        | Type.Class _ => Predicate.bot_pred
                        | Type.Array _ => Predicate.bot_pred)
                    | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
                    | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
                    | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
                    | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
                    | ShiftLeft => Predicate.bot_pred
                    | ShiftRightZeros => Predicate.bot_pred
                    | ShiftRightSigned => Predicate.bot_pred)
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.bot_pred)
              (Predicate.sup_pred
                (case xa of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer =>
                    (case xb of Eq => Predicate.bot_pred
                      | NotEq => Predicate.bot_pred
                      | LessThan => Predicate.bot_pred
                      | LessOrEqual => Predicate.bot_pred
                      | GreaterThan => Predicate.bot_pred
                      | GreaterOrEqual => Predicate.bot_pred
                      | Add =>
                        (case xc of Type.Void => Predicate.bot_pred
                          | Type.Boolean => Predicate.bot_pred
                          | Type.Integer =>
                            (case xd of Type.Void => Predicate.bot_pred
                              | Type.Boolean => Predicate.bot_pred
                              | Type.Integer => Predicate.single ()
                              | Type.Nt => Predicate.bot_pred
                              | Type.Class _ => Predicate.bot_pred
                              | Type.Array _ => Predicate.bot_pred)
                          | Type.Nt => Predicate.bot_pred
                          | Type.Class _ => Predicate.bot_pred
                          | Type.Array _ => Predicate.bot_pred)
                      | Subtract => Predicate.bot_pred
                      | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
                      | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
                      | BinOr => Predicate.bot_pred
                      | BinXor => Predicate.bot_pred
                      | ShiftLeft => Predicate.bot_pred
                      | ShiftRightZeros => Predicate.bot_pred
                      | ShiftRightSigned => Predicate.bot_pred)
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
                (Predicate.sup_pred
                  (case xa of Type.Void => Predicate.bot_pred
                    | Type.Boolean => Predicate.bot_pred
                    | Type.Integer =>
                      (case xb of Eq => Predicate.bot_pred
                        | NotEq => Predicate.bot_pred
                        | LessThan => Predicate.bot_pred
                        | LessOrEqual => Predicate.bot_pred
                        | GreaterThan => Predicate.bot_pred
                        | GreaterOrEqual => Predicate.bot_pred
                        | Add => Predicate.bot_pred
                        | Subtract =>
                          (case xc of Type.Void => Predicate.bot_pred
                            | Type.Boolean => Predicate.bot_pred
                            | Type.Integer =>
                              (case xd of Type.Void => Predicate.bot_pred
                                | Type.Boolean => Predicate.bot_pred
                                | Type.Integer => Predicate.single ()
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                            | Type.Nt => Predicate.bot_pred
                            | Type.Class _ => Predicate.bot_pred
                            | Type.Array _ => Predicate.bot_pred)
                        | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
                        | Mod => Predicate.bot_pred
                        | BinAnd => Predicate.bot_pred
                        | BinOr => Predicate.bot_pred
                        | BinXor => Predicate.bot_pred
                        | ShiftLeft => Predicate.bot_pred
                        | ShiftRightZeros => Predicate.bot_pred
                        | ShiftRightSigned => Predicate.bot_pred)
                    | Type.Nt => Predicate.bot_pred
                    | Type.Class _ => Predicate.bot_pred
                    | Type.Array _ => Predicate.bot_pred)
                  (Predicate.sup_pred
                    (case xa of Type.Void => Predicate.bot_pred
                      | Type.Boolean => Predicate.bot_pred
                      | Type.Integer =>
                        (case xb of Eq => Predicate.bot_pred
                          | NotEq => Predicate.bot_pred
                          | LessThan => Predicate.bot_pred
                          | LessOrEqual => Predicate.bot_pred
                          | GreaterThan => Predicate.bot_pred
                          | GreaterOrEqual => Predicate.bot_pred
                          | Add => Predicate.bot_pred
                          | Subtract => Predicate.bot_pred
                          | Mult =>
                            (case xc of Type.Void => Predicate.bot_pred
                              | Type.Boolean => Predicate.bot_pred
                              | Type.Integer =>
                                (case xd of Type.Void => Predicate.bot_pred
                                  | Type.Boolean => Predicate.bot_pred
                                  | Type.Integer => Predicate.single ()
                                  | Type.Nt => Predicate.bot_pred
                                  | Type.Class _ => Predicate.bot_pred
                                  | Type.Array _ => Predicate.bot_pred)
                              | Type.Nt => Predicate.bot_pred
                              | Type.Class _ => Predicate.bot_pred
                              | Type.Array _ => Predicate.bot_pred)
                          | Div => Predicate.bot_pred
                          | Mod => Predicate.bot_pred
                          | BinAnd => Predicate.bot_pred
                          | BinOr => Predicate.bot_pred
                          | BinXor => Predicate.bot_pred
                          | ShiftLeft => Predicate.bot_pred
                          | ShiftRightZeros => Predicate.bot_pred
                          | ShiftRightSigned => Predicate.bot_pred)
                      | Type.Nt => Predicate.bot_pred
                      | Type.Class _ => Predicate.bot_pred
                      | Type.Array _ => Predicate.bot_pred)
                    (Predicate.sup_pred
                      (case xa of Type.Void => Predicate.bot_pred
                        | Type.Boolean => Predicate.bot_pred
                        | Type.Integer =>
                          (case xb of Eq => Predicate.bot_pred
                            | NotEq => Predicate.bot_pred
                            | LessThan => Predicate.bot_pred
                            | LessOrEqual => Predicate.bot_pred
                            | GreaterThan => Predicate.bot_pred
                            | GreaterOrEqual => Predicate.bot_pred
                            | Add => Predicate.bot_pred
                            | Subtract => Predicate.bot_pred
                            | Mult => Predicate.bot_pred
                            | Div =>
                              (case xc of Type.Void => Predicate.bot_pred
                                | Type.Boolean => Predicate.bot_pred
                                | Type.Integer =>
                                  (case xd of Type.Void => Predicate.bot_pred
                                    | Type.Boolean => Predicate.bot_pred
                                    | Type.Integer => Predicate.single ()
                                    | Type.Nt => Predicate.bot_pred
                                    | Type.Class _ => Predicate.bot_pred
                                    | Type.Array _ => Predicate.bot_pred)
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                            | Mod => Predicate.bot_pred
                            | BinAnd => Predicate.bot_pred
                            | BinOr => Predicate.bot_pred
                            | BinXor => Predicate.bot_pred
                            | ShiftLeft => Predicate.bot_pred
                            | ShiftRightZeros => Predicate.bot_pred
                            | ShiftRightSigned => Predicate.bot_pred)
                        | Type.Nt => Predicate.bot_pred
                        | Type.Class _ => Predicate.bot_pred
                        | Type.Array _ => Predicate.bot_pred)
                      (Predicate.sup_pred
                        (case xa of Type.Void => Predicate.bot_pred
                          | Type.Boolean => Predicate.bot_pred
                          | Type.Integer =>
                            (case xb of Eq => Predicate.bot_pred
                              | NotEq => Predicate.bot_pred
                              | LessThan => Predicate.bot_pred
                              | LessOrEqual => Predicate.bot_pred
                              | GreaterThan => Predicate.bot_pred
                              | GreaterOrEqual => Predicate.bot_pred
                              | Add => Predicate.bot_pred
                              | Subtract => Predicate.bot_pred
                              | Mult => Predicate.bot_pred
                              | Div => Predicate.bot_pred
                              | Mod =>
                                (case xc of Type.Void => Predicate.bot_pred
                                  | Type.Boolean => Predicate.bot_pred
                                  | Type.Integer =>
                                    (case xd of Type.Void => Predicate.bot_pred
                                      | Type.Boolean => Predicate.bot_pred
                                      | Type.Integer => Predicate.single ()
                                      | Type.Nt => Predicate.bot_pred
                                      | Type.Class _ => Predicate.bot_pred
                                      | Type.Array _ => Predicate.bot_pred)
                                  | Type.Nt => Predicate.bot_pred
                                  | Type.Class _ => Predicate.bot_pred
                                  | Type.Array _ => Predicate.bot_pred)
                              | BinAnd => Predicate.bot_pred
                              | BinOr => Predicate.bot_pred
                              | BinXor => Predicate.bot_pred
                              | ShiftLeft => Predicate.bot_pred
                              | ShiftRightZeros => Predicate.bot_pred
                              | ShiftRightSigned => Predicate.bot_pred)
                          | Type.Nt => Predicate.bot_pred
                          | Type.Class _ => Predicate.bot_pred
                          | Type.Array _ => Predicate.bot_pred)
                        (Predicate.sup_pred
                          (case xa of Type.Void => Predicate.bot_pred
                            | Type.Boolean =>
                              (case xb of Eq => Predicate.bot_pred
                                | NotEq => Predicate.bot_pred
                                | LessThan => Predicate.bot_pred
                                | LessOrEqual => Predicate.bot_pred
                                | GreaterThan => Predicate.bot_pred
                                | GreaterOrEqual => Predicate.bot_pred
                                | Add => Predicate.bot_pred
                                | Subtract => Predicate.bot_pred
                                | Mult => Predicate.bot_pred
                                | Div => Predicate.bot_pred
                                | Mod => Predicate.bot_pred
                                | BinAnd =>
                                  (case xc of Type.Void => Predicate.bot_pred
                                    | Type.Boolean =>
                                      (case xd
of Type.Void => Predicate.bot_pred | Type.Boolean => Predicate.single ()
| Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
| Type.Class _ => Predicate.bot_pred | Type.Array _ => Predicate.bot_pred)
                                    | Type.Integer => Predicate.bot_pred
                                    | Type.Nt => Predicate.bot_pred
                                    | Type.Class _ => Predicate.bot_pred
                                    | Type.Array _ => Predicate.bot_pred)
                                | BinOr => Predicate.bot_pred
                                | BinXor => Predicate.bot_pred
                                | ShiftLeft => Predicate.bot_pred
                                | ShiftRightZeros => Predicate.bot_pred
                                | ShiftRightSigned => Predicate.bot_pred)
                            | Type.Integer => Predicate.bot_pred
                            | Type.Nt => Predicate.bot_pred
                            | Type.Class _ => Predicate.bot_pred
                            | Type.Array _ => Predicate.bot_pred)
                          (Predicate.sup_pred
                            (case xa of Type.Void => Predicate.bot_pred
                              | Type.Boolean => Predicate.bot_pred
                              | Type.Integer =>
                                (case xb of Eq => Predicate.bot_pred
                                  | NotEq => Predicate.bot_pred
                                  | LessThan => Predicate.bot_pred
                                  | LessOrEqual => Predicate.bot_pred
                                  | GreaterThan => Predicate.bot_pred
                                  | GreaterOrEqual => Predicate.bot_pred
                                  | Add => Predicate.bot_pred
                                  | Subtract => Predicate.bot_pred
                                  | Mult => Predicate.bot_pred
                                  | Div => Predicate.bot_pred
                                  | Mod => Predicate.bot_pred
                                  | BinAnd =>
                                    (case xc of Type.Void => Predicate.bot_pred
                                      | Type.Boolean => Predicate.bot_pred
                                      | Type.Integer =>
(case xd of Type.Void => Predicate.bot_pred | Type.Boolean => Predicate.bot_pred
  | Type.Integer => Predicate.single () | Type.Nt => Predicate.bot_pred
  | Type.Class _ => Predicate.bot_pred | Type.Array _ => Predicate.bot_pred)
                                      | Type.Nt => Predicate.bot_pred
                                      | Type.Class _ => Predicate.bot_pred
                                      | Type.Array _ => Predicate.bot_pred)
                                  | BinOr => Predicate.bot_pred
                                  | BinXor => Predicate.bot_pred
                                  | ShiftLeft => Predicate.bot_pred
                                  | ShiftRightZeros => Predicate.bot_pred
                                  | ShiftRightSigned => Predicate.bot_pred)
                              | Type.Nt => Predicate.bot_pred
                              | Type.Class _ => Predicate.bot_pred
                              | Type.Array _ => Predicate.bot_pred)
                            (Predicate.sup_pred
                              (case xa of Type.Void => Predicate.bot_pred
                                | Type.Boolean =>
                                  (case xb of Eq => Predicate.bot_pred
                                    | NotEq => Predicate.bot_pred
                                    | LessThan => Predicate.bot_pred
                                    | LessOrEqual => Predicate.bot_pred
                                    | GreaterThan => Predicate.bot_pred
                                    | GreaterOrEqual => Predicate.bot_pred
                                    | Add => Predicate.bot_pred
                                    | Subtract => Predicate.bot_pred
                                    | Mult => Predicate.bot_pred
                                    | Div => Predicate.bot_pred
                                    | Mod => Predicate.bot_pred
                                    | BinAnd => Predicate.bot_pred
                                    | BinOr =>
                                      (case xc
of Type.Void => Predicate.bot_pred
| Type.Boolean =>
  (case xd of Type.Void => Predicate.bot_pred
    | Type.Boolean => Predicate.single () | Type.Integer => Predicate.bot_pred
    | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
    | Type.Array _ => Predicate.bot_pred)
| Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
| Type.Class _ => Predicate.bot_pred | Type.Array _ => Predicate.bot_pred)
                                    | BinXor => Predicate.bot_pred
                                    | ShiftLeft => Predicate.bot_pred
                                    | ShiftRightZeros => Predicate.bot_pred
                                    | ShiftRightSigned => Predicate.bot_pred)
                                | Type.Integer => Predicate.bot_pred
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                              (Predicate.sup_pred
                                (case xa of Type.Void => Predicate.bot_pred
                                  | Type.Boolean => Predicate.bot_pred
                                  | Type.Integer =>
                                    (case xb of Eq => Predicate.bot_pred
                                      | NotEq => Predicate.bot_pred
                                      | LessThan => Predicate.bot_pred
                                      | LessOrEqual => Predicate.bot_pred
                                      | GreaterThan => Predicate.bot_pred
                                      | GreaterOrEqual => Predicate.bot_pred
                                      | Add => Predicate.bot_pred
                                      | Subtract => Predicate.bot_pred
                                      | Mult => Predicate.bot_pred
                                      | Div => Predicate.bot_pred
                                      | Mod => Predicate.bot_pred
                                      | BinAnd => Predicate.bot_pred
                                      | BinOr =>
(case xc of Type.Void => Predicate.bot_pred | Type.Boolean => Predicate.bot_pred
  | Type.Integer =>
    (case xd of Type.Void => Predicate.bot_pred
      | Type.Boolean => Predicate.bot_pred | Type.Integer => Predicate.single ()
      | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
      | Type.Array _ => Predicate.bot_pred)
  | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
  | Type.Array _ => Predicate.bot_pred)
                                      | BinXor => Predicate.bot_pred
                                      | ShiftLeft => Predicate.bot_pred
                                      | ShiftRightZeros => Predicate.bot_pred
                                      | ShiftRightSigned => Predicate.bot_pred)
                                  | Type.Nt => Predicate.bot_pred
                                  | Type.Class _ => Predicate.bot_pred
                                  | Type.Array _ => Predicate.bot_pred)
                                (Predicate.sup_pred
                                  (case xa of Type.Void => Predicate.bot_pred
                                    | Type.Boolean =>
                                      (case xb of Eq => Predicate.bot_pred
| NotEq => Predicate.bot_pred | LessThan => Predicate.bot_pred
| LessOrEqual => Predicate.bot_pred | GreaterThan => Predicate.bot_pred
| GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
| Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
| Div => Predicate.bot_pred | Mod => Predicate.bot_pred
| BinAnd => Predicate.bot_pred | BinOr => Predicate.bot_pred
| BinXor =>
  (case xc of Type.Void => Predicate.bot_pred
    | Type.Boolean =>
      (case xd of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.single ()
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
    | Type.Class _ => Predicate.bot_pred | Type.Array _ => Predicate.bot_pred)
| ShiftLeft => Predicate.bot_pred | ShiftRightZeros => Predicate.bot_pred
| ShiftRightSigned => Predicate.bot_pred)
                                    | Type.Integer => Predicate.bot_pred
                                    | Type.Nt => Predicate.bot_pred
                                    | Type.Class _ => Predicate.bot_pred
                                    | Type.Array _ => Predicate.bot_pred)
                                  (Predicate.sup_pred
                                    (case xa of Type.Void => Predicate.bot_pred
                                      | Type.Boolean => Predicate.bot_pred
                                      | Type.Integer =>
(case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
  | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
  | GreaterThan => Predicate.bot_pred | GreaterOrEqual => Predicate.bot_pred
  | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
  | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
  | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
  | BinOr => Predicate.bot_pred
  | BinXor =>
    (case xc of Type.Void => Predicate.bot_pred
      | Type.Boolean => Predicate.bot_pred
      | Type.Integer =>
        (case xd of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.single () | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
      | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
      | Type.Array _ => Predicate.bot_pred)
  | ShiftLeft => Predicate.bot_pred | ShiftRightZeros => Predicate.bot_pred
  | ShiftRightSigned => Predicate.bot_pred)
                                      | Type.Nt => Predicate.bot_pred
                                      | Type.Class _ => Predicate.bot_pred
                                      | Type.Array _ => Predicate.bot_pred)
                                    (Predicate.sup_pred
                                      (case xa
of Type.Void => Predicate.bot_pred | Type.Boolean => Predicate.bot_pred
| Type.Integer =>
  (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
    | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
    | GreaterThan => Predicate.bot_pred | GreaterOrEqual => Predicate.bot_pred
    | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
    | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
    | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
    | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
    | ShiftLeft =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer =>
          (case xd of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.single ()
            | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.bot_pred)
        | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | ShiftRightZeros => Predicate.bot_pred
    | ShiftRightSigned => Predicate.bot_pred)
| Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
| Type.Array _ => Predicate.bot_pred)
                                      (Predicate.sup_pred
(case xa of Type.Void => Predicate.bot_pred | Type.Boolean => Predicate.bot_pred
  | Type.Integer =>
    (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
      | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
      | GreaterThan => Predicate.bot_pred | GreaterOrEqual => Predicate.bot_pred
      | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
      | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
      | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
      | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
      | ShiftLeft => Predicate.bot_pred
      | ShiftRightZeros =>
        (case xc of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer =>
            (case xd of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single ()
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
      | ShiftRightSigned => Predicate.bot_pred)
  | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
  | Type.Array _ => Predicate.bot_pred)
(case xa of Type.Void => Predicate.bot_pred | Type.Boolean => Predicate.bot_pred
  | Type.Integer =>
    (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
      | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
      | GreaterThan => Predicate.bot_pred | GreaterOrEqual => Predicate.bot_pred
      | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
      | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
      | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
      | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
      | ShiftLeft => Predicate.bot_pred | ShiftRightZeros => Predicate.bot_pred
      | ShiftRightSigned =>
        (case xc of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer =>
            (case xd of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single ()
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred))
  | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
  | Type.Array _ => Predicate.bot_pred)))))))))))))))))));

fun wT_binop_i_i_i_i_o x xa xb xc =
  Predicate.sup_pred
    (case xb
      of Eq =>
        Predicate.bind
          (Predicate.if_pred
            (TypeRel.widen x xa xc orelse TypeRel.widen x xc xa))
          (fn () => Predicate.single Type.Boolean)
      | NotEq => Predicate.bot_pred | LessThan => Predicate.bot_pred
      | LessOrEqual => Predicate.bot_pred | GreaterThan => Predicate.bot_pred
      | GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
      | Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
      | Div => Predicate.bot_pred | Mod => Predicate.bot_pred
      | BinAnd => Predicate.bot_pred | BinOr => Predicate.bot_pred
      | BinXor => Predicate.bot_pred | ShiftLeft => Predicate.bot_pred
      | ShiftRightZeros => Predicate.bot_pred
      | ShiftRightSigned => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb of Eq => Predicate.bot_pred
        | NotEq =>
          Predicate.bind
            (Predicate.if_pred
              (TypeRel.widen x xa xc orelse TypeRel.widen x xc xa))
            (fn () => Predicate.single Type.Boolean)
        | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
        | GreaterThan => Predicate.bot_pred
        | GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
        | Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
        | Div => Predicate.bot_pred | Mod => Predicate.bot_pred
        | BinAnd => Predicate.bot_pred | BinOr => Predicate.bot_pred
        | BinXor => Predicate.bot_pred | ShiftLeft => Predicate.bot_pred
        | ShiftRightZeros => Predicate.bot_pred
        | ShiftRightSigned => Predicate.bot_pred)
      (Predicate.sup_pred
        (case xa of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer =>
            (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
              | LessThan =>
                (case xc of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single Type.Boolean
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | LessOrEqual => Predicate.bot_pred
              | GreaterThan => Predicate.bot_pred
              | GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
              | Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
              | Div => Predicate.bot_pred | Mod => Predicate.bot_pred
              | BinAnd => Predicate.bot_pred | BinOr => Predicate.bot_pred
              | BinXor => Predicate.bot_pred | ShiftLeft => Predicate.bot_pred
              | ShiftRightZeros => Predicate.bot_pred
              | ShiftRightSigned => Predicate.bot_pred)
          | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case xa of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer =>
              (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
                | LessThan => Predicate.bot_pred
                | LessOrEqual =>
                  (case xc of Type.Void => Predicate.bot_pred
                    | Type.Boolean => Predicate.bot_pred
                    | Type.Integer => Predicate.single Type.Boolean
                    | Type.Nt => Predicate.bot_pred
                    | Type.Class _ => Predicate.bot_pred
                    | Type.Array _ => Predicate.bot_pred)
                | GreaterThan => Predicate.bot_pred
                | GreaterOrEqual => Predicate.bot_pred
                | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
                | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
                | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
                | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
                | ShiftLeft => Predicate.bot_pred
                | ShiftRightZeros => Predicate.bot_pred
                | ShiftRightSigned => Predicate.bot_pred)
            | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case xa of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xb of Eq => Predicate.bot_pred
                  | NotEq => Predicate.bot_pred | LessThan => Predicate.bot_pred
                  | LessOrEqual => Predicate.bot_pred
                  | GreaterThan =>
                    (case xc of Type.Void => Predicate.bot_pred
                      | Type.Boolean => Predicate.bot_pred
                      | Type.Integer => Predicate.single Type.Boolean
                      | Type.Nt => Predicate.bot_pred
                      | Type.Class _ => Predicate.bot_pred
                      | Type.Array _ => Predicate.bot_pred)
                  | GreaterOrEqual => Predicate.bot_pred
                  | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
                  | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
                  | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
                  | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
                  | ShiftLeft => Predicate.bot_pred
                  | ShiftRightZeros => Predicate.bot_pred
                  | ShiftRightSigned => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
            (Predicate.sup_pred
              (case xa of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer =>
                  (case xb of Eq => Predicate.bot_pred
                    | NotEq => Predicate.bot_pred
                    | LessThan => Predicate.bot_pred
                    | LessOrEqual => Predicate.bot_pred
                    | GreaterThan => Predicate.bot_pred
                    | GreaterOrEqual =>
                      (case xc of Type.Void => Predicate.bot_pred
                        | Type.Boolean => Predicate.bot_pred
                        | Type.Integer => Predicate.single Type.Boolean
                        | Type.Nt => Predicate.bot_pred
                        | Type.Class _ => Predicate.bot_pred
                        | Type.Array _ => Predicate.bot_pred)
                    | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
                    | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
                    | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
                    | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
                    | ShiftLeft => Predicate.bot_pred
                    | ShiftRightZeros => Predicate.bot_pred
                    | ShiftRightSigned => Predicate.bot_pred)
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.bot_pred)
              (Predicate.sup_pred
                (case xa of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer =>
                    (case xb of Eq => Predicate.bot_pred
                      | NotEq => Predicate.bot_pred
                      | LessThan => Predicate.bot_pred
                      | LessOrEqual => Predicate.bot_pred
                      | GreaterThan => Predicate.bot_pred
                      | GreaterOrEqual => Predicate.bot_pred
                      | Add =>
                        (case xc of Type.Void => Predicate.bot_pred
                          | Type.Boolean => Predicate.bot_pred
                          | Type.Integer => Predicate.single Type.Integer
                          | Type.Nt => Predicate.bot_pred
                          | Type.Class _ => Predicate.bot_pred
                          | Type.Array _ => Predicate.bot_pred)
                      | Subtract => Predicate.bot_pred
                      | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
                      | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
                      | BinOr => Predicate.bot_pred
                      | BinXor => Predicate.bot_pred
                      | ShiftLeft => Predicate.bot_pred
                      | ShiftRightZeros => Predicate.bot_pred
                      | ShiftRightSigned => Predicate.bot_pred)
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
                (Predicate.sup_pred
                  (case xa of Type.Void => Predicate.bot_pred
                    | Type.Boolean => Predicate.bot_pred
                    | Type.Integer =>
                      (case xb of Eq => Predicate.bot_pred
                        | NotEq => Predicate.bot_pred
                        | LessThan => Predicate.bot_pred
                        | LessOrEqual => Predicate.bot_pred
                        | GreaterThan => Predicate.bot_pred
                        | GreaterOrEqual => Predicate.bot_pred
                        | Add => Predicate.bot_pred
                        | Subtract =>
                          (case xc of Type.Void => Predicate.bot_pred
                            | Type.Boolean => Predicate.bot_pred
                            | Type.Integer => Predicate.single Type.Integer
                            | Type.Nt => Predicate.bot_pred
                            | Type.Class _ => Predicate.bot_pred
                            | Type.Array _ => Predicate.bot_pred)
                        | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
                        | Mod => Predicate.bot_pred
                        | BinAnd => Predicate.bot_pred
                        | BinOr => Predicate.bot_pred
                        | BinXor => Predicate.bot_pred
                        | ShiftLeft => Predicate.bot_pred
                        | ShiftRightZeros => Predicate.bot_pred
                        | ShiftRightSigned => Predicate.bot_pred)
                    | Type.Nt => Predicate.bot_pred
                    | Type.Class _ => Predicate.bot_pred
                    | Type.Array _ => Predicate.bot_pred)
                  (Predicate.sup_pred
                    (case xa of Type.Void => Predicate.bot_pred
                      | Type.Boolean => Predicate.bot_pred
                      | Type.Integer =>
                        (case xb of Eq => Predicate.bot_pred
                          | NotEq => Predicate.bot_pred
                          | LessThan => Predicate.bot_pred
                          | LessOrEqual => Predicate.bot_pred
                          | GreaterThan => Predicate.bot_pred
                          | GreaterOrEqual => Predicate.bot_pred
                          | Add => Predicate.bot_pred
                          | Subtract => Predicate.bot_pred
                          | Mult =>
                            (case xc of Type.Void => Predicate.bot_pred
                              | Type.Boolean => Predicate.bot_pred
                              | Type.Integer => Predicate.single Type.Integer
                              | Type.Nt => Predicate.bot_pred
                              | Type.Class _ => Predicate.bot_pred
                              | Type.Array _ => Predicate.bot_pred)
                          | Div => Predicate.bot_pred
                          | Mod => Predicate.bot_pred
                          | BinAnd => Predicate.bot_pred
                          | BinOr => Predicate.bot_pred
                          | BinXor => Predicate.bot_pred
                          | ShiftLeft => Predicate.bot_pred
                          | ShiftRightZeros => Predicate.bot_pred
                          | ShiftRightSigned => Predicate.bot_pred)
                      | Type.Nt => Predicate.bot_pred
                      | Type.Class _ => Predicate.bot_pred
                      | Type.Array _ => Predicate.bot_pred)
                    (Predicate.sup_pred
                      (case xa of Type.Void => Predicate.bot_pred
                        | Type.Boolean => Predicate.bot_pred
                        | Type.Integer =>
                          (case xb of Eq => Predicate.bot_pred
                            | NotEq => Predicate.bot_pred
                            | LessThan => Predicate.bot_pred
                            | LessOrEqual => Predicate.bot_pred
                            | GreaterThan => Predicate.bot_pred
                            | GreaterOrEqual => Predicate.bot_pred
                            | Add => Predicate.bot_pred
                            | Subtract => Predicate.bot_pred
                            | Mult => Predicate.bot_pred
                            | Div =>
                              (case xc of Type.Void => Predicate.bot_pred
                                | Type.Boolean => Predicate.bot_pred
                                | Type.Integer => Predicate.single Type.Integer
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                            | Mod => Predicate.bot_pred
                            | BinAnd => Predicate.bot_pred
                            | BinOr => Predicate.bot_pred
                            | BinXor => Predicate.bot_pred
                            | ShiftLeft => Predicate.bot_pred
                            | ShiftRightZeros => Predicate.bot_pred
                            | ShiftRightSigned => Predicate.bot_pred)
                        | Type.Nt => Predicate.bot_pred
                        | Type.Class _ => Predicate.bot_pred
                        | Type.Array _ => Predicate.bot_pred)
                      (Predicate.sup_pred
                        (case xa of Type.Void => Predicate.bot_pred
                          | Type.Boolean => Predicate.bot_pred
                          | Type.Integer =>
                            (case xb of Eq => Predicate.bot_pred
                              | NotEq => Predicate.bot_pred
                              | LessThan => Predicate.bot_pred
                              | LessOrEqual => Predicate.bot_pred
                              | GreaterThan => Predicate.bot_pred
                              | GreaterOrEqual => Predicate.bot_pred
                              | Add => Predicate.bot_pred
                              | Subtract => Predicate.bot_pred
                              | Mult => Predicate.bot_pred
                              | Div => Predicate.bot_pred
                              | Mod =>
                                (case xc of Type.Void => Predicate.bot_pred
                                  | Type.Boolean => Predicate.bot_pred
                                  | Type.Integer =>
                                    Predicate.single Type.Integer
                                  | Type.Nt => Predicate.bot_pred
                                  | Type.Class _ => Predicate.bot_pred
                                  | Type.Array _ => Predicate.bot_pred)
                              | BinAnd => Predicate.bot_pred
                              | BinOr => Predicate.bot_pred
                              | BinXor => Predicate.bot_pred
                              | ShiftLeft => Predicate.bot_pred
                              | ShiftRightZeros => Predicate.bot_pred
                              | ShiftRightSigned => Predicate.bot_pred)
                          | Type.Nt => Predicate.bot_pred
                          | Type.Class _ => Predicate.bot_pred
                          | Type.Array _ => Predicate.bot_pred)
                        (Predicate.sup_pred
                          (case xa of Type.Void => Predicate.bot_pred
                            | Type.Boolean =>
                              (case xb of Eq => Predicate.bot_pred
                                | NotEq => Predicate.bot_pred
                                | LessThan => Predicate.bot_pred
                                | LessOrEqual => Predicate.bot_pred
                                | GreaterThan => Predicate.bot_pred
                                | GreaterOrEqual => Predicate.bot_pred
                                | Add => Predicate.bot_pred
                                | Subtract => Predicate.bot_pred
                                | Mult => Predicate.bot_pred
                                | Div => Predicate.bot_pred
                                | Mod => Predicate.bot_pred
                                | BinAnd =>
                                  (case xc of Type.Void => Predicate.bot_pred
                                    | Type.Boolean =>
                                      Predicate.single Type.Boolean
                                    | Type.Integer => Predicate.bot_pred
                                    | Type.Nt => Predicate.bot_pred
                                    | Type.Class _ => Predicate.bot_pred
                                    | Type.Array _ => Predicate.bot_pred)
                                | BinOr => Predicate.bot_pred
                                | BinXor => Predicate.bot_pred
                                | ShiftLeft => Predicate.bot_pred
                                | ShiftRightZeros => Predicate.bot_pred
                                | ShiftRightSigned => Predicate.bot_pred)
                            | Type.Integer => Predicate.bot_pred
                            | Type.Nt => Predicate.bot_pred
                            | Type.Class _ => Predicate.bot_pred
                            | Type.Array _ => Predicate.bot_pred)
                          (Predicate.sup_pred
                            (case xa of Type.Void => Predicate.bot_pred
                              | Type.Boolean => Predicate.bot_pred
                              | Type.Integer =>
                                (case xb of Eq => Predicate.bot_pred
                                  | NotEq => Predicate.bot_pred
                                  | LessThan => Predicate.bot_pred
                                  | LessOrEqual => Predicate.bot_pred
                                  | GreaterThan => Predicate.bot_pred
                                  | GreaterOrEqual => Predicate.bot_pred
                                  | Add => Predicate.bot_pred
                                  | Subtract => Predicate.bot_pred
                                  | Mult => Predicate.bot_pred
                                  | Div => Predicate.bot_pred
                                  | Mod => Predicate.bot_pred
                                  | BinAnd =>
                                    (case xc of Type.Void => Predicate.bot_pred
                                      | Type.Boolean => Predicate.bot_pred
                                      | Type.Integer =>
Predicate.single Type.Integer
                                      | Type.Nt => Predicate.bot_pred
                                      | Type.Class _ => Predicate.bot_pred
                                      | Type.Array _ => Predicate.bot_pred)
                                  | BinOr => Predicate.bot_pred
                                  | BinXor => Predicate.bot_pred
                                  | ShiftLeft => Predicate.bot_pred
                                  | ShiftRightZeros => Predicate.bot_pred
                                  | ShiftRightSigned => Predicate.bot_pred)
                              | Type.Nt => Predicate.bot_pred
                              | Type.Class _ => Predicate.bot_pred
                              | Type.Array _ => Predicate.bot_pred)
                            (Predicate.sup_pred
                              (case xa of Type.Void => Predicate.bot_pred
                                | Type.Boolean =>
                                  (case xb of Eq => Predicate.bot_pred
                                    | NotEq => Predicate.bot_pred
                                    | LessThan => Predicate.bot_pred
                                    | LessOrEqual => Predicate.bot_pred
                                    | GreaterThan => Predicate.bot_pred
                                    | GreaterOrEqual => Predicate.bot_pred
                                    | Add => Predicate.bot_pred
                                    | Subtract => Predicate.bot_pred
                                    | Mult => Predicate.bot_pred
                                    | Div => Predicate.bot_pred
                                    | Mod => Predicate.bot_pred
                                    | BinAnd => Predicate.bot_pred
                                    | BinOr =>
                                      (case xc
of Type.Void => Predicate.bot_pred
| Type.Boolean => Predicate.single Type.Boolean
| Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
| Type.Class _ => Predicate.bot_pred | Type.Array _ => Predicate.bot_pred)
                                    | BinXor => Predicate.bot_pred
                                    | ShiftLeft => Predicate.bot_pred
                                    | ShiftRightZeros => Predicate.bot_pred
                                    | ShiftRightSigned => Predicate.bot_pred)
                                | Type.Integer => Predicate.bot_pred
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                              (Predicate.sup_pred
                                (case xa of Type.Void => Predicate.bot_pred
                                  | Type.Boolean => Predicate.bot_pred
                                  | Type.Integer =>
                                    (case xb of Eq => Predicate.bot_pred
                                      | NotEq => Predicate.bot_pred
                                      | LessThan => Predicate.bot_pred
                                      | LessOrEqual => Predicate.bot_pred
                                      | GreaterThan => Predicate.bot_pred
                                      | GreaterOrEqual => Predicate.bot_pred
                                      | Add => Predicate.bot_pred
                                      | Subtract => Predicate.bot_pred
                                      | Mult => Predicate.bot_pred
                                      | Div => Predicate.bot_pred
                                      | Mod => Predicate.bot_pred
                                      | BinAnd => Predicate.bot_pred
                                      | BinOr =>
(case xc of Type.Void => Predicate.bot_pred | Type.Boolean => Predicate.bot_pred
  | Type.Integer => Predicate.single Type.Integer
  | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
  | Type.Array _ => Predicate.bot_pred)
                                      | BinXor => Predicate.bot_pred
                                      | ShiftLeft => Predicate.bot_pred
                                      | ShiftRightZeros => Predicate.bot_pred
                                      | ShiftRightSigned => Predicate.bot_pred)
                                  | Type.Nt => Predicate.bot_pred
                                  | Type.Class _ => Predicate.bot_pred
                                  | Type.Array _ => Predicate.bot_pred)
                                (Predicate.sup_pred
                                  (case xa of Type.Void => Predicate.bot_pred
                                    | Type.Boolean =>
                                      (case xb of Eq => Predicate.bot_pred
| NotEq => Predicate.bot_pred | LessThan => Predicate.bot_pred
| LessOrEqual => Predicate.bot_pred | GreaterThan => Predicate.bot_pred
| GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
| Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
| Div => Predicate.bot_pred | Mod => Predicate.bot_pred
| BinAnd => Predicate.bot_pred | BinOr => Predicate.bot_pred
| BinXor =>
  (case xc of Type.Void => Predicate.bot_pred
    | Type.Boolean => Predicate.single Type.Boolean
    | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
    | Type.Class _ => Predicate.bot_pred | Type.Array _ => Predicate.bot_pred)
| ShiftLeft => Predicate.bot_pred | ShiftRightZeros => Predicate.bot_pred
| ShiftRightSigned => Predicate.bot_pred)
                                    | Type.Integer => Predicate.bot_pred
                                    | Type.Nt => Predicate.bot_pred
                                    | Type.Class _ => Predicate.bot_pred
                                    | Type.Array _ => Predicate.bot_pred)
                                  (Predicate.sup_pred
                                    (case xa of Type.Void => Predicate.bot_pred
                                      | Type.Boolean => Predicate.bot_pred
                                      | Type.Integer =>
(case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
  | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
  | GreaterThan => Predicate.bot_pred | GreaterOrEqual => Predicate.bot_pred
  | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
  | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
  | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
  | BinOr => Predicate.bot_pred
  | BinXor =>
    (case xc of Type.Void => Predicate.bot_pred
      | Type.Boolean => Predicate.bot_pred
      | Type.Integer => Predicate.single Type.Integer
      | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
      | Type.Array _ => Predicate.bot_pred)
  | ShiftLeft => Predicate.bot_pred | ShiftRightZeros => Predicate.bot_pred
  | ShiftRightSigned => Predicate.bot_pred)
                                      | Type.Nt => Predicate.bot_pred
                                      | Type.Class _ => Predicate.bot_pred
                                      | Type.Array _ => Predicate.bot_pred)
                                    (Predicate.sup_pred
                                      (case xa
of Type.Void => Predicate.bot_pred | Type.Boolean => Predicate.bot_pred
| Type.Integer =>
  (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
    | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
    | GreaterThan => Predicate.bot_pred | GreaterOrEqual => Predicate.bot_pred
    | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
    | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
    | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
    | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
    | ShiftLeft =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.single Type.Integer
        | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | ShiftRightZeros => Predicate.bot_pred
    | ShiftRightSigned => Predicate.bot_pred)
| Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
| Type.Array _ => Predicate.bot_pred)
                                      (Predicate.sup_pred
(case xa of Type.Void => Predicate.bot_pred | Type.Boolean => Predicate.bot_pred
  | Type.Integer =>
    (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
      | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
      | GreaterThan => Predicate.bot_pred | GreaterOrEqual => Predicate.bot_pred
      | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
      | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
      | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
      | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
      | ShiftLeft => Predicate.bot_pred
      | ShiftRightZeros =>
        (case xc of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.single Type.Integer
          | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
      | ShiftRightSigned => Predicate.bot_pred)
  | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
  | Type.Array _ => Predicate.bot_pred)
(case xa of Type.Void => Predicate.bot_pred | Type.Boolean => Predicate.bot_pred
  | Type.Integer =>
    (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
      | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
      | GreaterThan => Predicate.bot_pred | GreaterOrEqual => Predicate.bot_pred
      | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
      | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
      | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
      | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
      | ShiftLeft => Predicate.bot_pred | ShiftRightZeros => Predicate.bot_pred
      | ShiftRightSigned =>
        (case xc of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.single Type.Integer
          | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred))
  | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
  | Type.Array _ => Predicate.bot_pred)))))))))))))))))));

end; (*struct BinOp*)

structure Expr : sig
  datatype ('a, 'b, 'c) exp = New of string |
    NewArray of Type.ty * ('a, 'b, 'c) exp | Cast of Type.ty * ('a, 'b, 'c) exp
    | InstanceOf of ('a, 'b, 'c) exp * Type.ty | Val of 'c Value.vala |
    BinOp of ('a, 'b, 'c) exp * BinOp.bop * ('a, 'b, 'c) exp | Var of 'a |
    LAss of 'a * ('a, 'b, 'c) exp | AAcc of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp
    | AAss of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    ALen of ('a, 'b, 'c) exp | FAcc of ('a, 'b, 'c) exp * string * string |
    FAss of ('a, 'b, 'c) exp * string * string * ('a, 'b, 'c) exp |
    Call of ('a, 'b, 'c) exp * string * ('a, 'b, 'c) exp list |
    Block of 'a * Type.ty * 'c Value.vala option * ('a, 'b, 'c) exp |
    Synchronized of 'b * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    InSynchronized of 'b * 'c * ('a, 'b, 'c) exp |
    Seq of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    Cond of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    While of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp | Throw of ('a, 'b, 'c) exp |
    TryCatch of ('a, 'b, 'c) exp * string * 'a * ('a, 'b, 'c) exp
  val final_i : ('a, 'b, 'c) exp -> unit Predicate.pred
  val final : ('a, 'b, 'c) exp -> bool
  val blocks :
    'a list ->
      Type.ty list -> 'b Value.vala list -> ('a, 'c, 'b) exp -> ('a, 'c, 'b) exp
  val is_val_i : ('a, 'b, 'c) exp -> unit Predicate.pred
  val is_val : ('a, 'b, 'c) exp -> bool
  val is_vals : ('a, 'b, 'c) exp list -> bool
  val the_Val : ('a, 'b, 'c) exp -> 'c Value.vala
  val equal_exp :
    'a HOL.equal -> 'b HOL.equal -> 'c HOL.equal -> ('a, 'b, 'c) exp HOL.equal
  val equal_expa :
    'a HOL.equal -> 'b HOL.equal -> 'c HOL.equal ->
      ('a, 'b, 'c) exp -> ('a, 'b, 'c) exp -> bool
  val extRet2J :
    ('a, 'b, 'c) exp -> 'c ExternalCall.extCallRet -> ('a, 'b, 'c) exp
  val is_Throws_i : ('a, 'b, 'c) exp list -> unit Predicate.pred
end = struct

datatype ('a, 'b, 'c) exp = New of string |
  NewArray of Type.ty * ('a, 'b, 'c) exp | Cast of Type.ty * ('a, 'b, 'c) exp |
  InstanceOf of ('a, 'b, 'c) exp * Type.ty | Val of 'c Value.vala |
  BinOp of ('a, 'b, 'c) exp * BinOp.bop * ('a, 'b, 'c) exp | Var of 'a |
  LAss of 'a * ('a, 'b, 'c) exp | AAcc of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  AAss of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  ALen of ('a, 'b, 'c) exp | FAcc of ('a, 'b, 'c) exp * string * string |
  FAss of ('a, 'b, 'c) exp * string * string * ('a, 'b, 'c) exp |
  Call of ('a, 'b, 'c) exp * string * ('a, 'b, 'c) exp list |
  Block of 'a * Type.ty * 'c Value.vala option * ('a, 'b, 'c) exp |
  Synchronized of 'b * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  InSynchronized of 'b * 'c * ('a, 'b, 'c) exp |
  Seq of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  Cond of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  While of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp | Throw of ('a, 'b, 'c) exp |
  TryCatch of ('a, 'b, 'c) exp * string * 'a * ('a, 'b, 'c) exp;

fun final_i x =
  (case x of New _ => Predicate.bot_pred | NewArray (_, _) => Predicate.bot_pred
    | Cast (_, _) => Predicate.bot_pred
    | InstanceOf (_, _) => Predicate.bot_pred | Val _ => Predicate.single ()
    | BinOp (_, _, _) => Predicate.bot_pred | Var _ => Predicate.bot_pred
    | LAss (_, _) => Predicate.bot_pred | AAcc (_, _) => Predicate.bot_pred
    | AAss (_, _, _) => Predicate.bot_pred | ALen _ => Predicate.bot_pred
    | FAcc (_, _, _) => Predicate.bot_pred
    | FAss (_, _, _, _) => Predicate.bot_pred
    | Call (_, _, _) => Predicate.bot_pred
    | Block (_, _, _, _) => Predicate.bot_pred
    | Synchronized (_, _, _) => Predicate.bot_pred
    | InSynchronized (_, _, _) => Predicate.bot_pred
    | Seq (_, _) => Predicate.bot_pred | Cond (_, _, _) => Predicate.bot_pred
    | While (_, _) => Predicate.bot_pred | Throw (New _) => Predicate.bot_pred
    | Throw (NewArray (_, _)) => Predicate.bot_pred
    | Throw (Cast (_, _)) => Predicate.bot_pred
    | Throw (InstanceOf (_, _)) => Predicate.bot_pred
    | Throw (Val Value.Unit) => Predicate.bot_pred
    | Throw (Val Value.Null) => Predicate.bot_pred
    | Throw (Val (Value.Bool _)) => Predicate.bot_pred
    | Throw (Val (Value.Intg _)) => Predicate.bot_pred
    | Throw (Val (Value.Addr _)) => Predicate.single ()
    | Throw (BinOp (_, _, _)) => Predicate.bot_pred
    | Throw (Var _) => Predicate.bot_pred
    | Throw (LAss (_, _)) => Predicate.bot_pred
    | Throw (AAcc (_, _)) => Predicate.bot_pred
    | Throw (AAss (_, _, _)) => Predicate.bot_pred
    | Throw (ALen _) => Predicate.bot_pred
    | Throw (FAcc (_, _, _)) => Predicate.bot_pred
    | Throw (FAss (_, _, _, _)) => Predicate.bot_pred
    | Throw (Call (_, _, _)) => Predicate.bot_pred
    | Throw (Block (_, _, _, _)) => Predicate.bot_pred
    | Throw (Synchronized (_, _, _)) => Predicate.bot_pred
    | Throw (InSynchronized (_, _, _)) => Predicate.bot_pred
    | Throw (Seq (_, _)) => Predicate.bot_pred
    | Throw (Cond (_, _, _)) => Predicate.bot_pred
    | Throw (While (_, _)) => Predicate.bot_pred
    | Throw (Throw _) => Predicate.bot_pred
    | Throw (TryCatch (_, _, _, _)) => Predicate.bot_pred
    | TryCatch (_, _, _, _) => Predicate.bot_pred);

fun final x1 = Predicate.holds (final_i x1);

fun blocks (va :: vsa) (t :: ts) (v :: vs) e =
  Block (va, t, SOME v, blocks vsa ts vs e)
  | blocks [] [] [] e = e;

fun is_val_i x =
  (case x of New _ => Predicate.bot_pred | NewArray (_, _) => Predicate.bot_pred
    | Cast (_, _) => Predicate.bot_pred
    | InstanceOf (_, _) => Predicate.bot_pred | Val _ => Predicate.single ()
    | BinOp (_, _, _) => Predicate.bot_pred | Var _ => Predicate.bot_pred
    | LAss (_, _) => Predicate.bot_pred | AAcc (_, _) => Predicate.bot_pred
    | AAss (_, _, _) => Predicate.bot_pred | ALen _ => Predicate.bot_pred
    | FAcc (_, _, _) => Predicate.bot_pred
    | FAss (_, _, _, _) => Predicate.bot_pred
    | Call (_, _, _) => Predicate.bot_pred
    | Block (_, _, _, _) => Predicate.bot_pred
    | Synchronized (_, _, _) => Predicate.bot_pred
    | InSynchronized (_, _, _) => Predicate.bot_pred
    | Seq (_, _) => Predicate.bot_pred | Cond (_, _, _) => Predicate.bot_pred
    | While (_, _) => Predicate.bot_pred | Throw _ => Predicate.bot_pred
    | TryCatch (_, _, _, _) => Predicate.bot_pred);

fun is_val x1 = Predicate.holds (is_val_i x1);

fun is_vals [] = true
  | is_vals (e :: es) = is_val e andalso is_vals es;

fun the_Val (Val v) = v;

fun equal_exp A_ B_ C_ = {equal = equal_expa A_ B_ C_} :
  ('a, 'b, 'c) exp HOL.equal
and equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (Throw exp) (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a)) (Seq (exp1, exp2))
    = false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (TryCatch (exp1, literal, a, exp2))
    = false
  | equal_expa A_ B_ C_ (Throw exp) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw expa) (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (InSynchronized (b, addr, exp)) =
    false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp)) (While (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp))
    (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (InSynchronized (b, addr, exp)) =
    false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp)) (Seq (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (Synchronized (b, exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a)) (While (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a))
    (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (Synchronized (b, exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a)) (Seq (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (InSynchronized (ba, addr, exp))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Synchronized (ba, exp1, exp2))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, aa, exp2))
    (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Block (aa, ty, optiona, exp))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw expa) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, expa)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (Block (a, ty, optiona, exp)) =
    false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (While (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (Block (a, ty, optiona, exp))
    = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (Cond (exp1, exp2, exp3))
    = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa))
    (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, expa))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2))
    (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literala, a, exp2))
    (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (Call (exp, literala, lista))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw expa) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (Call (exp, literal, lista)) =
    false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (While (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (Call (exp, literal, lista)) =
    false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (Cond (exp1, exp2, exp3)) =
    false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa))
    (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2))
    (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, expa))
    (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista))
    (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (FAss (exp1, literal1, literal2, exp2)) =
    false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (Throw exp) =
    false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2))
    (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2))
    (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw expa) (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (FAcc (exp, literal1, literal2)) =
    false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2)) (While (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2))
    (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (FAcc (exp, literal1, literal2)) =
    false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2)) (Seq (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, expa))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2))
    (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2))
    (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1a, literal2a, exp2))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1a, literal2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen exp) (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw expa) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen expa) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen exp) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen exp) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen exp) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen expa) (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen exp) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, expa)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen expa) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen expa) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (ALen exp) =
    false
  | equal_expa A_ B_ C_ (ALen exp) (FAss (exp1, literal1, literal2, exp2)) =
    false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen expa) (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3a)) (AAss (exp1, exp2, exp3)) =
    false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3a)) (Cond (exp1, exp2, exp3)) =
    false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp))
    (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a))
    (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (AAss (exp1, exp2, exp3))
    = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (Block (a, ty, optiona, exp))
    = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (AAss (exp1, exp2, exp3)) =
    false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (Call (exp, literal, lista)) =
    false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2))
    (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (ALen exp) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (ALen exp) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp)) (AAcc (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (InSynchronized (b, addr, exp)) =
    false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a)) (AAcc (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (Synchronized (b, exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (AAcc (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (Block (a, ty, optiona, exp)) =
    false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2)) (AAcc (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (FAcc (exp, literal1, literal2)) =
    false
  | equal_expa A_ B_ C_ (ALen exp) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, aa, exp2)) (LAss (a, exp)) =
    false
  | equal_expa A_ B_ C_ (LAss (aa, exp)) (TryCatch (exp1, literal, a, exp2)) =
    false
  | equal_expa A_ B_ C_ (Throw expa) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (aa, ty, optiona, expa)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (aa, expa)) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (LAss (a, exp))
    = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (FAss (exp1, literal1, literal2, exp2))
    = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2)) (LAss (a, exp)) =
    false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (FAcc (exp, literal1, literal2)) =
    false
  | equal_expa A_ B_ C_ (ALen expa) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, aa, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var aa) (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (aa, ty, optiona, exp)) (Var a) = false
  | equal_expa A_ B_ C_ (Var aa) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (ALen exp) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (aa, exp)) (Var a) = false
  | equal_expa A_ B_ C_ (Var aa) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3)) (BinOp (exp1, bop, exp2)) =
    false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a)) (Cond (exp1, exp2, exp3)) =
    false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp))
    (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a))
    (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (BinOp (exp1, bop, exp2))
    = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (Block (a, ty, optiona, exp))
    = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (BinOp (exp1, bop, exp2)) =
    false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (Call (exp, literal, lista)) =
    false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2))
    (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (ALen exp) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3)) (BinOp (exp1, bop, exp2)) =
    false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a)) (AAss (exp1, exp2, exp3)) =
    false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (Var a) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (Val vala) =
    false
  | equal_expa A_ B_ C_ (Val vala) (FAss (exp1, literal1, literal2, exp2)) =
    false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (ALen exp) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (Var a) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Var a) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2))
    (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw expa) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, ty)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa)) (InstanceOf (exp, ty))
    = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, ty)) (InSynchronized (b, addr, exp))
    = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (InstanceOf (exp, ty)) =
    false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (Synchronized (b, exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Block (a, tya, optiona, expa)) (InstanceOf (exp, ty)) =
    false
  | equal_expa A_ B_ C_ (InstanceOf (expa, tya)) (Block (a, ty, optiona, exp)) =
    false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista)) (InstanceOf (exp, ty)) =
    false
  | equal_expa A_ B_ C_ (InstanceOf (expa, ty)) (Call (exp, literal, lista)) =
    false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2))
    (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2)) (InstanceOf (exp, ty))
    = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, ty)) (FAcc (exp, literal1, literal2))
    = false
  | equal_expa A_ B_ C_ (ALen expa) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, ty)) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, ty)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (Var a) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (Var a) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (Val vala) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (Val vala) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2)) (Cast (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (TryCatch (exp1, literal, a, exp2)) =
    false
  | equal_expa A_ B_ C_ (Throw expa) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, expa)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa)) (Cast (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (Cast (ty, expa)) (InSynchronized (b, addr, exp)) =
    false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, tya, optiona, expa)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (tya, expa)) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, expa)) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (Cast (ty, exp))
    = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (FAss (exp1, literal1, literal2, exp2))
    = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2)) (Cast (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (Cast (ty, expa)) (FAcc (exp, literal1, literal2)) =
    false
  | equal_expa A_ B_ C_ (ALen expa) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, expa)) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, expa)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (Var a) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (Var a) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (Val vala) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (Val vala) = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, tya)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (tya, expa)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2)) (NewArray (ty, exp))
    = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (TryCatch (exp1, literal, a, exp2))
    = false
  | equal_expa A_ B_ C_ (Throw expa) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, expa)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa)) (NewArray (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (NewArray (ty, expa)) (InSynchronized (b, addr, exp)) =
    false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (NewArray (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (Synchronized (b, exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Block (a, tya, optiona, expa)) (NewArray (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (NewArray (tya, expa)) (Block (a, ty, optiona, exp)) =
    false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista)) (NewArray (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (NewArray (ty, expa)) (Call (exp, literal, lista)) =
    false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2))
    (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2)) (NewArray (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (NewArray (ty, expa)) (FAcc (exp, literal1, literal2)) =
    false
  | equal_expa A_ B_ C_ (ALen expa) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, expa)) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, expa)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (Var a) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (Var a) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (Val vala) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (Val vala) = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, tya)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (tya, expa)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (Cast (tya, expa)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (tya, expa)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literala, a, exp2)) (New literal) =
    false
  | equal_expa A_ B_ C_ (New literala) (TryCatch (exp1, literal, a, exp2)) =
    false
  | equal_expa A_ B_ C_ (Throw exp) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (exp, literala, lista)) (New literal) = false
  | equal_expa A_ B_ C_ (New literala) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (New literal) =
    false
  | equal_expa A_ B_ C_ (New literal) (FAss (exp1, literal1, literal2, exp2)) =
    false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (ALen exp) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (Var a) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Var a) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (Val vala) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Val vala) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literala, aa, exp2a))
    (TryCatch (exp1, literal, a, exp2)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso
      (((literala : string) = literal) andalso
        (HOL.eq A_ aa a andalso equal_expa A_ B_ C_ exp2a exp2))
  | equal_expa A_ B_ C_ (Throw expa) (Throw exp) = equal_expa A_ B_ C_ expa exp
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (While (exp1, exp2)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso equal_expa A_ B_ C_ exp2a exp2
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3a)) (Cond (exp1, exp2, exp3)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso
      (equal_expa A_ B_ C_ exp2a exp2 andalso equal_expa A_ B_ C_ exp3a exp3)
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (Seq (exp1, exp2)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso equal_expa A_ B_ C_ exp2a exp2
  | equal_expa A_ B_ C_ (InSynchronized (ba, addra, expa))
    (InSynchronized (b, addr, exp)) =
    HOL.eq B_ ba b andalso
      (HOL.eq C_ addra addr andalso equal_expa A_ B_ C_ expa exp)
  | equal_expa A_ B_ C_ (Synchronized (ba, exp1a, exp2a))
    (Synchronized (b, exp1, exp2)) =
    HOL.eq B_ ba b andalso
      (equal_expa A_ B_ C_ exp1a exp1 andalso equal_expa A_ B_ C_ exp2a exp2)
  | equal_expa A_ B_ C_ (Block (aa, tya, optionaa, expa))
    (Block (a, ty, optiona, exp)) =
    HOL.eq A_ aa a andalso
      (Type.equal_tya tya ty andalso
        (Option.equal_optiona (Value.equal_val C_) optionaa optiona andalso
          equal_expa A_ B_ C_ expa exp))
  | equal_expa A_ B_ C_ (Call (expa, literala, listaa))
    (Call (exp, literal, lista)) =
    equal_expa A_ B_ C_ expa exp andalso
      (((literala : string) = literal) andalso
        List.equal_lista (equal_exp A_ B_ C_) listaa lista)
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1a, literal2a, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso
      (((literal1a : string) = literal1) andalso
        (((literal2a : string) = literal2) andalso
          equal_expa A_ B_ C_ exp2a exp2))
  | equal_expa A_ B_ C_ (FAcc (expa, literal1a, literal2a))
    (FAcc (exp, literal1, literal2)) =
    equal_expa A_ B_ C_ expa exp andalso
      (((literal1a : string) = literal1) andalso
        ((literal2a : string) = literal2))
  | equal_expa A_ B_ C_ (ALen expa) (ALen exp) = equal_expa A_ B_ C_ expa exp
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3a)) (AAss (exp1, exp2, exp3)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso
      (equal_expa A_ B_ C_ exp2a exp2 andalso equal_expa A_ B_ C_ exp3a exp3)
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (AAcc (exp1, exp2)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso equal_expa A_ B_ C_ exp2a exp2
  | equal_expa A_ B_ C_ (LAss (aa, expa)) (LAss (a, exp)) =
    HOL.eq A_ aa a andalso equal_expa A_ B_ C_ expa exp
  | equal_expa A_ B_ C_ (Var aa) (Var a) = HOL.eq A_ aa a
  | equal_expa A_ B_ C_ (BinOp (exp1a, bopa, exp2a)) (BinOp (exp1, bop, exp2)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso
      (BinOp.equal_bop bopa bop andalso equal_expa A_ B_ C_ exp2a exp2)
  | equal_expa A_ B_ C_ (Val valaa) (Val vala) = Value.equal_vala C_ valaa vala
  | equal_expa A_ B_ C_ (InstanceOf (expa, tya)) (InstanceOf (exp, ty)) =
    equal_expa A_ B_ C_ expa exp andalso Type.equal_tya tya ty
  | equal_expa A_ B_ C_ (Cast (tya, expa)) (Cast (ty, exp)) =
    Type.equal_tya tya ty andalso equal_expa A_ B_ C_ expa exp
  | equal_expa A_ B_ C_ (NewArray (tya, expa)) (NewArray (ty, exp)) =
    Type.equal_tya tya ty andalso equal_expa A_ B_ C_ expa exp
  | equal_expa A_ B_ C_ (New literala) (New literal) =
    ((literala : string) = literal);

fun extRet2J e (ExternalCall.RetVal v) = Val v
  | extRet2J e (ExternalCall.RetExc a) = Throw (Val (Value.Addr a))
  | extRet2J e ExternalCall.RetStaySame = e;

fun is_Throws_i x =
  Predicate.sup_pred
    (case x of [] => Predicate.bot_pred | New _ :: _ => Predicate.bot_pred
      | NewArray (_, _) :: _ => Predicate.bot_pred
      | Cast (_, _) :: _ => Predicate.bot_pred
      | InstanceOf (_, _) :: _ => Predicate.bot_pred
      | Val _ :: _ => Predicate.bot_pred
      | BinOp (_, _, _) :: _ => Predicate.bot_pred
      | Var _ :: _ => Predicate.bot_pred
      | LAss (_, _) :: _ => Predicate.bot_pred
      | AAcc (_, _) :: _ => Predicate.bot_pred
      | AAss (_, _, _) :: _ => Predicate.bot_pred
      | ALen _ :: _ => Predicate.bot_pred
      | FAcc (_, _, _) :: _ => Predicate.bot_pred
      | FAss (_, _, _, _) :: _ => Predicate.bot_pred
      | Call (_, _, _) :: _ => Predicate.bot_pred
      | Block (_, _, _, _) :: _ => Predicate.bot_pred
      | Synchronized (_, _, _) :: _ => Predicate.bot_pred
      | InSynchronized (_, _, _) :: _ => Predicate.bot_pred
      | Seq (_, _) :: _ => Predicate.bot_pred
      | Cond (_, _, _) :: _ => Predicate.bot_pred
      | While (_, _) :: _ => Predicate.bot_pred
      | Throw (New _) :: _ => Predicate.bot_pred
      | Throw (NewArray (_, _)) :: _ => Predicate.bot_pred
      | Throw (Cast (_, _)) :: _ => Predicate.bot_pred
      | Throw (InstanceOf (_, _)) :: _ => Predicate.bot_pred
      | Throw (Val Value.Unit) :: _ => Predicate.bot_pred
      | Throw (Val Value.Null) :: _ => Predicate.bot_pred
      | Throw (Val (Value.Bool _)) :: _ => Predicate.bot_pred
      | Throw (Val (Value.Intg _)) :: _ => Predicate.bot_pred
      | Throw (Val (Value.Addr _)) :: _ => Predicate.single ()
      | Throw (BinOp (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (Var _) :: _ => Predicate.bot_pred
      | Throw (LAss (_, _)) :: _ => Predicate.bot_pred
      | Throw (AAcc (_, _)) :: _ => Predicate.bot_pred
      | Throw (AAss (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (ALen _) :: _ => Predicate.bot_pred
      | Throw (FAcc (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (FAss (_, _, _, _)) :: _ => Predicate.bot_pred
      | Throw (Call (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (Block (_, _, _, _)) :: _ => Predicate.bot_pred
      | Throw (Synchronized (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (InSynchronized (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (Seq (_, _)) :: _ => Predicate.bot_pred
      | Throw (Cond (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (While (_, _)) :: _ => Predicate.bot_pred
      | Throw (Throw _) :: _ => Predicate.bot_pred
      | Throw (TryCatch (_, _, _, _)) :: _ => Predicate.bot_pred
      | TryCatch (_, _, _, _) :: _ => Predicate.bot_pred)
    (case x of [] => Predicate.bot_pred | New _ :: es => Predicate.bot_pred
      | NewArray (_, _) :: es => Predicate.bot_pred
      | Cast (_, _) :: es => Predicate.bot_pred
      | InstanceOf (_, _) :: es => Predicate.bot_pred
      | Val _ :: es =>
        Predicate.bind (is_Throws_i es) (fn () => Predicate.single ())
      | BinOp (_, _, _) :: es => Predicate.bot_pred
      | Var _ :: es => Predicate.bot_pred
      | LAss (_, _) :: es => Predicate.bot_pred
      | AAcc (_, _) :: es => Predicate.bot_pred
      | AAss (_, _, _) :: es => Predicate.bot_pred
      | ALen _ :: es => Predicate.bot_pred
      | FAcc (_, _, _) :: es => Predicate.bot_pred
      | FAss (_, _, _, _) :: es => Predicate.bot_pred
      | Call (_, _, _) :: es => Predicate.bot_pred
      | Block (_, _, _, _) :: es => Predicate.bot_pred
      | Synchronized (_, _, _) :: es => Predicate.bot_pred
      | InSynchronized (_, _, _) :: es => Predicate.bot_pred
      | Seq (_, _) :: es => Predicate.bot_pred
      | Cond (_, _, _) :: es => Predicate.bot_pred
      | While (_, _) :: es => Predicate.bot_pred
      | Throw _ :: es => Predicate.bot_pred
      | TryCatch (_, _, _, _) :: es => Predicate.bot_pred);

end; (*struct Expr*)

structure Fifo : sig
  val fifo_push : 'a -> 'a list * 'a list -> 'a list * 'a list
  val fifo_empty : 'a list * 'a list
  val fifo_dequeue : 'a list * 'a list -> 'a * ('a list * 'a list)
  val fifo_enqueue : 'a -> 'a list * 'a list -> 'a list * 'a list
  val fifo_isEmpty : 'a HOL.equal -> 'a list * 'a list -> bool
end = struct

fun fifo_push x f = let
                      val (e, d) = f;
                    in
                      (e, x :: d)
                    end;

val fifo_empty : 'a list * 'a list = ([], []);

fun fifo_dequeue f =
  (case Product_Type.snd f
    of [] =>
      let
        val rp = List.rev (Product_Type.fst f);
      in
        (List.hd rp, ([], List.tl rp))
      end
    | a :: l => (a, (Product_Type.fst f, l)));

fun fifo_enqueue a f = (a :: Product_Type.fst f, Product_Type.snd f);

fun fifo_isEmpty A_ f =
  Product_Type.equal_proda (List.equal_list A_) (List.equal_list A_) f ([], []);

end; (*struct Fifo*)

structure Assoc_List : sig
  datatype ('a, 'b) assoc_list = Assoc_List of ('a * 'b) list
  val empty : ('a, 'b) assoc_list
  val impl_of : ('a, 'b) assoc_list -> ('a * 'b) list
  val lookup : 'a HOL.equal -> ('a, 'b) assoc_list -> 'a -> 'b option
  val update_with_aux :
    'b HOL.equal -> 'a -> 'b -> ('a -> 'a) -> ('b * 'a) list -> ('b * 'a) list
  val update_with :
    'a HOL.equal ->
      'b -> 'a -> ('b -> 'b) -> ('a, 'b) assoc_list -> ('a, 'b) assoc_list
  val update :
    'a HOL.equal -> 'a -> 'b -> ('a, 'b) assoc_list -> ('a, 'b) assoc_list
  val iteratei_aux :
    ('a -> bool) -> ('b -> 'c -> 'a -> 'a) -> ('b * 'c) list -> 'a -> 'a
end = struct

datatype ('a, 'b) assoc_list = Assoc_List of ('a * 'b) list;

val empty : ('a, 'b) assoc_list = Assoc_List [];

fun impl_of (Assoc_List x) = x;

fun lookup A_ al = Map.map_of A_ (impl_of al);

fun update_with_aux B_ v k f [] = [(k, f v)]
  | update_with_aux B_ v k f (p :: ps) =
    (if HOL.eq B_ (Product_Type.fst p) k then (k, f (Product_Type.snd p)) :: ps
      else p :: update_with_aux B_ v k f ps);

fun update_with A_ v k f al =
  Assoc_List (update_with_aux A_ v k f (impl_of al));

fun update A_ k v = update_with A_ v k (fn _ => v);

fun iteratei_aux c f [] sigma = sigma
  | iteratei_aux c f ((k, v) :: l) sigma =
    (if c sigma then iteratei_aux c f l (f k v sigma) else sigma);

end; (*struct Assoc_List*)

structure Trie_Impl : sig
  datatype ('a, 'b) trie = Trie of 'b option * ('a * ('a, 'b) trie) list
  val empty : ('a, 'b) trie
  val lookup : 'a HOL.equal -> ('a, 'b) trie -> 'a list -> 'b option
  val update : 'a HOL.equal -> ('a, 'b) trie -> 'a list -> 'b -> ('a, 'b) trie
  val iteratei_postfixed :
    ('a -> bool) ->
      ('b list -> 'c -> 'a -> 'a) -> ('b, 'c) trie -> 'a -> 'b list -> 'a
  val iteratei :
    ('a -> bool) -> ('b list -> 'c -> 'a -> 'a) -> ('b, 'c) trie -> 'a -> 'a
end = struct

datatype ('a, 'b) trie = Trie of 'b option * ('a * ('a, 'b) trie) list;

val empty : ('a, 'b) trie = Trie (NONE, []);

fun lookup A_ (Trie (vo, uu)) [] = vo
  | lookup A_ (Trie (uv, ts)) (k :: ks) =
    (case Map.map_of A_ ts k of NONE => NONE | SOME t => lookup A_ t ks);

fun update A_ (Trie (vo, ts)) [] v = Trie (SOME v, ts)
  | update A_ (Trie (vo, ts)) (k :: ks) v =
    Trie (vo, Assoc_List.update_with_aux A_ (Trie (NONE, [])) k
                (fn t => update A_ t ks v) ts);

fun iteratei_postfixed c f (Trie (vo, ts)) sigma ks =
  (if c sigma
    then Assoc_List.iteratei_aux c
           (fn k => fn t => fn sigmaa =>
             iteratei_postfixed c f t sigmaa (k :: ks))
           ts (case vo of NONE => sigma | SOME v => f ks v sigma)
    else sigma);

fun iteratei c f t sigma = iteratei_postfixed c f t sigma [];

end; (*struct Trie_Impl*)

structure Trie : sig
  datatype ('a, 'b) trie = Trie of ('a, 'b) Trie_Impl.trie
  val empty : ('a, 'b) trie
  val impl_of : ('a, 'b) trie -> ('a, 'b) Trie_Impl.trie
  val lookup : 'a HOL.equal -> ('a, 'b) trie -> 'a list -> 'b option
  val update : 'a HOL.equal -> 'a list -> 'b -> ('a, 'b) trie -> ('a, 'b) trie
  val iteratei :
    ('a -> bool) -> ('b list -> 'c -> 'a -> 'a) -> ('b, 'c) trie -> 'a -> 'a
end = struct

datatype ('a, 'b) trie = Trie of ('a, 'b) Trie_Impl.trie;

val empty : ('a, 'b) trie = Trie Trie_Impl.empty;

fun impl_of (Trie x) = x;

fun lookup A_ t = Trie_Impl.lookup A_ (impl_of t);

fun update A_ ks v t = Trie (Trie_Impl.update A_ (impl_of t) ks v);

fun iteratei c f t = Trie_Impl.iteratei c (f o List.rev) (impl_of t);

end; (*struct Trie*)

structure MapGA : sig
  val iti_sel :
    (('a option -> bool) ->
      ('b -> 'c -> 'd -> 'e option) -> 'f -> 'g option -> 'e option) ->
      'f -> ('b -> 'c -> 'e option) -> 'e option
  val sel_sel :
    ('a -> ('b -> 'c -> ('b * 'c) option) -> ('b * 'c) option) ->
      'a -> ('b -> 'c -> bool) -> ('b * 'c) option
  val ritoi_max :
    'b Orderings.linorder ->
      (('a option -> bool) ->
        ('b -> 'c -> 'd -> ('b * 'c) option) ->
          'e -> 'f option -> ('b * 'c) option) ->
        'e -> ('b -> 'c -> bool) -> ('b * 'c) option
  val iti_iterate :
    (('a -> bool) -> ('b -> 'c -> 'a -> 'a) -> 'd -> 'a -> 'a) ->
      ('b -> 'c -> 'a -> 'a) -> 'd -> 'a -> 'a
  val itoi_iterateo :
    (('a -> bool) -> ('b -> 'c -> 'a -> 'a) -> 'd -> 'a -> 'a) ->
      ('b -> 'c -> 'a -> 'a) -> 'd -> 'a -> 'a
  val it_map_to_list :
    (('a -> 'b -> ('a * 'b) list -> ('a * 'b) list) ->
      'c -> ('a * 'b) list -> ('a * 'b) list) ->
      'c -> ('a * 'b) list
  val gen_list_to_map_aux : ('a -> 'b -> 'c -> 'c) -> 'c -> ('a * 'b) list -> 'c
  val gen_list_to_map : 'a -> ('b -> 'c -> 'a -> 'a) -> ('b * 'c) list -> 'a
  val itoi_reverse_iterateo :
    (('a -> bool) -> ('b -> 'c -> 'a -> 'a) -> 'd -> 'a -> 'a) ->
      ('b -> 'c -> 'a -> 'a) -> 'd -> 'a -> 'a
  val rito_map_to_sorted_list :
    (('a -> 'b -> ('a * 'b) list -> ('a * 'b) list) ->
      'c -> ('a * 'b) list -> ('a * 'b) list) ->
      'c -> ('a * 'b) list
end = struct

fun iti_sel iti m f = iti Option.is_none (fn k => fn v => fn _ => f k v) m NONE;

fun sel_sel sel s p =
  sel s (fn k => fn v => (if p k v then SOME (k, v) else NONE));

fun ritoi_max B_ it m p =
  it Option.is_none
    (fn k => fn v => fn _ => (if p k v then SOME (k, v) else NONE))
    m
    NONE;

fun iti_iterate iti = iti (fn _ => true);

fun itoi_iterateo it = it (fn _ => true);

fun it_map_to_list iterate m =
  iterate (fn u => fn v => (fn a => (u, v) :: a)) m [];

fun gen_list_to_map_aux upd accs [] = accs
  | gen_list_to_map_aux upd accs ((k, v) :: l) =
    gen_list_to_map_aux upd (upd k v accs) l;

fun gen_list_to_map empt upd l = gen_list_to_map_aux upd empt (List.rev l);

fun itoi_reverse_iterateo it = it (fn _ => true);

fun rito_map_to_sorted_list iterate m =
  iterate (fn u => fn v => (fn a => (u, v) :: a)) m [];

end; (*struct MapGA*)

structure State : sig
  val hp : 'a * 'b -> 'a
  val lcl : 'a * 'b -> 'b
end = struct

fun hp x = Product_Type.fst x;

fun lcl x = Product_Type.snd x;

end; (*struct State*)

structure Lattices : sig
  type 'a semilattice_inf
  val order_semilattice_inf : 'a semilattice_inf -> 'a Orderings.order
  val inf : 'a semilattice_inf -> 'a -> 'a -> 'a
  type 'a semilattice_sup
  val order_semilattice_sup : 'a semilattice_sup -> 'a Orderings.order
  val sup : 'a semilattice_sup -> 'a -> 'a -> 'a
  type 'a lattice
  val semilattice_inf_lattice : 'a lattice -> 'a semilattice_inf
  val semilattice_sup_lattice : 'a lattice -> 'a semilattice_sup
  val inf_fun : 'b lattice -> ('a -> 'b) -> ('a -> 'b) -> 'a -> 'b
  val sup_fun : 'b lattice -> ('a -> 'b) -> ('a -> 'b) -> 'a -> 'b
  val inf_bool : bool -> bool -> bool
  val sup_bool : bool -> bool -> bool
  val minus_boola : bool -> bool -> bool
  val minus_bool : bool Arith.minus
  val minus_fun : 'b Arith.minus -> ('a -> 'b) -> ('a -> 'b) -> 'a -> 'b
  val semilattice_inf_bool : bool semilattice_inf
  val semilattice_sup_bool : bool semilattice_sup
  val lattice_bool : bool lattice
end = struct

type 'a semilattice_inf =
  {order_semilattice_inf : 'a Orderings.order, inf : 'a -> 'a -> 'a};
val order_semilattice_inf = #order_semilattice_inf :
  'a semilattice_inf -> 'a Orderings.order;
val inf = #inf : 'a semilattice_inf -> 'a -> 'a -> 'a;

type 'a semilattice_sup =
  {order_semilattice_sup : 'a Orderings.order, sup : 'a -> 'a -> 'a};
val order_semilattice_sup = #order_semilattice_sup :
  'a semilattice_sup -> 'a Orderings.order;
val sup = #sup : 'a semilattice_sup -> 'a -> 'a -> 'a;

type 'a lattice =
  {semilattice_inf_lattice : 'a semilattice_inf,
    semilattice_sup_lattice : 'a semilattice_sup};
val semilattice_inf_lattice = #semilattice_inf_lattice :
  'a lattice -> 'a semilattice_inf;
val semilattice_sup_lattice = #semilattice_sup_lattice :
  'a lattice -> 'a semilattice_sup;

fun inf_fun B_ f g = (fn x => inf (semilattice_inf_lattice B_) (f x) (g x));

fun sup_fun B_ f g = (fn x => sup (semilattice_sup_lattice B_) (f x) (g x));

fun inf_bool p q = p andalso q;

fun sup_bool p q = p orelse q;

fun minus_boola a b = a andalso not b;

val minus_bool = {minus = minus_boola} : bool Arith.minus;

fun minus_fun B_ a b = (fn x => Arith.minus B_ (a x) (b x));

val semilattice_inf_bool =
  {order_semilattice_inf = Orderings.order_bool, inf = inf_bool} :
  bool semilattice_inf;

val semilattice_sup_bool =
  {order_semilattice_sup = Orderings.order_bool, sup = sup_bool} :
  bool semilattice_sup;

val lattice_bool =
  {semilattice_inf_lattice = semilattice_inf_bool,
    semilattice_sup_lattice = semilattice_sup_bool}
  : bool lattice;

end; (*struct Lattices*)

structure DefAss : sig
  val hyperUn :
    ('a -> bool) option -> ('a -> bool) option -> ('a -> bool) option
  val hyperDiff1 :
    'a HOL.equal -> ('a -> bool) option -> 'a -> ('a -> bool) option
  val hyperInt :
    ('a -> bool) option -> ('a -> bool) option -> ('a -> bool) option
  val a : 'a HOL.equal -> ('a, 'b, 'c) Expr.exp -> ('a -> bool) option
  val a_s : 'a HOL.equal -> ('a, 'b, 'c) Expr.exp list -> ('a -> bool) option
  val hyper_isin : 'a -> ('a -> bool) option -> bool
  val d : 'a HOL.equal -> ('a, 'b, 'c) Expr.exp -> ('a -> bool) option -> bool
  val d_s :
    'a HOL.equal -> ('a, 'b, 'c) Expr.exp list -> ('a -> bool) option -> bool
end = struct

fun hyperUn a b =
  (case a of NONE => NONE
    | SOME aa =>
      (case b of NONE => NONE
        | SOME ba => SOME (Lattices.sup_fun Lattices.lattice_bool aa ba)));

fun hyperDiff1 A_ aa a =
  (case aa of NONE => NONE
    | SOME aaa =>
      SOME (Lattices.minus_fun Lattices.minus_bool aaa
             (Set.insert A_ a (Orderings.bot_fun Orderings.bot_bool))));

fun hyperInt a b =
  (case a of NONE => b
    | SOME aa =>
      (case b of NONE => SOME aa
        | SOME ba => SOME (Lattices.inf_fun Lattices.lattice_bool aa ba)));

fun a A_ (Expr.New c) = SOME (Orderings.bot_fun Orderings.bot_bool)
  | a A_ (Expr.NewArray (t, e)) = a A_ e
  | a A_ (Expr.Cast (c, e)) = a A_ e
  | a A_ (Expr.InstanceOf (e, t)) = a A_ e
  | a A_ (Expr.Val v) = SOME (Orderings.bot_fun Orderings.bot_bool)
  | a A_ (Expr.BinOp (e_1, bop, e_2)) = hyperUn (a A_ e_1) (a A_ e_2)
  | a A_ (Expr.Var v) = SOME (Orderings.bot_fun Orderings.bot_bool)
  | a A_ (Expr.LAss (v, e)) =
    hyperUn (SOME (Set.insert A_ v (Orderings.bot_fun Orderings.bot_bool)))
      (a A_ e)
  | a A_ (Expr.AAcc (aa, i)) = hyperUn (a A_ aa) (a A_ i)
  | a A_ (Expr.AAss (aa, i, e)) = hyperUn (hyperUn (a A_ aa) (a A_ i)) (a A_ e)
  | a A_ (Expr.ALen aa) = a A_ aa
  | a A_ (Expr.FAcc (e, f, d)) = a A_ e
  | a A_ (Expr.FAss (e_1, f, d, e_2)) = hyperUn (a A_ e_1) (a A_ e_2)
  | a A_ (Expr.Call (e, m, es)) = hyperUn (a A_ e) (a_s A_ es)
  | a A_ (Expr.Block (v, t, vo, e)) = hyperDiff1 A_ (a A_ e) v
  | a A_ (Expr.Synchronized (v, oa, e)) = hyperUn (a A_ oa) (a A_ e)
  | a A_ (Expr.InSynchronized (v, aa, e)) = a A_ e
  | a A_ (Expr.Seq (e_1, e_2)) = hyperUn (a A_ e_1) (a A_ e_2)
  | a A_ (Expr.Cond (e, e_1, e_2)) =
    hyperUn (a A_ e) (hyperInt (a A_ e_1) (a A_ e_2))
  | a A_ (Expr.While (b, e)) = a A_ b
  | a A_ (Expr.Throw e) = NONE
  | a A_ (Expr.TryCatch (e_1, c, v, e_2)) =
    hyperInt (a A_ e_1) (hyperDiff1 A_ (a A_ e_2) v)
and a_s A_ [] = SOME (Orderings.bot_fun Orderings.bot_bool)
  | a_s A_ (e :: es) = hyperUn (a A_ e) (a_s A_ es);

fun hyper_isin aa a = (case a of NONE => true | SOME ab => Set.member aa ab);

fun d A_ (Expr.New c) a = true
  | d A_ (Expr.NewArray (t, e)) a = d A_ e a
  | d A_ (Expr.Cast (c, e)) a = d A_ e a
  | d A_ (Expr.InstanceOf (e, t)) a = d A_ e a
  | d A_ (Expr.Val v) a = true
  | d A_ (Expr.BinOp (e_1, bop, e_2)) aa =
    d A_ e_1 aa andalso d A_ e_2 (hyperUn aa (a A_ e_1))
  | d A_ (Expr.Var v) a = hyper_isin v a
  | d A_ (Expr.LAss (v, e)) a = d A_ e a
  | d A_ (Expr.AAcc (aaa, i)) aa =
    d A_ aaa aa andalso d A_ i (hyperUn aa (a A_ aaa))
  | d A_ (Expr.AAss (aaa, i, e)) aa =
    d A_ aaa aa andalso
      (d A_ i (hyperUn aa (a A_ aaa)) andalso
        d A_ e (hyperUn (hyperUn aa (a A_ aaa)) (a A_ i)))
  | d A_ (Expr.ALen aa) a = d A_ aa a
  | d A_ (Expr.FAcc (e, f, da)) a = d A_ e a
  | d A_ (Expr.FAss (e_1, f, da, e_2)) aa =
    d A_ e_1 aa andalso d A_ e_2 (hyperUn aa (a A_ e_1))
  | d A_ (Expr.Call (e, m, es)) aa =
    d A_ e aa andalso d_s A_ es (hyperUn aa (a A_ e))
  | d A_ (Expr.Block (v, t, vo, e)) a =
    (if Option.is_none vo then d A_ e (hyperDiff1 A_ a v)
      else d A_ e
             (hyperUn a
               (SOME (Set.insert A_ v (Orderings.bot_fun Orderings.bot_bool)))))
  | d A_ (Expr.Synchronized (v, oa, e)) aa =
    d A_ oa aa andalso d A_ e (hyperUn aa (a A_ oa))
  | d A_ (Expr.InSynchronized (v, aa, e)) a = d A_ e a
  | d A_ (Expr.Seq (e_1, e_2)) aa =
    d A_ e_1 aa andalso d A_ e_2 (hyperUn aa (a A_ e_1))
  | d A_ (Expr.Cond (e, e_1, e_2)) aa =
    d A_ e aa andalso
      (d A_ e_1 (hyperUn aa (a A_ e)) andalso d A_ e_2 (hyperUn aa (a A_ e)))
  | d A_ (Expr.While (e, c)) aa = d A_ e aa andalso d A_ c (hyperUn aa (a A_ e))
  | d A_ (Expr.Throw e) a = d A_ e a
  | d A_ (Expr.TryCatch (e_1, c, v, e_2)) a =
    d A_ e_1 a andalso
      d A_ e_2
        (hyperUn a
          (SOME (Set.insert A_ v (Orderings.bot_fun Orderings.bot_bool))))
and d_s A_ [] a = true
  | d_s A_ (e :: es) aa = d A_ e aa andalso d_s A_ es (hyperUn aa (a A_ e));

end; (*struct DefAss*)

structure FWLock : sig
  val may_lock : 'a HOL.equal -> ('a * IntInf.int) option -> 'a -> bool
  val lock_lock : ('a * IntInf.int) option -> 'a -> ('a * IntInf.int) option
  val unlock_lock : ('a * IntInf.int) option -> ('a * IntInf.int) option
  val release_all :
    'a HOL.equal -> ('a * IntInf.int) option -> 'a -> ('a * IntInf.int) option
  val upd_lock :
    'a HOL.equal ->
      ('a * IntInf.int) option ->
        'a -> FWState.lock_action -> ('a * IntInf.int) option
  val has_locks : 'a HOL.equal -> ('a * IntInf.int) option -> 'a -> IntInf.int
  val upd_locks :
    'a HOL.equal ->
      ('a * IntInf.int) option ->
        'a -> FWState.lock_action list -> ('a * IntInf.int) option
  val acquire_locks :
    ('a * IntInf.int) option -> 'a -> IntInf.int -> ('a * IntInf.int) option
  val lock_action_ok :
    'a HOL.equal ->
      ('a * IntInf.int) option -> 'a -> FWState.lock_action -> bool
  val lock_actions_ok :
    'a HOL.equal ->
      ('a * IntInf.int) option -> 'a -> FWState.lock_action list -> bool
end = struct

fun may_lock A_ l t = (case l of NONE => true | SOME (ta, _) => HOL.eq A_ t ta);

fun lock_lock NONE t = SOME (t, (0 : IntInf.int))
  | lock_lock (SOME (ta, n)) t = SOME (ta, Arith.suc n);

fun unlock_lock NONE = NONE
  | unlock_lock (SOME (t, n)) =
    (if ((n : IntInf.int) = (0 : IntInf.int)) then NONE
      else SOME (t, Arith.minus_nat n (1 : IntInf.int)));

fun release_all A_ NONE t = NONE
  | release_all A_ (SOME (ta, n)) t =
    (if HOL.eq A_ t ta then NONE else SOME (ta, n));

fun upd_lock A_ l t FWState.Lock = lock_lock l t
  | upd_lock A_ l t FWState.Unlock = unlock_lock l
  | upd_lock A_ l t FWState.UnlockFail = l
  | upd_lock A_ l t FWState.ReleaseAcquire = release_all A_ l t;

fun has_locks A_ NONE t = (0 : IntInf.int)
  | has_locks A_ (SOME (ta, n)) t =
    (if HOL.eq A_ t ta then Arith.suc n else (0 : IntInf.int));

fun upd_locks A_ l t [] = l
  | upd_locks A_ la t (l :: ls) = upd_locks A_ (upd_lock A_ la t l) t ls;

fun acquire_locks l t m =
  (if ((m : IntInf.int) = (0 : IntInf.int)) then l
    else acquire_locks (lock_lock l t) t (Arith.minus_nat m (1 : IntInf.int)));

fun lock_action_ok A_ l t FWState.Lock = may_lock A_ l t
  | lock_action_ok A_ l t FWState.Unlock =
    IntInf.< ((0 : IntInf.int), has_locks A_ l t)
  | lock_action_ok A_ l t FWState.UnlockFail =
    not (IntInf.< ((0 : IntInf.int), has_locks A_ l t))
  | lock_action_ok A_ l t FWState.ReleaseAcquire = true;

fun lock_actions_ok A_ l t [] = true
  | lock_actions_ok A_ la t (l :: ls) =
    lock_action_ok A_ la t l andalso
      lock_actions_ok A_ (upd_lock A_ la t l) t ls;

end; (*struct FWLock*)

structure FWWait : sig
  val waiting : 'a FWState.wait_set_status option -> bool
end = struct

fun waiting (SOME (FWState.InWS w)) = true
  | waiting (SOME (FWState.PostWS w)) = false
  | waiting NONE = false;

end; (*struct FWWait*)

structure Code_Numeral : sig
  val div_code_numeral : int -> int -> int
  val mod_code_numeral : int -> int -> int
  val minus_code_numeral : int -> int -> int
end = struct

fun div_code_numeral n m =
  Product_Type.fst
    ((fn n => fn m => if m = 0 then (0, n) else
       (Int.div (n, m), Int.mod (n, m)))
      n m);

fun mod_code_numeral n m =
  Product_Type.snd
    ((fn n => fn m => if m = 0 then (0, n) else
       (Int.div (n, m), Int.mod (n, m)))
      n m);

fun minus_code_numeral n m = Int.max (n - m, 0 : int);

end; (*struct Code_Numeral*)

structure Random : sig
  val log : int -> int -> int
  val minus_shift : int -> int -> int -> int
  val next : int * int -> int * (int * int)
  val iterate : int -> ('a -> 'b -> 'a * 'b) -> 'a -> 'b -> 'a * 'b
  val range : int -> int * int -> int * (int * int)
  val select : 'a list -> int * int -> 'a * (int * int)
end = struct

fun log b i =
  (if Int.<= (b, 1) orelse Int.< (i, b) then 1
    else Int.+ (1, log b (Code_Numeral.div_code_numeral i b)));

fun minus_shift r k l =
  (if Int.< (k, l) then Code_Numeral.minus_code_numeral (Int.+ (r, k)) l
    else Code_Numeral.minus_code_numeral k l);

fun next (v, w) =
  let
    val k = Code_Numeral.div_code_numeral v 53668;
    val va =
      minus_shift 2147483563
        (Int.* (Code_Numeral.mod_code_numeral v 53668, 40014))
        (Int.* (k, 12211));
    val l = Code_Numeral.div_code_numeral w 52774;
    val wa =
      minus_shift 2147483399
        (Int.* (Code_Numeral.mod_code_numeral w 52774, 40692))
        (Int.* (l, 3791));
    val z = Int.+ (minus_shift 2147483562 va (Int.+ (wa, 1)), 1);
  in
    (z, (va, wa))
  end;

fun iterate k f x =
  (if ((k : Int.int) = 0) then (fn a => (x, a))
    else Product_Type.scomp (f x)
           (iterate (Code_Numeral.minus_code_numeral k 1) f));

fun range k =
  Product_Type.scomp
    (iterate (log 2147483561 k)
      (fn l =>
        Product_Type.scomp next
          (fn v => (fn a => (Int.+ (v, Int.* (l, 2147483561)), a))))
      1)
    (fn v => (fn a => (Code_Numeral.mod_code_numeral v k, a)));

fun select xs =
  Product_Type.scomp (range (IntInf.toInt (List.size_list xs)))
    (fn k => (fn a => (List.nth xs (IntInf.fromInt k), a)));

end; (*struct Random*)

structure Coinductive_List : sig
  datatype 'a llist = Lazy_llist of (unit -> ('a * 'a llist) option)
  val force : 'a llist -> ('a * 'a llist) option
  val lNil : 'a llist
  val lmap : ('a -> 'b) -> 'a llist -> 'b llist
  val lCons : 'a -> 'a llist -> 'a llist
  val lappend : 'a llist -> 'a llist -> 'a llist
  val llist_case : 'a -> ('b -> 'b llist -> 'a) -> 'b llist -> 'a
  val lconcat : 'a llist llist -> 'a llist
  val lfilter : ('a -> bool) -> 'a llist -> 'a llist
  val llist_of : 'a list -> 'a llist
end = struct

datatype 'a llist = Lazy_llist of (unit -> ('a * 'a llist) option);

fun force (Lazy_llist xs) = xs ();

val lNil : 'a llist = Lazy_llist (fn _ => NONE);

fun lmap f (Lazy_llist xs) =
  Lazy_llist (fn _ => Option.map (Product_Type.map_pair f (lmap f)) (xs ()));

fun lCons x xs = Lazy_llist (fn _ => SOME (x, xs));

fun lappend (Lazy_llist xs) ys =
  Lazy_llist
    (fn _ =>
      (case xs () of NONE => force ys
        | SOME (x, xsa) => SOME (x, lappend xsa ys)));

fun llist_case n c (Lazy_llist xs) =
  (case xs () of NONE => n | SOME (a, b) => c a b);

fun lconcat (Lazy_llist xss) =
  Lazy_llist
    (fn _ =>
      (case xss () of NONE => NONE
        | SOME (xs, xssa) => force (lappend xs (lconcat xssa))));

fun lfilter p (Lazy_llist xs) =
  Lazy_llist
    (fn _ =>
      (case xs () of NONE => NONE
        | SOME (x, ys) =>
          (if p x then SOME (x, lfilter p ys) else force (lfilter p ys))));

fun llist_of [] = lNil
  | llist_of (x :: xs) = lCons x (llist_of xs);

end; (*struct Coinductive_List*)

structure TLList : sig
  datatype ('a, 'b) tllist =
    Lazy_tllist of (unit -> (('a * ('a, 'b) tllist), 'b) Sum_Type.sum)
  val tmap : ('a -> 'b) -> ('c -> 'd) -> ('a, 'c) tllist -> ('b, 'd) tllist
  val force : ('a, 'b) tllist -> (('a * ('a, 'b) tllist), 'b) Sum_Type.sum
  val tfilter :
    (unit -> 'a) -> ('b -> bool) -> ('b, 'a) tllist -> ('b, 'a) tllist
  val terminal : ('a, 'b) tllist -> 'b
  val tllist_case :
    ('a -> 'b) -> ('c -> ('c, 'a) tllist -> 'b) -> ('c, 'a) tllist -> 'b
  val tllist_corec :
    'a -> ('a -> (('b * 'a), 'c) Sum_Type.sum) -> ('b, 'c) tllist
  val llist_of_tllist : ('a, 'b) tllist -> 'a Coinductive_List.llist
end = struct

datatype ('a, 'b) tllist =
  Lazy_tllist of (unit -> (('a * ('a, 'b) tllist), 'b) Sum_Type.sum);

fun tmap f g (Lazy_tllist xs) =
  Lazy_tllist
    (fn _ =>
      (case xs () of Sum_Type.Inl (x, ys) => Sum_Type.Inl (f x, tmap f g ys)
        | Sum_Type.Inr b => Sum_Type.Inr (g b)));

fun force (Lazy_tllist xs) = xs ();

fun tfilter b p (Lazy_tllist xs) =
  Lazy_tllist
    (fn _ =>
      (case xs ()
        of Sum_Type.Inl (x, xsa) =>
          (if p x then Sum_Type.Inl (x, tfilter b p xsa)
            else force (tfilter b p xsa))
        | Sum_Type.Inr a => Sum_Type.Inr a));

fun terminal (Lazy_tllist xs) =
  (case xs () of Sum_Type.Inl (_, a) => terminal a | Sum_Type.Inr b => b);

fun tllist_case n c (Lazy_tllist xs) =
  (case xs () of Sum_Type.Inl (a, b) => c a b | Sum_Type.Inr a => n a);

fun tllist_corec a f =
  Lazy_tllist
    (fn _ =>
      (case f a of Sum_Type.Inl (x, aa) => Sum_Type.Inl (x, tllist_corec aa f)
        | Sum_Type.Inr aa => Sum_Type.Inr aa));

fun llist_of_tllist (Lazy_tllist xs) =
  Coinductive_List.Lazy_llist
    (fn _ =>
      (case xs () of Sum_Type.Inl (x, ys) => SOME (x, llist_of_tllist ys)
        | Sum_Type.Inr _ => NONE));

end; (*struct TLList*)

structure RBT_add : sig
  val rm_iterateoi :
    ('a -> bool) -> ('b -> 'c -> 'a -> 'a) -> ('b, 'c) RBT_Impl.rbt -> 'a -> 'a
  val rm_reverse_iterateoi :
    ('a -> bool) -> ('b -> 'c -> 'a -> 'a) -> ('b, 'c) RBT_Impl.rbt -> 'a -> 'a
end = struct

fun rm_iterateoi c f RBT_Impl.Empty sigma = sigma
  | rm_iterateoi c f (RBT_Impl.Branch (col, l, k, v, r)) sigma =
    (if c sigma
      then let
             val sigmaa = rm_iterateoi c f l sigma;
           in
             (if c sigmaa then rm_iterateoi c f r (f k v sigmaa) else sigmaa)
           end
      else sigma);

fun rm_reverse_iterateoi c f RBT_Impl.Empty sigma = sigma
  | rm_reverse_iterateoi c f (RBT_Impl.Branch (col, l, k, v, r)) sigma =
    (if c sigma
      then let
             val sigmaa = rm_reverse_iterateoi c f r sigma;
           in
             (if c sigmaa then rm_reverse_iterateoi c f l (f k v sigmaa)
               else sigmaa)
           end
      else sigma);

end; (*struct RBT_add*)

structure While_Combinator : sig
  val while_option : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a option
  val whilea : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a
end = struct

fun while_option b c s = (if b s then while_option b c (c s) else SOME s);

fun whilea b c s = Option.the (while_option b c s);

end; (*struct While_Combinator*)

structure SemiType : sig
  val widen1p_i_i_o : 'a TypeRel.prog -> Type.ty -> Type.ty Predicate.pred
  val super : 'a TypeRel.prog -> Type.ty -> Type.ty
  val sup : 'a TypeRel.prog -> Type.ty -> Type.ty -> Type.ty Err.err
end = struct

fun widen1p_i_i_o x xa =
  Predicate.sup_pred
    (case xa of Type.Void => Predicate.bot_pred
      | Type.Boolean => Predicate.bot_pred | Type.Integer => Predicate.bot_pred
      | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
      | Type.Array Type.Void => Predicate.bot_pred
      | Type.Array Type.Boolean => Predicate.bot_pred
      | Type.Array Type.Integer => Predicate.bot_pred
      | Type.Array Type.Nt => Predicate.bot_pred
      | Type.Array (Type.Class xb) =>
        (if ((xb : string) = Type.object)
          then Predicate.single (Type.Class Type.object)
          else Predicate.bot_pred)
      | Type.Array (Type.Array _) => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xa of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array Type.Void => Predicate.bot_pred
        | Type.Array Type.Boolean => Predicate.bot_pred
        | Type.Array Type.Integer => Predicate.single (Type.Class Type.object)
        | Type.Array Type.Nt => Predicate.bot_pred
        | Type.Array (Type.Class _) => Predicate.bot_pred
        | Type.Array (Type.Array _) => Predicate.bot_pred)
      (Predicate.sup_pred
        (case xa of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.bot_pred
          | Type.Array Type.Void => Predicate.bot_pred
          | Type.Array Type.Boolean => Predicate.single (Type.Class Type.object)
          | Type.Array Type.Integer => Predicate.bot_pred
          | Type.Array Type.Nt => Predicate.bot_pred
          | Type.Array (Type.Class _) => Predicate.bot_pred
          | Type.Array (Type.Array _) => Predicate.bot_pred)
        (Predicate.sup_pred
          (case xa of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class _ => Predicate.bot_pred
            | Type.Array Type.Void => Predicate.single (Type.Class Type.object)
            | Type.Array Type.Boolean => Predicate.bot_pred
            | Type.Array Type.Integer => Predicate.bot_pred
            | Type.Array Type.Nt => Predicate.bot_pred
            | Type.Array (Type.Class _) => Predicate.bot_pred
            | Type.Array (Type.Array _) => Predicate.bot_pred)
          (Predicate.sup_pred
            (case xa of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class c =>
                Predicate.bind (TypeRel.subcls1_i_i_o x c)
                  (fn xb => Predicate.single (Type.Class xb))
              | Type.Array _ => Predicate.bot_pred)
            (case xa of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array t =>
                Predicate.bind
                  (Predicate.if_pred
                    (not (Type.equal_tya (Type.ground_type t) Type.Nt)))
                  (fn () =>
                    Predicate.bind (widen1p_i_i_o x t)
                      (fn xb => Predicate.single (Type.Array xb))))))));

fun super p (Type.Array Type.Integer) = Type.Class Type.object
  | super p (Type.Array Type.Boolean) = Type.Class Type.object
  | super p (Type.Array Type.Void) = Type.Class Type.object
  | super p (Type.Array (Type.Class c)) =
    (if ((c : string) = Type.object) then Type.Class Type.object
      else Type.Array (super p (Type.Class c)))
  | super p (Type.Array (Type.Array t)) = Type.Array (super p (Type.Array t))
  | super p (Type.Class c) =
    Type.Class (Product_Type.fst (Option.the (TypeRel.class p c)));

fun sup p t u =
  (if Type.is_refT t andalso Type.is_refT u
    then Err.Ok
           (if Type.equal_tya u Type.Nt then t
             else (if Type.equal_tya t Type.Nt then u
                    else While_Combinator.whilea
                           (fn z =>
                             not (Set.member (t, z)
                                   (fn (a, b) =>
                                     Predicate.holds
                                       (Transitive_Closure_Table.rtrancl_tab_FioB_i_i_i
 Type.equal_ty (widen1p_i_i_o p) [] a b))))
                           (super p) u))
    else (if Type.equal_tya t u then Err.Ok t else Err.Err));

end; (*struct SemiType*)

structure SetByMap : sig
  val s_ins : ('a -> unit -> 'b -> 'c) -> 'a -> 'b -> 'c
  val s_memb : ('a -> 'b -> 'c option) -> 'a -> 'b -> bool
  val s_empty : 'a -> 'a
  val s_delete : 'a -> 'a
end = struct

fun s_ins update x s = update x () s;

fun s_memb lookup x s = not (Option.is_none (lookup x s));

fun s_empty empt = empt;

fun s_delete delete = delete;

end; (*struct SetByMap*)

structure TrieMapImpl : sig
  val tm_empty : ('a, 'b) Trie.trie
  val tm_lookup : 'a HOL.equal -> 'a list -> ('a, 'b) Trie.trie -> 'b option
  val tm_update :
    'a HOL.equal -> 'a list -> 'b -> ('a, 'b) Trie.trie -> ('a, 'b) Trie.trie
  val tm_iteratei :
    ('a -> bool) ->
      ('b list -> 'c -> 'a -> 'a) -> ('b, 'c) Trie.trie -> 'a -> 'a
  val tm_iterate : ('a list -> 'b -> 'c -> 'c) -> ('a, 'b) Trie.trie -> 'c -> 'c
  val tm_to_list : ('a, 'b) Trie.trie -> ('a list * 'b) list
end = struct

val tm_empty : ('a, 'b) Trie.trie = Trie.empty;

fun tm_lookup A_ k t = Trie.lookup A_ t k;

fun tm_update A_ = Trie.update A_;

fun tm_iteratei x = Trie.iteratei x;

fun tm_iterate x = MapGA.iti_iterate tm_iteratei x;

fun tm_to_list x = MapGA.it_map_to_list tm_iterate x;

end; (*struct TrieMapImpl*)

structure RBTMapImpl : sig
  val rm_reverse_iterateoi :
    'b Orderings.linorder ->
      ('a -> bool) -> ('b -> 'c -> 'a -> 'a) -> ('b, 'c) RBT.rbt -> 'a -> 'a
  val rm_max :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('a -> 'b -> bool) -> ('a * 'b) option
  val rm_iterateoi :
    'b Orderings.linorder ->
      ('a -> bool) -> ('b -> 'c -> 'a -> 'a) -> ('b, 'c) RBT.rbt -> 'a -> 'a
  val rm_iteratei :
    'b Orderings.linorder ->
      ('a -> bool) -> ('b -> 'c -> 'a -> 'a) -> ('b, 'c) RBT.rbt -> 'a -> 'a
  val rm_sela :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('a -> 'b -> 'c option) -> 'c option
  val rm_sel :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('a -> 'b -> bool) -> ('a * 'b) option
  val rm_empty : 'a Orderings.linorder -> ('a, 'b) RBT.rbt
  val rm_delete :
    'a Orderings.linorder -> 'a -> ('a, 'b) RBT.rbt -> ('a, 'b) RBT.rbt
  val rm_lookup : 'a Orderings.linorder -> 'a -> ('a, 'b) RBT.rbt -> 'b option
  val rm_update :
    'a Orderings.linorder -> 'a -> 'b -> ('a, 'b) RBT.rbt -> ('a, 'b) RBT.rbt
  val rm_iterateo :
    'a Orderings.linorder ->
      ('a -> 'b -> 'c -> 'c) -> ('a, 'b) RBT.rbt -> 'c -> 'c
  val rm_iterate :
    'a Orderings.linorder ->
      ('a -> 'b -> 'c -> 'c) -> ('a, 'b) RBT.rbt -> 'c -> 'c
  val rm_reverse_iterateo :
    'a Orderings.linorder ->
      ('a -> 'b -> 'c -> 'c) -> ('a, 'b) RBT.rbt -> 'c -> 'c
  val rm_to_list : 'a Orderings.linorder -> ('a, 'b) RBT.rbt -> ('a * 'b) list
end = struct

fun rm_reverse_iterateoi B_ c f r =
  RBT_add.rm_reverse_iterateoi c f (RBT.impl_of B_ r);

fun rm_max A_ = MapGA.ritoi_max A_ (rm_reverse_iterateoi A_);

fun rm_iterateoi B_ c f r = RBT_add.rm_iterateoi c f (RBT.impl_of B_ r);

fun rm_iteratei B_ = rm_iterateoi B_;

fun rm_sela A_ = MapGA.iti_sel (rm_iteratei A_);

fun rm_sel A_ = MapGA.sel_sel (rm_sela A_);

fun rm_empty A_ = RBT.empty A_;

fun rm_delete A_ = RBT.delete A_;

fun rm_lookup A_ k m = RBT.lookup A_ m k;

fun rm_update A_ = RBT.insert A_;

fun rm_iterateo A_ = MapGA.itoi_iterateo (rm_iterateoi A_);

fun rm_iterate A_ = rm_iterateo A_;

fun rm_reverse_iterateo A_ =
  MapGA.itoi_reverse_iterateo (rm_reverse_iterateoi A_);

fun rm_to_list A_ = MapGA.rito_map_to_sorted_list (rm_reverse_iterateo A_);

end; (*struct RBTMapImpl*)

structure ToString : sig
  type 'a toString
  val toString : 'a toString -> 'a -> string
  val toString_literala : string -> string
  val ty_toString : Type.ty -> string
  val toString_ty : Type.ty -> string
  val toString_bool : bool -> string
  val digit_toString : IntInf.int -> string
  val int_toString : IntInf.int -> string list
  val toString_int : IntInf.int -> string
  val toString_word : 'a Type_Length.len -> 'a Word.word -> string
  val val_toString : 'a toString -> 'a Value.vala -> string
  val toString_vala : 'a toString -> 'a Value.vala -> string
  val bop_toString : BinOp.bop -> string
  val toString_bop : BinOp.bop -> string
  val exp_toString :
    'a toString -> 'b toString -> 'c toString -> ('a, 'b, 'c) Expr.exp -> string
  val toString_expa :
    'a toString -> 'b toString -> 'c toString -> ('a, 'b, 'c) Expr.exp -> string
  val toString_exp :
    'a toString -> 'b toString -> 'c toString -> ('a, 'b, 'c) Expr.exp toString
  val toString_funa : ('a -> 'b) -> string
  val toString_fun : ('a -> 'b) toString
  val toString_nata : IntInf.int -> string
  val toString_nat : IntInf.int toString
  val toString_val : 'a toString -> 'a Value.vala toString
  val toString_chara : char -> string
  val toString_char : char toString
  val list_toString : 'a toString -> string -> 'a list -> string list
  val toString_lista : 'a toString -> 'a list -> string
  val toString_list : 'a toString -> ('a list) toString
  val toString_proda : 'a toString -> 'b toString -> 'a * 'b -> string
  val toString_prod : 'a toString -> 'b toString -> ('a * 'b) toString
  val toString_rbt :
    'a Orderings.linorder * 'a toString -> 'b toString ->
      ('a, 'b) RBT.rbt -> string
  val toString_unita : unit -> string
  val toString_unit : unit toString
  val toString_trie : 'a toString -> 'b toString -> ('a, 'b) Trie.trie -> string
  val toString_finfuna :
    'a Card_Univ.card_UNIV * 'a HOL.equal * 'a Orderings.linorder *
      'a toString ->
      'b HOL.equal * 'b toString -> ('a, 'b) FinFun.finfun -> string
  val toString_finfun :
    'a Card_Univ.card_UNIV * 'a HOL.equal * 'a Orderings.linorder *
      'a toString ->
      'b HOL.equal * 'b toString -> ('a, 'b) FinFun.finfun toString
  val toString_literal : string toString
  val addr_loc_toString : Heap.addr_loc -> string
  val toString_addr_loc : Heap.addr_loc -> string
  val obs_event_toString :
    'a toString -> 'b toString -> ('a, 'b) Observable_Events.obs_event -> string
  val toString_obs_eventa :
    'a toString -> 'b toString -> ('a, 'b) Observable_Events.obs_event -> string
  val toString_obs_event :
    'a toString -> 'b toString -> ('a, 'b) Observable_Events.obs_event toString
  val toString_assoc_lista :
    'a toString -> 'b toString -> ('a, 'b) Assoc_List.assoc_list -> string
  val toString_assoc_list :
    'a toString -> 'b toString -> ('a, 'b) Assoc_List.assoc_list toString
end = struct

type 'a toString = {toString : 'a -> string};
val toString = #toString : 'a toString -> 'a -> string;

fun toString_literala s = s;

fun ty_toString Type.Void = "Void"
  | ty_toString Type.Boolean = "Boolean"
  | ty_toString Type.Integer = "Integer"
  | ty_toString Type.Nt = "NT"
  | ty_toString (Type.Class c) = String.concat ["Class ", toString_literala c]
  | ty_toString (Type.Array t) = String.concat [ty_toString t, "[]"];

fun toString_ty x = ty_toString x;

fun toString_bool b = (case b of true => "True" | false => "False");

fun digit_toString (0 : IntInf.int) = "0"
  | digit_toString (1 : IntInf.int) = "1"
  | digit_toString (2 : IntInf.int) = "2"
  | digit_toString (3 : IntInf.int) = "3"
  | digit_toString (4 : IntInf.int) = "4"
  | digit_toString (5 : IntInf.int) = "5"
  | digit_toString (6 : IntInf.int) = "6"
  | digit_toString (7 : IntInf.int) = "7"
  | digit_toString (8 : IntInf.int) = "8"
  | digit_toString (9 : IntInf.int) = "9";

fun int_toString n =
  (if IntInf.< (n, (0 : IntInf.int)) then "-" :: int_toString (IntInf.~ n)
    else (if IntInf.< (n, (10 : IntInf.int)) then [digit_toString n]
           else int_toString (Arith.div_int n (10 : IntInf.int)) @
                  [digit_toString (Arith.mod_int n (10 : IntInf.int))]));

fun toString_int i = String.concat (int_toString i);

fun toString_word A_ w = toString_int (Word.sint A_ w);

fun val_toString A_ Value.Unit = "Unit"
  | val_toString A_ Value.Null = "Null"
  | val_toString A_ (Value.Bool b) = String.concat ["Bool ", toString_bool b]
  | val_toString A_ (Value.Intg i) =
    String.concat
      ["Intg ",
        toString_word
          (Type_Length.len_bit0
            (Type_Length.len_bit0
              (Type_Length.len_bit0
                (Type_Length.len_bit0
                  (Type_Length.len_bit0 Type_Length.len_num1)))))
          i]
  | val_toString A_ (Value.Addr a) = String.concat ["Addr ", toString A_ a];

fun toString_vala A_ = val_toString A_;

fun bop_toString BinOp.Eq = "=="
  | bop_toString BinOp.NotEq = "!="
  | bop_toString BinOp.LessThan = "<"
  | bop_toString BinOp.LessOrEqual = "<="
  | bop_toString BinOp.GreaterThan = ">"
  | bop_toString BinOp.GreaterOrEqual = ">="
  | bop_toString BinOp.Add = "+"
  | bop_toString BinOp.Subtract = "-"
  | bop_toString BinOp.Mult = "*"
  | bop_toString BinOp.Div = "/"
  | bop_toString BinOp.Mod = "%"
  | bop_toString BinOp.BinAnd = "&"
  | bop_toString BinOp.BinOr = "|"
  | bop_toString BinOp.BinXor = "^"
  | bop_toString BinOp.ShiftLeft = "<<"
  | bop_toString BinOp.ShiftRightZeros = ">>"
  | bop_toString BinOp.ShiftRightSigned = ">>>";

fun toString_bop x = bop_toString x;

fun exp_toString A_ B_ C_ (Expr.New c) = String.concat ["new ", c]
  | exp_toString A_ B_ C_ (Expr.NewArray (t, e)) =
    String.concat ["new ", toString_ty t, "[", exp_toString A_ B_ C_ e, "]"]
  | exp_toString A_ B_ C_ (Expr.Cast (t, e)) =
    String.concat ["(", toString_ty t, ") (", exp_toString A_ B_ C_ e, ")"]
  | exp_toString A_ B_ C_ (Expr.InstanceOf (e, t)) =
    String.concat ["(", exp_toString A_ B_ C_ e, ") instanceof ", toString_ty t]
  | exp_toString A_ B_ C_ (Expr.Val v) =
    String.concat ["Val (", toString_vala C_ v, ")"]
  | exp_toString A_ B_ C_ (Expr.BinOp (e1, bop, e2)) =
    String.concat
      ["(", exp_toString A_ B_ C_ e1, ") ", toString_bop bop, " (",
        exp_toString A_ B_ C_ e2, ")"]
  | exp_toString A_ B_ C_ (Expr.Var v) = String.concat ["Var ", toString A_ v]
  | exp_toString A_ B_ C_ (Expr.LAss (v, e)) =
    String.concat [toString A_ v, " := (", exp_toString A_ B_ C_ e, ")"]
  | exp_toString A_ B_ C_ (Expr.AAcc (a, i)) =
    String.concat
      ["(", exp_toString A_ B_ C_ a, ")[", exp_toString A_ B_ C_ i, "]"]
  | exp_toString A_ B_ C_ (Expr.AAss (a, i, e)) =
    String.concat
      ["(", exp_toString A_ B_ C_ a, ")[", exp_toString A_ B_ C_ i, "] := (",
        exp_toString A_ B_ C_ e, ")"]
  | exp_toString A_ B_ C_ (Expr.ALen a) =
    String.concat ["(", exp_toString A_ B_ C_ a, ").length"]
  | exp_toString A_ B_ C_ (Expr.FAcc (e, f, d)) =
    String.concat ["(", exp_toString A_ B_ C_ e, ").", f, "{", d, "}"]
  | exp_toString A_ B_ C_ (Expr.FAss (ea, f, d, e)) =
    String.concat
      ["(", exp_toString A_ B_ C_ ea, ").", f, "{", d, "} := (",
        exp_toString A_ B_ C_ e, ")"]
  | exp_toString A_ B_ C_ (Expr.Call (e, m, es)) =
    String.concat
      (["(", exp_toString A_ B_ C_ e, ").", m, "("] @
        List.map (exp_toString A_ B_ C_) es @ [")"])
  | exp_toString A_ B_ C_ (Expr.Block (v, t, vo, e)) =
    String.concat
      (["{", toString A_ v, ":", toString_ty t] @
        (case vo of NONE => [] | SOME va => ["=", toString_vala C_ va]) @
          ["; ", exp_toString A_ B_ C_ e, "}"])
  | exp_toString A_ B_ C_ (Expr.Synchronized (v, ea, e)) =
    String.concat
      ["synchronized_", toString B_ v, "_(", exp_toString A_ B_ C_ ea, ") {",
        exp_toString A_ B_ C_ e, "}"]
  | exp_toString A_ B_ C_ (Expr.InSynchronized (v, ad, e)) =
    String.concat
      ["insynchronized_", toString B_ v, "_(", toString C_ ad, ") {",
        exp_toString A_ B_ C_ e, "}"]
  | exp_toString A_ B_ C_ (Expr.Seq (ea, e)) =
    String.concat [exp_toString A_ B_ C_ ea, "; ", exp_toString A_ B_ C_ e]
  | exp_toString A_ B_ C_ (Expr.Cond (eb, ea, e)) =
    String.concat
      ["if (", exp_toString A_ B_ C_ eb, ") { ", exp_toString A_ B_ C_ ea,
        " } else { ", exp_toString A_ B_ C_ e, "}"]
  | exp_toString A_ B_ C_ (Expr.While (ea, e)) =
    String.concat
      ["while (", exp_toString A_ B_ C_ ea, ") { ", exp_toString A_ B_ C_ e,
        " }"]
  | exp_toString A_ B_ C_ (Expr.Throw e) =
    String.concat ["throw (", exp_toString A_ B_ C_ e, ")"]
  | exp_toString A_ B_ C_ (Expr.TryCatch (ea, c, v, e)) =
    String.concat
      ["try { ", exp_toString A_ B_ C_ ea, " } catch (", c, " ", toString A_ v,
        ") { ", exp_toString A_ B_ C_ e, " }"];

fun toString_expa A_ B_ C_ = exp_toString A_ B_ C_;

fun toString_exp A_ B_ C_ = {toString = toString_expa A_ B_ C_} :
  ('a, 'b, 'c) Expr.exp toString;

fun toString_funa f = "fn";

val toString_fun = {toString = toString_funa} : ('a -> 'b) toString;

fun toString_nata n = toString_int n;

val toString_nat = {toString = toString_nata} : IntInf.int toString;

fun toString_val A_ = {toString = toString_vala A_} : 'a Value.vala toString;

fun toString_chara c = String.implode [c];

val toString_char = {toString = toString_chara} : char toString;

fun list_toString A_ sep [] = []
  | list_toString A_ sep [x] = [toString A_ x]
  | list_toString A_ sep (x :: v :: va) =
    toString A_ x :: sep :: list_toString A_ sep (v :: va);

fun toString_lista A_ xs =
  String.concat ("[" :: list_toString A_ "," xs @ ["]"]);

fun toString_list A_ = {toString = toString_lista A_} : ('a list) toString;

fun toString_proda A_ B_ =
  (fn (a, b) => String.concat ["(", toString A_ a, ", ", toString B_ b, ")"]);

fun toString_prod A_ B_ = {toString = toString_proda A_ B_} :
  ('a * 'b) toString;

fun toString_rbt (A1_, A2_) B_ t =
  String.concat
    (list_toString (toString_prod A2_ B_) ",\n" (RBTMapImpl.rm_to_list A1_ t));

fun toString_unita u = "()";

val toString_unit = {toString = toString_unita} : unit toString;

fun toString_trie A_ B_ t =
  toString_lista (toString_prod (toString_list A_) B_)
    (TrieMapImpl.tm_to_list t);

fun toString_finfuna (A1_, A2_, A3_, A4_) (B1_, B2_) f =
  String.concat
    ("(" ::
      toString B2_ (FinFun.finfun_default A1_ f) ::
        List.maps
          (fn x =>
            [",", toString A4_ x, "|->",
              toString B2_ (FinFun.finfun_apply A2_ f x)])
          (FinFun.finfun_to_list (A1_, A2_, A3_) B1_ f) @
          [")"]);

fun toString_finfun (A1_, A2_, A3_, A4_) (B1_, B2_) =
  {toString = toString_finfuna (A1_, A2_, A3_, A4_) (B1_, B2_)} :
  ('a, 'b) FinFun.finfun toString;

val toString_literal = {toString = toString_literala} : string toString;

fun addr_loc_toString (Heap.CField (c, f)) =
  String.concat ["CField ", f, "{", c, "}"]
  | addr_loc_toString (Heap.ACell n) =
    String.concat ["ACell ", toString_nata n];

fun toString_addr_loc x = addr_loc_toString x;

fun obs_event_toString A_ B_ (Observable_Events.ExternalCall (ad, m, vs, v)) =
  String.concat
    ["ExternalCall ", m, "(", toString_lista (toString_val A_) vs, ") = ",
      toString_vala A_ v]
  | obs_event_toString A_ B_ (Observable_Events.ReadMem (ad, al, v)) =
    String.concat
      ["ReadMem ", toString A_ ad, "@", toString_addr_loc al, "=",
        toString_vala A_ v]
  | obs_event_toString A_ B_ (Observable_Events.WriteMem (ad, al, v)) =
    String.concat
      ["WriteMem ", toString A_ ad, "@", toString_addr_loc al, "=",
        toString_vala A_ v]
  | obs_event_toString A_ B_ (Observable_Events.NewObj (ad, c)) =
    String.concat ["NewObj ", toString A_ ad, ":", c]
  | obs_event_toString A_ B_ (Observable_Events.NewArr (ad, t, n)) =
    String.concat
      ["NewArr ", toString A_ ad, ":", toString_ty t, "[", toString_nata n, "]"]
  | obs_event_toString A_ B_ (Observable_Events.ThreadStart t) =
    String.concat ["ThreadStart ", toString B_ t]
  | obs_event_toString A_ B_ (Observable_Events.ThreadJoin t) =
    String.concat ["ThreadJoin ", toString B_ t]
  | obs_event_toString A_ B_ (Observable_Events.SyncLock ad) =
    String.concat ["SyncLock ", toString A_ ad]
  | obs_event_toString A_ B_ (Observable_Events.SyncUnlock ad) =
    String.concat ["SyncUnlock ", toString A_ ad]
  | obs_event_toString A_ B_ (Observable_Events.ObsInterrupt t) =
    String.concat ["Interrupt ", toString B_ t]
  | obs_event_toString A_ B_ (Observable_Events.ObsInterrupted t) =
    String.concat ["Interrupted ", toString B_ t];

fun toString_obs_eventa A_ B_ = obs_event_toString A_ B_;

fun toString_obs_event A_ B_ = {toString = toString_obs_eventa A_ B_} :
  ('a, 'b) Observable_Events.obs_event toString;

fun toString_assoc_lista A_ B_ =
  toString_lista (toString_prod A_ B_) o Assoc_List.impl_of;

fun toString_assoc_list A_ B_ = {toString = toString_assoc_lista A_ B_} :
  ('a, 'b) Assoc_List.assoc_list toString;

end; (*struct ToString*)

structure Predicate_Compile_Alternative_Defs : sig
  val list_all2_FiiB_i_i :
    ('a -> 'b -> unit Predicate.pred) ->
      'a list -> 'b list -> unit Predicate.pred
end = struct

fun list_all2_FiiB_i_i p xa xb =
  Predicate.sup_pred
    (case xa
      of [] =>
        (case xb of [] => Predicate.single () | _ :: _ => Predicate.bot_pred)
      | _ :: _ => Predicate.bot_pred)
    (case xa of [] => Predicate.bot_pred
      | x :: xs =>
        (case xb of [] => Predicate.bot_pred
          | y :: ys =>
            Predicate.bind (list_all2_FiiB_i_i p xs ys)
              (fn () =>
                Predicate.bind (p x y) (fn () => Predicate.single ()))));

end; (*struct Predicate_Compile_Alternative_Defs*)

structure WellForm : sig
  val acyclic_class_hierarchy : 'a TypeRel.prog -> bool
  val wf_fdecl :
    'a TypeRel.prog -> string * (Type.ty * unit TypeRel.fmod_ext) -> bool
  val wf_mdecl :
    ('a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a)) -> bool) ->
      'a TypeRel.prog ->
        string -> string * (Type.ty list * (Type.ty * 'a option)) -> bool
  val wf_overriding_aux_aux_i_i_i_i_i :
    Type.ty ->
      Type.ty ->
        Type.ty list -> Type.ty list -> 'a TypeRel.prog -> unit Predicate.pred
  val wf_overriding_aux_i_i_i_i_i :
    Type.ty ->
      Type.ty list -> string -> string -> 'a TypeRel.prog -> unit Predicate.pred
  val wf_overriding_i_i_i :
    'a TypeRel.prog ->
      string ->
        string * (Type.ty list * (Type.ty * 'a option)) -> unit Predicate.pred
  val wf_overriding :
    'a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a option)) -> bool
  val wf_cdecl :
    ('a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a)) -> bool) ->
      'a TypeRel.prog ->
        string *
          (string *
            ((string * (Type.ty * unit TypeRel.fmod_ext)) list *
              (string * (Type.ty list * (Type.ty * 'a option))) list)) ->
          bool
  val wf_prog :
    ('a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a)) -> bool) ->
      'a TypeRel.prog -> bool
end = struct

fun acyclic_class_hierarchy p =
  List.list_all
    (fn (c, (d, (_, _))) =>
      (if not ((c : string) = Type.object) then not (TypeRel.subcls p d c)
        else true))
    (TypeRel.classes p);

fun wf_fdecl p = (fn (_, (t, _)) => TypeRel.is_type p t);

fun wf_mdecl wf_md p c =
  (fn (m, (ts, (t, ma))) =>
    List.list_all (TypeRel.is_type p) ts andalso
      (TypeRel.is_type p t andalso
        (case ma of NONE => ExternalCall.external_WT_defs c m ts t
          | SOME mb => wf_md p c (m, (ts, (t, mb))))));

fun wf_overriding_aux_aux_i_i_i_i_i x xa xb xc xd =
  Predicate.sup_pred
    (Predicate.bind
      (Predicate.not_pred
        (Predicate_Compile_Alternative_Defs.list_all2_FiiB_i_i
          (TypeRel.widen_i_i_i xd) xc xb))
      (fn () => Predicate.single ()))
    (Predicate.bind (Predicate.not_pred (TypeRel.widen_i_i_i xd xa x))
      (fn () => Predicate.single ()));

fun wf_overriding_aux_i_i_i_i_i x xa xb xc xd =
  Predicate.bind (TypeRel.method_i_i_i_o_o_o_o xd xc xb)
    (fn (ts, (t, (_, _))) =>
      Predicate.bind (wf_overriding_aux_aux_i_i_i_i_i t x xa ts xd)
        (fn () => Predicate.single ()));

fun wf_overriding_i_i_i x xa (xb, (xc, (xd, xe))) =
  Predicate.bind
    (Predicate.not_pred (wf_overriding_aux_i_i_i_i_i xd xc xb xa x))
    (fn () => Predicate.single ());

fun wf_overriding x1 x2 x3 = Predicate.holds (wf_overriding_i_i_i x1 x2 x3);

fun wf_cdecl wf_md p =
  (fn (c, (d, (fs, ms))) =>
    List.list_all (wf_fdecl p) fs andalso
      (Aux.distinct_fst Stringa.equal_literal fs andalso
        (List.list_all (wf_mdecl wf_md p c) ms andalso
          (Aux.distinct_fst Stringa.equal_literal ms andalso
            ((if not ((c : string) = Type.object)
               then TypeRel.is_class p d andalso
                      List.list_all (wf_overriding p d) ms
               else true) andalso
              (if ((c : string) = Type.thread)
                then List.member
                       (Product_Type.equal_prod Stringa.equal_literal
                         (Product_Type.equal_prod
                           (List.equal_list Type.equal_ty) Type.equal_ty))
                       (List.map (fn (m, (ts, (t, _))) => (m, (ts, t))) ms)
                       ("run()V", ([], Type.Void))
                else true))))));

fun wf_prog wf_md p =
  acyclic_class_hierarchy p andalso
    (Exceptions.wf_syscls p andalso
      (List.list_all (wf_cdecl wf_md p) (TypeRel.classes p) andalso
        Aux.distinct_fst Stringa.equal_literal (TypeRel.classes p)));

end; (*struct WellForm*)

structure WellType : sig
  val wT_Fii_oB_i_i_i_o :
    (Type.ty -> Type.ty -> Type.ty Predicate.pred) ->
      (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
        (string -> Type.ty option) ->
          (string, unit, 'a) Expr.exp -> Type.ty Predicate.pred
  val wTs_Fii_oB_i_i_i_o :
    (Type.ty -> Type.ty -> Type.ty Predicate.pred) ->
      (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
        (string -> Type.ty option) ->
          (string, unit, 'a) Expr.exp list -> (Type.ty list) Predicate.pred
  val wT_Fii_oB_i_i_i_i :
    (Type.ty -> Type.ty -> Type.ty Predicate.pred) ->
      (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
        (string -> Type.ty option) ->
          (string, unit, 'a) Expr.exp -> Type.ty -> unit Predicate.pred
  val is_lub_sup_i_i_i_o :
    'a TypeRel.prog -> Type.ty -> Type.ty -> Type.ty Predicate.pred
  val wT_code_i_i_i_o :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      (string -> Type.ty option) ->
        (string, unit, 'a) Expr.exp -> Type.ty Predicate.pred
end = struct

fun wT_Fii_oB_i_i_i_o is_lub x xa xb =
  (case xb
    of Expr.New c1 =>
      Predicate.bind (TypeRel.is_class_i_i x c1)
        (fn () => Predicate.single (Type.Class c1))
    | Expr.NewArray (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
        (fn () =>
          Predicate.bind (Predicate.if_pred (TypeRel.is_type x (Type.Array c1)))
            (fn () => Predicate.single (Type.Array c1)))
    | Expr.Cast (c1, c2) =>
      Predicate.bind (Predicate.if_pred (TypeRel.is_type x c1))
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xaa =>
              Predicate.bind
                (Predicate.if_pred
                  (TypeRel.widen x c1 xaa orelse TypeRel.widen x xaa c1))
                (fn () => Predicate.single c1)))
    | Expr.InstanceOf (c1, c2) =>
      Predicate.bind (Predicate.if_pred (TypeRel.is_type x c2))
        (fn () =>
          Predicate.bind (Type.is_refT_i c2)
            (fn () =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
                (fn xaa =>
                  Predicate.bind
                    (Predicate.if_pred
                      (TypeRel.widen x c2 xaa orelse TypeRel.widen x xaa c2))
                    (fn () => Predicate.single Type.Boolean))))
    | Expr.Val c1 =>
      (case Heap.typeof c1 of NONE => Predicate.bot_pred
        | SOME a => Predicate.single a)
    | Expr.BinOp (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
            (BinOp.wT_binop_i_i_i_i_o x xba c2))
    | Expr.Var c1 =>
      (case xa c1 of NONE => Predicate.bot_pred | SOME a => Predicate.single a)
    | Expr.LAss (c1, c2) =>
      Predicate.bind (Predicate.if_pred (not ((c1 : string) = "this")))
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xba =>
              (case xa c1 of NONE => Predicate.bot_pred
                | SOME t =>
                  Predicate.bind (TypeRel.widen_i_i_i x xba t)
                    (fn () => Predicate.single Type.Void))))
    | Expr.AAcc (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn a =>
              (case a of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array aa => Predicate.single aa)))
    | Expr.AAss (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn a =>
              (case a of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array t =>
                  Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
                    (fn xaa =>
                      Predicate.bind (TypeRel.widen_i_i_i x xaa t)
                        (fn () => Predicate.single Type.Void)))))
    | Expr.ALen c1 =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn a =>
          (case a of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.single Type.Integer))
    | Expr.FAcc (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xaa =>
          Predicate.bind (Type.is_class_type_of_i_o xaa)
            (fn xab =>
              Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i x xab c2 c3)
                (fn (t, _) => Predicate.single t)))
    | Expr.FAss (c1, c2, c3, c4) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c4)
            (fn xaa =>
              Predicate.bind (Type.is_class_type_of_i_o xba)
                (fn xaaa =>
                  Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i x xaaa c2 c3)
                    (fn (t, _) =>
                      Predicate.bind (TypeRel.widen_i_i_i x xaa t)
                        (fn () => Predicate.single Type.Void)))))
    | Expr.Call (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          Predicate.bind (Type.is_class_type_of_i_o xba)
            (fn xbb =>
              Predicate.bind (wTs_Fii_oB_i_i_i_o is_lub x xa c3)
                (fn xaa =>
                  Predicate.bind (TypeRel.method_i_i_i_o_o_o_o x xbb c2)
                    (fn (ts, (t, (_, _))) =>
                      Predicate.bind
                        (Predicate_Compile_Alternative_Defs.list_all2_FiiB_i_i
                          (TypeRel.widen_i_i_i x) xaa ts)
                        (fn () => Predicate.single t)))))
    | Expr.Block (c1, c2, c3, c4) =>
      Predicate.bind (Predicate.if_pred (TypeRel.is_type x c2))
        (fn () =>
          Predicate.bind
            (Predicate.if_pred
              (case c3 of NONE => true
                | SOME v =>
                  (case Heap.typeof v of NONE => false
                    | SOME t => TypeRel.widen x t c2)))
            (fn () =>
              wT_Fii_oB_i_i_i_o is_lub x
                (Fun.fun_upd Stringa.equal_literal xa c1 (SOME c2)) c4))
    | Expr.Synchronized ((), c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
        (fn xba =>
          Predicate.bind (Type.is_refT_i xba)
            (fn () =>
              Predicate.bind
                (Predicate.if_pred (not (Type.equal_tya xba Type.Nt)))
                (fn () => wT_Fii_oB_i_i_i_o is_lub x xa c3)))
    | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
    | Expr.Seq (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn _ => wT_Fii_oB_i_i_i_o is_lub x xa c2)
    | Expr.Cond (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 Type.Boolean)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xba =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3) (is_lub xba)))
    | Expr.While (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 Type.Boolean)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn _ => Predicate.single Type.Void))
    | Expr.Throw c1 =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn a =>
          (case a of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class c =>
              Predicate.bind
                (Transitive_Closure_Table.rtranclp_FioB_i_i
                  Stringa.equal_literal (TypeRel.subcls1_i_i_o x) c
                  Type.throwable)
                (fn () => Predicate.single Type.Void)
            | Type.Array _ => Predicate.bot_pred))
    | Expr.TryCatch (c1, c2, c3, c4) =>
      Predicate.bind
        (Transitive_Closure_Table.rtranclp_FioB_i_i Stringa.equal_literal
          (TypeRel.subcls1_i_i_o x) c2 Type.throwable)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn xba =>
              Predicate.bind
                (wT_Fii_oB_i_i_i_i is_lub x
                  (Fun.fun_upd Stringa.equal_literal xa c3
                    (SOME (Type.Class c2)))
                  c4 xba)
                (fn () => Predicate.single xba))))
and wTs_Fii_oB_i_i_i_o is_lub x xa xb =
  (case xb of [] => Predicate.single []
    | c1 :: c2 =>
      Predicate.bind (wTs_Fii_oB_i_i_i_o is_lub x xa c2)
        (fn xba =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn xc => Predicate.single (xc :: xba))))
and wT_Fii_oB_i_i_i_i is_lub x xa xb xc =
  (case xb
    of Expr.New _ =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class c1 =>
          Predicate.bind (TypeRel.is_class_i_i x c1)
            (fn () => Predicate.single ())
        | Type.Array _ => Predicate.bot_pred)
    | Expr.NewArray (_, c2) =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array c1 =>
          Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
            (fn () =>
              Predicate.bind
                (Predicate.if_pred (TypeRel.is_type x (Type.Array c1)))
                (fn () => Predicate.single ())))
    | Expr.Cast (_, c2) =>
      Predicate.bind (Predicate.if_pred (TypeRel.is_type x xc))
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xaa =>
              Predicate.bind
                (Predicate.if_pred
                  (TypeRel.widen x xc xaa orelse TypeRel.widen x xaa xc))
                (fn () => Predicate.single ())))
    | Expr.InstanceOf (c1, c2) =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean =>
          Predicate.bind (Predicate.if_pred (TypeRel.is_type x c2))
            (fn () =>
              Predicate.bind (Type.is_refT_i c2)
                (fn () =>
                  Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
                    (fn xaa =>
                      Predicate.bind
                        (Predicate.if_pred
                          (TypeRel.widen x c2 xaa orelse
                            TypeRel.widen x xaa c2))
                        (fn () => Predicate.single ()))))
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.Val c1 =>
      Predicate.bind
        (TypeRel.eq_i_i (Option.equal_option Type.equal_ty) (Heap.typeof c1)
          (SOME xc))
        (fn () => Predicate.single ())
    | Expr.BinOp (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
            (fn xaa =>
              Predicate.bind (BinOp.wT_binop_i_i_i_i_i x xba c2 xaa xc)
                (fn () => Predicate.single ())))
    | Expr.Var c1 =>
      Predicate.bind
        (TypeRel.eq_i_i (Option.equal_option Type.equal_ty) (xa c1) (SOME xc))
        (fn () => Predicate.single ())
    | Expr.LAss (c1, c2) =>
      (case xc
        of Type.Void =>
          Predicate.bind (Predicate.if_pred (not ((c1 : string) = "this")))
            (fn () =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
                (fn xba =>
                  (case xa c1 of NONE => Predicate.bot_pred
                    | SOME t =>
                      Predicate.bind (TypeRel.widen_i_i_i x xba t)
                        (fn () => Predicate.single ()))))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.AAcc (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 (Type.Array xc))
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
            (fn () => Predicate.single ()))
    | Expr.AAss (c1, c2, c3) =>
      (case xc
        of Type.Void =>
          Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
            (fn () =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
                (fn a =>
                  (case a of Type.Void => Predicate.bot_pred
                    | Type.Boolean => Predicate.bot_pred
                    | Type.Integer => Predicate.bot_pred
                    | Type.Nt => Predicate.bot_pred
                    | Type.Class _ => Predicate.bot_pred
                    | Type.Array t =>
                      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
                        (fn xaa =>
                          Predicate.bind (TypeRel.widen_i_i_i x xaa t)
                            (fn () => Predicate.single ())))))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.ALen c1 =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn a =>
              (case a of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.single ()))
        | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.FAcc (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xaa =>
          Predicate.bind (Type.is_class_type_of_i_o xaa)
            (fn xab =>
              Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i x xab c2 c3)
                (fn (_, _) => Predicate.single ())))
    | Expr.FAss (c1, c2, c3, c4) =>
      (case xc
        of Type.Void =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn xba =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c4)
                (fn xaa =>
                  Predicate.bind (Type.is_class_type_of_i_o xba)
                    (fn xaaa =>
                      Predicate.bind
                        (TypeRel.sees_field_i_i_i_o_o_i x xaaa c2 c3)
                        (fn (t, _) =>
                          Predicate.bind (TypeRel.widen_i_i_i x xaa t)
                            (fn () => Predicate.single ())))))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.Call (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          Predicate.bind (Type.is_class_type_of_i_o xba)
            (fn xbb =>
              Predicate.bind (wTs_Fii_oB_i_i_i_o is_lub x xa c3)
                (fn xaa =>
                  Predicate.bind (TypeRel.method_i_i_i_o_o_o_o x xbb c2)
                    (fn (ts, (_, (_, _))) =>
                      Predicate.bind
                        (Predicate_Compile_Alternative_Defs.list_all2_FiiB_i_i
                          (TypeRel.widen_i_i_i x) xaa ts)
                        (fn () => Predicate.single ())))))
    | Expr.Block (c1, c2, c3, c4) =>
      Predicate.bind
        (wT_Fii_oB_i_i_i_i is_lub x
          (Fun.fun_upd Stringa.equal_literal xa c1 (SOME c2)) c4 xc)
        (fn () =>
          Predicate.bind (Predicate.if_pred (TypeRel.is_type x c2))
            (fn () =>
              Predicate.bind
                (Predicate.if_pred
                  (case c3 of NONE => true
                    | SOME v =>
                      (case Heap.typeof v of NONE => false
                        | SOME t => TypeRel.widen x t c2)))
                (fn () => Predicate.single ())))
    | Expr.Synchronized ((), c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c3 xc)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xd =>
              Predicate.bind (Type.is_refT_i xd)
                (fn () =>
                  Predicate.bind
                    (Predicate.if_pred (not (Type.equal_tya xd Type.Nt)))
                    (fn () => Predicate.single ()))))
    | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
    | Expr.Seq (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 xc)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn _ => Predicate.single ()))
    | Expr.Cond (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 Type.Boolean)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xba =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
                (fn xd =>
                  Predicate.bind (is_lub xba xd)
                    (fn _ => Predicate.single ()))))
    | Expr.While (c1, c2) =>
      (case xc
        of Type.Void =>
          Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 Type.Boolean)
            (fn () =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
                (fn _ => Predicate.single ()))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.Throw c1 =>
      (case xc
        of Type.Void =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn a =>
              (case a of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class c =>
                  Predicate.bind
                    (Transitive_Closure_Table.rtranclp_FioB_i_i
                      Stringa.equal_literal (TypeRel.subcls1_i_i_o x) c
                      Type.throwable)
                    (fn () => Predicate.single ())
                | Type.Array _ => Predicate.bot_pred))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.TryCatch (c1, c2, c3, c4) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 xc)
        (fn () =>
          Predicate.bind
            (wT_Fii_oB_i_i_i_i is_lub x
              (Fun.fun_upd Stringa.equal_literal xa c3 (SOME (Type.Class c2)))
              c4 xc)
            (fn () =>
              Predicate.bind
                (Transitive_Closure_Table.rtranclp_FioB_i_i
                  Stringa.equal_literal (TypeRel.subcls1_i_i_o x) c2
                  Type.throwable)
                (fn () => Predicate.single ()))));

fun is_lub_sup_i_i_i_o x xa xb =
  (case SemiType.sup x xa xb of Err.Err => Predicate.bot_pred
    | Err.Ok a => Predicate.single a);

fun wT_code_i_i_i_o x xa xb = wT_Fii_oB_i_i_i_o (is_lub_sup_i_i_i_o x) x xa xb;

end; (*struct WellType*)

structure FWLocking : sig
  val redT_updLs :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * IntInf.int) option)) FinFun.finfun ->
        'b -> ('a, (FWState.lock_action list)) FinFun.finfun ->
                ('a, (('b * IntInf.int) option)) FinFun.finfun
  val acquire_all :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * IntInf.int) option)) FinFun.finfun ->
        'b -> ('a, IntInf.int) FinFun.finfun ->
                ('a, (('b * IntInf.int) option)) FinFun.finfun
  val lock_ok_las :
    'a Card_Univ.card_UNIV * 'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * IntInf.int) option)) FinFun.finfun ->
        'b -> ('a, (FWState.lock_action list)) FinFun.finfun -> bool
  val may_acquire_all :
    'a Card_Univ.card_UNIV * 'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * IntInf.int) option)) FinFun.finfun ->
        'b -> ('a, IntInf.int) FinFun.finfun -> bool
end = struct

fun redT_updLs A_ B_ ls t las =
  FinFun.finfun_comp (fn (l, a) => FWLock.upd_locks B_ l t a)
    (FinFun.finfun_Diag A_
      (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
      (List.equal_list FWState.equal_lock_action) ls las);

fun acquire_all A_ B_ ls t ln =
  FinFun.finfun_comp (fn (l, a) => FWLock.acquire_locks l t a)
    (FinFun.finfun_Diag A_
      (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
      Arith.equal_nat ls ln);

fun lock_ok_las (A1_, A2_) B_ ls t las =
  FinFun.finfun_All (A1_, A2_)
    (FinFun.finfun_comp (fn (l, a) => FWLock.lock_actions_ok B_ l t a)
      (FinFun.finfun_Diag A2_
        (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
        (List.equal_list FWState.equal_lock_action) ls las));

fun may_acquire_all (A1_, A2_) B_ ls t ln =
  FinFun.finfun_All (A1_, A2_)
    (FinFun.finfun_comp
      (fn (lock, n) =>
        (if IntInf.< ((0 : IntInf.int), n) then FWLock.may_lock B_ lock t
          else true))
      (FinFun.finfun_Diag A2_
        (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
        Arith.equal_nat ls ln));

end; (*struct FWLocking*)

structure JWellForm : sig
  val typeable_with_i_i_i_i :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      (string -> Type.ty option) ->
        (string, unit, 'a) Expr.exp -> Type.ty -> unit Predicate.pred
  val typeable_with :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      (string -> Type.ty option) ->
        (string, unit, 'a) Expr.exp -> Type.ty -> bool
  val wf_J_mdecl :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      string ->
        string *
          (Type.ty list *
            (Type.ty * (string list * (string, unit, 'a) Expr.exp))) ->
          bool
  val wf_J_prog :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog -> bool
end = struct

fun typeable_with_i_i_i_i x xa xb xc =
  Predicate.bind (WellType.wT_code_i_i_i_o x xa xb)
    (fn xaa =>
      Predicate.bind (TypeRel.widen_i_i_i x xaa xc)
        (fn () => Predicate.single ()));

fun typeable_with x1 x2 x3 x4 =
  Predicate.holds (typeable_with_i_i_i_i x1 x2 x3 x4);

fun wf_J_mdecl p c =
  (fn (_, (ts, (t, (pns, body)))) =>
    (((List.size_list ts) : IntInf.int) = (List.size_list pns)) andalso
      (List.distinct Stringa.equal_literal pns andalso
        (not (List.member Stringa.equal_literal pns "this") andalso
          (typeable_with p
             (Map.map_upds Stringa.equal_literal
               (Fun.fun_upd Stringa.equal_literal (fn _ => NONE) "this"
                 (SOME (Type.Class c)))
               pns ts)
             body t andalso
            DefAss.d Stringa.equal_literal body
              (SOME (Lattices.sup_fun Lattices.lattice_bool
                      (Set.insert Stringa.equal_literal "this"
                        (Orderings.bot_fun Orderings.bot_bool))
                      (List.set Stringa.equal_literal pns)))))));

fun wf_J_prog x = WellForm.wf_prog wf_J_mdecl x;

end; (*struct JWellForm*)

structure SmallStep : sig
  val extNTA2J :
    'a HOL.equal ->
      (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
        string * (string * 'a) ->
          (string, unit, 'a) Expr.exp * (string -> 'a Value.vala option)
  val red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o :
    'a HOL.equal * 'a Value.addr ->
      ('a -> 'b) ->
        ('b -> 'a) ->
          'c -> ('c -> string -> 'c * 'a option) ->
                  ('c -> Type.ty -> IntInf.int -> 'c * 'a option) ->
                    ('c -> 'a -> Type.ty option) ->
                      ('c -> 'a -> IntInf.int) ->
                        ('c ->
                          'a -> Heap.addr_loc ->
                                  'a Value.vala Predicate.pred) ->
                          ('c ->
                            'a -> Heap.addr_loc ->
                                    'a Value.vala -> 'c Predicate.pred) ->
                            (('a, (FWState.lock_action list)) FinFun.finfun *
                               (('b, (string * (string * 'a)), 'c)
                                  FWState.new_thread_action list *
                                 ('b FWState.conditional_action list *
                                   (('b, 'a) FWState.wait_set_action list *
                                     ('b FWState.interrupt_action list *
                                       ('a, 'b)
 Observable_Events.obs_event list)))) ->
                              ('a, (FWState.lock_action list)) FinFun.finfun *
                                (('b, 'd, 'c) FWState.new_thread_action list *
                                  ('b FWState.conditional_action list *
                                    (('b, 'a) FWState.wait_set_action list *
                                      ('b FWState.interrupt_action list *
('a, 'b) Observable_Events.obs_event list))))) ->
                              (string list * (string, unit, 'a) Expr.exp)
                                TypeRel.prog ->
                                'b -> (string, unit, 'a) Expr.exp ->
'c * (string -> 'a Value.vala option) ->
  ((('a, (FWState.lock_action list)) FinFun.finfun *
     (('b, 'd, 'c) FWState.new_thread_action list *
       ('b FWState.conditional_action list *
         (('b, 'a) FWState.wait_set_action list *
           ('b FWState.interrupt_action list *
             ('a, 'b) Observable_Events.obs_event list))))) *
    ((string, unit, 'a) Expr.exp * ('c * (string -> 'a Value.vala option))))
    Predicate.pred
  val reds_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o :
    'a HOL.equal * 'a Value.addr ->
      ('a -> 'b) ->
        ('b -> 'a) ->
          'c -> ('c -> string -> 'c * 'a option) ->
                  ('c -> Type.ty -> IntInf.int -> 'c * 'a option) ->
                    ('c -> 'a -> Type.ty option) ->
                      ('c -> 'a -> IntInf.int) ->
                        ('c ->
                          'a -> Heap.addr_loc ->
                                  'a Value.vala Predicate.pred) ->
                          ('c ->
                            'a -> Heap.addr_loc ->
                                    'a Value.vala -> 'c Predicate.pred) ->
                            (('a, (FWState.lock_action list)) FinFun.finfun *
                               (('b, (string * (string * 'a)), 'c)
                                  FWState.new_thread_action list *
                                 ('b FWState.conditional_action list *
                                   (('b, 'a) FWState.wait_set_action list *
                                     ('b FWState.interrupt_action list *
                                       ('a, 'b)
 Observable_Events.obs_event list)))) ->
                              ('a, (FWState.lock_action list)) FinFun.finfun *
                                (('b, 'd, 'c) FWState.new_thread_action list *
                                  ('b FWState.conditional_action list *
                                    (('b, 'a) FWState.wait_set_action list *
                                      ('b FWState.interrupt_action list *
('a, 'b) Observable_Events.obs_event list))))) ->
                              (string list * (string, unit, 'a) Expr.exp)
                                TypeRel.prog ->
                                'b -> (string, unit, 'a) Expr.exp list ->
'c * (string -> 'a Value.vala option) ->
  ((('a, (FWState.lock_action list)) FinFun.finfun *
     (('b, 'd, 'c) FWState.new_thread_action list *
       ('b FWState.conditional_action list *
         (('b, 'a) FWState.wait_set_action list *
           ('b FWState.interrupt_action list *
             ('a, 'b) Observable_Events.obs_event list))))) *
    ((string, unit, 'a) Expr.exp list *
      ('c * (string -> 'a Value.vala option))))
    Predicate.pred
end = struct

fun extNTA2J A_ p =
  (fn (c, (m, a)) =>
    let
      val (d, (_, (_, meth))) =
        TypeRel.method
          (Product_Type.equal_prod (List.equal_list Stringa.equal_literal)
            (Expr.equal_exp Stringa.equal_literal Product_Type.equal_unit A_))
          p c m;
      val (_, body) = Option.the meth;
    in
      (Expr.Block ("this", Type.Class d, SOME (Value.Addr a), body),
        (fn _ => NONE))
    end);

fun red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl xm xn xo
  xp xq xr heap_read heap_write xs xt xu xv (xw, xx) =
  (case xv
    of Expr.New c1 =>
      Predicate.sup_pred
        (case xo xw c1
          of (h, NONE) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw
                   (Expr.Val
                     (Value.Addr
                       (StartConfig.addr_of_sys_xcpt A2_ xn xo
                         "java/lang/OutOfMemoryError"))),
                  (h, xx)))
          | (h, SOME _) => Predicate.bot_pred)
        (case xo xw c1 of (h, NONE) => Predicate.bot_pred
          | (h, SOME a) =>
            Predicate.single
              (FWState.thread_action_to_thread_action A1_
                 Observable_Events.obs_action_obs_event
                 (FWState.ObsAction (Observable_Events.NewObj (a, c1)))
                 (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Val (Value.Addr a), (h, xx))))
    | Expr.NewArray (c1, c2) =>
      Predicate.sup_pred
        (case c2 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val Value.Unit => Predicate.bot_pred
          | Expr.Val Value.Null => Predicate.bot_pred
          | Expr.Val (Value.Bool _) => Predicate.bot_pred
          | Expr.Val (Value.Intg i) =>
            Predicate.sup_pred
              (Predicate.bind
                (Predicate.if_pred
                  (Word.word_sle
                    (Type_Length.len_bit0
                      (Type_Length.len_bit0
                        (Type_Length.len_bit0
                          (Type_Length.len_bit0
                            (Type_Length.len_bit0 Type_Length.len_num1)))))
                    (Word.zero_word
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0
                              (Type_Length.len0_bit0 Type_Length.len0_num1))))))
                    i))
                (fn () =>
                  (case xp xw c1
                          (IntInf.max (0,
                            (Word.sint
                              (Type_Length.len_bit0
                                (Type_Length.len_bit0
                                  (Type_Length.len_bit0
                                    (Type_Length.len_bit0
                                      (Type_Length.len_bit0
Type_Length.len_num1)))))
                              i)))
                    of (h, NONE) =>
                      Predicate.single
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (Expr.Throw
                             (Expr.Val
                               (Value.Addr
                                 (StartConfig.addr_of_sys_xcpt A2_ xn xo
                                   "java/lang/OutOfMemoryError"))),
                            (h, xx)))
                    | (h, SOME _) => Predicate.bot_pred)))
              (Predicate.sup_pred
                (Predicate.bind
                  (Predicate.if_pred
                    (Word.word_sless
                      (Type_Length.len_bit0
                        (Type_Length.len_bit0
                          (Type_Length.len_bit0
                            (Type_Length.len_bit0
                              (Type_Length.len_bit0 Type_Length.len_num1)))))
                      i (Word.zero_word
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0
                              (Type_Length.len0_bit0
                                (Type_Length.len0_bit0
                                  (Type_Length.len0_bit0
                                    Type_Length.len0_num1))))))))
                  (fn () =>
                    Predicate.single
                      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                        (Expr.Throw
                           (Expr.Val
                             (Value.Addr
                               (StartConfig.addr_of_sys_xcpt A2_ xn xo
                                 "java/lang/NegativeArraySizeException"))),
                          (xw, xx)))))
                (Predicate.bind
                  (Predicate.if_pred
                    (Word.word_sle
                      (Type_Length.len_bit0
                        (Type_Length.len_bit0
                          (Type_Length.len_bit0
                            (Type_Length.len_bit0
                              (Type_Length.len_bit0 Type_Length.len_num1)))))
                      (Word.zero_word
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0
                              (Type_Length.len0_bit0
                                (Type_Length.len0_bit0
                                  Type_Length.len0_num1))))))
                      i))
                  (fn () =>
                    (case xp xw c1
                            (IntInf.max (0,
                              (Word.sint
                                (Type_Length.len_bit0
                                  (Type_Length.len_bit0
                                    (Type_Length.len_bit0
                                      (Type_Length.len_bit0
(Type_Length.len_bit0 Type_Length.len_num1)))))
                                i)))
                      of (h, NONE) => Predicate.bot_pred
                      | (h, SOME a) =>
                        Predicate.single
                          (FWState.thread_action_to_thread_action A1_
                             Observable_Events.obs_action_obs_event
                             (FWState.ObsAction
                               (Observable_Events.NewArr
                                 (a, c1,
                                   IntInf.max (0,
                                     (Word.sint
                                       (Type_Length.len_bit0
 (Type_Length.len_bit0
   (Type_Length.len_bit0
     (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
                                       i)))))
                             (FinFun.Finfun_const [],
                               ([], ([], ([], ([], []))))),
                            (Expr.Val (Value.Addr a), (h, xx)))))))
          | Expr.Val (Value.Addr _) => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu c2 (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.NewArray (c1, e), (xf, yf)))))
    | Expr.Cast (c1, c2) =>
      Predicate.sup_pred
        (case c2 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            Predicate.sup_pred
              (case Heap.typeof_h A2_ xq (State.hp (xw, xx)) v
                of NONE => Predicate.bot_pred
                | SOME u =>
                  Predicate.bind
                    (Predicate.not_pred (TypeRel.widen_i_i_i xt u c1))
                    (fn () =>
                      Predicate.single
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (Expr.Throw
                             (Expr.Val
                               (Value.Addr
                                 (StartConfig.addr_of_sys_xcpt A2_ xn xo
                                   "java/lang/ClassCastException"))),
                            (xw, xx)))))
              (case Heap.typeof_h A2_ xq (State.hp (xw, xx)) v
                of NONE => Predicate.bot_pred
                | SOME u =>
                  Predicate.bind (TypeRel.widen_i_i_i xt u c1)
                    (fn () =>
                      Predicate.single
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (Expr.Val v, (xw, xx)))))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu c2 (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.Cast (c1, e), (xf, yf)))))
    | Expr.InstanceOf (c1, c2) =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            (case Heap.typeof_h A2_ xq (State.hp (xw, xx)) v
              of NONE => Predicate.bot_pred
              | SOME u =>
                Predicate.single
                  ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Val
                       (Value.Bool
                         (not (Value.equal_vala A1_ v Value.Null) andalso
                           TypeRel.widen xt u c2)),
                      (xw, xx))))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu c1 (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.InstanceOf (e, c2), (xf, yf)))))
    | Expr.Val _ => Predicate.bot_pred
    | Expr.BinOp (c1, c2, c3) =>
      Predicate.sup_pred
        (case c3 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val _ => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            (case c1 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val _ =>
                Predicate.single
                  ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.sup_pred
          (case c1 of Expr.New _ => Predicate.bot_pred
            | Expr.NewArray (_, _) => Predicate.bot_pred
            | Expr.Cast (_, _) => Predicate.bot_pred
            | Expr.InstanceOf (_, _) => Predicate.bot_pred
            | Expr.Val _ => Predicate.bot_pred
            | Expr.BinOp (_, _, _) => Predicate.bot_pred
            | Expr.Var _ => Predicate.bot_pred
            | Expr.LAss (_, _) => Predicate.bot_pred
            | Expr.AAcc (_, _) => Predicate.bot_pred
            | Expr.AAss (_, _, _) => Predicate.bot_pred
            | Expr.ALen _ => Predicate.bot_pred
            | Expr.FAcc (_, _, _) => Predicate.bot_pred
            | Expr.FAss (_, _, _, _) => Predicate.bot_pred
            | Expr.Call (_, _, _) => Predicate.bot_pred
            | Expr.Block (_, _, _, _) => Predicate.bot_pred
            | Expr.Synchronized (_, _, _) => Predicate.bot_pred
            | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
            | Expr.Seq (_, _) => Predicate.bot_pred
            | Expr.Cond (_, _, _) => Predicate.bot_pred
            | Expr.While (_, _) => Predicate.bot_pred
            | Expr.Throw (Expr.New _) => Predicate.bot_pred
            | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Addr a)) =>
              Predicate.single
                ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                  (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
            | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Var _) => Predicate.bot_pred
            | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
            | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
            | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
            | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          (Predicate.sup_pred
            (case c3 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val v =>
                (case c1 of Expr.New _ => Predicate.bot_pred
                  | Expr.NewArray (_, _) => Predicate.bot_pred
                  | Expr.Cast (_, _) => Predicate.bot_pred
                  | Expr.InstanceOf (_, _) => Predicate.bot_pred
                  | Expr.Val va =>
                    Predicate.sup_pred
                      (case BinOp.binop (A1_, A2_) xn xo c2 va v
                        of NONE => Predicate.bot_pred
                        | SOME (Sum_Type.Inl _) => Predicate.bot_pred
                        | SOME (Sum_Type.Inr a) =>
                          Predicate.single
                            ((FinFun.Finfun_const [],
                               ([], ([], ([], ([], []))))),
                              (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx))))
                      (case BinOp.binop (A1_, A2_) xn xo c2 va v
                        of NONE => Predicate.bot_pred
                        | SOME (Sum_Type.Inl vb) =>
                          Predicate.single
                            ((FinFun.Finfun_const [],
                               ([], ([], ([], ([], []))))),
                              (Expr.Val vb, (xw, xx)))
                        | SOME (Sum_Type.Inr _) => Predicate.bot_pred)
                  | Expr.BinOp (_, _, _) => Predicate.bot_pred
                  | Expr.Var _ => Predicate.bot_pred
                  | Expr.LAss (_, _) => Predicate.bot_pred
                  | Expr.AAcc (_, _) => Predicate.bot_pred
                  | Expr.AAss (_, _, _) => Predicate.bot_pred
                  | Expr.ALen _ => Predicate.bot_pred
                  | Expr.FAcc (_, _, _) => Predicate.bot_pred
                  | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                  | Expr.Call (_, _, _) => Predicate.bot_pred
                  | Expr.Block (_, _, _, _) => Predicate.bot_pred
                  | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                  | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                  | Expr.Seq (_, _) => Predicate.bot_pred
                  | Expr.Cond (_, _, _) => Predicate.bot_pred
                  | Expr.While (_, _) => Predicate.bot_pred
                  | Expr.Throw _ => Predicate.bot_pred
                  | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
            (Predicate.sup_pred
              (case c1 of Expr.New _ => Predicate.bot_pred
                | Expr.NewArray (_, _) => Predicate.bot_pred
                | Expr.Cast (_, _) => Predicate.bot_pred
                | Expr.InstanceOf (_, _) => Predicate.bot_pred
                | Expr.Val v =>
                  Predicate.bind
                    (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                      (A1_, A2_) xl xm xn xo xp xq xr heap_read heap_write xs xt
                      xu c3 (xw, xx))
                    (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
                      Predicate.single
                        ((xa, (xb, (xc, (xd, (xe, ye))))),
                          (Expr.BinOp (Expr.Val v, c2, e), (xf, yf))))
                | Expr.BinOp (_, _, _) => Predicate.bot_pred
                | Expr.Var _ => Predicate.bot_pred
                | Expr.LAss (_, _) => Predicate.bot_pred
                | Expr.AAcc (_, _) => Predicate.bot_pred
                | Expr.AAss (_, _, _) => Predicate.bot_pred
                | Expr.ALen _ => Predicate.bot_pred
                | Expr.FAcc (_, _, _) => Predicate.bot_pred
                | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                | Expr.Call (_, _, _) => Predicate.bot_pred
                | Expr.Block (_, _, _, _) => Predicate.bot_pred
                | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                | Expr.Seq (_, _) => Predicate.bot_pred
                | Expr.Cond (_, _, _) => Predicate.bot_pred
                | Expr.While (_, _) => Predicate.bot_pred
                | Expr.Throw _ => Predicate.bot_pred
                | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              (Predicate.bind
                (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                  (A1_, A2_) xl xm xn xo xp xq xr heap_read heap_write xs xt xu
                  c1 (xw, xx))
                (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
                  Predicate.single
                    ((xa, (xb, (xc, (xd, (xe, ye))))),
                      (Expr.BinOp (e, c2, c3), (xf, yf))))))))
    | Expr.Var c1 =>
      (case State.lcl (xw, xx) c1 of NONE => Predicate.bot_pred
        | SOME v =>
          Predicate.single
            ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
              (Expr.Val v, (xw, xx))))
    | Expr.LAss (c1, c2) =>
      Predicate.sup_pred
        (case c2 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Val Value.Unit,
                  (xw, Fun.fun_upd Stringa.equal_literal xx c1 (SOME v))))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu c2 (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.LAss (c1, e), (xf, yf)))))
    | Expr.AAcc (c1, c2) =>
      Predicate.sup_pred
        (case c2 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val _ => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            (case c1 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val _ =>
                Predicate.single
                  ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.sup_pred
          (case c1 of Expr.New _ => Predicate.bot_pred
            | Expr.NewArray (_, _) => Predicate.bot_pred
            | Expr.Cast (_, _) => Predicate.bot_pred
            | Expr.InstanceOf (_, _) => Predicate.bot_pred
            | Expr.Val _ => Predicate.bot_pred
            | Expr.BinOp (_, _, _) => Predicate.bot_pred
            | Expr.Var _ => Predicate.bot_pred
            | Expr.LAss (_, _) => Predicate.bot_pred
            | Expr.AAcc (_, _) => Predicate.bot_pred
            | Expr.AAss (_, _, _) => Predicate.bot_pred
            | Expr.ALen _ => Predicate.bot_pred
            | Expr.FAcc (_, _, _) => Predicate.bot_pred
            | Expr.FAss (_, _, _, _) => Predicate.bot_pred
            | Expr.Call (_, _, _) => Predicate.bot_pred
            | Expr.Block (_, _, _, _) => Predicate.bot_pred
            | Expr.Synchronized (_, _, _) => Predicate.bot_pred
            | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
            | Expr.Seq (_, _) => Predicate.bot_pred
            | Expr.Cond (_, _, _) => Predicate.bot_pred
            | Expr.While (_, _) => Predicate.bot_pred
            | Expr.Throw (Expr.New _) => Predicate.bot_pred
            | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Addr a)) =>
              Predicate.single
                ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                  (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
            | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Var _) => Predicate.bot_pred
            | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
            | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
            | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
            | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          (Predicate.sup_pred
            (case c2 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val v =>
                Predicate.sup_pred
                  (case v of Value.Unit => Predicate.bot_pred
                    | Value.Null => Predicate.bot_pred
                    | Value.Bool _ => Predicate.bot_pred
                    | Value.Intg i =>
                      (case c1 of Expr.New _ => Predicate.bot_pred
                        | Expr.NewArray (_, _) => Predicate.bot_pred
                        | Expr.Cast (_, _) => Predicate.bot_pred
                        | Expr.InstanceOf (_, _) => Predicate.bot_pred
                        | Expr.Val Value.Unit => Predicate.bot_pred
                        | Expr.Val Value.Null => Predicate.bot_pred
                        | Expr.Val (Value.Bool _) => Predicate.bot_pred
                        | Expr.Val (Value.Intg _) => Predicate.bot_pred
                        | Expr.Val (Value.Addr a) =>
                          Predicate.bind
                            (Predicate.if_pred
                              (Word.word_sle
                                (Type_Length.len_bit0
                                  (Type_Length.len_bit0
                                    (Type_Length.len_bit0
                                      (Type_Length.len_bit0
(Type_Length.len_bit0 Type_Length.len_num1)))))
                                (Word.zero_word
                                  (Type_Length.len0_bit0
                                    (Type_Length.len0_bit0
                                      (Type_Length.len0_bit0
(Type_Length.len0_bit0 (Type_Length.len0_bit0 Type_Length.len0_num1))))))
                                i))
                            (fn () =>
                              Predicate.bind
                                (Predicate.if_pred
                                  (IntInf.< (Word.sint
       (Type_Length.len_bit0
         (Type_Length.len_bit0
           (Type_Length.len_bit0
             (Type_Length.len_bit0
               (Type_Length.len_bit0 Type_Length.len_num1)))))
       i, (xr xw a))))
                                (fn () =>
                                  (case xq xw a of NONE => Predicate.bot_pred
                                    | SOME Type.Void => Predicate.bot_pred
                                    | SOME Type.Boolean => Predicate.bot_pred
                                    | SOME Type.Integer => Predicate.bot_pred
                                    | SOME Type.Nt => Predicate.bot_pred
                                    | SOME (Type.Class _) => Predicate.bot_pred
                                    | SOME (Type.Array _) =>
                                      Predicate.bind
(heap_read xw a
  (Heap.ACell
    (IntInf.max (0,
      (Word.sint
        (Type_Length.len_bit0
          (Type_Length.len_bit0
            (Type_Length.len_bit0
              (Type_Length.len_bit0
                (Type_Length.len_bit0 Type_Length.len_num1)))))
        i)))))
(fn x =>
  Predicate.single
    (FWState.thread_action_to_thread_action A1_
       Observable_Events.obs_action_obs_event
       (FWState.ObsAction
         (Observable_Events.ReadMem
           (a, Heap.ACell
                 (IntInf.max (0,
                   (Word.sint
                     (Type_Length.len_bit0
                       (Type_Length.len_bit0
                         (Type_Length.len_bit0
                           (Type_Length.len_bit0
                             (Type_Length.len_bit0 Type_Length.len_num1)))))
                     i))),
             x)))
       (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
      (Expr.Val x, (xw, xx)))))))
                        | Expr.BinOp (_, _, _) => Predicate.bot_pred
                        | Expr.Var _ => Predicate.bot_pred
                        | Expr.LAss (_, _) => Predicate.bot_pred
                        | Expr.AAcc (_, _) => Predicate.bot_pred
                        | Expr.AAss (_, _, _) => Predicate.bot_pred
                        | Expr.ALen _ => Predicate.bot_pred
                        | Expr.FAcc (_, _, _) => Predicate.bot_pred
                        | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                        | Expr.Call (_, _, _) => Predicate.bot_pred
                        | Expr.Block (_, _, _, _) => Predicate.bot_pred
                        | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                        | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                        | Expr.Seq (_, _) => Predicate.bot_pred
                        | Expr.Cond (_, _, _) => Predicate.bot_pred
                        | Expr.While (_, _) => Predicate.bot_pred
                        | Expr.Throw _ => Predicate.bot_pred
                        | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
                    | Value.Addr _ => Predicate.bot_pred)
                  (Predicate.sup_pred
                    (case v of Value.Unit => Predicate.bot_pred
                      | Value.Null => Predicate.bot_pred
                      | Value.Bool _ => Predicate.bot_pred
                      | Value.Intg i =>
                        (case c1 of Expr.New _ => Predicate.bot_pred
                          | Expr.NewArray (_, _) => Predicate.bot_pred
                          | Expr.Cast (_, _) => Predicate.bot_pred
                          | Expr.InstanceOf (_, _) => Predicate.bot_pred
                          | Expr.Val Value.Unit => Predicate.bot_pred
                          | Expr.Val Value.Null => Predicate.bot_pred
                          | Expr.Val (Value.Bool _) => Predicate.bot_pred
                          | Expr.Val (Value.Intg _) => Predicate.bot_pred
                          | Expr.Val (Value.Addr a) =>
                            Predicate.bind
                              (Predicate.if_pred
                                (Word.word_sless
                                   (Type_Length.len_bit0
                                     (Type_Length.len_bit0
                                       (Type_Length.len_bit0
 (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
                                   i (Word.zero_word
                                       (Type_Length.len0_bit0
 (Type_Length.len0_bit0
   (Type_Length.len0_bit0
     (Type_Length.len0_bit0
       (Type_Length.len0_bit0 Type_Length.len0_num1)))))) orelse
                                  IntInf.<= ((xr (State.hp (xw, xx))
       a), Word.sint
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i)))
                              (fn () =>
                                (case xq (State.hp (xw, xx)) a
                                  of NONE => Predicate.bot_pred
                                  | SOME Type.Void => Predicate.bot_pred
                                  | SOME Type.Boolean => Predicate.bot_pred
                                  | SOME Type.Integer => Predicate.bot_pred
                                  | SOME Type.Nt => Predicate.bot_pred
                                  | SOME (Type.Class _) => Predicate.bot_pred
                                  | SOME (Type.Array _) =>
                                    Predicate.single
                                      ((FinFun.Finfun_const [],
 ([], ([], ([], ([], []))))),
(Expr.Throw
   (Expr.Val
     (Value.Addr
       (StartConfig.addr_of_sys_xcpt A2_ xn xo
         "java/lang/ArrayIndexOutOfBoundsException"))),
  (xw, xx)))))
                          | Expr.BinOp (_, _, _) => Predicate.bot_pred
                          | Expr.Var _ => Predicate.bot_pred
                          | Expr.LAss (_, _) => Predicate.bot_pred
                          | Expr.AAcc (_, _) => Predicate.bot_pred
                          | Expr.AAss (_, _, _) => Predicate.bot_pred
                          | Expr.ALen _ => Predicate.bot_pred
                          | Expr.FAcc (_, _, _) => Predicate.bot_pred
                          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                          | Expr.Call (_, _, _) => Predicate.bot_pred
                          | Expr.Block (_, _, _, _) => Predicate.bot_pred
                          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                          | Expr.Seq (_, _) => Predicate.bot_pred
                          | Expr.Cond (_, _, _) => Predicate.bot_pred
                          | Expr.While (_, _) => Predicate.bot_pred
                          | Expr.Throw _ => Predicate.bot_pred
                          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
                      | Value.Addr _ => Predicate.bot_pred)
                    (case c1 of Expr.New _ => Predicate.bot_pred
                      | Expr.NewArray (_, _) => Predicate.bot_pred
                      | Expr.Cast (_, _) => Predicate.bot_pred
                      | Expr.InstanceOf (_, _) => Predicate.bot_pred
                      | Expr.Val Value.Unit => Predicate.bot_pred
                      | Expr.Val Value.Null =>
                        Predicate.single
                          ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                            (Expr.Throw
                               (Expr.Val
                                 (Value.Addr
                                   (StartConfig.addr_of_sys_xcpt A2_ xn xo
                                     "java/lang/NullPointerException"))),
                              (xw, xx)))
                      | Expr.Val (Value.Bool _) => Predicate.bot_pred
                      | Expr.Val (Value.Intg _) => Predicate.bot_pred
                      | Expr.Val (Value.Addr _) => Predicate.bot_pred
                      | Expr.BinOp (_, _, _) => Predicate.bot_pred
                      | Expr.Var _ => Predicate.bot_pred
                      | Expr.LAss (_, _) => Predicate.bot_pred
                      | Expr.AAcc (_, _) => Predicate.bot_pred
                      | Expr.AAss (_, _, _) => Predicate.bot_pred
                      | Expr.ALen _ => Predicate.bot_pred
                      | Expr.FAcc (_, _, _) => Predicate.bot_pred
                      | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                      | Expr.Call (_, _, _) => Predicate.bot_pred
                      | Expr.Block (_, _, _, _) => Predicate.bot_pred
                      | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                      | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                      | Expr.Seq (_, _) => Predicate.bot_pred
                      | Expr.Cond (_, _, _) => Predicate.bot_pred
                      | Expr.While (_, _) => Predicate.bot_pred
                      | Expr.Throw _ => Predicate.bot_pred
                      | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred))
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
            (Predicate.sup_pred
              (case c1 of Expr.New _ => Predicate.bot_pred
                | Expr.NewArray (_, _) => Predicate.bot_pred
                | Expr.Cast (_, _) => Predicate.bot_pred
                | Expr.InstanceOf (_, _) => Predicate.bot_pred
                | Expr.Val ab =>
                  Predicate.bind
                    (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                      (A1_, A2_) xl xm xn xo xp xq xr heap_read heap_write xs xt
                      xu c2 (xw, xx))
                    (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (i, (xf, yf))) =>
                      Predicate.single
                        ((xa, (xb, (xc, (xd, (xe, ye))))),
                          (Expr.AAcc (Expr.Val ab, i), (xf, yf))))
                | Expr.BinOp (_, _, _) => Predicate.bot_pred
                | Expr.Var _ => Predicate.bot_pred
                | Expr.LAss (_, _) => Predicate.bot_pred
                | Expr.AAcc (_, _) => Predicate.bot_pred
                | Expr.AAss (_, _, _) => Predicate.bot_pred
                | Expr.ALen _ => Predicate.bot_pred
                | Expr.FAcc (_, _, _) => Predicate.bot_pred
                | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                | Expr.Call (_, _, _) => Predicate.bot_pred
                | Expr.Block (_, _, _, _) => Predicate.bot_pred
                | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                | Expr.Seq (_, _) => Predicate.bot_pred
                | Expr.Cond (_, _, _) => Predicate.bot_pred
                | Expr.While (_, _) => Predicate.bot_pred
                | Expr.Throw _ => Predicate.bot_pred
                | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              (Predicate.bind
                (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                  (A1_, A2_) xl xm xn xo xp xq xr heap_read heap_write xs xt xu
                  c1 (xw, xx))
                (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (a, (xf, yf))) =>
                  Predicate.single
                    ((xa, (xb, (xc, (xd, (xe, ye))))),
                      (Expr.AAcc (a, c2), (xf, yf))))))))
    | Expr.AAss (c1, c2, c3) =>
      Predicate.sup_pred
        (case c3 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val _ => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            (case c2 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val _ =>
                (case c1 of Expr.New _ => Predicate.bot_pred
                  | Expr.NewArray (_, _) => Predicate.bot_pred
                  | Expr.Cast (_, _) => Predicate.bot_pred
                  | Expr.InstanceOf (_, _) => Predicate.bot_pred
                  | Expr.Val _ =>
                    Predicate.single
                      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                        (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
                  | Expr.BinOp (_, _, _) => Predicate.bot_pred
                  | Expr.Var _ => Predicate.bot_pred
                  | Expr.LAss (_, _) => Predicate.bot_pred
                  | Expr.AAcc (_, _) => Predicate.bot_pred
                  | Expr.AAss (_, _, _) => Predicate.bot_pred
                  | Expr.ALen _ => Predicate.bot_pred
                  | Expr.FAcc (_, _, _) => Predicate.bot_pred
                  | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                  | Expr.Call (_, _, _) => Predicate.bot_pred
                  | Expr.Block (_, _, _, _) => Predicate.bot_pred
                  | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                  | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                  | Expr.Seq (_, _) => Predicate.bot_pred
                  | Expr.Cond (_, _, _) => Predicate.bot_pred
                  | Expr.While (_, _) => Predicate.bot_pred
                  | Expr.Throw _ => Predicate.bot_pred
                  | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.sup_pred
          (case c2 of Expr.New _ => Predicate.bot_pred
            | Expr.NewArray (_, _) => Predicate.bot_pred
            | Expr.Cast (_, _) => Predicate.bot_pred
            | Expr.InstanceOf (_, _) => Predicate.bot_pred
            | Expr.Val _ => Predicate.bot_pred
            | Expr.BinOp (_, _, _) => Predicate.bot_pred
            | Expr.Var _ => Predicate.bot_pred
            | Expr.LAss (_, _) => Predicate.bot_pred
            | Expr.AAcc (_, _) => Predicate.bot_pred
            | Expr.AAss (_, _, _) => Predicate.bot_pred
            | Expr.ALen _ => Predicate.bot_pred
            | Expr.FAcc (_, _, _) => Predicate.bot_pred
            | Expr.FAss (_, _, _, _) => Predicate.bot_pred
            | Expr.Call (_, _, _) => Predicate.bot_pred
            | Expr.Block (_, _, _, _) => Predicate.bot_pred
            | Expr.Synchronized (_, _, _) => Predicate.bot_pred
            | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
            | Expr.Seq (_, _) => Predicate.bot_pred
            | Expr.Cond (_, _, _) => Predicate.bot_pred
            | Expr.While (_, _) => Predicate.bot_pred
            | Expr.Throw (Expr.New _) => Predicate.bot_pred
            | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Addr a)) =>
              (case c1 of Expr.New _ => Predicate.bot_pred
                | Expr.NewArray (_, _) => Predicate.bot_pred
                | Expr.Cast (_, _) => Predicate.bot_pred
                | Expr.InstanceOf (_, _) => Predicate.bot_pred
                | Expr.Val _ =>
                  Predicate.single
                    ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                      (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
                | Expr.BinOp (_, _, _) => Predicate.bot_pred
                | Expr.Var _ => Predicate.bot_pred
                | Expr.LAss (_, _) => Predicate.bot_pred
                | Expr.AAcc (_, _) => Predicate.bot_pred
                | Expr.AAss (_, _, _) => Predicate.bot_pred
                | Expr.ALen _ => Predicate.bot_pred
                | Expr.FAcc (_, _, _) => Predicate.bot_pred
                | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                | Expr.Call (_, _, _) => Predicate.bot_pred
                | Expr.Block (_, _, _, _) => Predicate.bot_pred
                | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                | Expr.Seq (_, _) => Predicate.bot_pred
                | Expr.Cond (_, _, _) => Predicate.bot_pred
                | Expr.While (_, _) => Predicate.bot_pred
                | Expr.Throw _ => Predicate.bot_pred
                | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
            | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Var _) => Predicate.bot_pred
            | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
            | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
            | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
            | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          (Predicate.sup_pred
            (case c1 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val _ => Predicate.bot_pred
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw (Expr.New _) => Predicate.bot_pred
              | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
              | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
              | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Val (Value.Addr a)) =>
                Predicate.single
                  ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
              | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Var _) => Predicate.bot_pred
              | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
              | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
              | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
            (Predicate.sup_pred
              (case c3 of Expr.New _ => Predicate.bot_pred
                | Expr.NewArray (_, _) => Predicate.bot_pred
                | Expr.Cast (_, _) => Predicate.bot_pred
                | Expr.InstanceOf (_, _) => Predicate.bot_pred
                | Expr.Val v =>
                  (case c2 of Expr.New _ => Predicate.bot_pred
                    | Expr.NewArray (_, _) => Predicate.bot_pred
                    | Expr.Cast (_, _) => Predicate.bot_pred
                    | Expr.InstanceOf (_, _) => Predicate.bot_pred
                    | Expr.Val va =>
                      Predicate.sup_pred
                        (case va of Value.Unit => Predicate.bot_pred
                          | Value.Null => Predicate.bot_pred
                          | Value.Bool _ => Predicate.bot_pred
                          | Value.Intg i =>
                            (case c1 of Expr.New _ => Predicate.bot_pred
                              | Expr.NewArray (_, _) => Predicate.bot_pred
                              | Expr.Cast (_, _) => Predicate.bot_pred
                              | Expr.InstanceOf (_, _) => Predicate.bot_pred
                              | Expr.Val Value.Unit => Predicate.bot_pred
                              | Expr.Val Value.Null => Predicate.bot_pred
                              | Expr.Val (Value.Bool _) => Predicate.bot_pred
                              | Expr.Val (Value.Intg _) => Predicate.bot_pred
                              | Expr.Val (Value.Addr a) =>
                                Predicate.bind
                                  (Predicate.if_pred
                                    (Word.word_sle
                                      (Type_Length.len_bit0
(Type_Length.len_bit0
  (Type_Length.len_bit0
    (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
                                      (Word.zero_word
(Type_Length.len0_bit0
  (Type_Length.len0_bit0
    (Type_Length.len0_bit0
      (Type_Length.len0_bit0 (Type_Length.len0_bit0 Type_Length.len0_num1))))))
                                      i))
                                  (fn () =>
                                    Predicate.bind
                                      (Predicate.if_pred
(IntInf.< (Word.sint
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i, (xr xw a))))
                                      (fn () =>
(case xq xw a of NONE => Predicate.bot_pred
  | SOME Type.Void => Predicate.bot_pred
  | SOME Type.Boolean => Predicate.bot_pred
  | SOME Type.Integer => Predicate.bot_pred | SOME Type.Nt => Predicate.bot_pred
  | SOME (Type.Class _) => Predicate.bot_pred
  | SOME (Type.Array t) =>
    (case Heap.typeof_h A2_ xq xw v of NONE => Predicate.bot_pred
      | SOME u =>
        Predicate.bind (TypeRel.widen_i_i_i xt u t)
          (fn () =>
            Predicate.bind
              (heap_write xw a
                 (Heap.ACell
                   (IntInf.max (0,
                     (Word.sint
                       (Type_Length.len_bit0
                         (Type_Length.len_bit0
                           (Type_Length.len_bit0
                             (Type_Length.len_bit0
                               (Type_Length.len_bit0 Type_Length.len_num1)))))
                       i))))
                v)
              (fn x =>
                Predicate.single
                  (FWState.thread_action_to_thread_action A1_
                     Observable_Events.obs_action_obs_event
                     (FWState.ObsAction
                       (Observable_Events.WriteMem
                         (a, Heap.ACell
                               (IntInf.max (0,
                                 (Word.sint
                                   (Type_Length.len_bit0
                                     (Type_Length.len_bit0
                                       (Type_Length.len_bit0
 (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
                                   i))),
                           v)))
                     (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Val Value.Unit, (x, xx)))))))))
                              | Expr.BinOp (_, _, _) => Predicate.bot_pred
                              | Expr.Var _ => Predicate.bot_pred
                              | Expr.LAss (_, _) => Predicate.bot_pred
                              | Expr.AAcc (_, _) => Predicate.bot_pred
                              | Expr.AAss (_, _, _) => Predicate.bot_pred
                              | Expr.ALen _ => Predicate.bot_pred
                              | Expr.FAcc (_, _, _) => Predicate.bot_pred
                              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                              | Expr.Call (_, _, _) => Predicate.bot_pred
                              | Expr.Block (_, _, _, _) => Predicate.bot_pred
                              | Expr.Synchronized (_, _, _) =>
                                Predicate.bot_pred
                              | Expr.InSynchronized (_, _, _) =>
                                Predicate.bot_pred
                              | Expr.Seq (_, _) => Predicate.bot_pred
                              | Expr.Cond (_, _, _) => Predicate.bot_pred
                              | Expr.While (_, _) => Predicate.bot_pred
                              | Expr.Throw _ => Predicate.bot_pred
                              | Expr.TryCatch (_, _, _, _) =>
                                Predicate.bot_pred)
                          | Value.Addr _ => Predicate.bot_pred)
                        (Predicate.sup_pred
                          (case va of Value.Unit => Predicate.bot_pred
                            | Value.Null => Predicate.bot_pred
                            | Value.Bool _ => Predicate.bot_pred
                            | Value.Intg i =>
                              (case c1 of Expr.New _ => Predicate.bot_pred
                                | Expr.NewArray (_, _) => Predicate.bot_pred
                                | Expr.Cast (_, _) => Predicate.bot_pred
                                | Expr.InstanceOf (_, _) => Predicate.bot_pred
                                | Expr.Val Value.Unit => Predicate.bot_pred
                                | Expr.Val Value.Null => Predicate.bot_pred
                                | Expr.Val (Value.Bool _) => Predicate.bot_pred
                                | Expr.Val (Value.Intg _) => Predicate.bot_pred
                                | Expr.Val (Value.Addr a) =>
                                  Predicate.bind
                                    (Predicate.if_pred
                                      (Word.word_sle
(Type_Length.len_bit0
  (Type_Length.len_bit0
    (Type_Length.len_bit0
      (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
(Word.zero_word
  (Type_Length.len0_bit0
    (Type_Length.len0_bit0
      (Type_Length.len0_bit0
        (Type_Length.len0_bit0
          (Type_Length.len0_bit0 Type_Length.len0_num1))))))
i))
                                    (fn () =>
                                      Predicate.bind
(Predicate.if_pred
  (IntInf.< (Word.sint
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0
                       (Type_Length.len_bit0 Type_Length.len_num1)))))
               i, (xr (State.hp (xw, xx)) a))))
(fn () =>
  (case xq (State.hp (xw, xx)) a of NONE => Predicate.bot_pred
    | SOME Type.Void => Predicate.bot_pred
    | SOME Type.Boolean => Predicate.bot_pred
    | SOME Type.Integer => Predicate.bot_pred
    | SOME Type.Nt => Predicate.bot_pred
    | SOME (Type.Class _) => Predicate.bot_pred
    | SOME (Type.Array t) =>
      (case Heap.typeof_h A2_ xq (State.hp (xw, xx)) v
        of NONE => Predicate.bot_pred
        | SOME u =>
          Predicate.bind (Predicate.not_pred (TypeRel.widen_i_i_i xt u t))
            (fn () =>
              Predicate.single
                ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                  (Expr.Throw
                     (Expr.Val
                       (Value.Addr
                         (StartConfig.addr_of_sys_xcpt A2_ xn xo
                           "java/lang/ArrayStoreException"))),
                    (xw, xx))))))))
                                | Expr.BinOp (_, _, _) => Predicate.bot_pred
                                | Expr.Var _ => Predicate.bot_pred
                                | Expr.LAss (_, _) => Predicate.bot_pred
                                | Expr.AAcc (_, _) => Predicate.bot_pred
                                | Expr.AAss (_, _, _) => Predicate.bot_pred
                                | Expr.ALen _ => Predicate.bot_pred
                                | Expr.FAcc (_, _, _) => Predicate.bot_pred
                                | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                                | Expr.Call (_, _, _) => Predicate.bot_pred
                                | Expr.Block (_, _, _, _) => Predicate.bot_pred
                                | Expr.Synchronized (_, _, _) =>
                                  Predicate.bot_pred
                                | Expr.InSynchronized (_, _, _) =>
                                  Predicate.bot_pred
                                | Expr.Seq (_, _) => Predicate.bot_pred
                                | Expr.Cond (_, _, _) => Predicate.bot_pred
                                | Expr.While (_, _) => Predicate.bot_pred
                                | Expr.Throw _ => Predicate.bot_pred
                                | Expr.TryCatch (_, _, _, _) =>
                                  Predicate.bot_pred)
                            | Value.Addr _ => Predicate.bot_pred)
                          (Predicate.sup_pred
                            (case va of Value.Unit => Predicate.bot_pred
                              | Value.Null => Predicate.bot_pred
                              | Value.Bool _ => Predicate.bot_pred
                              | Value.Intg i =>
                                (case c1 of Expr.New _ => Predicate.bot_pred
                                  | Expr.NewArray (_, _) => Predicate.bot_pred
                                  | Expr.Cast (_, _) => Predicate.bot_pred
                                  | Expr.InstanceOf (_, _) => Predicate.bot_pred
                                  | Expr.Val Value.Unit => Predicate.bot_pred
                                  | Expr.Val Value.Null => Predicate.bot_pred
                                  | Expr.Val (Value.Bool _) =>
                                    Predicate.bot_pred
                                  | Expr.Val (Value.Intg _) =>
                                    Predicate.bot_pred
                                  | Expr.Val (Value.Addr a) =>
                                    Predicate.bind
                                      (Predicate.if_pred
(Word.word_sless
   (Type_Length.len_bit0
     (Type_Length.len_bit0
       (Type_Length.len_bit0
         (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
   i (Word.zero_word
       (Type_Length.len0_bit0
         (Type_Length.len0_bit0
           (Type_Length.len0_bit0
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0 Type_Length.len0_num1)))))) orelse
  IntInf.<= ((xr (State.hp (xw, xx))
               a), Word.sint
                     (Type_Length.len_bit0
                       (Type_Length.len_bit0
                         (Type_Length.len_bit0
                           (Type_Length.len_bit0
                             (Type_Length.len_bit0 Type_Length.len_num1)))))
                     i)))
                                      (fn () =>
(case xq (State.hp (xw, xx)) a of NONE => Predicate.bot_pred
  | SOME Type.Void => Predicate.bot_pred
  | SOME Type.Boolean => Predicate.bot_pred
  | SOME Type.Integer => Predicate.bot_pred | SOME Type.Nt => Predicate.bot_pred
  | SOME (Type.Class _) => Predicate.bot_pred
  | SOME (Type.Array _) =>
    Predicate.single
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        (Expr.Throw
           (Expr.Val
             (Value.Addr
               (StartConfig.addr_of_sys_xcpt A2_ xn xo
                 "java/lang/ArrayIndexOutOfBoundsException"))),
          (xw, xx)))))
                                  | Expr.BinOp (_, _, _) => Predicate.bot_pred
                                  | Expr.Var _ => Predicate.bot_pred
                                  | Expr.LAss (_, _) => Predicate.bot_pred
                                  | Expr.AAcc (_, _) => Predicate.bot_pred
                                  | Expr.AAss (_, _, _) => Predicate.bot_pred
                                  | Expr.ALen _ => Predicate.bot_pred
                                  | Expr.FAcc (_, _, _) => Predicate.bot_pred
                                  | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                                  | Expr.Call (_, _, _) => Predicate.bot_pred
                                  | Expr.Block (_, _, _, _) =>
                                    Predicate.bot_pred
                                  | Expr.Synchronized (_, _, _) =>
                                    Predicate.bot_pred
                                  | Expr.InSynchronized (_, _, _) =>
                                    Predicate.bot_pred
                                  | Expr.Seq (_, _) => Predicate.bot_pred
                                  | Expr.Cond (_, _, _) => Predicate.bot_pred
                                  | Expr.While (_, _) => Predicate.bot_pred
                                  | Expr.Throw _ => Predicate.bot_pred
                                  | Expr.TryCatch (_, _, _, _) =>
                                    Predicate.bot_pred)
                              | Value.Addr _ => Predicate.bot_pred)
                            (case c1 of Expr.New _ => Predicate.bot_pred
                              | Expr.NewArray (_, _) => Predicate.bot_pred
                              | Expr.Cast (_, _) => Predicate.bot_pred
                              | Expr.InstanceOf (_, _) => Predicate.bot_pred
                              | Expr.Val Value.Unit => Predicate.bot_pred
                              | Expr.Val Value.Null =>
                                Predicate.single
                                  ((FinFun.Finfun_const [],
                                     ([], ([], ([], ([], []))))),
                                    (Expr.Throw
                                       (Expr.Val
 (Value.Addr
   (StartConfig.addr_of_sys_xcpt A2_ xn xo "java/lang/NullPointerException"))),
                                      (xw, xx)))
                              | Expr.Val (Value.Bool _) => Predicate.bot_pred
                              | Expr.Val (Value.Intg _) => Predicate.bot_pred
                              | Expr.Val (Value.Addr _) => Predicate.bot_pred
                              | Expr.BinOp (_, _, _) => Predicate.bot_pred
                              | Expr.Var _ => Predicate.bot_pred
                              | Expr.LAss (_, _) => Predicate.bot_pred
                              | Expr.AAcc (_, _) => Predicate.bot_pred
                              | Expr.AAss (_, _, _) => Predicate.bot_pred
                              | Expr.ALen _ => Predicate.bot_pred
                              | Expr.FAcc (_, _, _) => Predicate.bot_pred
                              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                              | Expr.Call (_, _, _) => Predicate.bot_pred
                              | Expr.Block (_, _, _, _) => Predicate.bot_pred
                              | Expr.Synchronized (_, _, _) =>
                                Predicate.bot_pred
                              | Expr.InSynchronized (_, _, _) =>
                                Predicate.bot_pred
                              | Expr.Seq (_, _) => Predicate.bot_pred
                              | Expr.Cond (_, _, _) => Predicate.bot_pred
                              | Expr.While (_, _) => Predicate.bot_pred
                              | Expr.Throw _ => Predicate.bot_pred
                              | Expr.TryCatch (_, _, _, _) =>
                                Predicate.bot_pred)))
                    | Expr.BinOp (_, _, _) => Predicate.bot_pred
                    | Expr.Var _ => Predicate.bot_pred
                    | Expr.LAss (_, _) => Predicate.bot_pred
                    | Expr.AAcc (_, _) => Predicate.bot_pred
                    | Expr.AAss (_, _, _) => Predicate.bot_pred
                    | Expr.ALen _ => Predicate.bot_pred
                    | Expr.FAcc (_, _, _) => Predicate.bot_pred
                    | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                    | Expr.Call (_, _, _) => Predicate.bot_pred
                    | Expr.Block (_, _, _, _) => Predicate.bot_pred
                    | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                    | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                    | Expr.Seq (_, _) => Predicate.bot_pred
                    | Expr.Cond (_, _, _) => Predicate.bot_pred
                    | Expr.While (_, _) => Predicate.bot_pred
                    | Expr.Throw _ => Predicate.bot_pred
                    | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
                | Expr.BinOp (_, _, _) => Predicate.bot_pred
                | Expr.Var _ => Predicate.bot_pred
                | Expr.LAss (_, _) => Predicate.bot_pred
                | Expr.AAcc (_, _) => Predicate.bot_pred
                | Expr.AAss (_, _, _) => Predicate.bot_pred
                | Expr.ALen _ => Predicate.bot_pred
                | Expr.FAcc (_, _, _) => Predicate.bot_pred
                | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                | Expr.Call (_, _, _) => Predicate.bot_pred
                | Expr.Block (_, _, _, _) => Predicate.bot_pred
                | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                | Expr.Seq (_, _) => Predicate.bot_pred
                | Expr.Cond (_, _, _) => Predicate.bot_pred
                | Expr.While (_, _) => Predicate.bot_pred
                | Expr.Throw _ => Predicate.bot_pred
                | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              (Predicate.sup_pred
                (case c2 of Expr.New _ => Predicate.bot_pred
                  | Expr.NewArray (_, _) => Predicate.bot_pred
                  | Expr.Cast (_, _) => Predicate.bot_pred
                  | Expr.InstanceOf (_, _) => Predicate.bot_pred
                  | Expr.Val ib =>
                    (case c1 of Expr.New _ => Predicate.bot_pred
                      | Expr.NewArray (_, _) => Predicate.bot_pred
                      | Expr.Cast (_, _) => Predicate.bot_pred
                      | Expr.InstanceOf (_, _) => Predicate.bot_pred
                      | Expr.Val ab =>
                        Predicate.bind
                          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                            (A1_, A2_) xl xm xn xo xp xq xr heap_read heap_write
                            xs xt xu c3 (xw, xx))
                          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf)))
                             =>
                            Predicate.single
                              ((xa, (xb, (xc, (xd, (xe, ye))))),
                                (Expr.AAss (Expr.Val ab, Expr.Val ib, e),
                                  (xf, yf))))
                      | Expr.BinOp (_, _, _) => Predicate.bot_pred
                      | Expr.Var _ => Predicate.bot_pred
                      | Expr.LAss (_, _) => Predicate.bot_pred
                      | Expr.AAcc (_, _) => Predicate.bot_pred
                      | Expr.AAss (_, _, _) => Predicate.bot_pred
                      | Expr.ALen _ => Predicate.bot_pred
                      | Expr.FAcc (_, _, _) => Predicate.bot_pred
                      | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                      | Expr.Call (_, _, _) => Predicate.bot_pred
                      | Expr.Block (_, _, _, _) => Predicate.bot_pred
                      | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                      | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                      | Expr.Seq (_, _) => Predicate.bot_pred
                      | Expr.Cond (_, _, _) => Predicate.bot_pred
                      | Expr.While (_, _) => Predicate.bot_pred
                      | Expr.Throw _ => Predicate.bot_pred
                      | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
                  | Expr.BinOp (_, _, _) => Predicate.bot_pred
                  | Expr.Var _ => Predicate.bot_pred
                  | Expr.LAss (_, _) => Predicate.bot_pred
                  | Expr.AAcc (_, _) => Predicate.bot_pred
                  | Expr.AAss (_, _, _) => Predicate.bot_pred
                  | Expr.ALen _ => Predicate.bot_pred
                  | Expr.FAcc (_, _, _) => Predicate.bot_pred
                  | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                  | Expr.Call (_, _, _) => Predicate.bot_pred
                  | Expr.Block (_, _, _, _) => Predicate.bot_pred
                  | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                  | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                  | Expr.Seq (_, _) => Predicate.bot_pred
                  | Expr.Cond (_, _, _) => Predicate.bot_pred
                  | Expr.While (_, _) => Predicate.bot_pred
                  | Expr.Throw _ => Predicate.bot_pred
                  | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
                (Predicate.sup_pred
                  (case c1 of Expr.New _ => Predicate.bot_pred
                    | Expr.NewArray (_, _) => Predicate.bot_pred
                    | Expr.Cast (_, _) => Predicate.bot_pred
                    | Expr.InstanceOf (_, _) => Predicate.bot_pred
                    | Expr.Val ab =>
                      Predicate.bind
                        (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                          (A1_, A2_) xl xm xn xo xp xq xr heap_read heap_write
                          xs xt xu c2 (xw, xx))
                        (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (i, (xf, yf))) =>
                          Predicate.single
                            ((xa, (xb, (xc, (xd, (xe, ye))))),
                              (Expr.AAss (Expr.Val ab, i, c3), (xf, yf))))
                    | Expr.BinOp (_, _, _) => Predicate.bot_pred
                    | Expr.Var _ => Predicate.bot_pred
                    | Expr.LAss (_, _) => Predicate.bot_pred
                    | Expr.AAcc (_, _) => Predicate.bot_pred
                    | Expr.AAss (_, _, _) => Predicate.bot_pred
                    | Expr.ALen _ => Predicate.bot_pred
                    | Expr.FAcc (_, _, _) => Predicate.bot_pred
                    | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                    | Expr.Call (_, _, _) => Predicate.bot_pred
                    | Expr.Block (_, _, _, _) => Predicate.bot_pred
                    | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                    | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                    | Expr.Seq (_, _) => Predicate.bot_pred
                    | Expr.Cond (_, _, _) => Predicate.bot_pred
                    | Expr.While (_, _) => Predicate.bot_pred
                    | Expr.Throw _ => Predicate.bot_pred
                    | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
                  (Predicate.bind
                    (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                      (A1_, A2_) xl xm xn xo xp xq xr heap_read heap_write xs xt
                      xu c1 (xw, xx))
                    (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (a, (xf, yf))) =>
                      Predicate.single
                        ((xa, (xb, (xc, (xd, (xe, ye))))),
                          (Expr.AAss (a, c2, c3), (xf, yf))))))))))
    | Expr.ALen c1 =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val Value.Unit => Predicate.bot_pred
          | Expr.Val Value.Null =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw
                   (Expr.Val
                     (Value.Addr
                       (StartConfig.addr_of_sys_xcpt A2_ xn xo
                         "java/lang/NullPointerException"))),
                  (xw, xx)))
          | Expr.Val (Value.Bool _) => Predicate.bot_pred
          | Expr.Val (Value.Intg _) => Predicate.bot_pred
          | Expr.Val (Value.Addr a) =>
            (case xq xw a of NONE => Predicate.bot_pred
              | SOME Type.Void => Predicate.bot_pred
              | SOME Type.Boolean => Predicate.bot_pred
              | SOME Type.Integer => Predicate.bot_pred
              | SOME Type.Nt => Predicate.bot_pred
              | SOME (Type.Class _) => Predicate.bot_pred
              | SOME (Type.Array _) =>
                Predicate.single
                  ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Val (Value.Intg (Word.Word_of_int (xr xw a))),
                      (xw, xx))))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu c1 (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (a, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))), (Expr.ALen a, (xf, yf)))))
    | Expr.FAcc (c1, c2, c3) =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val Value.Unit => Predicate.bot_pred
          | Expr.Val Value.Null =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw
                   (Expr.Val
                     (Value.Addr
                       (StartConfig.addr_of_sys_xcpt A2_ xn xo
                         "java/lang/NullPointerException"))),
                  (xw, xx)))
          | Expr.Val (Value.Bool _) => Predicate.bot_pred
          | Expr.Val (Value.Intg _) => Predicate.bot_pred
          | Expr.Val (Value.Addr a) =>
            Predicate.bind (heap_read xw a (Heap.CField (c3, c2)))
              (fn x =>
                Predicate.single
                  (FWState.thread_action_to_thread_action A1_
                     Observable_Events.obs_action_obs_event
                     (FWState.ObsAction
                       (Observable_Events.ReadMem (a, Heap.CField (c3, c2), x)))
                     (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Val x, (xw, xx))))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu c1 (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.FAcc (e, c2, c3), (xf, yf)))))
    | Expr.FAss (c1, c2, c3, c4) =>
      Predicate.sup_pred
        (case c4 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val _ => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            (case c1 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val _ =>
                Predicate.single
                  ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.sup_pred
          (case c1 of Expr.New _ => Predicate.bot_pred
            | Expr.NewArray (_, _) => Predicate.bot_pred
            | Expr.Cast (_, _) => Predicate.bot_pred
            | Expr.InstanceOf (_, _) => Predicate.bot_pred
            | Expr.Val _ => Predicate.bot_pred
            | Expr.BinOp (_, _, _) => Predicate.bot_pred
            | Expr.Var _ => Predicate.bot_pred
            | Expr.LAss (_, _) => Predicate.bot_pred
            | Expr.AAcc (_, _) => Predicate.bot_pred
            | Expr.AAss (_, _, _) => Predicate.bot_pred
            | Expr.ALen _ => Predicate.bot_pred
            | Expr.FAcc (_, _, _) => Predicate.bot_pred
            | Expr.FAss (_, _, _, _) => Predicate.bot_pred
            | Expr.Call (_, _, _) => Predicate.bot_pred
            | Expr.Block (_, _, _, _) => Predicate.bot_pred
            | Expr.Synchronized (_, _, _) => Predicate.bot_pred
            | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
            | Expr.Seq (_, _) => Predicate.bot_pred
            | Expr.Cond (_, _, _) => Predicate.bot_pred
            | Expr.While (_, _) => Predicate.bot_pred
            | Expr.Throw (Expr.New _) => Predicate.bot_pred
            | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Addr a)) =>
              Predicate.single
                ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                  (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
            | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Var _) => Predicate.bot_pred
            | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
            | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
            | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
            | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          (Predicate.sup_pred
            (case c4 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val v =>
                (case c1 of Expr.New _ => Predicate.bot_pred
                  | Expr.NewArray (_, _) => Predicate.bot_pred
                  | Expr.Cast (_, _) => Predicate.bot_pred
                  | Expr.InstanceOf (_, _) => Predicate.bot_pred
                  | Expr.Val Value.Unit => Predicate.bot_pred
                  | Expr.Val Value.Null =>
                    Predicate.single
                      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                        (Expr.Throw
                           (Expr.Val
                             (Value.Addr
                               (StartConfig.addr_of_sys_xcpt A2_ xn xo
                                 "java/lang/NullPointerException"))),
                          (xw, xx)))
                  | Expr.Val (Value.Bool _) => Predicate.bot_pred
                  | Expr.Val (Value.Intg _) => Predicate.bot_pred
                  | Expr.Val (Value.Addr a) =>
                    Predicate.bind (heap_write xw a (Heap.CField (c3, c2)) v)
                      (fn x =>
                        Predicate.single
                          (FWState.thread_action_to_thread_action A1_
                             Observable_Events.obs_action_obs_event
                             (FWState.ObsAction
                               (Observable_Events.WriteMem
                                 (a, Heap.CField (c3, c2), v)))
                             (FinFun.Finfun_const [],
                               ([], ([], ([], ([], []))))),
                            (Expr.Val Value.Unit, (x, xx))))
                  | Expr.BinOp (_, _, _) => Predicate.bot_pred
                  | Expr.Var _ => Predicate.bot_pred
                  | Expr.LAss (_, _) => Predicate.bot_pred
                  | Expr.AAcc (_, _) => Predicate.bot_pred
                  | Expr.AAss (_, _, _) => Predicate.bot_pred
                  | Expr.ALen _ => Predicate.bot_pred
                  | Expr.FAcc (_, _, _) => Predicate.bot_pred
                  | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                  | Expr.Call (_, _, _) => Predicate.bot_pred
                  | Expr.Block (_, _, _, _) => Predicate.bot_pred
                  | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                  | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                  | Expr.Seq (_, _) => Predicate.bot_pred
                  | Expr.Cond (_, _, _) => Predicate.bot_pred
                  | Expr.While (_, _) => Predicate.bot_pred
                  | Expr.Throw _ => Predicate.bot_pred
                  | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
            (Predicate.sup_pred
              (case c1 of Expr.New _ => Predicate.bot_pred
                | Expr.NewArray (_, _) => Predicate.bot_pred
                | Expr.Cast (_, _) => Predicate.bot_pred
                | Expr.InstanceOf (_, _) => Predicate.bot_pred
                | Expr.Val v =>
                  Predicate.bind
                    (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                      (A1_, A2_) xl xm xn xo xp xq xr heap_read heap_write xs xt
                      xu c4 (xw, xx))
                    (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
                      Predicate.single
                        ((xa, (xb, (xc, (xd, (xe, ye))))),
                          (Expr.FAss (Expr.Val v, c2, c3, e), (xf, yf))))
                | Expr.BinOp (_, _, _) => Predicate.bot_pred
                | Expr.Var _ => Predicate.bot_pred
                | Expr.LAss (_, _) => Predicate.bot_pred
                | Expr.AAcc (_, _) => Predicate.bot_pred
                | Expr.AAss (_, _, _) => Predicate.bot_pred
                | Expr.ALen _ => Predicate.bot_pred
                | Expr.FAcc (_, _, _) => Predicate.bot_pred
                | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                | Expr.Call (_, _, _) => Predicate.bot_pred
                | Expr.Block (_, _, _, _) => Predicate.bot_pred
                | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                | Expr.Seq (_, _) => Predicate.bot_pred
                | Expr.Cond (_, _, _) => Predicate.bot_pred
                | Expr.While (_, _) => Predicate.bot_pred
                | Expr.Throw _ => Predicate.bot_pred
                | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              (Predicate.bind
                (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                  (A1_, A2_) xl xm xn xo xp xq xr heap_read heap_write xs xt xu
                  c1 (xw, xx))
                (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
                  Predicate.single
                    ((xa, (xb, (xc, (xd, (xe, ye))))),
                      (Expr.FAss (e, c2, c3, c4), (xf, yf))))))))
    | Expr.Call (c1, c2, c3) =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            Predicate.sup_pred
              (Predicate.bind (Expr.is_Throws_i c3)
                (fn () =>
                  Predicate.single
                    ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                      (List.hd (List.dropWhile Expr.is_val c3), (xw, xx)))))
              (Predicate.sup_pred
                (case v of Value.Unit => Predicate.bot_pred
                  | Value.Null =>
                    Predicate.bind (Predicate.if_pred (Expr.is_vals c3))
                      (fn () =>
                        Predicate.single
                          ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                            (Expr.Throw
                               (Expr.Val
                                 (Value.Addr
                                   (StartConfig.addr_of_sys_xcpt A2_ xn xo
                                     "java/lang/NullPointerException"))),
                              (xw, xx))))
                  | Value.Bool _ => Predicate.bot_pred
                  | Value.Intg _ => Predicate.bot_pred
                  | Value.Addr _ => Predicate.bot_pred)
                (Predicate.sup_pred
                  (case v of Value.Unit => Predicate.bot_pred
                    | Value.Null => Predicate.bot_pred
                    | Value.Bool _ => Predicate.bot_pred
                    | Value.Intg _ => Predicate.bot_pred
                    | Value.Addr a =>
                      Predicate.bind (Predicate.if_pred (Expr.is_vals c3))
                        (fn () =>
                          (case xq (State.hp (xw, xx)) a
                            of NONE => Predicate.bot_pred
                            | SOME u =>
                              Predicate.bind (Type.is_class_type_of_i_o u)
                                (fn x =>
                                  Predicate.bind
                                    (ExternalCall.red_external_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_i_o_o_o
                                      (A1_, A2_) xl xm xn xo xp xq xr heap_read
                                      heap_write xt xu (State.hp (xw, xx)) a c2
                                      (List.map Expr.the_Val c3))
                                    (fn ((xfa, (xg, (xh, (xi, (xj, yj))))),
  (va, h))
                                       =>
                                      Predicate.bind
(TypeRel.method_i_i_i_o_o_o_o xt x c2)
(fn b =>
  (case b
    of (_, (_, (NONE, _))) =>
      Predicate.single
        (xs (xfa, (xg, (xh, (xi, (xj, yj))))),
          (Expr.extRet2J (Expr.Call (Expr.Val (Value.Addr a), c2, c3)) va,
            (h, State.lcl (xw, xx))))
    | (_, (_, (SOME _, _))) => Predicate.bot_pred)))))))
                  (Predicate.sup_pred
                    (case v of Value.Unit => Predicate.bot_pred
                      | Value.Null => Predicate.bot_pred
                      | Value.Bool _ => Predicate.bot_pred
                      | Value.Intg _ => Predicate.bot_pred
                      | Value.Addr a =>
                        Predicate.bind (Predicate.if_pred (Expr.is_vals c3))
                          (fn () =>
                            (case xq (State.hp (xw, xx)) a
                              of NONE => Predicate.bot_pred
                              | SOME u =>
                                Predicate.bind (Type.is_class_type_of_i_o u)
                                  (fn x =>
                                    Predicate.bind
                                      (TypeRel.method_i_i_i_o_o_o_o xt x c2)
                                      (fn b =>
(case b of (ts, (_, (NONE, d))) => Predicate.bot_pred
  | (ts, (_, (SOME (pns, body), d))) =>
    Predicate.bind
      (TypeRel.eq_i_i Arith.equal_nat (List.size_list c3) (List.size_list pns))
      (fn () =>
        Predicate.bind
          (TypeRel.eq_i_i Arith.equal_nat (List.size_list ts)
            (List.size_list pns))
          (fn () =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.blocks ("this" :: pns) (Type.Class d :: ts)
                   (Value.Addr a :: List.map Expr.the_Val c3) body,
                  (xw, xx)))))))))))
                    (Predicate.bind
                      (reds_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                        (A1_, A2_) xl xm xn xo xp xq xr heap_read heap_write xs
                        xt xu c3 (xw, xx))
                      (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (es, (xf, yf))) =>
                        Predicate.single
                          ((xa, (xb, (xc, (xd, (xe, ye))))),
                            (Expr.Call (Expr.Val v, c2, es), (xf, yf))))))))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu c1 (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.Call (e, c2, c3), (xf, yf)))))
    | Expr.Block (c1, c2, c3, c4) =>
      Predicate.sup_pred
        (case c4 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Val v, (xw, xx)))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu c4
            (xw, Fun.fun_upd Stringa.equal_literal xx c1 c3))
          (fn ((xk, (xaa, (xba, (xca, (xda, yd))))), (e, (h, l))) =>
            Predicate.single
              ((xk, (xaa, (xba, (xca, (xda, yd))))),
                (Expr.Block (c1, c2, l c1, e),
                  (h, Fun.fun_upd Stringa.equal_literal l c1 (xx c1))))))
    | Expr.Synchronized (c1, c2, c3) =>
      Predicate.sup_pred
        (case c2 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val Value.Unit => Predicate.bot_pred
          | Expr.Val Value.Null =>
            let
              val () = c1;
            in
              Predicate.single
                ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                  (Expr.Throw
                     (Expr.Val
                       (Value.Addr
                         (StartConfig.addr_of_sys_xcpt A2_ xn xo
                           "java/lang/NullPointerException"))),
                    (xw, xx)))
            end
          | Expr.Val (Value.Bool _) => Predicate.bot_pred
          | Expr.Val (Value.Intg _) => Predicate.bot_pred
          | Expr.Val (Value.Addr a) =>
            let
              val () = c1;
            in
              Predicate.single
                (FWState.thread_action_to_thread_action A1_
                   Observable_Events.obs_action_obs_event
                   (FWState.ObsAction (Observable_Events.SyncLock a))
                   (FWState.thread_action_to_thread_action A1_
                     Observable_Events.obs_action_obs_event
                     (FWState.LockAction (FWState.Lock, a))
                     (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))),
                  (Expr.InSynchronized ((), a, c3), (xw, xx)))
            end
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            let
              val () = c1;
            in
              Predicate.single
                ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                  (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
            end
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        let
          val () = c1;
        in
          Predicate.bind
            (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
              xm xn xo xp xq xr heap_read heap_write xs xt xu c2 (xw, xx))
            (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (oa, (xf, yf))) =>
              Predicate.single
                ((xa, (xb, (xc, (xd, (xe, ye))))),
                  (Expr.Synchronized ((), oa, c3), (xf, yf))))
        end
    | Expr.InSynchronized (c1, c2, c3) =>
      Predicate.sup_pred
        (case c3 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            let
              val () = c1;
            in
              Predicate.single
                (FWState.thread_action_to_thread_action A1_
                   Observable_Events.obs_action_obs_event
                   (FWState.ObsAction (Observable_Events.SyncUnlock c2))
                   (FWState.thread_action_to_thread_action A1_
                     Observable_Events.obs_action_obs_event
                     (FWState.LockAction (FWState.Unlock, c2))
                     (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))),
                  (Expr.Val v, (xw, xx)))
            end
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr ad)) =>
            let
              val () = c1;
            in
              Predicate.single
                (FWState.thread_action_to_thread_action A1_
                   Observable_Events.obs_action_obs_event
                   (FWState.ObsAction (Observable_Events.SyncUnlock c2))
                   (FWState.thread_action_to_thread_action A1_
                     Observable_Events.obs_action_obs_event
                     (FWState.LockAction (FWState.Unlock, c2))
                     (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))),
                  (Expr.Throw (Expr.Val (Value.Addr ad)), (xw, xx)))
            end
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        let
          val () = c1;
        in
          Predicate.bind
            (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
              xm xn xo xp xq xr heap_read heap_write xs xt xu c3 (xw, xx))
            (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
              Predicate.single
                ((xa, (xb, (xc, (xd, (xe, ye))))),
                  (Expr.InSynchronized ((), c2, e), (xf, yf))))
        end
    | Expr.Seq (c1, c2) =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val _ =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (c2, (xw, xx)))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu c1 (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))), (Expr.Seq (e, c2), (xf, yf)))))
    | Expr.Cond (c1, c2, c3) =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val Value.Unit => Predicate.bot_pred
          | Expr.Val Value.Null => Predicate.bot_pred
          | Expr.Val (Value.Bool true) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (c2, (xw, xx)))
          | Expr.Val (Value.Bool false) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (c3, (xw, xx)))
          | Expr.Val (Value.Intg _) => Predicate.bot_pred
          | Expr.Val (Value.Addr _) => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu c1 (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (b, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.Cond (b, c2, c3), (xf, yf)))))
    | Expr.While (c1, c2) =>
      Predicate.single
        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
          (Expr.Cond
             (c1, Expr.Seq (c2, Expr.While (c1, c2)), Expr.Val Value.Unit),
            (xw, xx)))
    | Expr.Throw c1 =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val Value.Unit => Predicate.bot_pred
          | Expr.Val Value.Null =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw
                   (Expr.Val
                     (Value.Addr
                       (StartConfig.addr_of_sys_xcpt A2_ xn xo
                         "java/lang/NullPointerException"))),
                  (xw, xx)))
          | Expr.Val (Value.Bool _) => Predicate.bot_pred
          | Expr.Val (Value.Intg _) => Predicate.bot_pred
          | Expr.Val (Value.Addr _) => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu c1 (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))), (Expr.Throw e, (xf, yf)))))
    | Expr.TryCatch (c1, c2, c3, c4) =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Val v, (xw, xx)))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.sup_pred
              (case xq (State.hp (xw, xx)) a of NONE => Predicate.bot_pred
                | SOME Type.Void => Predicate.bot_pred
                | SOME Type.Boolean => Predicate.bot_pred
                | SOME Type.Integer => Predicate.bot_pred
                | SOME Type.Nt => Predicate.bot_pred
                | SOME (Type.Class d) =>
                  Predicate.bind
                    (Predicate.not_pred
                      (Transitive_Closure_Table.rtranclp_FioB_i_i
                        Stringa.equal_literal (TypeRel.subcls1_i_i_o xt) d c2))
                    (fn () =>
                      Predicate.single
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (Expr.Throw (Expr.Val (Value.Addr a)), (xw, xx))))
                | SOME (Type.Array _) => Predicate.bot_pred)
              (case xq (State.hp (xw, xx)) a of NONE => Predicate.bot_pred
                | SOME Type.Void => Predicate.bot_pred
                | SOME Type.Boolean => Predicate.bot_pred
                | SOME Type.Integer => Predicate.bot_pred
                | SOME Type.Nt => Predicate.bot_pred
                | SOME (Type.Class d) =>
                  Predicate.bind
                    (Transitive_Closure_Table.rtranclp_FioB_i_i
                      Stringa.equal_literal (TypeRel.subcls1_i_i_o xt) d c2)
                    (fn () =>
                      Predicate.single
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (Expr.Block
                             (c3, Type.Class c2, SOME (Value.Addr a), c4),
                            (xw, xx))))
                | SOME (Type.Array _) => Predicate.bot_pred)
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu c1 (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.TryCatch (e, c2, c3, c4), (xf, yf))))))
and reds_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl xm xn
  xo xp xq xr heap_read heap_write xs xt xu xv (xw, xx) =
  Predicate.sup_pred
    (case xv of [] => Predicate.bot_pred
      | e :: es =>
        Predicate.bind
          (red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu e (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (ea, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))), (ea :: es, (xf, yf)))))
    (case xv of [] => Predicate.bot_pred
      | Expr.New _ :: es => Predicate.bot_pred
      | Expr.NewArray (_, _) :: es => Predicate.bot_pred
      | Expr.Cast (_, _) :: es => Predicate.bot_pred
      | Expr.InstanceOf (_, _) :: es => Predicate.bot_pred
      | Expr.Val v :: es =>
        Predicate.bind
          (reds_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) xl
            xm xn xo xp xq xr heap_read heap_write xs xt xu es (xw, xx))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (esa, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))), (Expr.Val v :: esa, (xf, yf))))
      | Expr.BinOp (_, _, _) :: es => Predicate.bot_pred
      | Expr.Var _ :: es => Predicate.bot_pred
      | Expr.LAss (_, _) :: es => Predicate.bot_pred
      | Expr.AAcc (_, _) :: es => Predicate.bot_pred
      | Expr.AAss (_, _, _) :: es => Predicate.bot_pred
      | Expr.ALen _ :: es => Predicate.bot_pred
      | Expr.FAcc (_, _, _) :: es => Predicate.bot_pred
      | Expr.FAss (_, _, _, _) :: es => Predicate.bot_pred
      | Expr.Call (_, _, _) :: es => Predicate.bot_pred
      | Expr.Block (_, _, _, _) :: es => Predicate.bot_pred
      | Expr.Synchronized (_, _, _) :: es => Predicate.bot_pred
      | Expr.InSynchronized (_, _, _) :: es => Predicate.bot_pred
      | Expr.Seq (_, _) :: es => Predicate.bot_pred
      | Expr.Cond (_, _, _) :: es => Predicate.bot_pred
      | Expr.While (_, _) :: es => Predicate.bot_pred
      | Expr.Throw _ :: es => Predicate.bot_pred
      | Expr.TryCatch (_, _, _, _) :: es => Predicate.bot_pred);

end; (*struct SmallStep*)

structure FWLockingThread : sig
  val upd_threadR :
    'a HOL.equal ->
      IntInf.int ->
        ('a * IntInf.int) option -> 'a -> FWState.lock_action -> IntInf.int
  val upd_threadRs :
    'a HOL.equal ->
      IntInf.int ->
        ('a * IntInf.int) option -> 'a -> FWState.lock_action list -> IntInf.int
  val redT_updLns :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * IntInf.int) option)) FinFun.finfun ->
        'b -> ('a, IntInf.int) FinFun.finfun ->
                ('a, (FWState.lock_action list)) FinFun.finfun ->
                  ('a, IntInf.int) FinFun.finfun
end = struct

fun upd_threadR A_ n l t FWState.ReleaseAcquire =
  IntInf.+ (n, FWLock.has_locks A_ l t)
  | upd_threadR A_ n l t FWState.Lock = n
  | upd_threadR A_ n l t FWState.Unlock = n
  | upd_threadR A_ n l t FWState.UnlockFail = n;

fun upd_threadRs A_ n l t [] = n
  | upd_threadRs A_ n l t (la :: las) =
    upd_threadRs A_ (upd_threadR A_ n l t la) (FWLock.upd_lock A_ l t la) t las;

fun redT_updLns A_ B_ ls t ln las =
  FinFun.finfun_comp (fn (l, (n, a)) => upd_threadRs B_ n l t a)
    (FinFun.finfun_Diag A_
      (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
      (Product_Type.equal_prod Arith.equal_nat
        (List.equal_list FWState.equal_lock_action))
      ls (FinFun.finfun_Diag A_ Arith.equal_nat
           (List.equal_list FWState.equal_lock_action) ln las));

end; (*struct FWLockingThread*)

structure Scheduler : sig
  val redT_updT :
    ('a -> 'b * ('c, IntInf.int) FinFun.finfun -> 'd -> 'd) ->
      'd -> ('a, 'b, 'e) FWState.new_thread_action -> 'd
  val redT_updTs :
    ('a -> 'b * ('c, IntInf.int) FinFun.finfun -> 'd -> 'd) ->
      'd -> ('a, 'b, 'e) FWState.new_thread_action list -> 'd
  val exec_updW :
    'c HOL.equal ->
      ('a -> 'b -> 'c -> 'd -> 'b option) ->
        ('b -> 'd -> 'c FWState.wait_set_status option) ->
          ('b -> 'c FWState.wait_set_status -> 'd -> 'd) ->
            ('b -> 'd -> 'd) ->
              (('b -> 'c FWState.wait_set_status -> 'd -> 'd) ->
                'd -> 'd -> 'd) ->
                'a -> 'b -> 'd -> ('b, 'c) FWState.wait_set_action -> 'd
  val exec_updWs :
    'c HOL.equal ->
      ('a -> 'b -> 'c -> 'd -> 'b option) ->
        ('b -> 'd -> 'c FWState.wait_set_status option) ->
          ('b -> 'c FWState.wait_set_status -> 'd -> 'd) ->
            ('b -> 'd -> 'd) ->
              (('b -> 'c FWState.wait_set_status -> 'd -> 'd) ->
                'd -> 'd -> 'd) ->
                'a -> 'b -> 'd -> ('b, 'c) FWState.wait_set_action list -> 'd
  val redT_updI :
    ('a -> 'b -> 'b) ->
      ('a -> 'b -> 'b) -> 'b -> 'a FWState.interrupt_action -> 'b
  val redT_updIs :
    ('a -> 'b -> 'b) ->
      ('a -> 'b -> 'b) -> 'b -> 'a FWState.interrupt_action list -> 'b
  val exec_upd :
    'b HOL.equal -> 'c HOL.equal -> 'g HOL.equal ->
      ('a -> 'b -> 'c -> 'd -> 'b option) ->
        ('b -> 'e -> ('f * ('g, IntInf.int) FinFun.finfun) option) ->
          ('b -> 'f * ('g, IntInf.int) FinFun.finfun -> 'e -> 'e) ->
            ('b -> 'd -> 'c FWState.wait_set_status option) ->
              ('b -> 'c FWState.wait_set_status -> 'd -> 'd) ->
                ('b -> 'd -> 'd) ->
                  (('b -> 'c FWState.wait_set_status -> 'd -> 'd) ->
                    'd -> 'd -> 'd) ->
                    ('b -> 'h -> 'h) ->
                      ('b -> 'h -> 'h) ->
                        'a -> ('g, (('b * IntInf.int) option)) FinFun.finfun *
                                (('e * 'i) * ('d * 'h)) ->
                                'b -> ('g, (FWState.lock_action list))
FinFun.finfun *
(('b, 'f, 'i) FWState.new_thread_action list *
  ('b FWState.conditional_action list *
    (('b, 'c) FWState.wait_set_action list *
      ('b FWState.interrupt_action list * 'j list)))) ->
'f -> 'i -> ('g, (('b * IntInf.int) option)) FinFun.finfun *
              (('e * 'i) * ('d * 'h))
  val execT :
    'a HOL.equal -> 'd HOL.equal -> 'j HOL.equal ->
      (('a, IntInf.int) FinFun.finfun -> 'b list) ->
        ('c ->
          ('a, (('d * IntInf.int) option)) FinFun.finfun *
            (('e * 'f) * ('g * 'h)) ->
            ('d * (((('a, (FWState.lock_action list)) FinFun.finfun *
                      (('d, 'i, 'f) FWState.new_thread_action list *
                        ('d FWState.conditional_action list *
                          (('d, 'j) FWState.wait_set_action list *
                            ('d FWState.interrupt_action list * 'b list))))) *
                     ('i * 'f)) option *
                    'c)) option) ->
          ('c -> 'd -> 'j -> 'g -> 'd option) ->
            ('d -> 'e -> ('i * ('a, IntInf.int) FinFun.finfun) option) ->
              ('d -> 'i * ('a, IntInf.int) FinFun.finfun -> 'e -> 'e) ->
                ('d -> 'g -> 'j FWState.wait_set_status option) ->
                  ('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                    ('d -> 'g -> 'g) ->
                      (('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                        'g -> 'g -> 'g) ->
                        ('d -> 'h -> 'h) ->
                          ('d -> 'h -> 'h) ->
                            'c -> ('a, (('d * IntInf.int) option))
                                    FinFun.finfun *
                                    (('e * 'f) * ('g * 'h)) ->
                                    ('c * ('d *
    ((('a, (FWState.lock_action list)) FinFun.finfun *
       (('d, 'i, 'f) FWState.new_thread_action list *
         ('d FWState.conditional_action list *
           (('d, 'j) FWState.wait_set_action list *
             ('d FWState.interrupt_action list * 'b list))))) *
      (('a, (('d * IntInf.int) option)) FinFun.finfun *
        (('e * 'f) * ('g * 'h)))))) option
  val exec_step :
    'a HOL.equal -> 'd HOL.equal -> 'j HOL.equal ->
      (('a, IntInf.int) FinFun.finfun -> 'b list) ->
        ('c ->
          ('a, (('d * IntInf.int) option)) FinFun.finfun *
            (('e * 'f) * ('g * 'h)) ->
            ('d * (((('a, (FWState.lock_action list)) FinFun.finfun *
                      (('d, 'i, 'f) FWState.new_thread_action list *
                        ('d FWState.conditional_action list *
                          (('d, 'j) FWState.wait_set_action list *
                            ('d FWState.interrupt_action list * 'b list))))) *
                     ('i * 'f)) option *
                    'c)) option) ->
          ('c -> 'd -> 'j -> 'g -> 'd option) ->
            ('d -> 'e -> ('i * ('a, IntInf.int) FinFun.finfun) option) ->
              ('d -> 'i * ('a, IntInf.int) FinFun.finfun -> 'e -> 'e) ->
                ('d -> 'g -> 'j FWState.wait_set_status option) ->
                  ('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                    ('d -> 'g -> 'g) ->
                      (('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                        'g -> 'g -> 'g) ->
                        ('d -> 'h -> 'h) ->
                          ('d -> 'h -> 'h) ->
                            'c * (('a, (('d * IntInf.int) option))
                                    FinFun.finfun *
                                   (('e * 'f) * ('g * 'h))) ->
                              ((('c * ('d *
(('a, (FWState.lock_action list)) FinFun.finfun *
  (('d, 'i, 'f) FWState.new_thread_action list *
    ('d FWState.conditional_action list *
      (('d, 'j) FWState.wait_set_action list *
        ('d FWState.interrupt_action list * 'b list))))))) *
                                 ('c * (('a, (('d * IntInf.int) option))
  FinFun.finfun *
 (('e * 'f) * ('g * 'h))))),
                                (('a, (('d * IntInf.int) option))
                                   FinFun.finfun *
                                  (('e * 'f) * ('g * 'h))))
                                Sum_Type.sum
  val exec_aux :
    'a HOL.equal -> 'd HOL.equal -> 'j HOL.equal ->
      (('a, IntInf.int) FinFun.finfun -> 'b list) ->
        ('c ->
          ('a, (('d * IntInf.int) option)) FinFun.finfun *
            (('e * 'f) * ('g * 'h)) ->
            ('d * (((('a, (FWState.lock_action list)) FinFun.finfun *
                      (('d, 'i, 'f) FWState.new_thread_action list *
                        ('d FWState.conditional_action list *
                          (('d, 'j) FWState.wait_set_action list *
                            ('d FWState.interrupt_action list * 'b list))))) *
                     ('i * 'f)) option *
                    'c)) option) ->
          ('c -> 'd -> 'j -> 'g -> 'd option) ->
            ('d -> 'e -> ('i * ('a, IntInf.int) FinFun.finfun) option) ->
              ('d -> 'i * ('a, IntInf.int) FinFun.finfun -> 'e -> 'e) ->
                ('d -> 'g -> 'j FWState.wait_set_status option) ->
                  ('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                    ('d -> 'g -> 'g) ->
                      (('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                        'g -> 'g -> 'g) ->
                        ('d -> 'h -> 'h) ->
                          ('d -> 'h -> 'h) ->
                            'c * (('a, (('d * IntInf.int) option))
                                    FinFun.finfun *
                                   (('e * 'f) * ('g * 'h))) ->
                              (('c * ('d * (('a, (FWState.lock_action list))
      FinFun.finfun *
     (('d, 'i, 'f) FWState.new_thread_action list *
       ('d FWState.conditional_action list *
         (('d, 'j) FWState.wait_set_action list *
           ('d FWState.interrupt_action list * 'b list))))))),
                                (('a, (('d * IntInf.int) option))
                                   FinFun.finfun *
                                  (('e * 'f) * ('g * 'h))))
                                TLList.tllist
  val exec :
    'a HOL.equal -> 'd HOL.equal -> 'j HOL.equal ->
      (('a, IntInf.int) FinFun.finfun -> 'b list) ->
        ('c ->
          ('a, (('d * IntInf.int) option)) FinFun.finfun *
            (('e * 'f) * ('g * 'h)) ->
            ('d * (((('a, (FWState.lock_action list)) FinFun.finfun *
                      (('d, 'i, 'f) FWState.new_thread_action list *
                        ('d FWState.conditional_action list *
                          (('d, 'j) FWState.wait_set_action list *
                            ('d FWState.interrupt_action list * 'b list))))) *
                     ('i * 'f)) option *
                    'c)) option) ->
          ('c ->
            'd -> ('a, (FWState.lock_action list)) FinFun.finfun *
                    (('d, 'i, 'f) FWState.new_thread_action list *
                      ('d FWState.conditional_action list *
                        (('d, 'j) FWState.wait_set_action list *
                          ('d FWState.interrupt_action list * 'b list)))) ->
                    'k option) ->
            ('c -> 'd -> 'j -> 'g -> 'd option) ->
              ('d -> 'e -> ('i * ('a, IntInf.int) FinFun.finfun) option) ->
                ('d -> 'i * ('a, IntInf.int) FinFun.finfun -> 'e -> 'e) ->
                  ('d -> 'g -> 'j FWState.wait_set_status option) ->
                    ('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                      ('d -> 'g -> 'g) ->
                        (('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                          'g -> 'g -> 'g) ->
                          ('d -> 'h -> 'h) ->
                            ('d -> 'h -> 'h) ->
                              'c -> ('a, (('d * IntInf.int) option))
                                      FinFun.finfun *
                                      (('e * 'f) * ('g * 'h)) ->
                                      ('k,
(('a, (('d * IntInf.int) option)) FinFun.finfun * (('e * 'f) * ('g * 'h))))
TLList.tllist
  val singleton2 : (unit -> 'a) -> 'a Predicate.pred -> 'a
  val the_only2 : (unit -> 'a) -> 'a Predicate.seq -> 'a
  val the2 : 'a Predicate.pred -> 'a
  val free_thread_id :
    ('a -> 'b -> ('c * ('d, IntInf.int) FinFun.finfun) option) ->
      'b -> 'a -> bool
  val thread_ok :
    ('a -> 'b -> ('c * ('d, IntInf.int) FinFun.finfun) option) ->
      'b -> ('a, 'c, 'e) FWState.new_thread_action -> bool
  val thread_oks :
    ('a -> 'b -> ('c * ('d, IntInf.int) FinFun.finfun) option) ->
      ('a -> 'c * ('d, IntInf.int) FinFun.finfun -> 'b -> 'b) ->
        'b -> ('a, 'c, 'e) FWState.new_thread_action list -> bool
  val cond_action_ok :
    'b HOL.equal -> 'd Card_Univ.card_UNIV * 'd HOL.equal ->
      ('a -> bool) ->
        ('b -> 'c -> ('a * ('d, IntInf.int) FinFun.finfun) option) ->
          ('b -> 'e -> 'f FWState.wait_set_status option) ->
            ('d, (('b * IntInf.int) option)) FinFun.finfun *
              (('c * 'g) * ('e * 'h)) ->
              'b -> 'b FWState.conditional_action -> bool
  val cond_action_oks :
    'b HOL.equal -> 'd Card_Univ.card_UNIV * 'd HOL.equal ->
      ('a -> bool) ->
        ('b -> 'c -> ('a * ('d, IntInf.int) FinFun.finfun) option) ->
          ('b -> 'e -> 'f FWState.wait_set_status option) ->
            ('d, (('b * IntInf.int) option)) FinFun.finfun *
              (('c * 'g) * ('e * 'h)) ->
              'b -> 'b FWState.conditional_action list -> bool
  val wset_actions_ok :
    'a HOL.equal -> 'c HOL.equal ->
      ('a -> 'b -> 'c FWState.wait_set_status option) ->
        'b -> 'a -> ('a, 'c) FWState.wait_set_action list -> bool
  val interrupt_action_ok :
    ('a -> 'b -> bool) -> 'b -> 'a FWState.interrupt_action -> bool
  val interrupt_actions_ok :
    ('a -> 'b -> bool) ->
      ('a -> 'b -> 'b) ->
        ('a -> 'b -> 'b) -> 'b -> 'a FWState.interrupt_action list -> bool
  val actions_ok :
    'b HOL.equal -> 'd Card_Univ.card_UNIV * 'd HOL.equal -> 'f HOL.equal ->
      ('a -> bool) ->
        ('b -> 'c -> ('a * ('d, IntInf.int) FinFun.finfun) option) ->
          ('b -> 'a * ('d, IntInf.int) FinFun.finfun -> 'c -> 'c) ->
            ('b -> 'e -> 'f FWState.wait_set_status option) ->
              ('b -> 'g -> bool) ->
                ('b -> 'g -> 'g) ->
                  ('b -> 'g -> 'g) ->
                    ('d, (('b * IntInf.int) option)) FinFun.finfun *
                      (('c * 'h) * ('e * 'g)) ->
                      'b -> ('d, (FWState.lock_action list)) FinFun.finfun *
                              (('b, 'a, 'h) FWState.new_thread_action list *
                                ('b FWState.conditional_action list *
                                  (('b, 'f) FWState.wait_set_action list *
                                    ('b FWState.interrupt_action list *
                                      'i list)))) ->
                              bool
  val step_thread :
    'b HOL.equal -> 'd Card_Univ.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          (('d, IntInf.int) FinFun.finfun -> 'f list) ->
            ('b -> 'g -> ('a * ('d, IntInf.int) FinFun.finfun) option) ->
              ('b -> 'a * ('d, IntInf.int) FinFun.finfun -> 'g -> 'g) ->
                ('b -> 'h -> 'e FWState.wait_set_status option) ->
                  ('b -> 'i -> bool) ->
                    ('b -> 'i -> 'i) ->
                      ('b -> 'i -> 'i) ->
                        (('d, (FWState.lock_action list)) FinFun.finfun *
                           (('b, 'a, 'c) FWState.new_thread_action list *
                             ('b FWState.conditional_action list *
                               (('b, 'e) FWState.wait_set_action list *
                                 ('b FWState.interrupt_action list *
                                   'f list)))) ->
                          'j) ->
                          ('d, (('b * IntInf.int) option)) FinFun.finfun *
                            (('g * 'c) * ('h * 'i)) ->
                            'b -> ('b * (((('d, (FWState.lock_action list))
     FinFun.finfun *
    (('b, 'a, 'c) FWState.new_thread_action list *
      ('b FWState.conditional_action list *
        (('b, 'e) FWState.wait_set_action list *
          ('b FWState.interrupt_action list * 'f list))))) *
   ('a * 'c)) option *
  'j)) option
  val active_threads :
    'b HOL.equal -> 'd Card_Univ.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          ('b -> 'g -> ('a * ('d, IntInf.int) FinFun.finfun) option) ->
            ('b -> 'a * ('d, IntInf.int) FinFun.finfun -> 'g -> 'g) ->
              (('b -> 'a * ('d, IntInf.int) FinFun.finfun -> 'h -> 'h) ->
                'g -> 'h -> 'h) ->
                ('b -> 'i -> 'e FWState.wait_set_status option) ->
                  ('b -> 'j -> bool) ->
                    ('b -> 'j -> 'j) ->
                      ('b -> 'j -> 'j) ->
                        'h -> ('b -> 'h -> 'h) ->
                                ('d, (('b * IntInf.int) option)) FinFun.finfun *
                                  (('g * 'c) * ('i * 'j)) ->
                                  'h
  val pick_wakeup_via_sel :
    'c HOL.equal ->
      ('a ->
        ('b -> 'c FWState.wait_set_status -> bool) ->
          ('b * 'c FWState.wait_set_status) option) ->
        'd -> 'b -> 'c -> 'a -> 'b option
end = struct

fun redT_updT thr_update ts (FWState.ThreadExists (v, va)) = ts
  | redT_updT thr_update ts (FWState.NewThread (t, x, m)) =
    thr_update t (x, FinFun.Finfun_const (0 : IntInf.int)) ts;

fun redT_updTs thr_update = List.foldl (redT_updT thr_update);

fun exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
  FWState.WokenUp = ws_delete t ws
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
    FWState.Notified = ws_delete t ws
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma ta
    ws (FWState.WakeUp t) =
    (case ws_lookup t ws of NONE => ws
      | SOME (FWState.InWS _) =>
        ws_update t (FWState.PostWS FWState.WSWokenUp) ws
      | SOME (FWState.PostWS _) => ws)
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
    (FWState.Suspend w) = ws_update t (FWState.InWS w) ws
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
    (FWState.NotifyAll w) =
    ws_iterate
      (fn ta => fn wa => fn wsa =>
        (if FWState.equal_wait_set_statusa C_ wa (FWState.InWS w)
          then ws_update ta (FWState.PostWS FWState.WSNotified) wsa else wsa))
      ws
      ws
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
    (FWState.Notify w) =
    (case pick_wakeup sigma t w ws of NONE => ws
      | SOME ta => ws_update ta (FWState.PostWS FWState.WSNotified) ws);

fun exec_updWs C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t =
  List.foldl
    (exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t);

fun redT_updI is_ins is_delete is (FWState.IsInterrupted (t, b)) = is
  | redT_updI is_ins is_delete is (FWState.ClearInterrupt t) = is_delete t is
  | redT_updI is_ins is_delete is (FWState.Interrupt t) = is_ins t is;

fun redT_updIs is_ins is_delete is (ia :: ias) =
  redT_updIs is_ins is_delete (redT_updI is_ins is_delete is ia) ias
  | redT_updIs is_ins is_delete is [] = is;

fun exec_upd B_ C_ G_ pick_wakeup thr_lookup thr_update ws_lookup ws_update
  ws_delete ws_iterate is_ins is_delete sigma s t ta x m =
  (FWLocking.redT_updLs G_ B_ (FWState.locks s) t (FWState.locks_a ta),
    ((thr_update t
        (x, FWLockingThread.redT_updLns G_ B_ (FWState.locks s) t
              (Product_Type.snd (Option.the (thr_lookup t (FWState.thr s))))
              (FWState.locks_a ta))
        (redT_updTs thr_update (FWState.thr s) (FWState.thr_a ta)),
       m),
      (exec_updWs C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma
         t (FWState.wset s) (FWState.wset_a ta),
        redT_updIs is_ins is_delete (FWState.interrupts s)
          (FWState.interrupt_a ta))));

fun execT A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
  ws_lookup ws_update ws_delete ws_iterate is_ins is_delete sigma s =
  Option.bind (schedule sigma s)
    (fn a =>
      (case a
        of (t, (NONE, sigmaa)) =>
          let
            val (x, ln) = Option.the (thr_lookup t (FWState.thr s));
            val ta =
              (FinFun.Finfun_const [], ([], ([], ([], ([], convert_RA ln)))));
            val sa =
              (FWLocking.acquire_all A_ D_ (FWState.locks s) t ln,
                ((thr_update t (x, FinFun.Finfun_const (0 : IntInf.int))
                    (FWState.thr s),
                   FWState.shr s),
                  (FWState.wset s, FWState.interrupts s)));
          in
            SOME (sigmaa, (t, (ta, sa)))
          end
        | (t, (SOME (ta, (x, m)), sigmaa)) =>
          SOME (sigmaa,
                 (t, (ta, exec_upd D_ J_ A_ pick_wakeup thr_lookup thr_update
                            ws_lookup ws_update ws_delete ws_iterate is_ins
                            is_delete sigma s t ta x m)))));

fun exec_step A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
  ws_lookup ws_update ws_delete ws_iterate is_ins is_delete (sigma, s) =
  (case execT A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
          ws_lookup ws_update ws_delete ws_iterate is_ins is_delete sigma s
    of NONE => Sum_Type.Inr s
    | SOME (sigmaa, (t, (ta, sa))) =>
      Sum_Type.Inl ((sigma, (t, ta)), (sigmaa, sa)));

fun exec_aux A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
  ws_lookup ws_update ws_delete ws_iterate is_ins is_delete sigma_s =
  TLList.tllist_corec sigma_s
    (exec_step A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
      ws_lookup ws_update ws_delete ws_iterate is_ins is_delete);

fun exec A_ D_ J_ convert_RA schedule output pick_wakeup thr_lookup thr_update
  ws_lookup ws_update ws_delete ws_iterate is_ins is_delete sigma s =
  TLList.tmap Option.the Fun.id
    (TLList.tfilter (fn _ => (raise Fail "undefined"))
      (fn q => not (Option.is_none q))
      (TLList.tmap (fn (sigmaa, (a, b)) => output sigmaa a b) Fun.id
        (exec_aux A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
          ws_lookup ws_update ws_delete ws_iterate is_ins is_delete
          (sigma, s))));

fun singleton2 dfault (Predicate.Seq f) =
  (case f () of Predicate.Empty => dfault ()
    | Predicate.Insert (x, p) =>
      (if Predicate.is_empty p then x
        else FinFun.code_abort (fn _ => singleton2 dfault (Predicate.Seq f)))
    | Predicate.Join (p, xq) =>
      (if Predicate.is_empty p then the_only2 dfault xq
        else (if Predicate.null xq then singleton2 dfault p
               else FinFun.code_abort
                      (fn _ => singleton2 dfault (Predicate.Seq f)))))
and the_only2 dfault (Predicate.Join (p, xq)) =
  (if Predicate.is_empty p then the_only2 dfault xq
    else (if Predicate.null xq then singleton2 dfault p
           else FinFun.code_abort
                  (fn _ => the_only2 dfault (Predicate.Join (p, xq)))))
  | the_only2 dfault (Predicate.Insert (x, p)) =
    (if Predicate.is_empty p then x
      else FinFun.code_abort
             (fn _ => the_only2 dfault (Predicate.Insert (x, p))))
  | the_only2 dfault Predicate.Empty = FinFun.code_abort dfault;

fun the2 a = singleton2 (fn _ => Predicate.not_unique a) a;

fun free_thread_id thr_lookup ts t = Option.is_none (thr_lookup t ts);

fun thread_ok thr_lookup ts (FWState.ThreadExists (t, b)) =
  not (Product_Type.equal_bool b (free_thread_id thr_lookup ts t))
  | thread_ok thr_lookup ts (FWState.NewThread (t, x, m)) =
    free_thread_id thr_lookup ts t;

fun thread_oks thr_lookup thr_update ts (ta :: tas) =
  thread_ok thr_lookup ts ta andalso
    thread_oks thr_lookup thr_update (redT_updT thr_update ts ta) tas
  | thread_oks thr_lookup thr_update ts [] = true;

fun cond_action_ok B_ (D1_, D2_) final thr_lookup ws_lookup s t FWState.Yield =
  true
  | cond_action_ok B_ (D1_, D2_) final thr_lookup ws_lookup s ta
    (FWState.Join t) =
    (case thr_lookup t (FWState.thr s) of NONE => true
      | SOME (x, ln) =>
        not (HOL.eq B_ ta t) andalso
          (final x andalso
            (HOL.eq (FinFun.equal_finfun (D1_, D2_) Arith.equal_nat) ln
               (FinFun.Finfun_const (0 : IntInf.int)) andalso
              Option.is_none (ws_lookup t (FWState.wset s)))));

fun cond_action_oks B_ (D1_, D2_) final thr_lookup ws_lookup s t cts =
  List.list_all (cond_action_ok B_ (D1_, D2_) final thr_lookup ws_lookup s t)
    cts;

fun wset_actions_ok A_ C_ ws_lookup ws t was =
  Option.equal_optiona (FWState.equal_wait_set_status C_) (ws_lookup t ws)
    (if List.member (FWState.equal_wait_set_action A_ C_) was FWState.Notified
      then SOME (FWState.PostWS FWState.WSNotified)
      else (if List.member (FWState.equal_wait_set_action A_ C_) was
                 FWState.WokenUp
             then SOME (FWState.PostWS FWState.WSWokenUp) else NONE));

fun interrupt_action_ok is_memb is (FWState.IsInterrupted (t, b)) =
  Product_Type.equal_bool b (is_memb t is)
  | interrupt_action_ok is_memb is (FWState.ClearInterrupt t) = true
  | interrupt_action_ok is_memb is (FWState.Interrupt t) = true;

fun interrupt_actions_ok is_memb is_ins is_delete is (ia :: ias) =
  interrupt_action_ok is_memb is ia andalso
    interrupt_actions_ok is_memb is_ins is_delete
      (redT_updI is_ins is_delete is ia) ias
  | interrupt_actions_ok is_memb is_ins is_delete is [] = true;

fun actions_ok B_ (D1_, D2_) F_ final thr_lookup thr_update ws_lookup is_memb
  is_ins is_delete s t ta =
  FWLocking.lock_ok_las (D1_, D2_) B_ (FWState.locks s) t
    (FWState.locks_a ta) andalso
    (thread_oks thr_lookup thr_update (FWState.thr s) (FWState.thr_a ta) andalso
      (cond_action_oks B_ (D1_, D2_) final thr_lookup ws_lookup s t
         (FWState.cond_a ta) andalso
        (wset_actions_ok B_ F_ ws_lookup (FWState.wset s) t
           (FWState.wset_a ta) andalso
          interrupt_actions_ok is_memb is_ins is_delete (FWState.interrupts s)
            (FWState.interrupt_a ta))));

fun step_thread B_ (D1_, D2_) E_ final r convert_RA thr_lookup thr_update
  ws_lookup is_memb is_ins is_delete update_state s t =
  (case thr_lookup t (FWState.thr s) of NONE => NONE
    | SOME (x, ln) =>
      (if HOL.eq (FinFun.equal_finfun (D1_, D2_) Arith.equal_nat) ln
            (FinFun.Finfun_const (0 : IntInf.int))
        then let
               val reds =
                 Predicate.bind (r t (x, FWState.shr s))
                   (fn (ta, (xa, m)) =>
                     (if actions_ok B_ (D1_, D2_) E_ final thr_lookup thr_update
                           ws_lookup is_memb is_ins is_delete s t ta
                       then Predicate.single (ta, (xa, m))
                       else Predicate.bot_pred));
             in
               (if Predicate.holds
                     (Predicate.bind reds (fn _ => Predicate.single ()))
                 then let
                        val (ta, (xa, m)) = the2 reds;
                      in
                        SOME (t, (SOME (ta, (xa, m)), update_state ta))
                      end
                 else NONE)
             end
        else (if FWLocking.may_acquire_all (D1_, D2_) B_ (FWState.locks s) t
                   ln andalso
                   not (FWWait.waiting (ws_lookup t (FWState.wset s)))
               then SOME (t, (NONE,
                               update_state
                                 (FinFun.Finfun_const [],
                                   ([], ([], ([], ([], convert_RA ln)))))))
               else NONE)));

fun active_threads B_ (D1_, D2_) E_ final r thr_lookup thr_update thr_iterate
  ws_lookup is_memb is_ins is_delete thr_empty thr_ins_dj
  (ls, ((ts, m), (ws, is))) =
  thr_iterate
    (fn t => fn (x, ln) => fn tsa =>
      (if HOL.eq (FinFun.equal_finfun (D1_, D2_) Arith.equal_nat) ln
            (FinFun.Finfun_const (0 : IntInf.int))
        then (if Predicate.holds
                   (Predicate.bind (r t (x, m))
                     (fn (ta, _) =>
                       Predicate.if_pred
                         (actions_ok B_ (D1_, D2_) E_ final thr_lookup
                           thr_update ws_lookup is_memb is_ins is_delete
                           (ls, ((ts, m), (ws, is))) t ta)))
               then thr_ins_dj t tsa else tsa)
        else (if not (FWWait.waiting (ws_lookup t ws)) andalso
                   FWLocking.may_acquire_all (D1_, D2_) B_ ls t ln
               then thr_ins_dj t tsa else tsa)))
    ts
    thr_empty;

fun pick_wakeup_via_sel C_ ws_sel sigma t w ws =
  Option.map Product_Type.fst
    (ws_sel ws
      (fn _ => fn wa => FWState.equal_wait_set_statusa C_ wa (FWState.InWS w)));

end; (*struct Scheduler*)

structure Random_Scheduler : sig
  val next_thread :
    ('a -> 'b list) -> int * int -> 'a -> ('b * (int * int)) option
  val random_scheduler :
    'b HOL.equal -> 'd Card_Univ.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          (('d, IntInf.int) FinFun.finfun -> 'f list) ->
            ('b -> 'g -> ('a * ('d, IntInf.int) FinFun.finfun) option) ->
              ('b -> 'a * ('d, IntInf.int) FinFun.finfun -> 'g -> 'g) ->
                (('b -> 'a * ('d, IntInf.int) FinFun.finfun -> 'h -> 'h) ->
                  'g -> 'h -> 'h) ->
                  ('b -> 'i -> 'e FWState.wait_set_status option) ->
                    ('b -> 'j -> bool) ->
                      ('b -> 'j -> 'j) ->
                        ('b -> 'j -> 'j) ->
                          'h -> ('b -> 'h -> 'h) ->
                                  ('h -> 'b list) ->
                                    int * int ->
                                      ('d, (('b * IntInf.int) option))
FinFun.finfun *
(('g * 'c) * ('i * 'j)) ->
('b * (((('d, (FWState.lock_action list)) FinFun.finfun *
          (('b, 'a, 'c) FWState.new_thread_action list *
            ('b FWState.conditional_action list *
              (('b, 'e) FWState.wait_set_action list *
                ('b FWState.interrupt_action list * 'f list))))) *
         ('a * 'c)) option *
        (int * int))) option
end = struct

fun next_thread thr_to_list seed active =
  let
    val ts = thr_to_list active;
  in
    (if List.null ts then NONE
      else SOME (Random.select (thr_to_list active) seed))
  end;

fun random_scheduler B_ (D1_, D2_) E_ final r convert_RA thr_lookup thr_update
  thr_iterate ws_lookup is_memb is_ins is_delete thr_empty thr_ins_dj
  thr_to_list seed s =
  Option.bind
    (next_thread thr_to_list seed
      (Scheduler.active_threads B_ (D1_, D2_) E_ final r thr_lookup thr_update
        thr_iterate ws_lookup is_memb is_ins is_delete thr_empty thr_ins_dj s))
    (fn (t, seeda) =>
      Scheduler.step_thread B_ (D1_, D2_) E_ final r convert_RA thr_lookup
        thr_update ws_lookup is_memb is_ins is_delete (fn _ => seeda) s t);

end; (*struct Random_Scheduler*)

structure ListSetImpl_Invar : sig
  val lsi_empty : 'a list
  val lsi_ins_dj : 'a -> 'a list -> 'a list
  val lsi_to_list : 'a list -> 'a list
end = struct

val lsi_empty : 'a list = [];

fun lsi_ins_dj x l = x :: l;

fun lsi_to_list x = Fun.id x;

end; (*struct ListSetImpl_Invar*)

structure Round_Robin : sig
  val enqueue_new_thread :
    ('a -> 'b -> 'b) -> 'b -> ('a, 'c, 'd) FWState.new_thread_action -> 'b
  val enqueue_new_threads :
    ('a -> 'b -> 'b) -> 'b -> ('a, 'c, 'd) FWState.new_thread_action list -> 'b
  val queue_rotate1 : ('a -> 'b -> 'b) -> ('b -> 'a * 'b) -> 'b -> 'b
  val round_robin_update_state :
    'a HOL.equal ->
      ('a -> 'b -> 'b) ->
        ('b -> 'a * 'b) ->
          IntInf.int ->
            'b * IntInf.int ->
              'a -> ('c, (FWState.lock_action list)) FinFun.finfun *
                      (('a, 'd, 'e) FWState.new_thread_action list *
                        ('a FWState.conditional_action list *
                          (('a, 'f) FWState.wait_set_action list *
                            ('a FWState.interrupt_action list * 'g list)))) ->
                      'b * IntInf.int
  val round_robin_reschedule :
    'b HOL.equal -> 'd Card_Univ.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          (('d, IntInf.int) FinFun.finfun -> 'f list) ->
            ('b -> 'g -> ('a * ('d, IntInf.int) FinFun.finfun) option) ->
              ('b -> 'a * ('d, IntInf.int) FinFun.finfun -> 'g -> 'g) ->
                ('b -> 'h -> 'e FWState.wait_set_status option) ->
                  ('b -> 'i -> bool) ->
                    ('b -> 'i -> 'i) ->
                      ('b -> 'i -> 'i) ->
                        ('b -> 'j -> 'j) ->
                          ('j -> 'b * 'j) ->
                            ('b -> 'j -> 'j) ->
                              'b -> 'j -> IntInf.int ->
    ('d, (('b * IntInf.int) option)) FinFun.finfun * (('g * 'c) * ('h * 'i)) ->
      ('b * (((('d, (FWState.lock_action list)) FinFun.finfun *
                (('b, 'a, 'c) FWState.new_thread_action list *
                  ('b FWState.conditional_action list *
                    (('b, 'e) FWState.wait_set_action list *
                      ('b FWState.interrupt_action list * 'f list))))) *
               ('a * 'c)) option *
              ('j * IntInf.int))) option
  val round_robin :
    'b HOL.equal -> 'd Card_Univ.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          (('d, IntInf.int) FinFun.finfun -> 'f list) ->
            ('b -> 'g -> ('a * ('d, IntInf.int) FinFun.finfun) option) ->
              ('b -> 'a * ('d, IntInf.int) FinFun.finfun -> 'g -> 'g) ->
                ('b -> 'h -> 'e FWState.wait_set_status option) ->
                  ('b -> 'i -> bool) ->
                    ('b -> 'i -> 'i) ->
                      ('b -> 'i -> 'i) ->
                        ('j -> bool) ->
                          ('b -> 'j -> 'j) ->
                            ('j -> 'b * 'j) ->
                              ('b -> 'j -> 'j) ->
                                IntInf.int ->
                                  'j * IntInf.int ->
                                    ('d, (('b * IntInf.int) option))
                                      FinFun.finfun *
                                      (('g * 'c) * ('h * 'i)) ->
                                      ('b *
(((('d, (FWState.lock_action list)) FinFun.finfun *
    (('b, 'a, 'c) FWState.new_thread_action list *
      ('b FWState.conditional_action list *
        (('b, 'e) FWState.wait_set_action list *
          ('b FWState.interrupt_action list * 'f list))))) *
   ('a * 'c)) option *
  ('j * IntInf.int))) option
  val round_robin_start :
    'a -> ('b -> 'a -> 'a) -> IntInf.int -> 'b -> 'a * IntInf.int
end = struct

fun enqueue_new_thread queue_enqueue ts (FWState.ThreadExists (t, b)) = ts
  | enqueue_new_thread queue_enqueue ts (FWState.NewThread (t, x, m)) =
    queue_enqueue t ts;

fun enqueue_new_threads queue_enqueue =
  List.foldl (enqueue_new_thread queue_enqueue);

fun queue_rotate1 queue_enqueue queue_dequeue =
  (fn (a, b) => queue_enqueue a b) o queue_dequeue;

fun round_robin_update_state A_ queue_enqueue queue_dequeue n0 (queue, n) t ta =
  let
    val queuea = enqueue_new_threads queue_enqueue queue (FWState.thr_a ta);
  in
    (if ((n : IntInf.int) = (0 : IntInf.int)) orelse
          List.member (FWState.equal_conditional_action A_) (FWState.cond_a ta)
            FWState.Yield
      then (queue_rotate1 queue_enqueue queue_dequeue queuea, n0)
      else (queuea, Arith.minus_nat n (1 : IntInf.int)))
  end;

fun round_robin_reschedule B_ (D1_, D2_) E_ final r convert_RA thr_lookup
  thr_update ws_lookup is_memb is_ins is_delete queue_enqueue queue_dequeue
  queue_push t0 queue n0 s =
  let
    val (t, queuea) = queue_dequeue queue;
  in
    (if HOL.eq B_ t t0 then NONE
      else (case Scheduler.step_thread B_ (D1_, D2_) E_ final r convert_RA
                   thr_lookup thr_update ws_lookup is_memb is_ins is_delete
                   (round_robin_update_state B_ queue_enqueue queue_dequeue n0
                     (queue_push t queuea, n0) t)
                   s t
             of NONE =>
               round_robin_reschedule B_ (D1_, D2_) E_ final r convert_RA
                 thr_lookup thr_update ws_lookup is_memb is_ins is_delete
                 queue_enqueue queue_dequeue queue_push t0
                 (queue_enqueue t queuea) n0 s
             | SOME a => SOME a))
  end;

fun round_robin B_ (D1_, D2_) E_ final r convert_RA thr_lookup thr_update
  ws_lookup is_memb is_ins is_delete queue_isEmpty queue_enqueue queue_dequeue
  queue_push n0 (queue, n) s =
  (if queue_isEmpty queue then NONE
    else let
           val (t, queuea) = queue_dequeue queue;
         in
           (case Scheduler.step_thread B_ (D1_, D2_) E_ final r convert_RA
                   thr_lookup thr_update ws_lookup is_memb is_ins is_delete
                   (round_robin_update_state B_ queue_enqueue queue_dequeue n0
                     (queue_push t queuea, n) t)
                   s t
             of NONE =>
               round_robin_reschedule B_ (D1_, D2_) E_ final r convert_RA
                 thr_lookup thr_update ws_lookup is_memb is_ins is_delete
                 queue_enqueue queue_dequeue queue_push t
                 (queue_enqueue t queuea) n0 s
             | SOME a => SOME a)
         end);

fun round_robin_start queue_empty queue_enqueue n0 t =
  (queue_enqueue t queue_empty, n0);

end; (*struct Round_Robin*)

structure State_Refinement : sig
  val jinja_output :
    'a -> 'b -> ('c, (FWState.lock_action list)) FinFun.finfun *
                  (('b, 'd, 'e) FWState.new_thread_action list *
                    ('b FWState.conditional_action list *
                      (('b, 'c) FWState.wait_set_action list *
                        ('b FWState.interrupt_action list *
                          ('c, 'b) Observable_Events.obs_event list)))) ->
                  ('b * ('c, 'b) Observable_Events.obs_event list) option
  val start_state_refine :
    'a Value.addr -> 'h HOL.equal ->
      ('a -> 'b) ->
        'c -> ('c -> string -> 'c * 'a option) ->
                'd -> ('b -> 'e * ('a, IntInf.int) FinFun.finfun -> 'd -> 'd) ->
                        'f -> 'g -> (string ->
                                      string ->
Type.ty list -> Type.ty -> 'h -> 'a Value.vala list -> 'e) ->
                                      'h TypeRel.prog ->
string ->
  string ->
    'a Value.vala list ->
      ('a, (('b * IntInf.int) option)) FinFun.finfun * (('d * 'c) * ('f * 'g))
end = struct

fun jinja_output sigma t ta =
  (if List.null (FWState.obs_a ta) then NONE else SOME (t, FWState.obs_a ta));

fun start_state_refine A_ H_ addr2thread_id empty_heap new_obj thr_empty
  thr_update ws_empty is_empty f p c m vs =
  let
    val (d, (ts, (t, ma))) = TypeRel.method H_ p c m;
  in
    (FinFun.Finfun_const NONE,
      ((thr_update (StartConfig.start_tid A_ addr2thread_id empty_heap new_obj)
          (f d m ts t (Option.the ma) vs, FinFun.Finfun_const (0 : IntInf.int))
          thr_empty,
         StartConfig.start_heap A_ empty_heap new_obj),
        (ws_empty, is_empty)))
  end;

end; (*struct State_Refinement*)

structure ListMapImpl : sig
  val lm_empty : ('a, 'b) Assoc_List.assoc_list
  val lm_lookup :
    'a HOL.equal -> 'a -> ('a, 'b) Assoc_List.assoc_list -> 'b option
  val lm_update :
    'a HOL.equal ->
      'a -> 'b -> ('a, 'b) Assoc_List.assoc_list ->
                    ('a, 'b) Assoc_List.assoc_list
  val list_to_lm :
    'a HOL.equal -> ('a * 'b) list -> ('a, 'b) Assoc_List.assoc_list
end = struct

val lm_empty : ('a, 'b) Assoc_List.assoc_list = Assoc_List.empty;

fun lm_lookup A_ k m = Assoc_List.lookup A_ m k;

fun lm_update A_ = Assoc_List.update A_;

fun list_to_lm A_ = MapGA.gen_list_to_map lm_empty (lm_update A_);

end; (*struct ListMapImpl*)

structure SC_Collections : sig
  datatype heapobj =
    Obj of
      string *
        (char, (string, IntInf.int Value.vala) Assoc_List.assoc_list) Trie.trie
    | Arr of
        Type.ty * IntInf.int *
          (string, IntInf.int Value.vala) Assoc_List.assoc_list *
          (IntInf.int, IntInf.int Value.vala) RBT.rbt
  val init_fields :
    ((string * string) * Type.ty) list ->
      (char, (string, IntInf.int Value.vala) Assoc_List.assoc_list) Trie.trie
  val blank : 'a TypeRel.prog -> string -> heapobj
  val obj_ty : heapobj -> Type.ty
  val new_Addr : (IntInf.int, heapobj) RBT.rbt -> IntInf.int option
  val init_fields_array :
    (string * Type.ty) list ->
      (string, IntInf.int Value.vala) Assoc_List.assoc_list
  val init_cells :
    Type.ty -> IntInf.int -> (IntInf.int, IntInf.int Value.vala) RBT.rbt
  val blank_arr : 'a TypeRel.prog -> Type.ty -> IntInf.int -> heapobj
  val sc_new_arr :
    'a TypeRel.prog ->
      (IntInf.int, heapobj) RBT.rbt ->
        Type.ty ->
          IntInf.int -> (IntInf.int, heapobj) RBT.rbt * IntInf.int option
  val sc_new_obj :
    'a TypeRel.prog ->
      (IntInf.int, heapobj) RBT.rbt ->
        string -> (IntInf.int, heapobj) RBT.rbt * IntInf.int option
  val sc_typeof_addr :
    (IntInf.int, heapobj) RBT.rbt -> IntInf.int -> Type.ty option
  val sc_array_length :
    (IntInf.int, heapobj) RBT.rbt -> IntInf.int -> IntInf.int
  val sc_heap_read_i_i_i_o :
    (IntInf.int, heapobj) RBT.rbt ->
      IntInf.int -> Heap.addr_loc -> IntInf.int Value.vala Predicate.pred
  val sc_heap_write_i_i_i_i_o :
    (IntInf.int, heapobj) RBT.rbt ->
      IntInf.int ->
        Heap.addr_loc ->
          IntInf.int Value.vala -> (IntInf.int, heapobj) RBT.rbt Predicate.pred
end = struct

datatype heapobj =
  Obj of
    string *
      (char, (string, IntInf.int Value.vala) Assoc_List.assoc_list) Trie.trie
  | Arr of
      Type.ty * IntInf.int *
        (string, IntInf.int Value.vala) Assoc_List.assoc_list *
        (IntInf.int, IntInf.int Value.vala) RBT.rbt;

fun init_fields fDTs =
  List.foldr
    (fn (a, b) =>
      let
        val (f, d) = a;
      in
        (fn t => fn fields =>
          let
            val fa = String.explode f;
          in
            TrieMapImpl.tm_update Stringa.equal_char fa
              (ListMapImpl.lm_update Stringa.equal_literal d
                (Value.default_val t)
                (case TrieMapImpl.tm_lookup Stringa.equal_char fa fields
                  of NONE => ListMapImpl.lm_empty | SOME lm => lm))
              fields
          end)
      end
        b)
    fDTs TrieMapImpl.tm_empty;

fun blank p c =
  Obj (c, init_fields
            (List.map (fn (fd, (t, _)) => (fd, t)) (TypeRel.fields p c)));

fun obj_ty (Obj (c, f)) = Type.Class c
  | obj_ty (Arr (t, f, si, e)) = Type.Array t;

fun new_Addr h =
  SOME (case RBTMapImpl.rm_max Arith.linorder_nat h (fn _ => fn _ => true)
         of NONE => (0 : IntInf.int)
         | SOME (a, _) => IntInf.+ (a, (1 : IntInf.int)));

fun init_fields_array x =
  (ListMapImpl.list_to_lm Stringa.equal_literal o
    List.map (fn (f, t) => (f, Value.default_val t)))
    x;

fun init_cells t n =
  List.foldl
    (fn cells => fn i =>
      RBTMapImpl.rm_update Arith.linorder_nat i (Value.default_val t) cells)
    (RBTMapImpl.rm_empty Arith.linorder_nat) (List.upt (0 : IntInf.int) n);

fun blank_arr p t n =
  Arr (t, n,
        init_fields_array
          (List.map
            (fn (a, b) => let
                            val (f, _) = a;
                          in
                            (fn (ta, _) => (f, ta))
                          end
                            b)
            (TypeRel.fields p Type.object)),
        init_cells t n);

fun sc_new_arr p h t n =
  (case new_Addr h of NONE => (h, NONE)
    | SOME a =>
      (RBTMapImpl.rm_update Arith.linorder_nat a (blank_arr p t n) h, SOME a));

fun sc_new_obj p h c =
  (case new_Addr h of NONE => (h, NONE)
    | SOME a =>
      (RBTMapImpl.rm_update Arith.linorder_nat a (blank p c) h, SOME a));

fun sc_typeof_addr h a =
  Option.map obj_ty (RBTMapImpl.rm_lookup Arith.linorder_nat a h);

fun sc_array_length h a =
  let
    val Arr (_, si, _, _) =
      Option.the (RBTMapImpl.rm_lookup Arith.linorder_nat a h);
  in
    si
  end;

fun sc_heap_read_i_i_i_o x xa xb =
  Predicate.sup_pred
    (case xb
      of Heap.CField (d, f) =>
        (case RBTMapImpl.rm_lookup Arith.linorder_nat xa x
          of NONE => Predicate.bot_pred
          | SOME (Obj (_, fs)) =>
            (case TrieMapImpl.tm_lookup Stringa.equal_char (String.explode f) fs
              of NONE => Predicate.bot_pred
              | SOME fsa =>
                (case ListMapImpl.lm_lookup Stringa.equal_literal d fsa
                  of NONE => Predicate.bot_pred | SOME a => Predicate.single a))
          | SOME (Arr (_, _, _, _)) => Predicate.bot_pred)
      | Heap.ACell _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb of Heap.CField (_, _) => Predicate.bot_pred
        | Heap.ACell n =>
          (case RBTMapImpl.rm_lookup Arith.linorder_nat xa x
            of NONE => Predicate.bot_pred
            | SOME (Obj (_, _)) => Predicate.bot_pred
            | SOME (Arr (_, si, _, el)) =>
              Predicate.bind (Predicate.if_pred (IntInf.< (n, si)))
                (fn () =>
                  Predicate.single
                    (Option.the
                      (RBTMapImpl.rm_lookup Arith.linorder_nat n el)))))
      (case xb
        of Heap.CField (xba, f) =>
          (case RBTMapImpl.rm_lookup Arith.linorder_nat xa x
            of NONE => Predicate.bot_pred
            | SOME (Obj (_, _)) => Predicate.bot_pred
            | SOME (Arr (_, _, fa, _)) =>
              (case ListMapImpl.lm_lookup Stringa.equal_literal f fa
                of NONE => Predicate.bot_pred
                | SOME v =>
                  (if ((xba : string) = Type.object) then Predicate.single v
                    else Predicate.bot_pred)))
        | Heap.ACell _ => Predicate.bot_pred));

fun sc_heap_write_i_i_i_i_o x xa xb xc =
  Predicate.sup_pred
    (case xb
      of Heap.CField (d, f) =>
        (case RBTMapImpl.rm_lookup Arith.linorder_nat xa x
          of NONE => Predicate.bot_pred
          | SOME (Obj (c, fs)) =>
            Predicate.single
              (RBTMapImpl.rm_update Arith.linorder_nat xa
                (Obj (c, TrieMapImpl.tm_update Stringa.equal_char
                           (String.explode f)
                           (ListMapImpl.lm_update Stringa.equal_literal d xc
                             (case TrieMapImpl.tm_lookup Stringa.equal_char
                                     (String.explode f) fs
                               of NONE => ListMapImpl.lm_empty
                               | SOME fsa => fsa))
                           fs))
                x)
          | SOME (Arr (_, _, _, _)) => Predicate.bot_pred)
      | Heap.ACell _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb of Heap.CField (_, _) => Predicate.bot_pred
        | Heap.ACell n =>
          (case RBTMapImpl.rm_lookup Arith.linorder_nat xa x
            of NONE => Predicate.bot_pred
            | SOME (Obj (_, _)) => Predicate.bot_pred
            | SOME (Arr (t, si, f, el)) =>
              Predicate.single
                (RBTMapImpl.rm_update Arith.linorder_nat xa
                  (Arr (t, si, f,
                         RBTMapImpl.rm_update Arith.linorder_nat n xc el))
                  x)))
      (case xb
        of Heap.CField (xba, f) =>
          (case RBTMapImpl.rm_lookup Arith.linorder_nat xa x
            of NONE => Predicate.bot_pred
            | SOME (Obj (_, _)) => Predicate.bot_pred
            | SOME (Arr (t, si, fa, el)) =>
              (if ((xba : string) = Type.object)
                then Predicate.single
                       (RBTMapImpl.rm_update Arith.linorder_nat xa
                         (Arr (t, si,
                                ListMapImpl.lm_update Stringa.equal_literal f xc
                                  fa,
                                el))
                         x)
                else Predicate.bot_pred))
        | Heap.ACell _ => Predicate.bot_pred));

end; (*struct SC_Collections*)

structure RBTSetImpl : sig
  val rs_ins :
    'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> ('a, unit) RBT.rbt
  val rs_memb : 'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> bool
  val rs_empty : 'a Orderings.linorder -> ('a, unit) RBT.rbt
  val rs_delete :
    'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> ('a, unit) RBT.rbt
end = struct

fun rs_ins A_ = SetByMap.s_ins (RBTMapImpl.rm_update A_);

fun rs_memb A_ = SetByMap.s_memb (RBTMapImpl.rm_lookup A_);

fun rs_empty A_ = SetByMap.s_empty (RBTMapImpl.rm_empty A_);

fun rs_delete A_ = SetByMap.s_delete (RBTMapImpl.rm_delete A_);

end; (*struct RBTSetImpl*)

structure J_Execute : sig
  val sc_red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o :
    (string list * (string, unit, IntInf.int) Expr.exp) TypeRel.prog ->
      IntInf.int ->
        ((string, unit, IntInf.int) Expr.exp *
          (string -> IntInf.int Value.vala option)) *
          (IntInf.int, SC_Collections.heapobj) RBT.rbt ->
          (((IntInf.int, (FWState.lock_action list)) FinFun.finfun *
             ((IntInf.int,
                ((string, unit, IntInf.int) Expr.exp *
                  (string -> IntInf.int Value.vala option)),
                (IntInf.int, SC_Collections.heapobj) RBT.rbt)
                FWState.new_thread_action list *
               (IntInf.int FWState.conditional_action list *
                 ((IntInf.int, IntInf.int) FWState.wait_set_action list *
                   (IntInf.int FWState.interrupt_action list *
                     (IntInf.int, IntInf.int)
                       Observable_Events.obs_event list))))) *
            (((string, unit, IntInf.int) Expr.exp *
               (string -> IntInf.int Value.vala option)) *
              (IntInf.int, SC_Collections.heapobj) RBT.rbt))
            Predicate.pred
  val sc_rr_J_start_state :
    IntInf.int ->
      'a TypeRel.prog -> (IntInf.int list * IntInf.int list) * IntInf.int
  val exec_J_rr :
    IntInf.int ->
      (string list * (string, unit, IntInf.int) Expr.exp) TypeRel.prog ->
        string ->
          string ->
            IntInf.int Value.vala list ->
              ((IntInf.int *
                 (IntInf.int, IntInf.int) Observable_Events.obs_event list),
                ((IntInf.int, ((IntInf.int * IntInf.int) option))
                   FinFun.finfun *
                  (((IntInf.int,
                      (((string, unit, IntInf.int) Expr.exp *
                         (string -> IntInf.int Value.vala option)) *
                        (IntInf.int, IntInf.int) FinFun.finfun))
                      RBT.rbt *
                     (IntInf.int, SC_Collections.heapobj) RBT.rbt) *
                    ((IntInf.int, IntInf.int FWState.wait_set_status) RBT.rbt *
                      (IntInf.int, unit) RBT.rbt))))
                TLList.tllist
  val sc_rnd_J_start_state : int * int -> int * int
  val exec_J_rnd :
    int * int ->
      (string list * (string, unit, IntInf.int) Expr.exp) TypeRel.prog ->
        string ->
          string ->
            IntInf.int Value.vala list ->
              ((IntInf.int *
                 (IntInf.int, IntInf.int) Observable_Events.obs_event list),
                ((IntInf.int, ((IntInf.int * IntInf.int) option))
                   FinFun.finfun *
                  (((IntInf.int,
                      (((string, unit, IntInf.int) Expr.exp *
                         (string -> IntInf.int Value.vala option)) *
                        (IntInf.int, IntInf.int) FinFun.finfun))
                      RBT.rbt *
                     (IntInf.int, SC_Collections.heapobj) RBT.rbt) *
                    ((IntInf.int, IntInf.int FWState.wait_set_status) RBT.rbt *
                      (IntInf.int, unit) RBT.rbt))))
                TLList.tllist
end = struct

fun sc_red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o p t ((e, xs), h) =
  Predicate.bind
    (SmallStep.red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
      (Arith.equal_nat, MM.addr_nat) (fn x => x) (fn x => x)
      (RBTMapImpl.rm_empty Arith.linorder_nat) (SC_Collections.sc_new_obj p)
      (SC_Collections.sc_new_arr p) SC_Collections.sc_typeof_addr
      SC_Collections.sc_array_length SC_Collections.sc_heap_read_i_i_i_o
      SC_Collections.sc_heap_write_i_i_i_i_o
      (FWState.convert_extTA (SmallStep.extNTA2J Arith.equal_nat p)) p t e
      (h, xs))
    (fn (ta, (ea, (ha, xsa))) => Predicate.single (ta, ((ea, xsa), ha)));

fun sc_rr_J_start_state n0 p =
  Round_Robin.round_robin_start Fifo.fifo_empty Fifo.fifo_enqueue n0
    (StartConfig.start_tid MM.addr_nat (fn x => x)
      (RBTMapImpl.rm_empty Arith.linorder_nat) (SC_Collections.sc_new_obj p));

fun exec_J_rr n0 p c m vs =
  Scheduler.exec Arith.equal_nat Arith.equal_nat Arith.equal_nat
    (Observable_Events.convert_RA (Arith.equal_nat, MM.addr_nat))
    (Round_Robin.round_robin Arith.equal_nat
      (Card_Univ.card_UNIV_nat, Arith.equal_nat) Arith.equal_nat
      (fn (e, _) => Expr.final e)
      (sc_red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o p)
      (Observable_Events.convert_RA (Arith.equal_nat, MM.addr_nat))
      (RBTMapImpl.rm_lookup Arith.linorder_nat)
      (RBTMapImpl.rm_update Arith.linorder_nat)
      (RBTMapImpl.rm_lookup Arith.linorder_nat)
      (RBTSetImpl.rs_memb Arith.linorder_nat)
      (RBTSetImpl.rs_ins Arith.linorder_nat)
      (RBTSetImpl.rs_delete Arith.linorder_nat)
      (Fifo.fifo_isEmpty Arith.equal_nat) Fifo.fifo_enqueue Fifo.fifo_dequeue
      Fifo.fifo_push n0)
    State_Refinement.jinja_output
    (Scheduler.pick_wakeup_via_sel Arith.equal_nat
      (RBTMapImpl.rm_sel Arith.linorder_nat))
    (RBTMapImpl.rm_lookup Arith.linorder_nat)
    (RBTMapImpl.rm_update Arith.linorder_nat)
    (RBTMapImpl.rm_lookup Arith.linorder_nat)
    (RBTMapImpl.rm_update Arith.linorder_nat)
    (RBTMapImpl.rm_delete Arith.linorder_nat)
    (RBTMapImpl.rm_iterate Arith.linorder_nat)
    (RBTSetImpl.rs_ins Arith.linorder_nat)
    (RBTSetImpl.rs_delete Arith.linorder_nat) (sc_rr_J_start_state n0 p)
    (State_Refinement.start_state_refine MM.addr_nat
      (Product_Type.equal_prod (List.equal_list Stringa.equal_literal)
        (Expr.equal_exp Stringa.equal_literal Product_Type.equal_unit
          Arith.equal_nat))
      (fn x => x) (RBTMapImpl.rm_empty Arith.linorder_nat)
      (SC_Collections.sc_new_obj p) (RBTMapImpl.rm_empty Arith.linorder_nat)
      (RBTMapImpl.rm_update Arith.linorder_nat)
      (RBTMapImpl.rm_empty Arith.linorder_nat)
      (RBTSetImpl.rs_empty Arith.linorder_nat)
      (fn ca => fn _ => fn ts => fn _ => fn (pns, body) => fn vsa =>
        (Expr.blocks ("this" :: pns) (Type.Class ca :: ts) (Value.Null :: vsa)
           body,
          (fn _ => NONE)))
      p c m vs);

fun sc_rnd_J_start_state seed = seed;

fun exec_J_rnd seed p c m vs =
  Scheduler.exec Arith.equal_nat Arith.equal_nat Arith.equal_nat
    (Observable_Events.convert_RA (Arith.equal_nat, MM.addr_nat))
    (Random_Scheduler.random_scheduler Arith.equal_nat
      (Card_Univ.card_UNIV_nat, Arith.equal_nat) Arith.equal_nat
      (fn (e, _) => Expr.final e)
      (sc_red_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o p)
      (Observable_Events.convert_RA (Arith.equal_nat, MM.addr_nat))
      (RBTMapImpl.rm_lookup Arith.linorder_nat)
      (RBTMapImpl.rm_update Arith.linorder_nat)
      (RBTMapImpl.rm_iterate Arith.linorder_nat)
      (RBTMapImpl.rm_lookup Arith.linorder_nat)
      (RBTSetImpl.rs_memb Arith.linorder_nat)
      (RBTSetImpl.rs_ins Arith.linorder_nat)
      (RBTSetImpl.rs_delete Arith.linorder_nat) ListSetImpl_Invar.lsi_empty
      ListSetImpl_Invar.lsi_ins_dj ListSetImpl_Invar.lsi_to_list)
    State_Refinement.jinja_output
    (Scheduler.pick_wakeup_via_sel Arith.equal_nat
      (RBTMapImpl.rm_sel Arith.linorder_nat))
    (RBTMapImpl.rm_lookup Arith.linorder_nat)
    (RBTMapImpl.rm_update Arith.linorder_nat)
    (RBTMapImpl.rm_lookup Arith.linorder_nat)
    (RBTMapImpl.rm_update Arith.linorder_nat)
    (RBTMapImpl.rm_delete Arith.linorder_nat)
    (RBTMapImpl.rm_iterate Arith.linorder_nat)
    (RBTSetImpl.rs_ins Arith.linorder_nat)
    (RBTSetImpl.rs_delete Arith.linorder_nat) (sc_rnd_J_start_state seed)
    (State_Refinement.start_state_refine MM.addr_nat
      (Product_Type.equal_prod (List.equal_list Stringa.equal_literal)
        (Expr.equal_exp Stringa.equal_literal Product_Type.equal_unit
          Arith.equal_nat))
      (fn x => x) (RBTMapImpl.rm_empty Arith.linorder_nat)
      (SC_Collections.sc_new_obj p) (RBTMapImpl.rm_empty Arith.linorder_nat)
      (RBTMapImpl.rm_update Arith.linorder_nat)
      (RBTMapImpl.rm_empty Arith.linorder_nat)
      (RBTSetImpl.rs_empty Arith.linorder_nat)
      (fn ca => fn _ => fn ts => fn _ => fn (pns, body) => fn vsa =>
        (Expr.blocks ("this" :: pns) (Type.Class ca :: ts) (Value.Null :: vsa)
           body,
          (fn _ => NONE)))
      p c m vs);

end; (*struct J_Execute*)

structure Code_Generation : sig
  val shr : 'a * (('b * 'c) * ('d * 'e)) -> 'c
  val thr : 'a * (('b * 'c) * ('d * 'e)) -> 'b
  val purge :
    (('a * ('b, 'c) Observable_Events.obs_event list), 'd) TLList.tllist ->
      Numeral_Type.num1 Numeral_Type.bit0 Numeral_Type.bit0 Numeral_Type.bit0
        Numeral_Type.bit0
        Numeral_Type.bit0
        Word.word
        Coinductive_List.llist
  val terminal : ('a, 'b) TLList.tllist -> 'b
  val j_Program :
    (string *
      (string *
        ((string * (Type.ty * unit TypeRel.fmod_ext)) list *
          (string *
            (Type.ty list *
              (Type.ty *
                (string list *
                  (string, unit, IntInf.int) Expr.exp) option))) list))) list ->
      (string list * (string, unit, IntInf.int) Expr.exp) TypeRel.prog
  val llist_case :
    'a -> ('b -> 'b Coinductive_List.llist -> 'a) ->
            'b Coinductive_List.llist -> 'a
  val tllist_case :
    ('a -> 'b) ->
      ('c -> ('c, 'a) TLList.tllist -> 'b) -> ('c, 'a) TLList.tllist -> 'b
  val heapobj_toString : SC_Collections.heapobj -> string
  val toString_heapobja : SC_Collections.heapobj -> string
  val toString_heapobj : SC_Collections.heapobj ToString.toString
  val heap_toString : (IntInf.int, SC_Collections.heapobj) RBT.rbt -> string
  val trace_toString :
    IntInf.int * (IntInf.int, IntInf.int) Observable_Events.obs_event list ->
      string
  val llist_of_tllist : ('a, 'b) TLList.tllist -> 'a Coinductive_List.llist
  val thread_toString :
    (IntInf.int,
      (((string, unit, IntInf.int) Expr.exp *
         (string -> IntInf.int Value.vala option)) *
        (IntInf.int, IntInf.int) FinFun.finfun))
      RBT.rbt ->
      string
end = struct

fun shr x = FWState.shr x;

fun thr x = FWState.thr x;

fun purge run =
  Coinductive_List.lmap
    (fn Observable_Events.ExternalCall (_, _, Value.Intg i :: _, _) => i)
    (Coinductive_List.lfilter
      (fn a =>
        (case a of Observable_Events.ExternalCall (_, m, [], _) => false
          | Observable_Events.ExternalCall (_, m, Value.Unit :: lista, _) =>
            false
          | Observable_Events.ExternalCall (_, m, Value.Null :: lista, _) =>
            false
          | Observable_Events.ExternalCall (_, m, Value.Bool _ :: lista, _) =>
            false
          | Observable_Events.ExternalCall (_, m, [Value.Intg _], _) =>
            ((m : string) = "~print(I)V")
          | Observable_Events.ExternalCall (_, m, Value.Intg _ :: _ :: _, _) =>
            false
          | Observable_Events.ExternalCall (_, m, Value.Addr _ :: lista, _) =>
            false
          | Observable_Events.ReadMem (_, _, _) => false
          | Observable_Events.WriteMem (_, _, _) => false
          | Observable_Events.NewObj (_, _) => false
          | Observable_Events.NewArr (_, _, _) => false
          | Observable_Events.ThreadStart _ => false
          | Observable_Events.ThreadJoin _ => false
          | Observable_Events.SyncLock _ => false
          | Observable_Events.SyncUnlock _ => false
          | Observable_Events.ObsInterrupt _ => false
          | Observable_Events.ObsInterrupted _ => false))
      (Coinductive_List.lconcat
        (Coinductive_List.lmap (Coinductive_List.llist_of o Product_Type.snd)
          (TLList.llist_of_tllist run))));

fun terminal x = TLList.terminal x;

fun j_Program x = TypeRel.Program x;

fun llist_case x = Coinductive_List.llist_case x;

fun tllist_case x = TLList.tllist_case x;

fun heapobj_toString (SC_Collections.Obj (c, fs)) =
  String.concat
    ["(Obj ", ToString.toString_literala c, ", ",
      ToString.toString_trie ToString.toString_char
        (ToString.toString_assoc_list ToString.toString_literal
          (ToString.toString_val ToString.toString_nat))
        fs,
      ")"]
  | heapobj_toString (SC_Collections.Arr (t, si, fs, el)) =
    String.concat
      ["([", ToString.toString_nata si, "]", ToString.toString_ty t, ", ",
        ToString.toString_assoc_lista ToString.toString_literal
          (ToString.toString_val ToString.toString_nat) fs,
        ", ",
        ToString.toString_lista (ToString.toString_val ToString.toString_nat)
          (List.map Product_Type.snd
            (RBTMapImpl.rm_to_list Arith.linorder_nat el)),
        ")"];

fun toString_heapobja x = heapobj_toString x;

val toString_heapobj = {toString = toString_heapobja} :
  SC_Collections.heapobj ToString.toString;

fun heap_toString x =
  ToString.toString_rbt (Arith.linorder_nat, ToString.toString_nat)
    toString_heapobj x;

fun trace_toString x =
  ToString.toString_proda ToString.toString_nat
    (ToString.toString_list
      (ToString.toString_obs_event ToString.toString_nat ToString.toString_nat))
    x;

fun llist_of_tllist x = TLList.llist_of_tllist x;

fun thread_toString x =
  ToString.toString_rbt (Arith.linorder_nat, ToString.toString_nat)
    (ToString.toString_prod
      (ToString.toString_prod
        (ToString.toString_exp ToString.toString_literal ToString.toString_unit
          ToString.toString_nat)
        ToString.toString_fun)
      (ToString.toString_finfun
        (Card_Univ.card_UNIV_nat, Arith.equal_nat, Arith.linorder_nat,
          ToString.toString_nat)
        (Arith.equal_nat, ToString.toString_nat)))
    x;

end; (*struct Code_Generation*)
